/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
/*
 * @(#)calculadora.java	2.0 02/04/20
 *
 * Copyright 2002 Marta Oliver & Jos Luis Abreu.
 * Use is free. (Uso gratuito)
 */

/**
 * A command line calculator.
 * (Una calculadora que funciona desde la lnea de comandos.)
 * calculadora.java, 1999.
 *
 * @author Jos Luis Abreu
 * @author Marta Oliver
 * @version 	2.0, 20 Apr 2002
 */
 
package com.jla.calc;

import java.io.*;
import com.mja.util.BasicStr;

public class calculadora {

    /**
     * Los operadores que reconocer la calculadora.
     */
	private static final String[] operadores = {"+","-","*","/","=","-x","1/x"};
    /**
     * El valor acumulado. 
     * Se usa como primer operando al realizar una operacin binaria
     * o como el operando de una operacin unaria.
     */
	private double acumulado;
    /**
     * El segundo operando de una operacin binaria.
     */
	private double entrada;
    /**
     * El operador. Debe ser uno de los elementos de operadores.
     */
	private String operador;
	
    /**
     * El constructor nico sin parmetros.
     */
	public calculadora() {
		limpiar();
	}

    /**
     * Pone en ceros la entrada y el acumulador y pone como operador "=".
     */
	public void limpiar() { 
		entrada=0.0; 
		acumulado=0.0; 
		operador="=";
	} 
	
    /**
     * Determina si la cadena s es un operador para esta calculadora.
     * Los operadores son "+", "-", "*", "/", "=", "-x" y "1/x"
     * @return true si s es uno de los operadores y false en caso contrario.
     */
	public static boolean esOperador(String s) {
		for (int i=0;i<operadores.length;i++) {
			if (operadores[i].equals(s)) {
				return true;
			}
		}
		return false;
	}

    /**
     * Determina si la cadena s es un operador binario para esta calculadora.
     * Los operadores binarios son "+", "-", "*" y "/".
     * @return true si s es uno de los operadores binarios y false en caso contrario.
     */
	public static boolean esOperadorBinario(String s) {
		return "+".equals(s)||"-".equals(s)||"*".equals(s)||"/".equals(s);
		
	}

    /**
     * Analiza la cadena s, la convierte en un double y asigna el valor a entrada.
     * Lanza una excepcin si s no tiene formato de nmero decimal.
     */
	public void entrarValor(String s) throws NumberFormatException {
//		String str=BasicStr.replace(s,",",".");
		entrada=Double.valueOf(s).doubleValue();
	}
	
    /**
     * Define el operador activo como op si la cadena op es un operador 
     * de la calculadora y lanza una excepcin en caso contrario.
     */
	public void definirOperador(String op) throws Exception {
		for (int i=0;i<operadores.length;i++) {
			if (operadores[i].equals(op)) {
				operador=op;
				return;
			}
		}
		throw new Exception("operador desconocido");
	}
	
    /**
     * Realiza la operacin usando los valores actuales de 
     * acumulador, operador y entrada y pone el resultado en acumulador.
     * Puede lanzar una excepcin si la operacin es invlida,
     * por ejemplo si el operador="/" y entrada=0.0. 
     */
	public void realizarOperacin() throws Exception { 
		if (operador.equals("=")) {
			acumulado  = entrada;
		} else if (operador.equals("-x")) {
			acumulado  = -entrada;
		} else if (operador.equals("1/x")) {
			if (entrada !=0.0) {
				acumulado = 1/entrada;
			} else {
				throw new Exception("intento de dividir entre cero.");
			}
		} else if (operador.equals("+")) {
			acumulado += entrada;
		} else if (operador.equals("*")) {
			acumulado *= entrada;
		} else if (operador.equals("-")) {
			acumulado -= entrada;
		} else if (operador.equals("/")) {
			if (entrada !=0.0) {
				acumulado /= entrada;
			} else {
				throw new Exception("intento de dividir entre cero.");
			}
		}
	}
	
    /**
     * Devuelve el valor actual de entrada.
     */
	public String entrada() { 
		return BasicStr.DoubleToString(entrada,20,false,".");
//		return Double.toString(redondear(entrada));
	}

    /**
     * Devuelve el valor actual de acumulador.
     */
	public String resultado() { 
		return BasicStr.DoubleToString(acumulado,20,false,".");
//		return Double.toString(redondear(acumulado)); 
	}
	
    /**
     * Devuelve el valor actual de operador.
     */
	public String operador() {
		return operador;
	}

    /**
     * Redondea x a 8 decimales.
     */
	private static double redondear(double x) {
		double pres=100000000.0;
		return (Math.round(x*pres)/pres);
	}


    /**
     * Crea un objeto calculadora, pide al usuario entrada de un nmero 
     * y realiza la operacin por defecto que es "=", con lo cual
     * pasa al acumulador el valor de la entrada.
     * Mientras no ocurra una excepcin:
     * pone el resultado como entrada,
     * pide al usuario un operador,
     * si el operador es binario pide otro nmero,
     * realiza la operacin,
     * escribe el resultado en la consola.
     */
	public static void main(String[] args) {
		BufferedReader br=new BufferedReader(new InputStreamReader(System.in));
		calculadora C=new calculadora();
		try	{
			System.out.print("operando : "); 
			C.entrarValor(br.readLine());
			C.realizarOperacin();
			while (true) {
				C.entrarValor(C.resultado());
				System.out.print("operador : ");
				C.definirOperador(br.readLine());
				if (C.esOperadorBinario(C.operador())) {
					System.out.print("operando : "); 
					C.entrarValor(br.readLine());
				}
				C.realizarOperacin();
				System.out.println("resultado= "+C.resultado());
			} 
		} catch (Exception e) {
			System.out.println(e);
		}
	}

}