/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.jla.desc2.descartes;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import com.jla.desc2.util.*;
import com.jla.desc2.math.*;
import com.jla.desc2.gui.*;
import com.jla.desc2.parser.*;

public class R2Space extends Panel implements ActionListener, MouseListener, MouseMotionListener, KeyListener, Runnable {

	private static final double infNotExp=0.00000001, supNotExp=9999999999.9;

	int w=-1,h=-1;
	int Ox,Oy;
	double esc;

	private Descartes D;
	private static final int bck=0, net=1, net10=2, axe=3, tex=4;
	private Color[] DC;
	private Color bC;
	private int Oxa,Oya;
	private double esca;
	private Control[] control;
	private volatile Control aC;
	private mjaGraph[] graph;
	private Image back,front,img;
	private Graphics bg,fg,ig;
	private Graphics[] bga,tga,fga;
	private boolean netOn, net10On, axesOn, textOn, fillInFront;
	private Thread thr;
	private Panel animCtrl;
	private mjaButton[] b;
	volatile boolean running;

	public R2Space(Descartes D) {
		this.D=D;
		setLayout(null);
		animCtrl=new Panel();
		animCtrl.setBounds(0,0,120,20);
		add(animCtrl);
		b=new mjaButton[6];
		animCtrl.setLayout(new GridLayout(1,b.length));
		for (int i=0;i<b.length;i++) {
			animCtrl.add(b[i]=new mjaButton(i));
			b[i].addActionListener(this);
		}
		animCtrl.setVisible(false);
		addMouseListener(this);
		addMouseMotionListener(this);
		addKeyListener(this);
	}

	R2 injection(double x,double y) {
		return new R2((double)(x-Ox)/esc,(double)(Oy-y)/esc);
	}

	R2 projection(R2 p) {
		return new R2(Ox+esc*p.x,Oy-esc*p.y);
	}

	double X(int size,double v) {
		double x=Ox+esc*v;
		if (x<-size) { x=-size; }
		if (x>w+size) { x=w+size; }
		return x;
	}
	double Y(int size, double v) {
		double y=Oy-esc*v;
		if (y<-size) { y=-size; }
		if (y>h+size) { y=h+size; }
		return y;
	}

	int XX(int size,double v) {
		return (int) Math.round(X(size,v));
	}
	int YY(int size, double v) {
		return (int) Math.round(Y(size,v));
	}

	public boolean inside(R2 q) {
		return -w<=q.ix()&&q.ix()<=2*w&&-h<=q.iy()&&q.iy()<=2*h ;
	}

	Panel initializeObjects(config cfg) {
		netOn=cfg.isOn[net];
		net10On=cfg.isOn[net10];
		axesOn=cfg.isOn[axe];
		textOn=cfg.isOn[tex];
		DC=cfg.color;
		bC=DC[bck];
		// create controls
		int nC=0;
		for (int i=0;i<cfg.cc.length;i++) {
			if (cfg.cc[i].type==data.c_graphic) {
				nC++;
			}
		}
		control=new Control[nC];
		int k=0;
		for (int i=0;i<cfg.cc.length;i++) {
			if (cfg.cc[i].type==data.c_graphic) {
				control[k]=cfg.cc[i].ctrl;
				control[k++].Initialize(D);
			}
		}
		// create graphs
		graph=new mjaGraph[cfg.gc.length];
		int N=0;
		for (int i=0;i<graph.length;i++) {
			if (cfg.gc[i].type<0) {
				graph[i]=null;
			} else {
				graph[i]=new mjaGraph(D,cfg.gc[i]);
				if (graph[i].gc.isVisible) {
					N++;
				}
			}
		}
		Panel ecp=null;
		if (N>0) {
			ecp=new Panel();
			ecp.setLayout(new GridLayout(1,N));
			for (int i=0;i<graph.length;i++) {
				if (graph[i].gc.isVisible) {
					graph[i].TF=new TextField(graph[i].gc.formula);
					graph[i].TF.setForeground(graph[i].gc.gcolor);
					if (mjaStr.osIsWindows()) {
						graph[i].TF.setFont(mjaGui.Courier);
					} else {
						graph[i].TF.setFont(mjaGui.Courier_11);
					}
					if (graph[i].gc.isEditable) {
						graph[i].TF.setEditable(true);
						graph[i].TF.addKeyListener(this);
					} else {
						graph[i].TF.setEditable(false);
					}
					graph[i].TF.setBackground(bC);
					ecp.add(graph[i].TF);
				}
			}
		}
		return ecp;
	}

/* ---------------------- Graphic Stuff --------------------- */
	private Graphics g;

	private boolean initializeGraphicStuff() {
		boolean bgch=false;
		if (img==null||w!=getSize().width||h!=getSize().height) {
			g=getGraphics();
			w=getSize().width;
			h=getSize().height;
			back=createImage(w,h);
			bg=back.getGraphics();
			bg.setFont(mjaGui.Courier);
			img=createImage(w,h);
			ig=img.getGraphics();
			ig.setFont(mjaGui.Courier);
			front=createImage(w,h);
			fg=front.getGraphics();
			fg.setFont(mjaGui.Courier);
			bga=new Graphics[1];
			bga[0]=bg;
			bgch=true;
		}
		fillInFront=false;
		for (int i=0;i<graph.length;i++) {
			if (graph[i]!=null && !graph[i].gc.isInBack) {
				if (graph[i].gc.type==data.fill) {
					fillInFront=true;
					break;
				}
			}
		}
		if (fillInFront) {
			fga=new Graphics[2];
			fga[0]=fg;
			fga[1]=ig;
			tga=new Graphics[3];
			tga[0]=bg;
			tga[1]=fg;
			tga[2]=ig;
		} else {
			fga=new Graphics[1];
			fga[0]=ig;
			tga=new Graphics[2];
			tga[0]=bg;
			tga[1]=ig;
		}
		bgch=spaceChanged()|bgch;
		return bgch;
	}

	private boolean spaceChanged() {
		Ox=(int)Math.round(D.p.value("Ox")+(double)w/2.0);
		if (D.cfg.version0.startsWith("1")&& !D.cfg.hasVisibleEquation()) {
			Oy=(int)Math.round(D.p.value("Oy")+(double)(h-24)/2.0);
		} else {
			Oy=(int)Math.round(D.p.value("Oy")+(double)h/2.0);
		}
		esc=D.p.value("escala",config.Escala0);
		// la escala est limitada a ser mayor que 1E-6 y menor que 1E6.
		if (esc<0.000001) {
			esc=0.000001;
		}
		if (esc>1000000) {
			esc=1000000;
		}
		boolean changed=esc!=esca||Ox!=Oxa||Oy!=Oya;
		esca=esc; Oxa=Ox; Oya=Oy;
		return changed;
	}

	private void drawBackground() {
		bg.setColor(data.adaptColor(D,DC[bck]));
		bg.fillRect(0,0,w,h);
		double rsc=esc;
		int dec=0;
		while (rsc>(w+h)) { rsc/=10; dec++; }
		while (rsc<(w+h)/10) { rsc*=10;	}
		if (netOn) {
			bg.setColor(data.adaptColor(D,DC[net]));
			drawMarks(rsc/10,-1);
			if (net10On) {
				bg.setColor(data.adaptColor(D,DC[net10]));
				drawMarks(rsc,-1);
			}
		}
		if (axesOn) {
			bg.setFont(mjaGui.Courier);
			bg.setColor(data.adaptColor(D,DC[axe]));
			if (D.cfg.x_On) {
				bg.drawLine(0,Oy,w,Oy);
			}
			if (D.cfg.y_On) {
				bg.drawLine(Ox,0,Ox,h);
			}
			if (D.cfg.numbers) {
				if (rsc>(w+h)/2) {
					drawNumbers(rsc/5,(rsc<=esc)?dec+1:dec);
				} else if (rsc>(w+h)/4) {
					drawNumbers(rsc/2,(rsc<=esc)?dec+1:dec);
				} else {
					drawNumbers(rsc,dec);
				}
			}
			drawMarks(rsc,3);
			rsc/=2; drawMarks(rsc,2);
			rsc/=5; drawMarks(rsc,1);
			bg.setColor(data.adaptColor(D,DC[axe]));
			int xL=Ox-2-bg.getFontMetrics(bg.getFont()).stringWidth(D.cfg.y_axis);
			int yL=bg.getFontMetrics(bg.getFont()).getAscent();
			if (D.cfg.y_On) {
				bg.drawString(D.cfg.y_axis,xL,yL);
			}
			xL=w-3-bg.getFontMetrics(bg.getFont()).stringWidth(D.cfg.x_axis);
			yL=Oy+yL;
			if (D.cfg.x_On) {
				bg.drawString(D.cfg.x_axis,xL,yL);
			}
		}
		for (int i=0;i<graph.length;i++) {
			if (graph[i]!=null && graph[i].gc.isInBack) {
				drawGraph(D.p,graph[i]);
			}
		}
	}

	private void drawMarks(double rsc,int sz) {
		int x1=0,x2=w,y1=0,y2=h;
		if (sz>=0) {
			x1=Ox-sz; x2=Ox+sz; y1=Oy-sz; y2=Oy+sz;
		}
		int x,y;
		if ((D.cfg.x_On&&sz>0) || sz<0) {
			for (int i=-(int)Math.round(Ox/rsc);(x=Ox+(int)Math.round(i*rsc))<w;i++) {
				bg.drawLine(x,y1,x,y2);
			}
		}
		if ((D.cfg.y_On&&sz>0) || sz<0) {
			for (int i=-(int)Math.round(Oy/rsc);(y=Oy+(int)Math.round(i*rsc))<h;i++) {
				bg.drawLine(x1,y,x2,y);
			}
		}
	}

	private void drawNumbers(double rsc,int dec) {
		int x,y;
		if (D.cfg.x_On) {
			for (int i=-(int)Math.round(Ox/rsc);(x=Ox+(int)Math.round(i*rsc))<w;i++) {
				bg.drawString(mjaStr.DoubleToString((double)i*rsc/esc,4,false),x+2,Oy-2);
//				bg.drawString(mjaStr.notExp((double)i*rsc/esc,infNotExp,supNotExp),x+2,Oy-2);
			}
		}
		if (D.cfg.y_On) {
			for (int i=-(int)Math.round(Oy/rsc);(y=Oy+(int)Math.round(i*rsc))<h;i++) {
				bg.drawString(mjaStr.DoubleToString(-(double)i*rsc/esc,4,false),Ox+2,y-2);
//				bg.drawString(mjaStr.notExp(-(double)i*rsc/esc,infNotExp,supNotExp),Ox+2,y-2);
			}
		}
	}

	void draw(boolean starting,boolean clearing,boolean ini) {
		if (initializeGraphicStuff()||clearing) {
			drawBackground();
		}
		if (fillInFront) {
			fg.setColor(data.adaptColor(D,DC[bck]));
			fg.fillRect(0,0,w,h);
		}
		if (ini&&D.cfg.animCfg.active) {
			alg=new Algorithm(D.p,false,D.cfg.animCfg.s_do,D.cfg.animCfg.s_ini,D.cfg.animCfg.s_while);
			alg.initialize();
		}
		if (starting && D.cfg.animCfg.active) {
			stop();
			start();
		} else {
			if (thr==null||!thr.isAlive()||pause) {
				drawFront();
			}
		}
	}

	private void animDraw() {
		D.doCalculations(false);
		if (initializeGraphicStuff()) {
			drawBackground();
		}
		if (fillInFront) {
			fg.setColor(data.adaptColor(D,DC[bck]));
			fg.fillRect(0,0,w,h);
		}
		drawFront();
	}

	private void drawFront() {
		ig.drawImage(back,0,0,this);
		for (int i=0;i<control.length;i++) {
			control[i].adjust();
		}
		for (int i=0;i<graph.length;i++) {
			if (graph[i]!=null && (!graph[i].gc.isInBack)) {
				drawGraph(D.p,graph[i]);
			}
		}
		for (int i=0;i<control.length;i++) {
			if (control[i].hasTrace) {
				control[i].drawControl(true,bga);
			}
			control[i].drawControl(false,fga);
		}
		g.drawImage(img,0,0,this);
	}

/* --------------------- Graphs drawing -------------------- */

	private void drawGraph(Parser p, mjaGraph G) {
		for (int i=0;i<bga.length;i++) {
			if (G.gc.isInBack) {
				bga[i].setColor(data.adaptColor(D,G.gc.gcolor));
			} else {
				bga[i].setColor(data.adaptColor(D,G.gc.traceColor));
			}
		}
		for (int i=0;i<fga.length;i++) {
			fga[i].setColor(data.adaptColor(D,G.gc.gcolor));
		}
		if (G.sparz!=null) {
			Node N=p.Var(G.sparz.name);
			synchronized(N) {
				double v=N.r;
				int sN=(int)Math.round(G.sparz.NN.Evaluate(32));
				double si=G.sparz.Ni.Evaluate(0);
				double sf=G.sparz.Nf.Evaluate(1);
				for (int n=0;n<=sN;n++) {
					N.r=si+(sf-si)*(double)n/(double)sN;
					drawGraph(G);
				}
				N.r=v;
			}
		} else {
			drawGraph(G);
		}
	}

	private void drawGraph(mjaGraph G) {
		if (G.gc.isInBack) {
			G.draw(back,bga);
		} else if (G.gc.hasTrace) {
			G.draw(front,tga);
		} else {
			G.draw(front,fga);
		}
	}

/* ------------------------ Coordinates ----------------------- */

	private void writeCoordinates(int x, int y) {
		if (textOn) {
			Graphics g=getGraphics();
			g.setFont(mjaGui.Courier);
			g.setColor(data.adaptColor(D,DC[tex]));
			g.drawOval(x-2,y-2,4,4);
			FontMetrics fm=g.getFontMetrics();
			int pp=0;
			while (esc>Math.pow(10,pp)) { pp++; }
			if (pp>0) {
				pp++;
			}
			String str="("+mjaStr.DoubleToString((double)(x-Ox)/esc,pp)
						+","+mjaStr.DoubleToString((double)(Oy-y)/esc,pp)+")";
			int sw=fm.stringWidth(str);
			int posx=x-sw/2;
			int posy=y-12;
			if (posx<0) { posx=0; }
			int ld=getSize().width-sw;
			if (posx>ld) { posx=ld; }
			int li=fm.getAscent();
			if (posy<li) { posy=li; }
			int ls=getSize().height-fm.getDescent();
			if (posy>ls) { posy=ls; }
			g.drawString(str,posx,posy);
		}
	}

/* ---------------------- paint Canvas ---------------------- */

	public void paint(Graphics g) {
		if (img!=null) { g.drawImage(img,0,0,this); }
	}

	public void update(Graphics g) {
		paint(g);
	}

/* ------------------------ Runnable ---------------------- */

	volatile boolean pause,onestep;

	void start() {
		pause=false;
		onestep=false;
		if (D.cfg.animCfg.controls) {
			animCtrl.setBounds((w-animCtrl.getSize().width)/2,h-animCtrl.getSize().height,animCtrl.getSize().width,animCtrl.getSize().height);
			b[mjaButton.play].type=mjaButton.pause;
			animCtrl.setVisible(true);
		}
		thr=new Thread(this);
		thr.start();
	}

	void stop() {
		animCtrl.setVisible(false);
		D.animEnded();
		running=false;
		if (thr!=null) {
			thr.stop();
			thr=null;
		}
	}

	private Algorithm alg;
	private volatile boolean stay;

	public void run() {
		D.animStarted();
		running=true;
		Node delay=D.p.Analyse(D.cfg.animCfg.s_delay,"60");
		animDraw();
		stay=false;
		try {
			while (true) {
				do {
					long T=System.currentTimeMillis()+(long) Math.round(delay.Evaluate(50));
					thr.sleep(20);
					if (!pause||onestep) {
						stay=false;
						alg.step();
						onestep=false;
						animDraw();
					}
					while (System.currentTimeMillis()<T) {
						thr.sleep(20);
					}
				} while (alg.repeat());
				if (!stay && !D.cfg.animCfg.loop) {
					break;
				}
			}
		} catch (InterruptedException ie) {
		}
		stop();
	}

	static void limitExceeded() {
		mjaGui.ask(null,"Descartes"," Se ha excedidio el lmite de\n 100000 repeticiones en un algoritmo!","ok",null);
	}

	void backStep() {
		try {
			alg.backstep();
			animDraw();
		} catch (Exception e) {
			limitExceeded();
		}
	}

	void toEnd() {
		try {
			alg.toEnd();
			stay=true;
			animDraw();
		} catch (Exception e) {
			limitExceeded();
		}
	}

	void setActiveControl(int ix) {
		if (ix<0) { ix=control.length-1; }
		if (ix>=control.length) { ix=0; }
		for (int i=0;i<control.length;i++) {
			control[i].isActive=false;
		}
		control[ix].isActive=true;
		aC=control[ix];
		acix=ix;
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getSource()==b[mjaButton.play]) {
			pause=!pause;
		} else if (e.getSource()==b[mjaButton.rewind]) {
			boolean auxpause=pause;
			stop();
			alg.initialize();
			animDraw();
			start();
			pause=auxpause;
		} else if (e.getSource()==b[mjaButton.back]) {
			pause=true;
			backStep();
		} else if (e.getSource()==b[mjaButton.next]) {
			pause=true;
			onestep=true;
		} else if (e.getSource()==b[mjaButton.toend]) {
			pause=true;
			toEnd();
		} else if (e.getSource()==b[mjaButton.stop]) {
			stop();
		}
		if (pause) {
			b[mjaButton.play].type=mjaButton.play;
		} else {
			b[mjaButton.play].type=mjaButton.pause;
		}
		b[mjaButton.play].repaint();
	}

/* ----------------- MouseListener ----------------- */

	public void mousePressed(MouseEvent ev) {
		if (ev.getModifiers()==ev.BUTTON3_MASK) {
			D.showExternalSpinners();
		} else {
			if (aC!=null) {
				aC.isActive=false;
				D.draw(false,false);
			}
			aC=null;
			for (int i=0;i<control.length;i++) {
				R2 t=new R2(ev.getX(),ev.getY());
				if (projection(control[i].cpos).distance(t)<control[i].size+1) {
					setActiveControl(i);
					D.draw(false,false);
					return;
				}
			}
			writeCoordinates(ev.getX(),ev.getY());
		}
	}

	public void mouseReleased(MouseEvent ev) {
		if (aC!=null) {
			D.draw(false,false);
			requestFocus();
		} else {
			repaint();
		}
	}
	public void mouseClicked(MouseEvent ev) {
		if (ev.getModifiers()==ev.BUTTON3_MASK) {
			D.showExternalSpinners();
		} else if (ev.getClickCount()==2 && (D.inNippeEditor())) {
			D.edit();
		}
	}
	public void mouseEntered(MouseEvent ev) { }
	public void mouseExited(MouseEvent ev) { }

/* ----------------- MouseMotionListener ----------------- */

	public void mouseDragged(MouseEvent ev) {
		if (aC!=null) {
			aC.setControlPos(ev.getX(),ev.getY());
			D.draw(false,false);
		} else {
			g.drawImage(img,0,0,this);
			writeCoordinates(ev.getX(),ev.getY());
		}
	}
	public void mouseMoved(MouseEvent ev) { }

/* --------------------- KeyListener -------------------- */

	public void keyTyped (KeyEvent e) {
	}

	private int acix=0;

	public void keyPressed (KeyEvent e) {
		if (e.getSource()==this && control.length>0) {
			if (aC==null) {
				setActiveControl(0);
			}
			if (aC!=null) {
				Control aC0=aC;
				R2 p=aC.getControlPos();
				switch (e.getKeyCode()) {
				case KeyEvent.VK_LEFT: p.x-=1; break;
				case KeyEvent.VK_RIGHT: p.x+=1; break;
				case KeyEvent.VK_UP: p.y-=1; break;
				case KeyEvent.VK_DOWN: p.y+=1; break;
				case KeyEvent.VK_TAB:
					if (e.getModifiers()==KeyEvent.SHIFT_MASK) {
						acix-=1;
						if (acix<0) {
							acix=control.length-1;
						}
					} else {
						acix+=1;
						if (acix>=control.length) {
							acix=0;
						}
					}
					setActiveControl(acix);
					e.consume();
					break;
				}
				aC0.setControlPos(p.x,p.y);
			}
			D.draw(false,false);
		} else if ((e.getSource() instanceof TextField) && (e.getKeyCode()==KeyEvent.VK_ENTER)) {
			for (int i=0;i<graph.length;i++) {
				if (graph[i].TF==e.getSource()) {
					if (graph[i].gc.isEditable) {
						graph[i].gc.formula=graph[i].TF.getText();
						graph[i].analyse();
						if (graph[i].gc.isInBack) {
							drawBackground();
						}
						D.draw(false,false);
					}
					break;
				}
			}
		}
	}

	public void keyReleased (KeyEvent e) {
	}

}