/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.jla.sound;


import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.FloatControl;
import javax.sound.sampled.Line;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.UnsupportedAudioFileException;

import javazoom.spi.mpeg.sampled.file.MpegAudioFileReader;

public class SoundPlayer {

	static final Hashtable hashAudios = new Hashtable(); 
	
//	/**
//	 * OSCAR
//	 * Reproduce y asocia un archivo de sonido a un ID
//	 * @param audioID el ID para nombrar el archivo de audio
//	 * @param Mp3File El URL del archivo a reproducir
//	 * @return
//	 * @throws IOException
//	 */
//	public static String play2(String audioID,URL Mp3File) throws IOException {
//		try {
//			MpegAudioFileReader re = new MpegAudioFileReader();
//			if(Mp3File == null )
//				throw new IOException("El URL es null");
//			AudioInputStream audioInputStream = re.getAudioInputStream(Mp3File);
//			hashAudios.put(audioID,audioInputStream);
//			AudioPlayer.player.start(audioInputStream);
//			return audioID; 
//		} catch (UnsupportedAudioFileException e) {
//			throw new IOException("Formato del archivo no soportado");
//		} 	
//	}
	
	private static AudioContainerPack[] getAudioPackArray(String audioID) {
		AudioContainerPack[] pack;
		if (audioID!=null) {
			pack=new AudioContainerPack[1];
			pack[0]= (AudioContainerPack) hashAudios.get(audioID);
		} else {
			Enumeration en=hashAudios.elements();
			pack=new AudioContainerPack[hashAudios.size()];
			for (int i=0;i<pack.length;i++) {
				if (en.hasMoreElements()) {
					pack[i]=(AudioContainerPack) en.nextElement();
				}
			}
		}
		return pack;
	}
	
	public static boolean isPaused(String audioID) {
		AudioContainerPack[] pack=getAudioPackArray(audioID);
		for (int i=0;i<pack.length;i++) {
			if (pack[i].isAlive() && pack[i].isPaused()) {
				return true;
			}
		}
		return false;
	}

	public static boolean isPlaying(String audioID) {
		AudioContainerPack[] pack=getAudioPackArray(audioID);
		for (int i=0;i<pack.length;i++) {
			if (pack[i].isAlive()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * OSCAR
	 * Detiene y libera recursos del audio asociado a audioID si existe.
	 * @param audioID
	 */

	public static void stop(String audioID){
		if (audioID==null) {
			stopAll();
		} else {
			AudioContainerPack pack=(AudioContainerPack) hashAudios.get(audioID);
			try {
				AudioInputStream stream = pack.getAudioInputStream();
				Thread thread = pack.getThread();
				SourceDataLine line = pack.getLine();
				
			
				if(thread!=null){
					if(thread.isAlive()){
						thread.stop();
					}
				}
				
				if (line != null){
					line.drain();
					line.stop();
					line.close();
				}
				
				if(stream == null){
					stream.close();
				}
				hashAudios.remove(audioID);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	} 

	/**
	 * OSCAR 
	 * Regresa true si se ha leido por completo el archivo de sonido asociado a el audioID.
	 * Adems si se termino ya trata de liberar recursos cerrando el stream. 
	 * @param audioID 
	 * @return Si el medio ya termino de reproducirse
	 */
	public static boolean isFinished(String audioID){
		AudioContainerPack pack = (AudioContainerPack) hashAudios.get(audioID);
		
		if(pack==null){
			return true;
		}
		AudioInputStream stream = pack.getAudioInputStream();
		Thread thread = pack.getThread();
		
		int quedan = 0;
		boolean termino = true;
		try {
			if(stream!=null){
				quedan = stream.available();
				termino = (!thread.isAlive()) && (quedan<=0);
				if(termino)
					stop(audioID);
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		return termino;
	}
	
	/**
	 * Detiene y cierra todos los streams de audio
	 *
	 */
	public static void stopAll(){
		Hashtable aDetener = hashAudios;
		Enumeration enumeration = aDetener.keys();
		
		while(enumeration.hasMoreElements()){
			stop((String) enumeration.nextElement());
		}
	}
    /**
     * OSCAR
     * Un pseudo-garbagecollector 
     *
     */
	public static void my_gc(){
		Hashtable aDetener      = hashAudios;
		Enumeration enumeration = aDetener.keys();
		
		while(enumeration.hasMoreElements()){
			String id = (String) enumeration.nextElement(); 
			isFinished(id);
		}
	}
	
		/**
	 * Cambia el Volumen de la reproduccin para el archivo de sonido asociado a
	 * audioID.
	 * 
	 * @param audioID
	 * @param newVolume El nuevo volumen. Este debe de estar en el intervalo [0,1]
	 */
	public static void setVolumeLevel(String audioID, float newVolume) {
		AudioContainerPack[] pack=getAudioPackArray(audioID);
		for (int i=0;i<pack.length;i++) {
			if(pack[i]!=null) {
				try {
					SourceDataLine line = pack[i].getLine();
					FloatControl m_gainControl = (FloatControl) line.getControl(FloatControl.Type.MASTER_GAIN);
					float min = m_gainControl.getMinimum();
					float max = m_gainControl.getMaximum();
			
					double minGainDB = min;
					double ampGainDB = ((10.0f / 20.0f) * max) - min;
					double cste = Math.log(10.0) / 20;
					double valueDB = minGainDB + (1 / cste)
							* Math.log(1 + (Math.exp(cste * ampGainDB) - 1) * newVolume);
					pack[i].setVolumen((float) newVolume);
					m_gainControl.setValue((float) valueDB);
				} catch (Exception e) {
		//			System.out.println("Exception Occurred "+e);
				}
			}
		}
	}


	/**
	 * Pausa o reanuda la reproduccion del archivo asociado a <b>audioID</b>
	 * @param audioID
	 */
	public static void pause(String audioID){
		AudioContainerPack[] pack=getAudioPackArray(audioID);
		for (int i=0;i<pack.length;i++) {
			if(pack[i]!=null) {
				Thread thread = pack[i].getThread();
				if(thread!=null && thread.isAlive()){
					if (pack[i].isPaused()) {
						pack[i].setPaused(false);
						thread.resume();
					} else {
						pack[i].setPaused(true);
						thread.suspend();
					}
				}
//				synchronized (pack[i].getLock()) {
//					pack[i].setPaused(!pack[i].isPaused());
//					pack[i].getLock().notifyAll();
//				}

			} 
		}
	}
	
	
	public static  void setInLoop(String audioID,boolean loop){
		AudioContainerPack pack = (AudioContainerPack) hashAudios.get(audioID);
		if(pack == null)
			return;
		pack.setLoop(loop);
	}
	
	/**
	 * OSCAR Reproduce y asocia un archivo de sonido a un ID
	 * 
	 * @param audioID
	 *            el ID para nombrar el archivo de audio
	 * @param Mp3File
	 *            El URL del archivo a reproducir
	 * 
	 * 
	 * @throws IOException
	 * */
	public static void play(String audioID, URL Mp3File){
		play(audioID, Mp3File,false);
	}
	
	
	/**
	 * OSCAR Reproduce y asocia un archivo de sonido a un ID. En este mtodo se
	 * puede especificar si se desea que sea un ciclo la reproducin.
	 * 
	 * @param audioID
	 *            el ID para nombrar el archivo de audio
	 * @param Mp3File
	 *            El URL del archivo a reproducir
	 * @param loop
	 *            Si el archivo debe reproducirce como un loop
	 * 
	 * @throws IOException
	 */
	public static void play(final String audioID, URL Mp3File, boolean loop) {
		try {
			
			MpegAudioFileReader re = new MpegAudioFileReader();
			if(Mp3File == null )
				throw new IOException("El URL es null");
			
			AudioInputStream in = re.getAudioInputStream(Mp3File);
			AudioFormat baseFormat = in.getFormat();
			final AudioFormat decodedFormat = new AudioFormat(
					AudioFormat.Encoding.PCM_SIGNED,
					baseFormat.getSampleRate(), 16, baseFormat.getChannels(),
					baseFormat.getChannels() * 2, baseFormat.getSampleRate(),
					false);
			final AudioInputStream din = 
				AudioSystem.getAudioInputStream(decodedFormat, in);
			DataLine.Info info = 
				new DataLine.Info(SourceDataLine.class,decodedFormat);
			final SourceDataLine line = (SourceDataLine) 
				AudioSystem.getLine(info);
			final AudioContainerPack pack = new AudioContainerPack(Mp3File,din,null,line);
			pack.setLoop(loop);
			
			if (line != null) {
				Runnable runnable = new Runnable() {
					AudioContainerPack pack2 = pack;
					public void run() {
						try {
							pack2.getLine().open(decodedFormat);
							final byte[] data = new byte[4096];
							// Start
							pack2.getLine().start();
							int nBytesRead = pack2.getAudioInputStream().read(data, 0, data.length);
							pack2.getAudioInputStream().mark(1024);
							synchronized (pack2.getLock()) {
								while (nBytesRead!= -1) {
									while (pack2.isPaused()) {
										if(pack2.getLine().isRunning()) {
											pack2.getLine().stop();
										}
										try {
											pack2.getLock().wait();
										} catch (InterruptedException e) {}
									}
								
									if(!pack2.getLine().isRunning()) {
										pack2.getLine().start();
									}
									pack2.getLine().write(data, 0, nBytesRead);
									nBytesRead = pack2.getAudioInputStream().read(data, 0, data.length);
								}
							}
							pack2.getLine().drain();
							//  Stop
							pack2.getLine().stop();
							pack2.getLine().close();
							pack2.getAudioInputStream().close();
						} catch (Exception e) {
							e.printStackTrace();
						} finally {
							if (pack2.getAudioInputStream() != null) {
								try {
									pack2.getAudioInputStream().close();
								} catch (IOException e) {}
								
							}
						}
						if(pack2.isLoop()){
							play(audioID, pack2.getFile(),true);
							setVolumeLevel(audioID, pack2.getVolumen());
						}

						
					}
				};
				Thread actual = new Thread(runnable); 
				pack.setThread(actual);
				
				hashAudios.put(audioID, pack);
				actual.start();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	


	
	public final static class AudioContainerPack {
		URL file;
		AudioInputStream audioInputStream;
		Thread thread;
		SourceDataLine line;
		Object lock = new Object();
		boolean paused;
		boolean loop;
		float volumen = 1;
		
		public AudioContainerPack(URL file, AudioInputStream audioInputStream, Thread thread, SourceDataLine line) {
			super();
			this.file = file;
			this.audioInputStream = audioInputStream;
			this.thread = thread;
			this.line = line;
			
		}

		public AudioInputStream getAudioInputStream() {
			return audioInputStream;
		}

		public SourceDataLine getLine() {
			return line;
		}

		public Object getLock() {
			return lock;
		}
		
		public boolean isAlive () {
			if(thread!=null){
				if(thread.isAlive()){
					return true;
				}
			}
			return false;	
		}

		public Thread getThread() {
			return thread;
		}

		public void setAudioInputStream(AudioInputStream audioInputStream) {
			this.audioInputStream = audioInputStream;
		}

		public void setLine(SourceDataLine line) {
			this.line = line;
		}

		public void setThread(Thread thread) {
			this.thread = thread;
		}

		public boolean isPaused() {
			return paused;
		}

		public void setPaused(boolean paused) {
			this.paused = paused;
		}

		public boolean isLoop() {
			return loop;
		}

		public void setLoop(boolean loop) {
			this.loop = loop;
		}

		public URL getFile() {
			return file;
		}

		public void setFile(URL file) {
			this.file = file;
		}

		public float getVolumen() {
			return volumen;
		}

		public void setVolumen(float volumen) {
			this.volumen = volumen;
		}
		
		
	}
	
}