package com.jla.versionManager;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.StringWriter;
import java.lang.reflect.Constructor;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.text.DateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;

import javax.swing.JOptionPane;
import javax.swing.table.DefaultTableModel;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import com.mja.file.mjaFile;
import com.jla.versionManager.gui.UpdateLibraryQuestion;

public class DescartesVersionManager {
	private static Hashtable<String, String> entities = null;
	private String pathCfgManager = "resources/version_manager/cfgManager.xml";
	private String pathCfgManagerDir = "{USR_HOME}/.descartes/version_manager";
	
	
	private String urlVersionDataRemote;
	private String pathVersionDataLocal;
	private Hashtable<String, LibraryVersionManager> managers = new Hashtable<String, LibraryVersionManager>();
	
	private int maxTimeToTimeOut = 500;


	
	public DescartesVersionManager() throws IOException {
		try{
			prepareEntities();
			InputStream cfgMIS = this.getClass().getClassLoader().getResourceAsStream(pathCfgManager);
			DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
			Document doc = dBuilder.parse(cfgMIS);
			
			Element verRemote = (Element) doc.getElementsByTagName("versionInfoRemote").item(0);
			this.urlVersionDataRemote = processLocalEntities(verRemote.getAttribute("url"));
			
			Element verLocal = (Element) doc.getElementsByTagName("versionInfoLocal").item(0);
			this.pathVersionDataLocal = processLocalEntities(verLocal.getAttribute("path"));
	
			NodeList managersCfg = doc.getElementsByTagName("manager");
			for (int i = 0; i < managersCfg.getLength(); i++) {
				Element managerCfg = (Element) managersCfg.item(i);
				String classToIns = managerCfg.getAttribute("class");
				Class managerClass = Class.forName(classToIns);
				Constructor construct = managerClass.getConstructor(new Class[]{String.class,String.class,String.class,String.class,String.class,String.class});
				
				Element descEle = (Element) managerCfg.getElementsByTagName("description").item(0);
				Object[] params = {
						managerCfg.getAttribute("id"),
						managerCfg.getAttribute("name"),
						managerCfg.getAttribute("urlFileToDownload"),
						processLocalEntities(managerCfg.getAttribute("urlFileDefault")),
						managerCfg.getAttribute("version"),
						descEle.getTextContent()
					};
				LibraryVersionManager manager  = (LibraryVersionManager) construct.newInstance(params);
				
				manager.processOthersCfg(managerCfg);
				this.addLibraryVersionManager(manager);
			}
		} catch (Exception e) {
			e.printStackTrace();
//			IOException et = new IOException("No se pudo cargar el arcivo de configuracion ["+pathCfgManager+"] :\n",e);
			IOException et = new IOException();
			et.fillInStackTrace();
			throw et;
		}
		
	}



	public void checkUpdates(){
		Properties localVersionDes	= new Properties();
		Properties cfgMgVersionDes	= getVersionFromCfgManager();
		
		try{ //CheckUpdates from JAR
			File fileVersionLocal		= new File(pathVersionDataLocal);
			if(!fileVersionLocal.exists()){
				appendToLogFile("\n"+"File ["+pathVersionDataLocal+"] don't exist. Load from cfgManager.xml ");
				if(!createVersionFileFromCfgManager(fileVersionLocal,cfgMgVersionDes))
					return;
			}
			localVersionDes.load(new FileInputStream(fileVersionLocal));
			Hashtable<String, LibraryVersionManager> mangersJar = new Hashtable<String, LibraryVersionManager>();
			
			Enumeration<String> keys =  this.managers.keys();
			while (keys.hasMoreElements()) {
				String LibId = (String) keys.nextElement();
				LibraryVersionManager actMng = this.managers.get(LibId); 
				URL localResource = this.getClass().getClassLoader().getResource(actMng.getUrlFileDefault());
				LibraryVersionManager tmpMng  = new LibraryVersionManager(
							actMng.getId(), 
							actMng.getLibraryName(),
							localResource.toExternalForm(),
							localResource.toExternalForm(),
							actMng.getCurrentVersion(),
							actMng.getDescription()
						); 
				mangersJar.put(LibId, tmpMng); 
			}
			
			Properties tmpProp = (Properties) cfgMgVersionDes.clone();
			
			tmpProp.setProperty(
					"debug.pathToVersionUpdateFile",
					this.pathCfgManager
					);
			tmpProp.setProperty("question.message", 
					"Hay actualizaciones de las bibliotecas del interprete disponibles dentro del editor Quieres actualizar a estas bibliotecas?");
			tmpProp.setProperty("question.title",
					"Actualizar bibliotecas desde editor");
			
			checkUpdatesFor(tmpProp,localVersionDes,mangersJar);
		} catch (IOException e){
		}
		
		
		
		Properties remoteVersionDes = new Properties();
		try{ //Check if remote file is avaible an updates from REMOTE
			URL remoteUrl = new URL(urlVersionDataRemote);
			URLConnection con = remoteUrl.openConnection();
			con.setConnectTimeout(maxTimeToTimeOut);
			con.setReadTimeout(maxTimeToTimeOut );
			remoteVersionDes.load(con.getInputStream());
			
			remoteVersionDes.setProperty(
					"debug.pathToVersionUpdateFile",
					urlVersionDataRemote
					);
			remoteVersionDes.setProperty("question.message", 
					"Hay actulizaciones de las bibliotecas del interprete disponibles en internet Quieres descargar estas actualizaciones?");
			remoteVersionDes.setProperty("question.title",
					"Actualizar bibliotecas desde internet");
				
			checkUpdatesFor(remoteVersionDes,localVersionDes,this.managers);
		}catch (IOException e) {
			appendToLogFile("\n"+"Can't load ["+urlVersionDataRemote+"] "+e.getMessage());
		}
		

	}
	
	
	
	
	
	private void checkUpdatesFor(Properties remoteVerForUpdate, Properties localVerForUpdate, Hashtable<String, LibraryVersionManager> mngsForUpdate) {
		try {
			
			String msgUpdate	= remoteVerForUpdate.getProperty("question.message");
			String titleUpdate	= remoteVerForUpdate.getProperty("question.title");
			String nameFileProp = remoteVerForUpdate.getProperty("debug.pathToVersionUpdateFile");
			
			String[] tableHeader = new String[]{"Library name","Ver. (Update/Local)"};
			
			Vector<Object[]> tableDataV = new Vector<Object[]>();
			Vector<LibraryVersionManager> listManagers = new Vector<LibraryVersionManager>();
		
			Enumeration<LibraryVersionManager> managerIt = mngsForUpdate.elements();
		
			String toLog  = "";
			while (managerIt.hasMoreElements()) {
				LibraryVersionManager manager = managerIt.nextElement();
				String id = manager.getId();
				String verKey = id+".version";
				String verValR = remoteVerForUpdate.getProperty(verKey);
				String verValL = localVerForUpdate.getProperty(verKey);
				if(verValR == null){
					verValR = "0";
					toLog += "["+verKey+"] ";
					System.out.println("There is a error in keys, see version manager log file ");
				}
				
				boolean isUpdatable = verValL != null && verValL.compareToIgnoreCase(verValR) < 0; 				
				if(isUpdatable){
					tableDataV.add(new Object[]{manager.getLibraryName(), verValR+" > "+verValL});
					listManagers.add(manager);
				}
				
			}
		
			if(toLog.length() > 0 ){ //SAVE LOG FOR ERROS
				final StringWriter sWriter = new StringWriter();
				remoteVerForUpdate.list(new PrintStream(new OutputStream() {
					public void write(int b) throws IOException {
						sWriter.write(b);
					}
				}));
				toLog = 
						"There is keys in local file : ["+nameFileProp+"] without pair in remote file.\n"+
						"Keys not found : "+toLog+"\n"+
						"List of remote keys : \n"+sWriter.toString();		
				appendToLogFile("\n"+toLog);
			}
		
			
			if(tableDataV.size() <= 0)
				return; //There isn't library to update 
			
			int[] idxsSel = new int[0];
			UpdateLibraryQuestion updateQ = new UpdateLibraryQuestion();
			updateQ.setLibraryData(tableHeader,(Object[][])tableDataV.toArray(new Object[tableDataV.size()][]), true);
			updateQ.setVisible(true);
			boolean algoActualizado = 	
					updateQ.getSelectedCloseBtn() != JOptionPane.OK_OPTION ||
					(idxsSel = updateQ.getSelected()).length > 0 ;
			updateQ.dispose();		
			if(!algoActualizado)
				return; //No library selected to update
			
				
			algoActualizado = false;
			java.awt.Frame frm = new java.awt.Frame("Actualizando bibliotecas") {

				public void paint(java.awt.Graphics g) {
					super.paint(g);
					g.setColor(java.awt.Color.black);
					g.setFont(new java.awt.Font("Serif",java.awt.Font.PLAIN,19));
					g.drawString("Un momento por favor...",20,80);
					if (getTitle().startsWith("Bibl")) {
						g.drawString("Bibliotecas actualizadas.",20,120);
					} else {
						g.drawString("Estamos actualizando bibliotecas.",20,120);
					}
				}
			};
			frm.setBounds(200,150,420,200);
			frm.setLocationRelativeTo(null);
			frm.setVisible(true);
			frm.setCursor(new java.awt.Cursor(java.awt.Cursor.WAIT_CURSOR));
			
			for (int i = 0; i < idxsSel.length; i++) {
				LibraryVersionManager manager = listManagers.get(idxsSel[i]);
				URL remoteLib = new URL(manager.getUrlFileToDownload());
				try{ // Check for remote file online
					URLConnection con = remoteLib.openConnection();
					con.setConnectTimeout(maxTimeToTimeOut); 
					con.setReadTimeout(maxTimeToTimeOut);
					con.connect();
				} catch(IOException e){
					System.out.println(
							"Library '"+manager.getLibraryName()+"' not found\n"+
							"\t URL '"+manager.getUrlFileToDownload()+"'"
							);
					remoteLib = null;
				}
				if(remoteLib != null){
					File fDefault = new File(manager.getUrlFileDefault());
					String strFDestPath = pathCfgManagerDir+File.separator+fDefault.getName();
					File fDest = new File(strFDestPath);
					File fDestBck = new File(strFDestPath+"~");
					try{
						if(fDest.exists()){
							fDest.renameTo(fDestBck);
						}
						mjaFile.updateFromURL(remoteLib, strFDestPath, false);
						fDestBck.delete();
						String id = manager.getId();
						String verKey = id+".version";
						String verValR = remoteVerForUpdate.getProperty(verKey);
						localVerForUpdate.setProperty(verKey,verValR);
						algoActualizado = true;
					} catch (IOException uriE){
						uriE.printStackTrace();
						
					} finally{
						if(fDestBck.exists()){
							fDestBck.renameTo(fDest);
						}
					}
				} 
			}
			frm.setTitle("Bibliotecas actualizadas");
			frm.repaint();
			frm.setCursor(null);
			for (int i=0;i<7000000;i++) {
				double a=Math.sin(i);
			}
			frm.setVisible(false);
			frm.dispose();
			if(!algoActualizado)
				return; //We couldn't update any library
			
			String comments = getCommentsForNewLocalVersionProp(false);
			localVerForUpdate.store(new FileOutputStream(pathVersionDataLocal), comments);
			
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}



	private boolean createVersionFileFromCfgManager(File toStrore,Properties fromCfgManager){
		File parent = toStrore.getParentFile();
		Properties res = new Properties();
		if(!parent.exists()){ parent.mkdirs(); }
		String comments = getCommentsForNewLocalVersionProp(true);
		try{
			Enumeration<Object> keys = fromCfgManager.keys();
			while (keys.hasMoreElements()) {
				Object id = keys.nextElement();
				res.put(id, "0");
			}
			res.store(new FileOutputStream(toStrore), comments);
		}catch(IOException e){
			appendToLogFile("\n"+"Can't store ["+pathVersionDataLocal+"] with versions from cfgManager.\n"+e.getMessage());
			return false; 
		}
		return true;
	}
	
	



	private String getCommentsForNewLocalVersionProp(boolean fromLocal) {
		Date a = new Date();
		DateFormat dFormat = DateFormat.getDateInstance(DateFormat.MEDIUM); 
		String strDate = dFormat.format(a);
		
		String res = "Auto generated by "+this.getClass()+"\n";
		if(fromLocal)
			res += "From "+pathCfgManager;
		res += "Update at "+strDate; 
		return  res;
				
	}



	public File writeLibrary(LibraryVersionManager manager, File destDir) throws IOException{
		if(destDir == null || !destDir.isDirectory()){
			throw new IOException("The directory ["+destDir+"] is invalid");
		}
		
		if(manager == null){
			throw new IOException("The Library manager can't be null");
		}
		
		File origFile = getLibraryFromLocalFor(manager);
		if(origFile == null){
			throw new IOException("Can't get library file for ["+manager.getId()+"]");
		}
		File destFile = new File(destDir, origFile.getName());
		if(copyFile(origFile, destFile)){
			manager.libraryWritte(destFile);
			return destFile;
		} else {
			return null;
		}
	}
	
	
	
	private void doUpdateLibs (){
		
		
		
	}
	
	private Properties getVersionFromCfgManager(){
		LibraryVersionManager[] managers = getLibraryVersionManagers();
		Properties resP = new Properties();
		for (int i = 0; i < managers.length; i++) {
			LibraryVersionManager manager = managers[i];
			resP.put(
					manager.getId()+".version", 
					manager.getCurrentVersion());
		}
		return resP;
	}
	
	
	
	
	
	public void addLibraryVersionManager(LibraryVersionManager manager){
		managers.put(manager.getId(), manager);
	}
	
	
	
	public File getLibraryFromLocalFor(LibraryVersionManager manager) throws IOException{
		if(manager == null)
			throw new IOException("The parameter 'manager' con't be null");
		
		File fDefault = new File(manager.getUrlFileDefault());
		File file = new File(pathCfgManagerDir+File.separator+fDefault.getName());
		if(!file.exists()){
			URL uDefault = this.getClass().getResource(manager.getUrlFileDefault());
			if(uDefault == null){
				uDefault = this.getClass().getClassLoader().getResource(manager.getUrlFileDefault());
			}
			if(uDefault == null)
				throw new IOException("Resource '"+manager.getUrlFileDefault()+"' not found");

			com.mja.file.mjaFile.copyFromURL(uDefault,file.getAbsolutePath(),false);
		}
		return file;
	}
	
	public LibraryVersionManager getVersionManagerId(String libraryID){
		return managers.get(libraryID);
	}
	
	public LibraryVersionManager[] getLibraryVersionManagers(){
		Vector<LibraryVersionManager> tmpV = new Vector<LibraryVersionManager>(managers.values());
		return tmpV.toArray(new LibraryVersionManager[managers.size()]);
	}
	private class SelectionTableModel extends DefaultTableModel {
		private boolean[][] cfgCellEditable = {{true}}; 
		
		public Class getColumnClass(int c) {
			return getValueAt(0, c).getClass();
		}

		public boolean isCellEditable(int row, int column) {
			int nCols = this.getColumnCount();
			int nRows = this.getRowCount();
			if(cfgCellEditable.length != nRows || cfgCellEditable[0].length != nCols){
				rebuildArrayCfgEditable();
			}
			return cfgCellEditable[row][column];
		}
		
		public void setCellEditable(int row, int column, boolean isEditable){
			int nCols = this.getColumnCount();
			int nRows = this.getRowCount();
			if(cfgCellEditable.length < nRows || cfgCellEditable[0].length < nCols){
				rebuildArrayCfgEditable();
			}
			
			cfgCellEditable[row][column] = isEditable;
		}
		
		
		private void rebuildArrayCfgEditable() {
			int nCols = this.getColumnCount();
			int nRows = this.getRowCount();
			boolean[][] cfgCellEditableAnt = cfgCellEditable;
			
			nRows = Math.max(1,nRows);
			nCols = Math.max(1,nCols);
			
			cfgCellEditable = new boolean[nRows][nCols];
			for (int i = 0; i < cfgCellEditableAnt.length; i++) {
				Arrays.fill(cfgCellEditable[i], true);
			}
			
			int minNRows = Math.min(cfgCellEditableAnt.length, cfgCellEditable.length);
			int minNCols = Math.min(cfgCellEditableAnt[0].length, cfgCellEditable[0].length);
			
			for (int i = 0; i < minNRows ; i++) {
				for (int j = 0; j < minNCols; j++) {
					cfgCellEditable[i][j] = cfgCellEditableAnt[i][j]; 
				}
			}
		}
	}


	
	
	private void prepareEntities(){
		entities = new Hashtable<String, String>();
		
		String usrHome = System.getProperty("user.home");
		usrHome=usrHome.replace("\\","/");
		entities.put("{USR_HOME}", usrHome);
		
		String usrDir = System.getProperty("user.dir");
		usrDir=usrDir.replace("\\","/");
		entities.put("{USR_DIR}", usrDir);

//		String mngHome = processLocalEntities("./updates");
		String mngHome = processLocalEntities(pathCfgManagerDir);
		entities.put("{MNG_VER_HOME}", mngHome);
		pathCfgManagerDir = mngHome;
		
	}
	
	private String processLocalEntities(String strToProcess){
		Enumeration<String> keys = entities.keys();
		String result = strToProcess;
		while(keys.hasMoreElements()){
			String key = keys.nextElement();
			String value = entities.get(key);
			result = result.replace(key,value);
		}
		return result;
	}

	private boolean copyFile(File orig, File dest) throws IOException{
		InputStream in = new FileInputStream(orig);
		OutputStream out = new FileOutputStream(dest);
		
		byte[] buf = new byte[1024];
		int len;
		while ((len = in.read(buf)) > 0) {
		   out.write(buf, 0, len);
		}
		in.close();
		out.close(); 
		return true;
	}
	
	
	private boolean appendToLogFile(String strlog){
		FileWriter out = null;
		try {
			Date a = new Date();
			DateFormat dFormat = DateFormat.getDateInstance(DateFormat.FULL); 
			String strDate = dFormat.format(a);
			String toLog = "****** ["+strDate+"] *************************************\n"+strlog+"\n";  
			File logoDir=new File(pathCfgManagerDir);
			logoDir.mkdirs();
			String logFilePath = pathCfgManagerDir+File.separator+"Descartes_version_manager.log";
		    out = new FileWriter(logFilePath, true);
		    out.write(toLog);
		}catch (IOException e) {
		    e.printStackTrace();
		}finally{
		    if(out != null){
		        try { out.close(); } catch (IOException e) {e.printStackTrace();}
		    }
		}
		
		return true;
	}
	
}













