/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.algebra;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;

import com.mja.lang.*;
import com.mja.parser.*;
import com.mja.text.*;
import com.mja.textedit.EditorCanvas;
import com.mja.util.TFont;

public class AEditorCanvas extends EditorCanvas {
	
	public static final Color darkBlue=new Color(0x224488);

	private ANodeConfig an_cfg;
	private Parser parser;
	private String por_str="";
	private boolean por_siempre=false;
	private boolean parenthesis_always=true;
	private TFont tf;
	private DescartesA DA;

	public AEditorCanvas(DescartesA A,Frame parent,translator Tr,Parser p,TFont tf,Color c,
						ANodeConfig an_cfg,String por_str,boolean por_siempre,boolean parenthesis_always) {
		super(A,parent,Tr,null,p,tf,c,false,true,true);
		DA=A;
		this.tf=tf;
		this.parser=p;
		this.an_cfg=an_cfg;
		this.por_str=por_str;
		this.por_siempre=por_siempre;
		this.parenthesis_always=parenthesis_always;
	}
	
    public synchronized void update(Graphics g) {
    	g.setColor(getBackground());
    	g.fillRect(0,0,getSize().width,getSize().height);
    	paint(g);
    }
	public void paint(Graphics g) {
		super.paint(g);
		updateScreen();
	}
	
	/* ------------------------- mouse Events ---------------------- */

	private boolean mouseDragged;
	private int dx0,dy0;
	private Formula PF;
    private Node MN,PN,PNparent,PNsister,RN, IN;
    private Formula auxFL,auxFR;
    private ANode AMN, ANP, ANparent, ANsister, ANR;
    private int xe=0,ye;
    private boolean startedLeft,eraseSign,isNegative;
	
	public void mousePressed(MouseEvent e) {
		if (!isEnabled()) return;
//System.out.println("AEC.mousePressed "+e.getClickCount());
		mouseDragged=false;
		eraseSign=false;
		PF=null;
		if (!isEditing() && mainF!=null) {
			IN=null;
			MN=mainF.getNode();
			AMN=new ANode(MN,an_cfg);
			PF=mainF.getFormulaAt(e.getX(),e.getY());
			if (PF!=null) {
				PN=PF.getNode();
				PNparent=(Node) MN.getParent(PN);
		    	dx0=e.getX()-PF.getLocation().x;
				dy0=e.getY()-PF.getLocation().y;
				boolean addSignM=false;
				if (PNparent!=null && PNparent.isUnaryOp()) {
					PN=PNparent;
					PF=PN.F;
					PNparent=(Node) MN.getParent(PN);
					addSignM=true;
				}
				if (PNparent!=null) {
					ANparent=new ANode(PNparent,an_cfg);
					ANP=new ANode(PN,an_cfg);
					String signP="+",signM="-";
					if (addSignM) {
						dx0+=PNparent.F.getStringWidth(this,signM);
					}
					startedLeft=true;
					Formula auxF=((Node) PN.cloneNode()).toFormula(parser,false,por_str,por_siempre,parenthesis_always);
					auxFL=auxF;
					auxFR=auxF;
					boolean ok=false;
					if (AMN.isComp()) {
						IN=MN;
						if (MN==PNparent) {
							if (PN==MN.getRight()[0]) {
								startedLeft=false;
								RN=(Node) MN.getLeft();
							} else {
								RN=(Node) MN.getRight()[0];
							}
							ok=true;
						} else if (MN==MN.getParent(PNparent)) {
							if (PNparent==MN.getRight()[0]) {
								startedLeft=false;
								RN=(Node) MN.getLeft();
							} else {
								RN=(Node) MN.getRight()[0];
							}
							ANparent=new ANode(PNparent,an_cfg);
							xe=0;
							if (ANparent.isSum()||ANparent.isSub()) {						//	A + B
								isNegative=false;
								if (PN.getType()==Node.Us && PN.getSymb()==Node.MENOS) {
									isNegative=true;
									auxF=((Node) PN.getRight()[0].cloneNode()).toFormula(parser,false,por_str,por_siempre,parenthesis_always);
									auxFL=auxF;
									auxFR=auxF;
//									Formula xF=new Formula(parser);
//									xF.insertString("(");
//									xF.insertFormula(auxF,false);
//									xF.insertString(")");
//									auxF=xF;
								}
								if (ANparent.isSum()) {
									eraseSign=true;
									if (PN==PNparent.getRight()[0]) {		// PN = B
										PNsister=(Node) PNparent.getLeft();
									} else if (PN==PNparent.getLeft()) {	// PN = A
										PNsister=(Node) PNparent.getRight()[0];
									}
									auxFL=new Formula(parser,tf);
									auxFR=new Formula(parser,tf);
									if (isNegative) {
//System.out.println("S1");
										auxFL.insertString(signM);
										auxFR.insertString(signP);
									} else {
//System.out.println("S2");
										auxFL.insertString(signP);
										auxFR.insertString(signM);
										dx0+=(PNparent.F.getStringWidth(this,signP)+PNparent.F.getStringWidth(this,signM))/2;
									}
									auxFL.insertFormula(auxF,false);
									auxFR.insertFormula(auxF,false);
									if (!startedLeft) {
										Formula aux=auxFR;
										auxFR=auxFL;
										auxFL=aux;
									}
									ok=true;
								} else if (ANparent.isSub()) {				//	A - B
									if (PN==PNparent.getRight()[0]) {		// PN = B
										eraseSign=true;
										PNsister=(Node) PNparent.getLeft();
										if (startedLeft) {
											auxFL=new Formula(parser,tf);
											if (isNegative) {
//System.out.println("R1");
												auxFL.insertString(signP);
											} else {
//System.out.println("R2");
												auxFL.insertString(signM);
												dx0+=PNparent.F.getStringWidth(this,signM);
											}
											auxFL.insertFormula(auxF,false);
											auxFR=new Formula(parser,tf);
											if (isNegative) {
//System.out.println("R3");
												auxFR.insertString(signM);
											} else {
//System.out.println("R4");
												auxFR.insertString(signP);
//												dx0+=PNparent.F.getStringWidth(this,signP);
											}
											auxFR.insertFormula(auxF,false);
										} else {
											auxFR=new Formula(parser,tf);
											if (isNegative) {
												auxFR.insertString(signP);
											} else {
												auxFR.insertString(signM);
												dx0+=PNparent.F.getStringWidth(this,signM);
											}
											auxFR.insertFormula(auxF,false);
											auxFL=new Formula(parser,tf);
											if (isNegative) {
												auxFL.insertString(signM);
											} else {
												auxFL.insertString(signP);
											}
											auxFL.insertFormula(auxF,false);
										}
										ok=true;
									} else if (PN==PNparent.getLeft()) {	//	PN = A
										PNsister=(Node) PNparent.getRight()[0];
										if (startedLeft) {
											auxFL=new Formula(parser,tf);
											if (isNegative) {
												auxFL.insertString(signM);
//System.out.println("L1");
											} else {
//System.out.println("L2");
												auxFL.insertString(signP);
												dx0+=PNparent.F.getStringWidth(this,signP);
											}
											auxFL.insertFormula(auxF,false);
											auxFR=new Formula(parser,tf);
											if (isNegative) {
												auxFR.insertString(signP);
											} else {
												auxFR.insertString(signM);
											}
											auxFR.insertFormula(auxF,false);
										} else {
											auxFR=new Formula(parser,tf);
											if (isNegative) {
//System.out.println("L3");
												auxFR.insertString(signM);
											} else {
//System.out.println("L4");
												auxFR.insertString(signP);
												dx0+=PNparent.F.getStringWidth(this,signP);
											}
											auxFR.insertFormula(auxF,false);
											auxFL=new Formula(parser,tf);
											if (isNegative) {
												auxFL.insertString(signP);
											} else {
												auxFL.insertString(signM);
											}
											auxFL.insertFormula(auxF,false);
										}
										ok=true;
									} else {
										ok=false;
									}
								}
							} else if (ANparent.isMul()) {				//	A * B 	
								eraseSign=true;
								if (PN==PNparent.getLeft()) {			// PN = A
									PNsister=(Node) PNparent.getRight()[0];
									if (startedLeft) {
										auxFR=AMN.Div(AMN.Num(1),new ANode(PN,an_cfg)).N.toFormula();
									} else {
										auxFL=AMN.Div(AMN.Num(1),new ANode(PN,an_cfg)).N.toFormula();
									}
									ok=true;
								} else if (PN==PNparent.getRight()[0]) {// PN = B
									PNsister=(Node) PNparent.getLeft();
									if (startedLeft) {
										auxFR=AMN.Div(AMN.Num(1),new ANode(PN,an_cfg)).N.toFormula();
									} else {
										auxFL=AMN.Div(AMN.Num(1),new ANode(PN,an_cfg)).N.toFormula();
									}
									ok=true;
								} else {
									ok=false;
								}
							} else if (ANparent.isDiv()) {				//	A / B 	
								eraseSign=false;
								if (PN==PNparent.getLeft()) {			// PN = A
									PNsister=(Node) PNparent.getRight()[0];
									if (startedLeft) {
										auxFR=AMN.Div(AMN.Num(1),new ANode(PN,an_cfg)).N.toFormula();
									} else {
										auxFL=AMN.Div(AMN.Num(1),new ANode(PN,an_cfg)).N.toFormula();
									}
									ok=true;
								} else if (PN==PNparent.getRight()[0]) {// PN = B
									PNsister=(Node) PNparent.getLeft();
									if (startedLeft) {
										auxFL=AMN.Div(AMN.Num(1),new ANode(PN,an_cfg)).N.toFormula();
									} else {
										auxFR=AMN.Div(AMN.Num(1),new ANode(PN,an_cfg)).N.toFormula();
									}
									ok=true;
								} else {
									ok=false;
								}
								dy0+=12;
							} else {	// otra operacin
								ok=false;
							}
						} else {
							ok=false;
						}
						if (ok) {
							xe=(MN.getRight()[0].F.getLocation().x+MN.getLeft().F.getLocation().x+MN.getLeft().F.getFormulaWidth(this))/2;
						}
					} else {
						ok=false;
					}
					if (!ok) {  // casos fuera de la ecuacin 
						if (ANparent.isSum()) {
							ok=true;
						} else if (ANparent.isSub()) {
							ok=true;
						} else if (ANparent.isMul()) {
							ok=true;
						} else {
							ok=false;
						}
						if (ok) {
							IN=PNparent;
							if (PN==PNparent.getRight()[0]) {
								startedLeft=false;
								RN=(Node) PNparent.getLeft();
							} else {
								RN=(Node) PNparent.getRight()[0];
							}
							xe=(PNparent.getRight()[0].F.getLocation().x+PNparent.getLeft().F.getLocation().x+PNparent.getLeft().F.getFormulaWidth(this))/2;
						}
					}
					if (ok) {
						auxFL.setFonts(PF.getTFont());
						auxFR.setFonts(PF.getTFont());
						auxFL.setColors(null,Color.black);
						auxFR.setColors(null,Color.black);
						ye=e.getY();
					} else {
						PF=null;
					}
				} // PNparent !=null
			}	// PF !=null
		} else {
			super.mousePressed(e);
		}
	}
	
    public void mouseDragged(MouseEvent e) {
    	if (!isEnabled()) return;
		if (!isEditing() && PF!=null && IN!=null) {
			mouseDragged=true;
    		unmarkSegment();
			IN.F.setColors(Color.white,darkBlue);
    		if (eraseSign) {
				PNparent.F.setColors(Color.white,Color.white);
				PNsister.F.setColors(Color.white,darkBlue);
    		} else {
    		}
			PF.setColors(Color.white,Color.white);
			Graphics gmp=DA.MP.getGraphics();
			Point P=getLocation();
			if (e.getX()<=xe) {
				int offsetH=0;
				for (int i=DA.VEC.size()-2;i>=0;i--) {
					AEditorCanvas PAEC=((AEditorCanvas) DA.VEC.elementAt(i));
					offsetH+=PAEC.getSize().height;
					PAEC.paint(PAEC.getGraphics());
					auxFL.locate(this,e.getX()-dx0,e.getY()-dy0+offsetH);
					auxFL.doDraw((Graphics2D) PAEC.getGraphics(),this);
				}
				paint(getGraphics());
				auxFL.locate(this,e.getX()-dx0,e.getY()-dy0);
				auxFL.doDraw((Graphics2D) getGraphics(),this);
				gmp.setColor(DescartesA.eqBC);
				gmp.fillRect(0,0,DA.ScrP.getSize().width-18,DA.ScrP.getSize().height);
				auxFL.locate(this,e.getX()-dx0+P.x+1,e.getY()-dy0+P.y+DA.H-DA.dH+1-DA.jsb.getValue());
				auxFL.doDraw((Graphics2D) gmp,this);
			} else {
				int offsetH=0;
				for (int i=DA.VEC.size()-2;i>=0;i--) {
					AEditorCanvas PAEC=((AEditorCanvas) DA.VEC.elementAt(i));
					offsetH+=PAEC.getSize().height;
					PAEC.paint(PAEC.getGraphics());
					auxFR.locate(this,e.getX()-dx0,e.getY()-dy0+offsetH);
					auxFR.doDraw((Graphics2D) PAEC.getGraphics(),this);
				}
				paint(getGraphics());
				auxFR.locate(this,e.getX()-dx0,e.getY()-dy0);
				auxFR.doDraw((Graphics2D) getGraphics(),this);
				gmp.setColor(DescartesA.eqBC);
				gmp.fillRect(0,0,DA.ScrP.getSize().width-18,DA.ScrP.getSize().height);
				auxFR.locate(this,e.getX()-dx0+P.x+1,e.getY()-dy0+P.y+DA.H-DA.dH+1-DA.jsb.getValue());
				auxFR.doDraw((Graphics2D) gmp,this);
			}
		} else {
	    	super.mouseDragged(e);
		}
    }
    
    public void mouseReleased(MouseEvent e) {
    	if (!isEnabled()) return;
    	if (!isEditing() && PF!=null && IN!=null) {
    		IN.F.setColors(getBackground(),Color.black);
			paint(getGraphics());
			if (mouseDragged) {
				boolean ok=false;
				if (e.getX()>xe&&startedLeft) {
					ok=true;
				} else if (e.getX()<=xe&&!startedLeft) {
					ok=true;
				}
				if (ok) {
					ANR=new ANode(RN,an_cfg);
					callActionListeners("formula_dragged");
				}
			}
    	} else {
    		super.mouseReleased(e);
    	}
    }


	public ANode performDragOperation(boolean onlyAssoc,boolean userSelectsOption,boolean showResult,
									boolean showRule,int x,int y,String title) {
		ANode A=AMN;
		ANode R=A.R;
		ANode L=A.L;
		if (A.isComp()) {
			if (ANP.equals(L) && ANR.equals(R)) {
				return A.Equal(A.Num(0),A.Sub(R,L));
			} else if (ANP.equals(R) && ANR.equals(L)) {
				return A.Equal(A.Sub(L,R),A.Num(0));
			} else if (ANR.equals(R)) {
				if (L.isSum()) {
					if (ANP.equals(L.R)) {					// A + B = C => A = C - B
						return A.Equal(L.L,A.Sub(R,L.R));
					} else if (ANP.equals(L.L)) {			// A + B = C => B = C - A
						return A.Equal(L.R,A.Sub(R,L.L));
					} 
				} else if (L.isSub()) {						// A - B = C => A = C + B
					if (ANP.equals(L.R)) {
						return A.Equal(L.L,A.Sum(R,L.R));
					} else if (ANP.equals(L.L)) {			// A - B = C => -B = C - A
						return A.Equal(A.Minus(L.R),A.Sub(R,L.L));
					} 
				} else if (L.isMul()) {
					if (ANP.equals(L.L)) {					// A * B = C => B = C / A
						return A.Equal(L.R,A.Div(R,L.L));
					} else if (ANP.equals(L.R)) {			// A * B = C => A = C / B
						return A.Equal(L.L,A.Div(R,L.R));
					}
				} else if (L.isDiv()) {
					if (ANP.equals(L.L)) {					// A / B = C => 1 / B = C / A 
						return A.Equal(A.Div(A.Num(1),L.R),A.Div(R,L.L));
					} else if (ANP.equals(L.R)) {			// A / B = C => A = B * C
						return A.Equal(L.L,A.Mul(L.R,R));
					}
				}
			} else if (ANR.equals(L)) {
				if (R.isSum()) {
					if (ANP.equals(R.L)) {					// A = B + C => A - B = C
						if (isNegative) {
							return A.Equal(A.Sum(L,A.Minus(R.L)),R.R);
						} else {
							return A.Equal(A.Sub(L,R.L),R.R);
						}
					} else if (ANP.equals(R.R)) {			// A = B + C => A - C = B
						return A.Equal(A.Sub(L,R.R),R.L);
					}
				} else if (R.isSub()) {
					if (ANP.equals(R.L)) {					// A = B - C => A - B = - C
						return A.Equal(A.Sub(L,R.L),A.Minus(R.R));
					} else if (ANP.equals(R.R)) {			// A = B - C => A + C = B
						return A.Equal(A.Sum(L,R.R),R.L);
					} 
				} else if (R.isMul()) {
					if (ANP.equals(R.L)) {					// A = B * C => A / B = C
						return A.Equal(A.Div(L,R.L),R.R);
					} else if (ANP.equals(R.R)) {			// A = B * C => A / C = B
						return A.Equal(A.Div(L,R.R),R.L);
					}
				} else if (R.isDiv()) {
					if (ANP.equals(R.L)) {					// A = B / C => A / B = 1 / C
						return A.Equal(A.Div(L,R.L),A.Div(A.Num(1),R.R));
					} else if (ANP.equals(R.R)) {			// A = B / C => A * C = B
						return A.Equal(A.Mul(L,R.R),R.L);
					}
				}
			}
		} 
		ANode newAN=null;
		if (ANparent.isSum()) {
			if (PN==PNparent.getLeft()&&RN==PNparent.getRight()[0]) {		// A + B = B + A
				newAN=A.Sum(ANR,ANP);
			} else if (RN==PNparent.getLeft()&&PN==PNparent.getRight()[0]) {// B + A = A + B
				newAN=A.Sum(ANP,ANR);
			}
		} else if (ANparent.isSub()) {
			if (PN==PNparent.getLeft()&&RN==PNparent.getRight()[0]) {		// A - B = -B + A
				newAN=A.Sum(A.Minus(ANR),ANP);
			} else if (RN==PNparent.getLeft()&&PN==PNparent.getRight()[0]) { // B - A = -A + B
				newAN=A.Sum(A.Minus(ANP),ANR);
			}
		} else if (ANparent.isMul()) {
			if (PN==PNparent.getLeft()&&RN==PNparent.getRight()[0]) {		// A * B = B * A
				if ((!PN.isNumber())|RN.isNumber()) {
					newAN=A.Mul(ANR,ANP);
				}
			} else if (RN==PNparent.getLeft()&&PN==PNparent.getRight()[0]) {// B * A = A * B
				if ((!RN.isNumber())|PN.isNumber()) {
					newAN=A.Mul(ANP,ANR);
				}
			}
		}
		if (newAN!=null) {
			Node auxMN=(Node) MN.cloneAndSubstitute(PNparent,newAN.N);
			return new ANode(auxMN,an_cfg);
		}
		return null;
	}
	
}