/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.algebra;


import java.util.*;

import com.mja.parser.*;
import com.mja.util.*;
import com.mja.gui.*;

public class ANode {

	protected Node N;
	protected String symbol;
	protected double value;
	protected ANode L,R;
	protected ANodeConfig an_cfg;
	private boolean animate=true;
	private static Parser parser=new Parser(null);
	public static boolean notablesAntes=true;

	protected ANode() {
	}

	public ANode(Node N,ANodeConfig an_cfg) {
		this.N=(Node) N.cloneNode();
		this.N.setDecimals(an_cfg.getDecimals());
		this.an_cfg=an_cfg;
		symbol=getPrettySymbol(N.getSymbol());
		value=N.Evaluate(N.getDouble());
		if (N.getLeft()!=null) {
			L=new ANode((Node) N.getLeft(),an_cfg);
		}
		if (N.getRight()!=null && N.getRight().length==1 && N.getRight()[0]!=null) {
			R=new ANode((Node) N.getRight()[0],an_cfg);
		}
	}

	public ANode(String s,ANodeConfig an_cfg) {
		this.an_cfg=an_cfg;
		N=new Node(parser,Node.Vi,Node.VAR,s);
		symbol=s;
		value=0;
		N.setDecimals(an_cfg.getDecimals());
	}

	public ANode(double x,ANodeConfig an_cfg) {
		this(new Node(parser,x,an_cfg.getDecimals()),an_cfg);
	}

/* ---------------------------------------------------------------------- */

	public Node getNode() {
		return N;
	}

	public boolean animate() {
		return animate;
	}

	public String toExpression(boolean allP) {
		return N.toExpression(allP);
	}

	boolean equals(ANode AN) {
		return toExpression(true).equals(AN.toExpression(true));
	}

	public static String getPrettySymbol(String s) {
		if ("+/-".equals(s)) {
			return "\u00b1";
		} else if ("*".equals(s)) {
			return "\u00d7";
		} else if ("/".equals(s)) {
			return "\u00f7";
		} else if ("-".equals(s)) {
			return "\u2012";
		} else if ("-/+".equals(s)) {
			return "\u2213";
		} else if ("sqrt".equals(s)) {
			return "\u221A";
		} else if ("\\".equals(s)) {
			return "\u221A";
		} else if ("=?".equals(s)) {
			return "\u225f";
		} else if ("#".equals(s) || "!=".equals(s)) {
			return "\u2260";
		} else if ("==".equals(s)) {
			return "<=>";
//			return "\u2261";
		} else if ("!==".equals(s)) {
			return "\u2262";
		} else if ("<=".equals(s)) {
			return "\u2264";
		} else if (">=".equals(s)) {
			return "\u2265";
		}
		return s;
	}

	public static String getPrettyExpression(String s) {
		String expr=BasicStr.replace(s,"+/-","\u00b1");
		expr=BasicStr.replace(expr,"*","");
		expr=BasicStr.replace(expr,"/","\u00f7");
		expr=BasicStr.replace(expr,"-","\u2012");
		expr=BasicStr.replace(expr,"-/+","\u2213");
		expr=BasicStr.replace(expr,"sqrt","\u221A");
		expr=BasicStr.replace(expr,"\\","\u221A");
		expr=BasicStr.replace(expr,"=?","\u225f");
		expr=BasicStr.replace(expr,"#","\u2260");
		expr=BasicStr.replace(expr,"==","<=>");
//		expr=BasicStr.replace(expr,"==","\u2261");
		expr=BasicStr.replace(expr,"!==","\u2262");
		expr=BasicStr.replace(expr,"<=","\u2264");
		expr=BasicStr.replace(expr,">=","\u2265");
		return expr;
	}

/* -------------------------- Basic Operations -------------------------- */

	public ANode Copy(ANode n) {
		return new ANode((Node) n.N.cloneNode(),an_cfg);
	}

	public ANode Num(double x) {
		if (x>=0) {
			return new ANode(new Node(parser,x,an_cfg.getDecimals()),an_cfg);
		} else {
			return Minus(new ANode(new Node(parser,-x,an_cfg.getDecimals()),an_cfg));
		}
	}

	public ANode Abs(ANode n) {
		return new ANode ((Node) Node.Abs(n.N),an_cfg);
	}

	public ANode Minus(ANode n) {
		if (an_cfg.simplify_minuses && n.isUnaryMinus()) { // agregado el 8 de abril del 2006.
			return new ANode(n.R.N,an_cfg); 
		} else {
			return new ANode ((Node) Node.Minus(n.N),an_cfg);
		}
	}

	public ANode Inv(ANode n) {
		return Div(Num(1),n);
	}


	public ANode Sum(ANode m,ANode n) {
		if (an_cfg.simplify_zero_one) {  // agregado el 8 de abril del 2006.
			if (m.isZero()) {
				return new ANode(n.N,an_cfg);
			} else if (n.isZero()) {
				return new ANode(m.N,an_cfg);
			}
		}
		if (an_cfg.simplify_minuses && n.isUnaryMinus()) { // agregado el 8 de abril del 2006.
			return new ANode ((Node) Node.Subtract(m.N,n.R.N),an_cfg);
		} else {
			return new ANode ((Node) Node.Sum(m.N,n.N),an_cfg);
		}
	}

	public ANode Sub(ANode m,ANode n) {
		if (an_cfg.simplify_zero_one) {  // agregado el 8 de abril del 2006.
			if (n.isZero()) {
				return new ANode(m.N,an_cfg);
			} else if (m.isZero()) {
				return Minus(n);
			}
		}
		if (an_cfg.simplify_minuses && n.isUnaryMinus()) { // agregado el 8 de abril del 2006.
			return new ANode ((Node) Node.Sum(m.N,n.R.N),an_cfg);
		} else {
			return new ANode (Node.Subtract(m.N,n.N),an_cfg);
		}
	}

	public ANode Mul(ANode m,ANode n) {
		return Mul(m,n,an_cfg.simplify_zero_one);
	}

	public ANode Mul(ANode m,ANode n,boolean simplify_0_1) {
		if (m.isOne() && simplify_0_1) { // agregado el 20 de abril del 2006.
			return new ANode (n.N,an_cfg);
		} else if (n.isOne() && simplify_0_1) { // agregado el 20 de abril del 2006.
			return new ANode (m.N,an_cfg);
		} else if ((m.isZero()||n.isZero() ) && simplify_0_1) { // agregado el 20 de abril del 2006.
			return Num(0);
		} else {
			return new ANode ((Node) Node.Multiply(m.N,n.N),an_cfg);
		}
	}

	public ANode Div(ANode num,ANode den) {
		return new ANode ((Node) Node.Divide(num.N,den.N),an_cfg);
	}

	public ANode Exp(ANode m,ANode n) {
		return new ANode ((Node) Node.Exponentiate(m.N,n.N),an_cfg);
	}

	public ANode Root(ANode m,ANode n) {
		return new ANode ((Node) Node.Root(m.N,n.N),an_cfg);
	}

	public ANode Sqrt(ANode n) {
		return new ANode ((Node) Node.Sqrt(n.N),an_cfg);
	}

	public ANode Equal(ANode m,ANode n) {
		return new ANode ((Node) Node.Equal(m.N,n.N),an_cfg);
	}

//	public ANode Comp(ANode m,ANode n) {
//		return new ANode ((Node) Node.Comp(m.N,n.N),an_cfg);
//	}


/* ---------------------- ANode clasification ---------------------- */

	public boolean isVar() {
		return N.getType()==Node.Vi;
	}

	public boolean isUnaryOperator() {
		boolean itis= (!isPositiveNumber() && R!=null) && 
					   ( (L==null) || 
						  ( opIs("\\") && (L.symbol.equals("2") || L.symbol.equals("")) ) || 
						  ( opIs("^") && R.symbol.equals("2") )
				   );
		return itis;
	}

	public boolean isBinaryOperator() {
		return (!isPositiveNumber() && L!=null && R!=null);
	}

/* ---------------------- Other "is" Functions ---------------------- */

	boolean opIs(String os) {
		return getPrettySymbol(os).equals(symbol);
	}

	public static boolean isNumber(String s) {
		if (BasicStr.hasContent(s)) {
			for (int i=0;i<s.length();i++) {
				if (!Node.isDecimal(s.charAt(i))) {
					return false;
				}
			}
			return true;
		}
		return false;
	}

	public boolean isNumber() {	
		return isPositiveNumber()||
			   (isUnaryMinus() && R.isPositiveNumber())||
			   (isUnaryPlus() && R.isPositiveNumber()) ;
	}

	boolean isIntegerFraction() {
		return isDiv() && L.isInteger() && R.isInteger();
	}

	boolean isSignedVar() {
		return isVar() || ((isUnaryMinus()&&R.isVar()));
	}

	boolean isInvariant() {
		return equals(performOperation());
	}

	boolean isCyclic() {
		ANode AN=performOperation();
		return equals(AN.performOperation());
	}

	public boolean isSimple() {
		if (isVar()) {
			return true;
		} else if (isUnaryOperator()) {
			return  isInvariant() && R.isSimple();
		} else if (isBinaryOperator()) {
			if (isInvariant()) {
				return L.isSimple() && R.isSimple();
			} else if (isCyclic()) {
				if (L.isSimple()&&R.isSimple()) {
					if (isSum()) {
						if (!L.isPositiveNumber()) {
							return true;
						}
						if (R.symbol.equals("x")) {
							return true;
						}
					} else if (isMul() && L.isVar()) {
						if (R.symbol.equals("x")) {
							return true;
						}
						if (L.isPositiveNumber()) {
							return true;
						}
						if (R.isExp()&&R.L.symbol.equals("x")) {
							return true;
						}
					}
				}
			}
			if (isExp()) {
				if (L.isVar() && R.isSimple()) {
					return true;
				}
			}
		}
		return false;
	}

/* ------------------------ Var Clasification ----------------------- */

	public boolean isAbstractNumber() {
		return !isNumber(symbol);
	}

	public boolean isPositiveNumber() {
		return isNumber(symbol);
	}

	public boolean isInteger() {
		return (symbol!=null) && isPositiveNumber() && (symbol.indexOf(".")<0);
	}

	public boolean isEven() {
		return isPositiveNumber() && isInteger() && (((int) value) % 2==0);
	}

	public boolean isOdd() {
		return isPositiveNumber() && isInteger() && (((int) value) % 1==0);
	}
	
	public boolean isOne() {
		return isInteger() && (symbol.equals("1") && ((int) value==1));
	}

	public boolean isZero() {
		return isInteger() && (symbol.equals("0") && ((int) value==0));
	}

/* ------------------ Unary Operator Clasification ------------------ */

	boolean isUnaryMinus() {
		return isUnaryOperator() && opIs("-");
	}

	boolean isUnaryPlus() {
		return isUnaryOperator() && opIs("+");  // @@ nuevo
	}

	boolean isSqr() {
		return isUnaryOperator() && 
					( opIs("sqr") || 
					  ( opIs("^") && R.symbol.equals("2") )
					);
	}

	boolean isSqrt() {
		return isUnaryOperator() && 
				  ( opIs("sqrt") || ( opIs("\\") && (L.symbol.equals("2") || L.symbol.equals("")) ) );
	}

	boolean isAbs() {
		return isUnaryOperator() && opIs("abs");
	}

/* ------------------ Binary Operator Clasification ----------------- */

	boolean isSum() {
		return isBinaryOperator() && opIs("+");
	}

	boolean isSub() {
		return isBinaryOperator() && opIs("-");
	}

	boolean isMul() {
		return isBinaryOperator() && opIs("*");
	}

	boolean isDiv() {
		return isBinaryOperator() && opIs("/");
	}

	boolean isExp() {
		return isBinaryOperator() && opIs("^");
	}

	boolean isRoot() {
		return isBinaryOperator() && opIs("\\");
	}

	boolean isEqual() {
		return isBinaryOperator() && opIs("=");
	}

	boolean isComp() {
// "\u225f".equals(getPrettySymbol(symbol)) || Node.symbol(symbol)==Node.IGUAL
		return isBinaryOperator() && ( opIs("=")||opIs("<")||opIs(">") );
	}

/* ------------------------- Perform Operation ---------------------- */

	void addResultado(Vector V,ANode AN,String expl) {
		addResultado(V,AN,expl,"");
	}

	void addResultado(Vector V,ANode AN,String expl,String cond) {
		addResultado(V,new Resultado(AN,expl,cond));
	}

	void addResultado(Vector V,Resultado r) {
		if (r!=null) {
			ANode AN=r.getANode();
			String expr=AN.toExpression(true);
			for (int i=0;i<V.size();i++) {
				Resultado R=(Resultado) V.elementAt(i);
				ANode ANi=R.getANode();
				if (expr.equals(ANi.toExpression(true))) {
					return;
				}
			}
			V.addElement(r);
		}
	}

	Vector V=new Vector();

	Vector getResultados() {
		performOperation();
		return V;
	}

	protected ANode performOperation() {
		return performOperation(0,false,false,false,false,false,0,0,"Operaciones posibles");
	}


	public ANode performOperation(int N,boolean onlyAssoc,boolean assocOnDblClick,boolean userSelectsOption,boolean showResult,
									boolean showRule,int x,int y,String title) {
		V=new Vector();
		if (onlyAssoc) {
			Resultado r=doAssociation(onlyAssoc);
			if (r!=null) {
				return r.getANode();
			} else {
				return this;
			}
		}
		ANode AN=this;
		if (assocOnDblClick) {
			Resultado r0=doAssociation(true);
			if (!r0.getANode().equals(this)) {
				AN=r0.getANode();
				V.addElement(r0);
				return AN;
			}
		}
//System.out.println("Performing on "+toExpression(false));
		if (isVar()) {
			return AN;
		}
		if (isUnaryOperator()) {
			(new U(AN)).perform(V);
			if (R.isUnaryOperator()) {
				(new UU(AN)).perform(V);
			} else if (L!=null && L.isUnaryOperator()) {
				(new UU(AN)).perform(V);
			} 
			Resultado r=(new ANode_U(AN)).perform_U();
			if (r!=null) {
				V.addElement(r);
			}
		} 
		if (isBinaryOperator()) {
			AN.doBinaryOperation(V);
		}
		countResults=V.size();
		if (countResults>0) {
			int n=N;
			if (countResults>1) {
				if (userSelectsOption) {
					n=UserSelection.selectOption(parser,AN,V,x,y,showResult,showRule,title);
				} else {
//					n=UserSelection.selectOption(parser,AN,V,x,y,true,true,title);
				}
			}
			if (n>=0 && n<V.size()) {
				Resultado r=(Resultado) V.elementAt(n);
				return r.getANode();
			} else {
				return null;
			}
		}
		return AN;
	}
	
	public int countResults;

	void doBinaryOperation(Vector V) {

	// agregado el 8 de abril de 2009 ---
		addResultado(V,(new ANode_BB(this)).perform_BN(V));
		addResultado(V,(new ANode_BB(this)).perform_NB(V));
	// ---
		if ((L.isPositiveNumber()||R.isPositiveNumber()) ) {
			// m bop B  || A bop n  ||  m bop n
			ANode_N ANN=new ANode_N(this);
			if (L.isPositiveNumber() && R.isPositiveNumber()) {
				// m bop n
				ANN.perform_NN(V);
			} 
			if (L.isPositiveNumber()) {
				// m bop B
				ANN.perform_NA_0(V);
				if (R.isUnaryOperator()) {
					ANN.perform_NU(V);
				}
				ANN.perform_NA_1(V);
			} 
			if (R.isPositiveNumber()) {
				// A bop n
				ANN.perform_AN_0(V);
				if (L.isUnaryOperator()) {
					ANN.perform_UN(V);
				}
				ANN.perform_AN_1(V);
			}
		}

		// A bop B
		(new ANode_AA(this)).perform_AA_0(V);

		if (L.isUnaryOperator() && R.isUnaryOperator()) {
			// (uop A) bop (uop B)
			addResultado(V,(new ANode_UU(this)).perform_UU());
		}
		if (R.isUnaryOperator()) { 			// (7)	A bop (uop B)
			//  A bop (uop B)
			addResultado(V,(new ANode_AU(this)).perform_AU());
		}
		if (L.isUnaryOperator()) {
			// (uop A) bop B
			addResultado(V,(new ANode_UA(this)).perform_UA());
		}

		if (L.isBinaryOperator() && R.isBinaryOperator() ) {
			// (A bop B) bop (C bop D)
			if (notablesAntes) {	// agregado el 8 de abril de 2009
				addResultado(V,(new ANode_BB(this)).perform_BB_notables(V));
			}
			addResultado(V,(new ANode_BB(this)).perform_BB(V));
		}
		if (L.isBinaryOperator()) {
			// (A bop B) bop C
			addResultado(V,(new ANode_BA(this)).perform_BA(V));
		}
		if (R.isBinaryOperator()) {
			//  A bop (B bop C)
			addResultado(V,(new ANode_AB(this)).perform_AB(V));
		}

		// A bop B
		(new ANode_AA(this)).perform_AA_1(V);
		
		if (L.isBinaryOperator() && R.isBinaryOperator() ) {
			// (A bop B) bop (C bop D)
			addResultado(V,(new ANode_BB(this)).perform_BB_notables(V));
		}
	}

/* ----------------------- Algorithms ----------------------- */

	int maxCommonDivisor(int n,int d) {
		if (n==0) {
			return d;
		}
		int i=2;
		int p=(n>=d)?n:d;
		int q=(n>=d)?d:n;
		int r=p-(int)Math.floor(p/q)*q;
		while ((q>1)&(r>0)) {
			p=(r==0)?p:q;
			q=(r==0)?q:r;
			r=p-(int)Math.floor(p/q)*q;
		}
		return q;
	}

	int minCommonMultiple(int n1,int n2) {
		return (n1*n2)/maxCommonDivisor(n1,n2);
	}

/* ------------------------- Div ------------------------- */

	void divideIntegers(Vector V) {
		if (isIntegerFraction()) {
			ANode AN=this;
			int n=(int) Math.round(L.value);
			int d=(int) Math.round(R.value);
			if (d!=0) {
				int mcd=maxCommonDivisor(n,d);
				if (mcd==d) {
					AN=Num((double)n/(double)d);
					addResultado(V,new Resultado(AN,"(n*d)/d = n",""));
				} else if (mcd>1) {
					AN=Div(Num(n/mcd),Num(d/mcd));
					addResultado(V,new Resultado(AN,"(m*c)/(n*c)=m/n",""));
				} else {
//					addResultado(V,new Resultado(AN,"m/n=m/n","si MCD(m,n)=1"));
					addResultado(V,new Resultado(new ANode(AN.N.Evaluate(0),an_cfg),"m/n=r","si n#0"));
				}
			}
		}
	}

	public ANode userModified(Parser p) {
		String expr=toExpression(true);
		String new_expr=mjaText.getInfo("Modificar",expr+" = ","aceptar","cancelar",false,expr);
		if (BasicStr.hasContent(new_expr)) {
			try {
				Node nN=p.Analyse(new_expr);
				if (nN.Evaluate()==N.Evaluate()) {
					return new ANode(nN,an_cfg);
				}
			} catch (Exception e) {
				java.awt.Toolkit.getDefaultToolkit().beep();
			}
		}
		return this;
	}

	void error(String msg) {
		System.out.println("Error: "+msg+" in "+toExpression(true));
	}

/* ------------- Order (association and commutation) ------------- */

//  problema: Cmo conseguir el finalOpANode sin usarlo como esttico?

	static ANode finalOpANode;

	public static Node getFinalOpNode() {
		if (finalOpANode!=null) {
			return finalOpANode.N;
		}
		return null;
	}

	public Resultado doAssociation(boolean recursive) {
		ANode AN=this;
		String msg="";
		finalOpANode=this;
		if (isSum()&&L.isSum()) {
			if (recursive) {
				AN=Sum(L.L,Sum(L.R,R).doAssociation(recursive).getANode());
			} else {
				AN=Sum(L.L,Sum(L.R,R));
			}
			msg="(A + B) + C = A + (B + C)";
		} else if (isSum()&&L.isSub()) {
			if (AN.an_cfg.simplify_minuses) {
				if (recursive) {
					AN=Sub(L.L,Sub(L.R,R).doAssociation(recursive).getANode());
				} else {
					AN=Sub(L.L,Sub(L.R,R));
				}
				msg="(A - B) + C = A - (B - C)";
			} else {
				if (recursive) {
					AN=Sum(L.L,Sum(Minus(L.R),R).doAssociation(recursive).getANode());
				} else {
					AN=Sum(L.L,Sum(Minus(L.R),R));
				}
				msg="(A - B) + C = A + (-B + C)";
			}

		} else if (isSub()&&(L.isSum())) {			//
			if (recursive) {
				AN=Sum(L.L,Sub(L.R,R).doAssociation(recursive).getANode());
			} else {
				AN=Sum(L.L,Sub(L.R,R));
			}
			msg="(A + B) - C = A + (B - C)";
		} else if (isSub()&&(L.isSub())) {
			if (AN.an_cfg.simplify_minuses) {
				if (recursive) {
					AN=Sub(L.L,Sum(L.R,R).doAssociation(recursive).getANode());
				} else {
					AN=Sub(L.L,Sum(L.R,R));
				}
				msg="(A - B) - C = A - (B + C)";
			} else {
				if (recursive) {
					AN=Sum(L.L,Sum(Minus(L.R),Minus(R)).doAssociation(recursive).getANode());
				} else {
					AN=Sum(L.L,Sum(Minus(L.R),Minus(R)));
				}
				msg="(A - B) - C = A + (-B + (-C))";
			}
		} else if (isMul()&&L.isMul()) {
			if (recursive) {
				AN=Mul(L.L,Mul(L.R,R,false).doAssociation(recursive).getANode(),false);
			} else {
				AN=Sum(L.L,Sum(L.R,R));
			}
			msg="(A * B) * C = A * (B * C)";
		} else if (isSum()&&R.isSum()) {
			if (recursive) {
				AN=Sum(Sum(L,R.L).doAssociation(recursive).getANode(),R.R);
			} else {
				AN=Sum(Sum(L,R.L),R.R);
			}
			msg="A + (B + C) = (A + B) + C";
		} else if (isSum()&&R.isSub()) {
			if (recursive) {
				AN=Sub(Sum(L,R.L).doAssociation(recursive).getANode(),R.R);
			} else {
				AN=Sub(Sum(L,R.L),R.R);
			}
			msg="A + (B - C) = (A + B) - C";
		} else if (isSub()&&(R.isSum())) {
			if (recursive) {
				AN=Sub(Sub(L,R.L).doAssociation(recursive).getANode(),R.R);
			} else {
				AN=Sub(Sub(L,R.L),R.R);
			}
			msg="A - (B + C) = (A - B) - C";
		} else if (isSub()&&(R.isSub())) {
			if (recursive) {
				AN=Sum(Sub(L,R.L).doAssociation(recursive).getANode(),R.R);
			} else {
				AN=Sum(Sub(L,R.L),R.R);
			}
			msg="A - (B - C) = (A - B) + C";
		} else if (isMul()&&R.isMul()) {
			if (recursive) {
				AN=Mul(Mul(L,R.L,false).doAssociation(recursive).getANode(),R.R,false);
			} else {
				AN=Mul(Mul(L,R.L,false),R.R,false);
			}
			msg="A * (B * C) = (A * B) * C";
		}
		AN.animate=false;
		return new Resultado(AN,msg);
	}

	static boolean isPerfectSqr(int n) {
		double s=Math.sqrt((double) n);
		return Math.round(s)==s;
	}

	Vector getVector() {
		return V;
	}

	public Node findModifiedNode(ANode AN) {
		Node n;
		if (L!=null&&AN.L!=null) {
			n=L.findModifiedNode(AN.L);
			if (n!=null) {
				return n;
			}
		}
		if (R!=null&&AN.R!=null) {
			n=R.findModifiedNode(AN.R);
			if (n!=null) {
				return n;
			}
		}
		Vector V=this.getResultados();
		for (int i=0;i<V.size();i++) {
			Resultado r = (Resultado) V.elementAt(i);
			if (r!=null) {
				ANode an=r.getANode();
				if (an!=null && an.equals(AN)) {
					return N;
				}
			}
		}
		return null;
	}

	public ANode findModifiedANode(ANode AN) {
		ANode an;
		if (L!=null&&AN.L!=null) {
			an=L.findModifiedANode(AN.L);
			if (an!=null) {
				return an;
			}
		}
		if (R!=null&&AN.R!=null) {
			an=R.findModifiedANode(AN.R);
			if (an!=null) {
				return an;
			}
		}
		Vector V=this.getResultados();
		for (int i=0;i<V.size();i++) {
			Resultado r = (Resultado) V.elementAt(i);
			if (r!=null) {
				an=r.getANode();
				if (an!=null && an.equals(AN)) {
					return this;
				}
			}
		}
		return null;
	}

}

class Resultado {

	private ANode AN;
	private String expl;
	private String cond;

	Resultado (ANode AN,String expl,String cond) {
		this(AN,expl);
		this.cond=cond;
	}

	Resultado (ANode AN,String expl) {
		this.AN=AN;
		this.expl=expl;
	}

	ANode getANode() {
		return AN;
	}

	String getExpl(boolean pretty) {
		if (pretty) {
			return ANode.getPrettyExpression(expl);
		} else {
			return expl;
		}
	}

	String getCond() {
		return cond;
	}

}