/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.algebra;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import com.mja.parser.*;
import com.mja.text.*;
import com.mja.util.*;
import com.mja.file.mjaFont;

import com.mja.textedit.EditorCanvas;

public class UserSelection extends Dialog implements MouseListener, WindowListener {

	private static final int dx=320;
	private static final int dy=66;
	private static Frame F=new Frame();
	private static UserSelection W;
	private static Color darkBlue=new Color(0x0044aa);
	private static Color darkRed=new Color(0xc02200);
	private static Color extraLightGray=new Color(0xe0e0e0);
	private static Color veryLightGray=new Color(0xd0d0d0);

	private Label[] lb,clb;
	private EditorCanvas[] RulEC;
	private EditorCanvas[] ResEC;
	private Font font;
	private TFont Tf=new TFont(font);
	private int userSelection=0;
	private int x,y, N;
	private Parser parser;
	
	public static void showFormula(Parser p,ANode AN,int x,int y) {
		if (W!=null) {
			W.dispose();
		}
		W=new UserSelection(p,"Objetivo");
		Vector V=new Vector();
		V.addElement(new Resultado(AN,"",""));
		W.rebuild(AN,V,x,y,true,false,false);
		W.setModal(true);
		W.setVisible(true);
		W.dispose();
	}

	public static int selectOption(Parser p,ANode AN,Vector V,int x,int y,boolean showResult,boolean showRule,String title) {
		if (W!=null) {
			W.dispose();
		}
		W=new UserSelection(p,title);
		W.rebuild(AN,V,x,y,showResult,showRule,true);
		W.setModal(true);
		W.setVisible(true);
		int n=W.userSelection;
		return n;
	}

	public UserSelection (Parser parser,String title) {
		super(F,title);
		this.parser=parser;
		addWindowListener(this);
		font=mjaFont.makeFont("SansSerif",Font.PLAIN,17);
		Tf=new TFont(font);
	}

	private void rebuild(ANode AN,Vector V,int x,int y, boolean showResult, boolean showRule,boolean bothSides) {
		this.x=x;
		this.y=y;
		removeAll();
		ResEC=new EditorCanvas[V.size()];
		RulEC=new EditorCanvas[V.size()];
		lb=new Label[V.size()];
		clb=new Label[V.size()];
		Panel ResP=new Panel();
		Panel RulP=new Panel();
		Panel ConP=new Panel();
		ResP.setLayout(new GridLayout(ResEC.length,1));
		ConP.setLayout(new GridLayout(clb.length,1));
		RulP.setLayout(new GridLayout(RulEC.length,1));
		boolean cond_exists=false;
		boolean is_equation=false;
		for (int i=0;i<lb.length;i++) {
			Resultado r=(Resultado) V.elementAt(i);
			String expr=r.getANode().toExpression(true);
			Node N=parser.Analyse(expr);
			Node N0=N;
			if (bothSides) {
				N=new Node(parser,Node.B0,Node.IGUAL,"=");
				if (AN.N.getSymb()==Node.IGUAL) {
					N=new Node(parser,Node.B0,Node.IGUAL,"==");
					is_equation=true;
				}
				N.setLeft(AN.N);
				N.setRight(N0);
			}
			Formula f=N.toFormula(parser,false,"",false,false);

			ResEC[i]=new EditorCanvas(null,F,null,null,parser,Tf,darkBlue,false,true,true);
			ResEC[i].setSize(dx,dy);
			ResEC[i].setBackground(extraLightGray);
			ResEC[i].setEnabled(false);
			ResEC[i].addMouseListener(this);
			ResEC[i].setMainFormula(f);
			ResP.add(ResEC[i]);

			RulEC[i]=new EditorCanvas(null,F,null,null,parser,Tf,Color.black,false,true,true);
			RulEC[i].setSize(dx,dy);
			RulEC[i].setBackground(Color.lightGray);
			RulEC[i].setEnabled(false);
			RulEC[i].addMouseListener(this);
			if (BasicStr.hasContent(r.getExpl(false))) {
				try {
					Node NN=parser.Analyse("UserSelection",r.getExpl(false),true,false);
					Formula F=NN.toFormula(parser,false,"",false,false);
					RulEC[i].setMainFormula(F);
					RulP.add(RulEC[i]);
				} catch (Exception ex) {
					lb[i]=new Label("   "+r.getExpl(false));
					lb[i].setAlignment(Label.LEFT);
					lb[i].setBackground(Color.lightGray);
					lb[i].setForeground(Color.black);
					lb[i].setFont(font);
					lb[i].addMouseListener(this);
					RulP.add(lb[i]);
				}
			} else {
				RulEC[i].setMainFormula(f);
				RulP.add(RulEC[i]);
			}

			cond_exists|=BasicStr.hasContent(r.getCond());
			clb[i]=new Label(r.getCond());
			clb[i].setAlignment(Label.LEFT);
			clb[i].setBackground(veryLightGray);
			clb[i].setForeground(darkRed);
			clb[i].setFont(font);
			clb[i].addMouseListener(this);
			ConP.add(clb[i]);

		}
		N=0;
		if (showRule) {
			N++;
		}
		if (cond_exists) {
			N++;
		}
		if (showResult) {
			N++;
			if (N<2 && is_equation) {
				N=2;
			}
		}
		setLayout(new GridLayout(1,N));
		if (showResult) {
			add(ResP);
		}
		if (showRule) {
			add(RulP);
		}
		if (cond_exists) {
			add(ConP);
		}
		int dX=4+N*dx;
		int dY=60+dy*ResEC.length;
		int X=Math.min(x,Toolkit.getDefaultToolkit().getScreenSize().width-dX);
		int Y=Math.min(y,Toolkit.getDefaultToolkit().getScreenSize().height-dY-30);
		W.setBounds(X,Y,dX,dY);
	}

	public void mouseClicked(MouseEvent e) {
	}
	public void mousePressed(MouseEvent e) {
		for (int i=0;i<lb.length;i++) {
			if (RulEC[i]==e.getSource()||lb[i]==e.getSource()
				||clb[i]==e.getSource()||ResEC[i]==e.getSource()) {
				userSelection=i;
				setVisible(false);
				break;
			}
		}
	}
	public void mouseReleased(MouseEvent e) {
	}
	public void mouseEntered(MouseEvent e) {
	}
	public void mouseExited(MouseEvent e) {
	}

	public void windowOpened(WindowEvent e) {
	}
	public void windowClosed(WindowEvent e) {
	}
	public void windowClosing(WindowEvent e) {
		userSelection=-1;
		setVisible(false);
	}
	public void windowActivated(WindowEvent e) {
		FontMetrics fm=getFontMetrics(font);
		for (int i=0;i<ResEC.length;i++) {
			ResEC[i].start();
			ResEC[i].updateScreen();
			RulEC[i].start();
			RulEC[i].updateScreen();
		}
//		int DX=50,DY=25;
//		for (int i=0;i<ResEC.length;i++) {
//			ResEC[i].updateSize();
//			DX=Math.max(DX,ResEC[i].getSize().width);
//			DY=Math.max(DY,ResEC[i].getSize().height);
//			RulEC[i].updateSize();
//			DX=Math.max(DX,RulEC[i].getSize().width);
//			DY=Math.max(DY,RulEC[i].getSize().height);
//			DX=Math.max(DX,16+fm.stringWidth(clb[i].getText()));
//			DY=Math.max(DY,fm.getHeight()+8);
//		}
//		int dX=4+N*DX;
//		int dY=60+DY*ResEC.length;
//		int X=Math.min(x,Toolkit.getDefaultToolkit().getScreenSize().width-dX);
//		int Y=Math.min(y,Toolkit.getDefaultToolkit().getScreenSize().height-dY-30);
//		W.setBounds(X,Y,dX,dY);
	}
	public void windowDeactivated(WindowEvent e) {
		for (int i=0;i<ResEC.length;i++) {
			ResEC[i].stop();
			RulEC[i].stop();
		}
	}
	public void windowIconified(WindowEvent e) {
	}
	public void windowDeiconified(WindowEvent e) {
	}


}