/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.descartes;
/*------------------------------------------------------------*/
/*                          Action                            */
/*                         (c)  MJA                           */
/*                     Septiembre de 2002                     */
/*------------------------------------------------------------*/
import com.jla.nippe.*;
import com.mja.lang.*;
import com.mja.gui.*;
import com.mja.parser.*;
import java.util.*;
import java.applet.*;
import java.net.*;
import java.awt.*;
import com.mja.util.BasicStr;
import com.jla.sound.SoundPlayer;


public class Action implements Runnable {

	static final int[] actions0={ data.empty, data.message, data.calculate, data.openURL, data.openScene, data.about, data.config, data.ini, data.clear, data.anim, data.initAnim};
	static final int[] actionsS={ data.empty, data.message, data.calculate, data.openURL, data.openScene, data.about, data.config, data.ini, data.clear, data.anim, data.initAnim, data.play };
	static int[] actions=actions0;

	static final int[] modes = { data.once, data.alternate, data.always };

	private Node condN;
	private int action;
	private URL codebase;
	private String name,param,condition;
	private boolean ready=true, doit=false;
	private Descartes D;
	private int mode=data.alternate;
	private Font font;
	private int msgLocation;
	private Algorithm A;
	private String audio_file;

	public Action(Descartes D,String name,String saction,String param,Font font,int loc) {
		this(D,name,data.always,"1",saction,param,font,loc);
	}

	public Action(Descartes D,String name,int mode,String condition,String saction,String param,Font font,int loc) {
		this.D=D;
		this.name=name;
		this.condition=condition;
		condN=D.p.Analyse(condition,0);
		this.mode=mode;
		this.action=ActionIndex(D.Tr,saction);
		this.param=param;
		this.font=font;
		this.msgLocation=loc;
		prepareCalc();
		if (this.action==data.play) {
			String[] params = BasicStr.getElements(param," ");
			if (params.length>0) {
				String first=params[0];
				audio_file = first;
				if (audio_file.toLowerCase().indexOf(".mp3")>0) {
					D.addFilename(audio_file);
				}
			}
		}
	}
	
	public String getName() {
		return name;
	}
	
	public int getAction() {
		return action;
	}

	private void prepareCalc() {
		if (action==data.calculate) {
			A=new Algorithm("Action."+getName(),D.p,false,param,"","");
		}
	}

	private boolean firstTime=true;
	
	public boolean isCalculation() {
		return action==data.calculate;
	}

	public boolean checkCondition() {
		boolean ok=false;
		if (mode==data.once) {
			if (firstTime) {
				if (condN!=null && condN.Evaluate(0)>0) {
					ok=true;
					firstTime=false;
				}
			}
		} else if (mode==data.alternate) {
			if (condN!=null && condN.Evaluate(0)>0) {
				if (ready) {
					ok=true;
					ready=false;
				}
			} else {
				ready=true;
			}
		} else if (mode==data.always) {
			if (condN!=null && condN.Evaluate(0)>0) {
				ok=true;
			}
		}
		return ok;
	}

	public static int ActionIndex(translator Tr,String s) {
		for (int i=0;i<actions.length;i++) {
			if (Tr.equals(s,actions[i])) {
				return actions[i];
			}
		}
		return actions[0];
	}

	public static void SoundPlayerUpdate (Parser p) {
		p.Var("_EnPausa_").setDouble((SoundPlayer.isPaused(null))?1:0);
		p.Var("_Reproduciendo_").setDouble((SoundPlayer.isPlaying(null))?1:0);
	}
	
	static String player_id="";
	static Thread thr;


	static void stopPlaying() {
		if (BasicStr.hasContent(player_id)) {
			SoundPlayer.stop(player_id);
			player_id="";
		}
	}
	
	static void play(Descartes D,String param, String id) {
		if(!D.inWeb()){
			System.out.println("Descartes: El sonido no est disponible durante la edicin");
			return;
		}
		//OSCAR
		//liberamos recursos si se puede.
		SoundPlayer.my_gc();
		String[] params = BasicStr.getElements(param," ");
		if (params.length>0) {
			int n0=1;
			String audio_id=id;
			String first=params[0];
			String audio_file = first;
			if (audio_file.toLowerCase().indexOf(".mp3")<0) {
				try {
					Node N=D.p.Analyse(first);
					N.Evaluate();
					if (N.getStr()!=null) {
						audio_file=N.getStr();
						D.addFilename(audio_file);
					}
				} catch (Exception ex) {
				}
			}
			if (audio_file.toLowerCase().indexOf(".mp3")>=0) {
				if (audio_id==null) {
					audio_id=audio_file;
				} else {
					audio_id+="."+audio_file;
				}
			} else {
				n0=0;
				audio_id=null;
			}
			// Si no esta vivo alguno asociado, empezamos uno, 
			// a menos que se llame explcitamente con pausa o alto
			if (audio_id!=null && SoundPlayer.isFinished(audio_id)) {
				try {
					for (int i = n0 ; i < params.length;i++){
						String tmpParam = params[i].trim();
						if("pause".equalsIgnoreCase(tmpParam)||"pausa".equalsIgnoreCase(tmpParam)){
							return;
						}
						if("stop".equalsIgnoreCase(tmpParam)||"alto".equalsIgnoreCase(tmpParam)){
							return;
						}
					}
						SoundPlayer.play(audio_id, new URL(D.getDocumentBase(),audio_file));
						for (int i = n0 ; i < params.length;i++){
							String tmpParam = params[i].trim();
							int indexVol = tmpParam.indexOf("volumen");
							if("loop".equalsIgnoreCase(tmpParam)){
								SoundPlayer.setInLoop(audio_id,true);
							}
							
							if (indexVol >= 0) {
								indexVol = tmpParam.indexOf("=");
								tmpParam = tmpParam.substring(indexVol + 1);
								Node newVol = D.p.Analyse(tmpParam);
								newVol.Evaluate();
								SoundPlayer.setVolumeLevel(audio_id, (float) newVol.getDouble());
							}
						}
				} catch (Exception e) {
					e.printStackTrace();
				}
			} else {
				for (int i = n0 ; i < params.length;i++){
					String tmpParam = params[i].trim();
					if("pause".equalsIgnoreCase(tmpParam)||"pausa".equalsIgnoreCase(tmpParam)){
						SoundPlayer.pause(audio_id);
						SoundPlayerUpdate(D.p);
						return;
					}
					if("stop".equalsIgnoreCase(tmpParam)||"alto".equalsIgnoreCase(tmpParam)){
						SoundPlayer.stop(audio_id);
						SoundPlayerUpdate(D.p);
						return;
					}
	
					int indexVol = tmpParam.indexOf("volumen"); 
					if (indexVol >= 0) {
						try {
							indexVol = tmpParam.indexOf("=");
							tmpParam = tmpParam.substring(indexVol + 1);
							Node newVol = D.p.Analyse(tmpParam);
							newVol.Evaluate();
							SoundPlayer.setVolumeLevel(audio_id, (float) newVol.getDouble());
							return;
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				}
				// habia uno y no estaba terminado 
				//lo detenemos
				SoundPlayer.stop(audio_id);
			}
		}
		SoundPlayerUpdate(D.p);
	}
	
	public void run() {
		try {
			while (true) {
				Thread.sleep(50);
				SoundPlayerUpdate(D.p);
				if (D.p.Var("_Reproduciendo_").getDouble()==0) {
					D.refresh(false);
					break;
				}
			}
		} catch (InterruptedException ie) {
		}
		thr=null;
	}

	public void execute() {
		String msg="";
		if (condN!=null && condition!=null) {
			msg+=D.Tr.getTr(data.condition)+"='"+condition+"'\n";
		}
		String str_action=D.Tr.getTr(action);
		if (BasicStr.hasContent(str_action)) {
			msg+=D.Tr.getTr(data.action)+"='"+str_action+"'\n";
		}
		if (BasicStr.hasContent(param)) {
			StringTokenizer st=new StringTokenizer(param,"\n");
			if (st.hasMoreTokens()) {
				msg+=st.nextToken()+"\n";
			}
		}
		if (action==data.message) {
			String parsedParam=substituteExpressions(param);
			mjaText.msg(D.mainP,"",parsedParam,font,msgLocation);
		} else if (action==data.calculate) {
			try {
				A.initialize();
				A.perform();
				D.updateTextFieldsFromNodes();
			} catch(Exception e) {
				System.out.println("Exception performing algorithm");
			}
		} else if (action==data.play) {
			play(D,param,name);
			if (thr==null) {
				thr=new Thread(this);
				thr.start();
			}
		} else if (action==data.openURL) {
			D.scene.updateScene(false);
			String target="_blank";
			String rparam=new String(param);
			if (BasicStr.hasContent(param)) {
				String lcp=rparam.toLowerCase();
				int ix=lcp.indexOf(" target=");
				if (ix>0) {
					target=rparam.substring(ix+8);
					rparam=rparam.substring(0,ix);
				}
				rparam=substituteExpressions(rparam);
			}
			if (param.startsWith("javascript:")) {
				try {
					URL url=new URL(rparam);
//					System.out.println(url.toExternalForm());
					D.getAppletContext().showDocument(url,"_self");
				} catch (MalformedURLException mue) {
					System.out.println(mue);
				}
			} else {
				try {
					Applet A=D.getAppletFromContext(target);
					if (A!=null) {
//						System.out.println("target "+target+" is applet:"+A);
					} else {
						D.getAppletContext().showDocument(new URL(rparam),target);
					}
				} catch(Exception e) {
					try {
//						if (param.startsWith("javascript:")) {
//							D.getAppletContext().showDocument(new URL("file","",rparam),target);
//						} else {
							D.getAppletContext().showDocument(new URL(D.getDocumentBase(),rparam),target);
//						}
					} catch (Exception e2) {
						System.out.println("error opening "+param);
						e2.printStackTrace();
					}
				}
			}
		} else if (action==data.openScene) {
//			D.scene.updateScene(false);
			if (param!=null && param.startsWith("_nippe_")) {
				String int_param=substituteExpressions(param);
				comm c=new comm(D);
				c.processCommand(int_param);
			} else if (param!=null && param.equals("_cerrar_")) {
				D.closeScene();
			} else {
				if (param.indexOf(" target=")>0) {
					int ix=param.indexOf(" target=");
					String target=param.substring(ix+" target=".length());
					String filename=param.substring(0,ix);
					if ("_self".equals(target)) {
						D.showScene(param);
					} else {
						try {
							D.runScene(filename,target);
						} catch (Exception e) {
							System.out.println("error opening "+param);
							e.printStackTrace();
						}
					}
				} else {
					try {
						D.showScene(param);
					} catch(Exception e) {
						System.out.println("error opening "+param);
						e.printStackTrace();
					}
				}
			}
		} else if (action==data.about) {
			D.about(false);
		} else if (action==data.config) {
			D.edit(null,false);
		} else if (action==data.ini) {
			D.restart(true,false);
		} else if (action==data.clear) {
			D.clear(false);
		} else if (action==data.anim) {
			D.toggleAnimation(false);
		} else if (action==data.initAnim) {
			D.initAnimation(false);
		}
	}
	
	private String substituteExpressions(String rparam) {
		StringTokenizer st=new StringTokenizer(rparam,"[]");
		String s=st.nextToken();
		while (st.hasMoreTokens()) {
			String varname=st.nextToken();
			if (D.p.isArray(varname)) {
				Node N=D.p.Array(varname);
				double[] ra=D.p.getRealArray(varname);
				String[] sa=D.p.getStringArray(varname);
				s+="{";
				for (int i=0;i<ra.length;i++) {
					if (i>0) {
						s+=",";
					}
					if (sa[i]!=null) {
						s+=sa[i];
					} else {
						s+=BasicStr.DoubleToString(ra[i],6,false,".");
					}
				}
				s+="}";
			} else if (D.p.isMatrix(varname)) {
				Node N=D.p.Matrix(varname);
				double[][] rm=D.p.getRealMatrix(varname);
				String[][] sm=D.p.getStringMatrix(varname);
				s+="{";
				for (int j=0;j<rm.length;j++) {
					if (j>0) {
						s+=",";
					}
					s+="{";
					for (int i=0;i<rm[j].length;i++) {
						if (i>0) {
							s+=",";
						}
						if (sm[i][j]!=null) {
							s+=sm[i][j];
						} else {
							s+=BasicStr.DoubleToString(rm[i][j],6,false,".");
						}
					}
					s+="}";
				}
				s+="}";
			} else if (D.p.isVar(varname)) {
				if (D.p.svalue(varname)!=null) {
					s+=D.p.svalue(varname);
				} else {
					s+=BasicStr.DoubleToString(D.p.rvalue(varname),6,false,".");
				}
			} else {
				try {
					Node tmpNode = D.p.Analyse(varname);
					tmpNode.Evaluate();
					s+=(tmpNode.getStr()!=null)?tmpNode.getStr():BasicStr.DoubleToString(tmpNode.getDouble(),6,false);
				} catch (Exception e) {
					s+="["+varname+"]";
				}
			}
			if (st.hasMoreTokens()) {
				s+=st.nextToken();
			}
		}
		return s;
	}

}