/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.descartes;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import javax.swing.JTextField;
import javax.swing.JScrollBar;
import com.mja.lang.*;
import com.mja.util.*;
import com.mja.math.*;
import com.mja.text.*;
import com.mja.gui.*;
import com.mja.img.imageCanvas;
import com.mja.cmp.*;
import com.mja.descgui.*;
import com.mja.descgui.edit.*;
import com.mja.parser.*;
import com.mja.file.*;
import java.awt.image.*;
import javax.swing.*;

public class Scene extends Panel implements ActionListener, MouseListener, KeyListener, Runnable, AdjustmentListener {
											

	private imageCanvas IC;

	mjaGraph[] graph=new mjaGraph[0];
	Space[] spa=new Space[0];

	private Descartes D;
	private Thread thr;
	private Panel animCtrl;
	private mjaButton[] b;
	private Algorithm animAlg;
	private volatile boolean pleaseStop, running, pause, onestep, stay;
	private Vector numCtrl,SceneNumCtrl;
	private TitleCanvas titcan;

	private Panel view;
	
	TextCanvas EC;
//	private JScrollBar jsb;
	private Scrollbar jsb;
	private Panel jp;
	private static Frame pf=new Frame();
	private static TFont TF;
	private doed dd;
	private Panel mainPanel;
	
	public static Color sceneBackColor=new Color(0xffffff);

	public Scene(Descartes D,doed dd) {
		this.D=D;
		this.dd=dd;
		setBackground(sceneBackColor);
		setLayout(new BorderLayout());

		b=new mjaButton[6];
		animCtrl=new Panel();
		animCtrl.setLayout(new GridLayout(1,b.length));
		animCtrl.setVisible(false);
		for (int i=0;i<b.length;i++) {
			animCtrl.add(b[i]=new mjaButton(i));
			b[i].addActionListener(this);
		}

		String pleca=D.getParameter("pleca");
		if (BasicStr.hasContent(pleca)) {
			Frame PF=pf;
			if (D.parentFrame!=null) {
				PF=D.parentFrame;
			}
			titcan=new TitleCanvas(D.Tr,PF,D,D.imagesHT,this,pleca);
			if (!(D instanceof Item)) {
				add("North",titcan);
			}
		}

		if (D instanceof Arquimedes) {
			constructArquimedesScene();
		} else {
			mainPanel=new Panel();
			mainPanel.setLayout(null);
			mainPanel.setBackground(sceneBackColor);
			add("Center",mainPanel);
			mainPanel.add(animCtrl);
			addMouseListener(this);
			addKeyListener(this);
		}

	}
	
	public Space getSpaceByPrefix(String str) {
		int ix=str.indexOf(".");
		if (ix>0) {
			String name=str.substring(0,ix);
			for (int i=0;i<spa.length;i++) {
				if (name.equals(spa[i].getSpaceName())) {
					return spa[i];
				}
			}
		}
		return null;
	}
	
	public int getTitleHeight() {
		if (titcan!=null) {
			return titcan.height;
		} else {
			return 0;
		}
	}
	
	private void constructArquimedesScene() {
		String license=licenseA;
		if (!D.cfg.useCC) {
			license="";
		}
		Frame PF=pf;
		if (D.parentFrame!=null) {
			PF=D.parentFrame;
		}
		if (D instanceof Arquimedes && D.inWeb()) {
			PF=null;
		}
		TF=new TFont(mjaFont.makeFont("SansSerif",Font.PLAIN,17));
		if (dd!=null && dd.getEditorCanvas()!=null) {
			EC=dd.getEditorCanvas();
		} else {
			EC=new TextCanvas(D,PF,D.Tr,D.p,TF,Color.black,true,false);
		}
		EC.setExprDefaults(2,true,D.cfg.language);
		EC.setBackground(sceneBackColor);
		EC.setForeground(Color.darkGray);
		
		Text T=new Text(D.p,TF);
		T.setTFont(new TFont(new Font("SansSerif",Font.PLAIN,17)));
		String strtxt=D.getParameter("rtf");
		if (BasicStr.hasContent(strtxt)) {
			if (D.inWeb() && !(D instanceof Item) && BasicStr.hasContent(license)) {
				strtxt=strtxt.substring(0,strtxt.lastIndexOf("}"))+license.substring(license.indexOf("{")+1);
				Image ccimg=D.getImage(getClass().getResource("/resources/logos/cc_by_nc_sa.png"));
				try {
					MediaTracker mt=new MediaTracker(this);
					mt.addImage(ccimg,0);
					mt.waitForID(0);
				} catch (Exception e) {
					e.printStackTrace();
				}
				IC=new imageCanvas(ccimg,EC);
				IC.setHyperlink(D,"http://creativecommons.org/licenses/by-nc-sa/2.5/es/");
			}
			RTF.parseRTF(strtxt,T);
		} else {
			T.insertString("Escribir aqu");
		}
		T.setTopAndBottomMargin(20);
		EC.setText(T);
		jp=new Panel();
		jp.setBackground(sceneBackColor);
		jp.setLayout(null);
		add("Center",jp);
		jp.add(EC);
//		jsb=new JScrollBar(JScrollBar.VERTICAL);
		jsb=new Scrollbar(Scrollbar.VERTICAL);
//		jsb.setBackground(new Color(0xdde6ee));
		jsb.setBackground(Descartes.basicBlue);
		if (!(D instanceof Item)) {
			add("East",jsb);
			jsb.addAdjustmentListener(this);
			EC.jsb=jsb;
		}
		EC.addMouseListener(this);
		EC.addKeyListener(this);
		EC.addActionListener(this);
	}
	
	public Object getEC() {
		return EC;
	}
	
//	public void setText(String strtxt) {
//		Text T=new Text(D.p,TF);
//		RTF.parseRTF(strtxt,T);
//		EC.setText(T);
//	}
	
	public String getText() {
		if (EC!=null) {
			Text text=EC.getText();
			RTF rtf=new RTF(text);
			return rtf.toRTF(false);
		} else {
			return "";
		}
	}
	
	public int getTextHeight() {
		if (EC!=null) {
			return EC.getTextHeight();
		} else {
			return 0;
		}
	}
	
	public String getPleca() {
		if (titcan!=null) {
			return titcan.getConfig();
		} else {
			return "";
		}
	}
	
	private static final String licenseA=
		"{\\f0\\fs34 __________________________________________________________________________________ \\par \\fs22 "+
		"                                       Los contenidos de esta unidad didctica interactiva estn bajo una  {\\*\\hyperlink licencia Creative Commons|http://creativecommons.org/licenses/by-nc-sa/2.5/es/}, si no se indica lo contrario.\\par "+
		"                                       La unidad didctica fue creada con Descartes, que es un producto de cdigo abierto del  {\\*\\hyperlink Ministerio de Educacin de Espaa|http://www.ite.educacion.es/} y\\par "+
		"                                       el {\\*\\hyperlink Instituto de Matemticas|http://arquimedes.matem.unam.mx/}  de la Universidad Nacional Autnoma de Mxico, cedido bajo licencia {\\*\\hyperlink EUPL v 1.1|http://joinup.ec.europa.eu/software/page/eupl/licence-eupl}, con {\\*\\hyperlink cdigo en Java|http://arquimedes.matem.unam.mx/Descartes5/source/}.";

	private int EC_width,EC_height;
	
	public void setBounds(int x, int y, int w, int h) {
		super.setBounds(x,y,w,h);
		if (EC!=null) {
			EC_width=w-20;
			EC_height=h;
			if (EC.getSize().width!=EC_width || EC.getSize().height!=EC_height) {
				EC.setBounds(0,0,EC_width,EC_height);
				EC.setEnabled(true);
				EC.start();
				if (D.inWeb()) {
					if (IC!=null) {
						IC.setBounds(27,EC.getCanvasHeight()-58,88,31);
					}
				}
			}
		}
	}
	
	public void toTeX() {
		for (int i=0; i<spa.length;i++) {
			java.io.File TeXdir=new java.io.File("TeX");
			if (!TeXdir.exists()) {
				TeXdir.mkdirs();
			}
			if (spa[i].isVisible()) {
				String str=spa[i].toTeX(false);
				if (str!="") {
					mjaFile.save("TeX/"+spa[i].getSpaceName()+"_color.tex",str);
				}
				str=spa[i].toTeX(true);
				if (str!="") {
					mjaFile.save("TeX/"+spa[i].getSpaceName()+"_grises.tex",str);
				}
			}
		}
	}
	
	public void saveScene(String filename,String format) {
		Cursor wc=new Cursor(Cursor.WAIT_CURSOR);
		D.setCursor(wc);
		setCursor(wc);
		boolean isEditing=false;
		BufferedImage SceneIma=null;
		Image ccimg=null;
		if (D instanceof Arquimedes) {
			isEditing=EC.isEditing();
			EC.compactCanvas(true);
			EC.updateScreen();
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {
			}
			SceneIma=EC.getBufferedImage();
			ccimg=Toolkit.getDefaultToolkit().getImage(getClass().getResource("/resources/images/pieDescartes.png"));
			try {
				MediaTracker mt=new MediaTracker(this);
				mt.addImage(ccimg,0);
				mt.waitForID(0);
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else {
			SceneIma=new BufferedImage(D.getSize().width,D.getSize().height,BufferedImage.TYPE_4BYTE_ABGR);
		}
		Graphics gr=SceneIma.getGraphics();
		int W0=SceneIma.getWidth(this);
		W0=(D instanceof Arquimedes)?W0+20:W0;
		int H0=SceneIma.getHeight(this);
		double percentage=1.0;
		int W=(int) Math.round(W0*percentage);
		int H=(int) Math.round(H0*percentage);
		int X0=(int) Math.round(W0*(1-percentage)/2);
		int Y0=0;
		if (D instanceof Arquimedes) {
			Y0=titcan.getSize().height;
		}

		if (D instanceof Arquimedes) {
			for (int i=0;i<SceneNumCtrl.size();i++) {
				NumericControl nc=(NumericControl) SceneNumCtrl.elementAt(i);
				Image ncima=createImage(nc.getSize().width,nc.getSize().height);
				Graphics ncg=ncima.getGraphics();
				nc.paintAll(ncg); 
				gr.drawImage(ncima,nc.getLocation().x,nc.getLocation().y,this);
			}
			for (int i=0; i<spa.length;i++) {
				if (spa[i].isVisible()) {
					int spx=spa[i].getLocation().x,spy=spa[i].getLocation().y;
					gr.drawImage(spa[i].main,spx,spy,this);
					for (int j=0;j<numCtrl.size();j++) {
						NumericControl nc=(NumericControl) numCtrl.elementAt(j);
						if (nc.isVisible() && getSpace(nc.getSpace())==spa[i]) {
							Image ncima=createImage(nc.getSize().width,nc.getSize().height);
							Graphics ncg=ncima.getGraphics();
							nc.paint(ncg); 
							gr.drawImage(ncima,spx+nc.getLocation().x,spy+nc.getLocation().y,this);
						} 
					}
					for (int j=0;j<spa[i].TAC.length;j++) {
						Image tacima=createImage(spa[i].TAC[j].getSize().width,spa[i].TAC[j].getSize().height);
						Graphics tacg=tacima.getGraphics();
						spa[i].TAC[j].paint(tacg); 
						gr.drawImage(tacima,spx+spa[i].TAC[j].getLocation().x,spy+spa[i].TAC[j].getLocation().y,this);
					}
				}
			}
		} else {
			D.paintAll(gr);
		}
		boolean docc=D.cfg.useCC;
		int dh=(docc)?60:0;
		BufferedImage bima;
		if (format.equals("png")) {
			bima=new BufferedImage(W,Y0+H+dh,BufferedImage.TYPE_4BYTE_ABGR);
		} else {
			bima=new BufferedImage(W,Y0+H+dh,BufferedImage.TYPE_3BYTE_BGR);
		}
		Graphics gbima=bima.getGraphics();
		Graphics2D gbima2=(Graphics2D) gbima;
		Stroke bs=new BasicStroke(1,BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND);
		gbima2.setStroke(bs);
		gbima2.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
	    gbima2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
	    gbima2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
	    gbima2.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
	    gbima2.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
	    gbima2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
	    gbima2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE);
		gbima2.setColor(Color.white);
		gbima2.fillRect(0,0,W0+20,Y0+H+dh);
		if (D instanceof Arquimedes) {
			titcan.paint(gbima2);
		}
		gbima2.drawImage(SceneIma,X0,Y0,W,H,this);
		if (docc) {
			gbima2.drawImage(ccimg,X0+(W-775)/2,Y0+H,this);
		}
		mjaFile.save(filename,bima,format);
		bima.flush();
		bima=null;
		D.setCursor(null);
		setCursor(null);
	}

	public boolean hayAnim() {
		return thr!=null;
	}
	
	protected void prepareSpace() {
		if (D instanceof Arquimedes) {
			EC.removeAll();
			EC.add(animCtrl);
			if (IC!=null) {
				EC.add(IC);
			}
		} else {
			mainPanel.removeAll();
			mainPanel.add(animCtrl);
		}
	}
	
	public String getNewSpaceName() {
		int n=0;
		String newname="E"+Integer.toString(n++);
		String proposedname=newname;
		do {
			proposedname=newname;
			for (int i=0;i<D.cfg.sc.length;i++) {
				if (newname.equals(D.cfg.sc[i].s_value[spaceConfig.id])) {
					newname="E"+Integer.toString(n++);
				}
			}
		} while (!newname.equals(proposedname));
		return newname;
	}
	
	public String getNewControlName() {
		int n=0;
		String newname="n"+Integer.toString(n++);
		String proposedname=newname;
		do {
			proposedname=newname;
			for (int i=0;i<D.cfg.cc.length;i++) {
				if (newname.equals(D.cfg.cc[i].s_value[controlConfig.ixid])) {
					newname="n"+Integer.toString(n++);
				}
			}
		} while (!newname.equals(proposedname));
		return newname;
	}
	
	public void addEditObjectToText(editObject eo,String cid) {
		if (eo instanceof spaceConfig) {
			D.cfg.addSpace((spaceConfig) eo);
			D.cfg.sc[D.cfg.sc.length-1].s_value[spaceConfig.id]=getNewSpaceName();
			D.cfg.sc[D.cfg.sc.length-1].s_value[spaceConfig.spaceID]=cid;
			D.renewEditorConfig();
		} else if (eo instanceof controlConfig) {
			D.cfg.addControl((controlConfig) eo);
			D.cfg.cc[D.cfg.cc.length-1].s_value[controlConfig.ixcontrolID]=cid;
			D.renewEditorConfig();
		}
	}

	void addSpaces() {
		prepareSpace();
		spa=new Space[D.cfg.sc.length];
		for (int i=spa.length-1;i>=0;i--) {
			D.cfg.sc[i].updateInfo(D.Tr);
			spa[i]=D.cfg.sc[i].makeSpace(D);
			spa[i].scene=this;
			if (D instanceof Arquimedes) {
				spa[i].TextBound=EC.updateTextComponent(spa[i],spa[i].getSpaceID());
			}
		}
		if (D instanceof Arquimedes) {
			for (int i=spa.length-1;i>=0;i--) {
				if (spa[i].TextBound) {
					spa[i].setLocation(-4000,-4000);
				}
				EC.add(spa[i]);
			}
		} else {
			for (int i=spa.length-1;i>=0;i--) {
				mainPanel.add(spa[i]);
			}
		}
	}

	private mjaGraph[] sceneGraphs;

	void addGraphs() {
		// create graphic objects
		Vector SceneGraphsVector=new Vector();
		graph=new mjaGraph[D.cfg.allgR2c.length];
		for (int i=0;i<graph.length;i++) {
			if (D.cfg.allgR2c[i].type<0) {
				graph[i]=null;
			} else {
				graph[i]=new mjaGraph(D,i);
				boolean inScene=true;
				for (int j=0;j<spa.length;j++) {
					if (graph[i].cs==spa[j]) {
						inScene=false;
						break;
					}
				}
				if (inScene) {
					SceneGraphsVector.add(graph[i]);
				}
			}
		}
		sceneGraphs=new mjaGraph[SceneGraphsVector.size()];
		for (int i=0;i<sceneGraphs.length;i++) {
			sceneGraphs[i]=(mjaGraph) SceneGraphsVector.elementAt(i);
		}
	}


	void defineNumericControls(NumericControl[] nc) {
int n=0;
//System.out.println("defineNumericControls "+n); n++;
		numCtrl=new Vector();
		SceneNumCtrl=new Vector();
		if (nc!=null) {
			for (int j=0;j<nc.length;j++) {
//System.out.println("defineNumericControls "+n); n++;
				if (nc[j]!=null && (nc[j].isInterior() || nc[j].inScenario() )) {
					if (nc[j].inScenario()) {
						SceneNumCtrl.addElement(nc[j]);
					} else {
						Space sp=getSpace(nc[j].getSpace());
						if (sp!=null) {
							numCtrl.addElement(nc[j]);
						} 
					}
					if (D instanceof Arquimedes) {
						if (nc[j]!=null && (nc[j].isInterior() || nc[j].inScenario() )) {
							if (nc[j].inScenario()) {
								if (EC.updateTextComponent(nc[j],nc[j].getControlID())) {
									EC.add(nc[j]);
								}
							} else {
								Space sp=getSpace(nc[j].getSpace());
								if (sp!=null) {
									sp.add(nc[j]);
								} 
							}
							try {
								nc[j].locate();
							} catch (Exception e) {
								System.out.println(e.getMessage()+" "+j);
							}
						}
					} else if (nc[j]!=null && nc[j].isInterior()) {
						Space sp=getSpace(spaceConfig.dosD,nc[j].getSpace());
						if (sp==null) {
							sp=getSpace(spaceConfig.tresD,nc[j].getSpace());
						}
						if (sp!=null) {
							sp.add(nc[j]);
							try {
								nc[j].locate();
							} catch (Exception e) {
								System.out.println(e.getMessage()+" "+j);
							}
						}
					}
					nc[j].updateFromNode();
					nc[j].updateFont();
				}
			}
		}
	}

	void initAnimation() {
		animAlg.initialize();
	}

	void toggleAnimation() {
		if (running) {
			pause=!pause;
		} else {
			startAnimation();
		}
		updateAnimationButtons();
	}

/* --------------------- Package Methods -------------------- */

	boolean hasEquationsPanel() {
		return countEqPanelElements()>0;
	}
	
	private boolean R2SpaceExists(String s) {
		for (int i=0;i<spa.length;i++) {
			if (spa[i]!=null && spa[i].spcfg.type==spaceConfig.dosD && ("".equals(s) || spa[i].spcfg.getName().equals(s)) ) {
				return true;
			} 
		}
		return false;
	}
	
	private int countEqPanelElements() {
		int N=0;
		for (int i=0;i<D.cfg.auxiliar.length;i++) {
			if (D.cfg.auxiliar[i].type==data.var && D.cfg.auxiliar[i].isEditable) {
				N++;
			}
		}
		for (int i=0;i<D.cfg.allgR2c.length;i++) {
			if (D.cfg.allgR2c[i].type>=0 && D.cfg.allgR2c[i].isVisible() && R2SpaceExists(D.cfg.allgR2c[i].getSpace()) /*&& graph[i].cs!=null*/) {
				N++;
			}
		}
		return N;
	}

	Panel getVisibleEquationsPanel() {
		Panel ecp=null;
		int N=countEqPanelElements();
		if ( N>0 ) {
			ecp=new Panel();
			ecp.setLayout(new GridLayout(1,N,0,0));
			int j=0;
			for (int i=0;i<graph.length;i++) {
				if (graph[i].cs!=null && graph[i].gc.isVisible()) {
					graph[i].TF=new JTextField(graph[i].gc.formula);
					graph[i].TF.setForeground(graph[i].gc.gcolor.getAdaptedColor());
					graph[i].TF.setFont(mjaFont.makeFont("Monospaced",Font.PLAIN,BasicStr.fontSizeForHeight(D.cfg.rows_height)));
					if (graph[i].gc.isEditable) {
						graph[i].TF.setEditable(true);
						graph[i].TF.addActionListener(this);
					} else {
						graph[i].TF.setEditable(false);
					}
					graph[i].TF.setBackground(graph[i].cs.backColor.getAdaptedColor());
					ecp.add(graph[i].TF);
					j++;
				}
			}
			for (int i=0;i<D.cfg.auxiliar.length;i++) {
				if (D.cfg.auxiliar[i].type==data.var && D.cfg.auxiliar[i].isEditable) {
					D.cfg.auxiliar[i].exprEd=new mjaLabelAndText(this,0.1,0.9);
					D.cfg.auxiliar[i].exprEd.addActionListener(this);
					D.cfg.auxiliar[i].exprEd.setInfo(
						D.cfg.auxiliar[i].getName()+" =",
						D.cfg.auxiliar[i].s_value[auxConfig.ixexpr]);
					D.cfg.auxiliar[i].exprEd.updateFonts(D.cfg.rows_height);
					ecp.add(D.cfg.auxiliar[i].exprEd);
				}
			}
		}
		return ecp;
	}

	void preSizeSpaces() {
		for (int i=spa.length-1;i>=0;i--) {
			spa[i].preSize();
		}
	}
	
	void start(boolean run) {
		setCursor(new Cursor(Cursor.WAIT_CURSOR));
		for (int i=0;i<graph.length;i++) {
			if (graph[i]!=null) {
				graph[i].initialize();
			}
		}
		running=false;
		animCtrl.setLocation((getSize().width-120)/2,getSize().height-22);
		animCtrl.setSize(120,20);
		if (D.cfg.animC.active) {
			animAlg=new Algorithm("Scene.Animation",D.p,false,D.cfg.animC.s_value[animConfig.ixdo],
				D.cfg.animC.s_value[animConfig.ixini],D.cfg.animC.s_value[animConfig.ixwhile]);
		}
		doAlgorithms(false);
		doEventsReturnRestarted();
		for (int i=spa.length-1;i>=0;i--) {
			spa[i].start();
		}
		updateScene(true);
		startAnimation(run);
		setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	}
	
	private volatile boolean updatingScene;

	private boolean updatingScene() {
		return updatingScene;
	}
	
	void showSolutions() {
		for (int i=0;i<spa.length;i++) {
			spa[i].showSolutions();
		}
	}
	
	void showUserTexts() {
		for (int i=0;i<spa.length;i++) {
			spa[i].showUserTexts();
		}
	}
	
	private boolean firstTime=true;
	
	void updateScene(boolean clear) {
//System.out.println("Scene.updateScene IN");
		if (updatingScene()) {
//			new Exception().printStackTrace();
		} else {
			updatingScene=true;
			if (clear && EC!=null) {
				EC.updateScreen();
			}
			D.updateTextFieldsFromNodes(!isRunning()); 
			for (int i=0;i<spa.length;i++) {
				spa[i].locate();
				spa[i].updateSpace(clear,true);
				if (!spa[i].isMoving()||spa[i].isDragging()) {
					spa[i].paintSpace();
				}
			}
			for (int i=0;i<numCtrl.size();i++) {
				NumericControl nc = (NumericControl) numCtrl.elementAt(i);
				try {
					nc.locate();
				} catch (Exception e) {
					System.out.println(e.getMessage()+" "+i);
				}
			}  
			updatingScene=false;
			firstTime=false;
		}
//System.out.println("Scene.updateScene OUT");
	}

	void paintSceneSpaces() {
		for (int i=0;i<spa.length;i++) {
			if (!spa[i].isMoving()||spa[i].isDragging()) {
				spa[i].paint(spa[i].getGraphics());
			}
		}
	}
	
	void paintSceneControls() {
		for (int i=0;i<SceneNumCtrl.size();i++) {
			NumericControl nc = (NumericControl) SceneNumCtrl.elementAt(i);
			nc.paint(nc.getGraphics());
		}  
	}
	
	void paintControls() {
		for (int i=0;i<numCtrl.size();i++) {
			NumericControl nc = (NumericControl) numCtrl.elementAt(i);
			nc.paint(nc.getGraphics());
		}  
	}
	
	void refreshScene(boolean updateFromControls) {
		if (updateFromControls) {
			D.updateNodesFromControls();
		}
		doAlgorithms(false);
		if (!doEventsReturnRestarted()) {
			updateScene(firstTime);
		}
	}
	
	void updateNodesFromTexts() {
		for (int i=0;i<numCtrl.size();i++) {
			Object O=numCtrl.elementAt(i);
			if (O!=null) {
				NumericControl nc=(NumericControl) O;
				nc.updateFromText();
			}
		}
		for (int i=0;i<SceneNumCtrl.size();i++) {
			Object O=SceneNumCtrl.elementAt(i);
			if (O!=null) {
				NumericControl nc=(NumericControl) O;
				nc.updateFromText();
			}
		}
	}

	void updateTextsFromNodes(boolean force) {
		for (int i=0;i<numCtrl.size();i++) {
			if (numCtrl.elementAt(i)!=null) {
				NumericControl nc=(NumericControl) numCtrl.elementAt(i);
				if ( /*force ||*/ !nc.isBeingEdited() ) {
					nc.updateFromNode();
				}
			}
		}
		for (int i=0;i<SceneNumCtrl.size();i++) {
			if (SceneNumCtrl.elementAt(i)!=null) {
				NumericControl nc=(NumericControl) SceneNumCtrl.elementAt(i);
				if ( force || !nc.isBeingEdited() ) {
					nc.updateFromNode();
				}
			}
		}
	}

	void showTextSolutions() {
		for (int i=0;i<spa.length;i++) {
			spa[i].showSolutions();
		}
	}

	void doAlgorithms(boolean firstTime) {
//new Exception("doAlgoritms firstTime="+firstTime).printStackTrace();
		try {
			D.p.doAlgorithms(firstTime);
			D.sendChanges();	// new 2011-09-09
		} catch (Exception e) {
			limitExceeded(e);
		}
	}

	private boolean doEventsReturnRestarted() {
//new Exception().printStackTrace();
		boolean restarted=false;
		Enumeration ae=D.eventActions.elements();
		while (ae.hasMoreElements()) {
			Action a=(Action) ae.nextElement();
			if (a.checkCondition()) {
				D.informLMS("evento",a.getName());
				a.execute();
				restarted|=(a.getAction()==data.ini);
			}
		}
		return restarted;
	}

	static void limitExceeded(Exception e) {
		System.out.println(" Se ha excedido el lmite de 100000 repeticiones en un algoritmo!");
		e.printStackTrace();
//		mjaText.ask("Descartes"," Se ha excedido el lmite de\n 100000 repeticiones en un algoritmo!","ok",null);
	}

	boolean animationIsRunning() {
		return running && !pause;
	}

	void startAnimation(boolean run) {
		if (run&&D.cfg.animC.active && D.cfg.animC.autorun) {
			startAnimation();
		}
	}

	private void startAnimation() {
		if (D.cfg.animC.controls) {
			b[mjaButton.play].type=mjaButton.pause;
			animCtrl.setVisible(true);
		}
		if (thr!=null) {
			pleaseStop=true;
			try {
				int n=0;
				while (thr!=null && pleaseStop() && n<20) {
					Thread.sleep(50);
					n++;
				}
			} catch (InterruptedException e) {
			}
			if (thr!=null) {
				System.out.println("thread forced to stop");
				thr.stop();
				thr=null;
			}
		}
		thr=new Thread(this);
//		thr.setPriority(thr.getThreadGroup().getMaxPriority());
		thr.start();
	}

	Space getSpace(int type,String spname) {
		for (int i=0;i<D.cfg.sc.length;i++) {
			if (D.cfg.sc[i]!=null && D.cfg.sc[i].getName().equals(spname)) {
				return spa[i];
			}
		}
		for (int i=0;i<D.cfg.sc.length;i++) {
			if (D.cfg.sc[i]!=null && D.cfg.sc[i].type==type) {
				return spa[i];
			}
		}
		if (spa.length>0) {
			return spa[0];
		} else {
			return null;
		}
	}

	Space getSpace(String spname) {
		for (int i=0;i<D.cfg.sc.length;i++) {
			if (D.cfg.sc[i].getName().equals(spname)) {
				return spa[i];
			}
		}
		return null;
	}

	Space getSpace(int type) {
		for (int i=0;i<D.cfg.sc.length;i++) {
			if (D.cfg.sc[i].type==type) {
				return spa[i];
			}
		}
		return null;
	}

/* ---------------------- Runnable --------------------- */

	void stopAnimation() {
		pleaseStop=true;
		while (isRunning()) {
			try {
				Thread.sleep(10);
			} catch (Exception e) {
			}
		}
	}
	
	void stopAnimations() {
		stopAnimation();
		if (spa!=null) {
			for (int i=spa.length-1;i>=0;i--) {
				spa[i].stopMoving();
			}
		}
	}

	private boolean pleaseStop() {
		return pleaseStop;
	}
	
	private volatile boolean stopping=false;
	
	private boolean isStopping() {
		return stopping;
	}
	
	public void registerTextsInParser() {
		if (spa!=null) {
			for (int i=0;i<spa.length;i++) {
				spa[i].registerTextsInParser();
			}
		}
	}

	public void updateTextsFromNodes() {
		if (spa!=null) {
			for (int i=0;i<spa.length;i++) {
				spa[i].updateTextsFromNodes();
			}
		}
	}

	void stop() {
		stopping=true;
		stopAnimations();
		if (D instanceof Arquimedes && EC!=null) {
			EC.stop();
		}
		if (spa!=null) {
			for (int i=spa.length-1;i>=0;i--) {
				spa[i].stop();
			}
		}
		if (numCtrl!=null) {
			for (int i=numCtrl.size()-1;i>=0;i--) {
				((NumericControl) numCtrl.elementAt(i)).stop();
			}
		}
		if (SceneNumCtrl!=null) {
			for (int i=SceneNumCtrl.size()-1;i>=0;i--) {
				((NumericControl) SceneNumCtrl.elementAt(i)).stop();
			}
		}
		stopping=false;
	}
	
	boolean isRunning() {
		return running;
	}

	public void run() {
		pleaseStop=false;
		running=true;
		onestep=false;
		stay=false;
		pause=false;
		updateAnimationButtons();
		Node delay=D.p.Analyse(D.cfg.animC.s_value[animConfig.ixdelay],"60");
		try {
			animAlg.initialize();
			int n=0;
			while (!pleaseStop()) {
				thr.sleep(20);
				do {
					thr.sleep(Math.max(20,(long) Math.round(delay.Evaluate(50))));
					if (!pause ||onestep) {
						stay=false;
						animAlg.step();
						onestep=false;
						if (!isStopping()) {
//							refreshScene(false);
							doAlgorithms(false);
							if (!doEventsReturnRestarted()) {
								updateScene(false);
							}
						}
					}
				} while (!pleaseStop() && animAlg.repeat());
				if (!stay && !D.cfg.animC.loop) {
					break;
				}
			}
		} catch (InterruptedException ie) {
		}
		running=false;
		if (!isStopping()) {
			animCtrl.setVisible(false);
			updateAnimationButtons();
			updateScene(false); // despus de poner running=false para que el doFront pinte las componentes.
		}
		thr=null;
	}
	
//	public void update(Graphics g) {
//		paint(g);
//	}

/* -------------------- ActionListener ------------------- */

	public void actionPerformed(ActionEvent e) {
		if (e.getSource()==b[mjaButton.play]) {
			toggleAnimation();
		} else if (e.getSource()==b[mjaButton.rewind]) {
			animAlg.initialize();
			if (pause) {
				refreshScene(true);
			}
		} else if (e.getSource()==b[mjaButton.back]) {
			pause=true;
			backStep();
		} else if (e.getSource()==b[mjaButton.next]) {
			pause=true;
			onestep=true;
		} else if (e.getSource()==b[mjaButton.toend]) {
			pause=true;
			toEnd();
		} else if (e.getSource()==b[mjaButton.stop]) {
			stopAnimation();
		} else if (e.getSource() instanceof mjaLabelAndText) {
			for (int i=0;i<D.cfg.auxiliar.length;i++) {
				if (D.cfg.auxiliar[i].type==data.var&&D.cfg.auxiliar[i].isEditable) {
					D.p.defineExpr(D.cfg.auxiliar[i].getName(),D.cfg.auxiliar[i].exprEd.getInfo());
				}
			}
			D.clear(false);	// agregado para la ver 3.919, 19 de mayo de 2008
			refreshScene(false);
		} else if ((e.getSource() instanceof JTextField)) {
			for (int i=0;i<graph.length;i++) {
				if (graph[i].gc.isEditable) {
					graph[i].gc.formula=graph[i].TF.getText();
					graph[i].analyse();
					if (graph[i].gc.isInBack) {
						graph[i].cs.doBack(false);
					}
				}
			}
			refreshScene(true);
		} else if (e.getSource()==EC) {
			if (EC.isEditing()) {
//				int y=EC.getCursorY();
				int y=EC.getText().getY(this,EC.getText().getCursorPos());
				if (y-50<jsb.getValue()) {
					jsb.setValue(y-50);
				} else if (jsb.getValue()<y-(jp.getSize().height-25)) {
					jsb.setValue(y-(jp.getSize().height-25));
				}
				jsb.setMaximum(Math.max(EC.getCanvasHeight()-jp.getSize().height+240,0));
				jsb.setVisibleAmount(250);
				if (EC.getLocation().y!=-jsb.getValue()) {
					EC.setLocation(0,-jsb.getValue());
					EC.repaint();
				}
			}
		}
	}
	
/* --------------------- Private Methods -------------------- */

	private void updateAnimationButtons() {
		if (!animationIsRunning()) {
			b[mjaButton.play].type=mjaButton.play;
		} else {
			b[mjaButton.play].type=mjaButton.pause;
		}
//		b[mjaButton.play].repaint();
		b[mjaButton.play].paintAll(b[mjaButton.play].getGraphics());
		D.updateAnimationButtons();
	}

	private void backStep() {
		try {
			animAlg.backstep();
			refreshScene(true);
		} catch (Exception e) {
			limitExceeded(e);
		}
	}

	private void toEnd() {
		try {
			animAlg.toEnd();
			refreshScene(true);
			stay=true;
		} catch (Exception e) {
			limitExceeded(e);
		}
	}

	/* -------------- Mouse Listener ------------- */

	public void mousePressed(MouseEvent e) {
	}

	public void mouseReleased(MouseEvent e) {
	}
	public void mouseClicked(MouseEvent e) {
		if ((e.getModifiers()==e.BUTTON3_MASK || (BasicStr.ShiftOn && BasicStr.CtrlOn)) /*&& !(D instanceof Arquimedes)*/)  {
			D.showExternalSpinners();
		} else if (e.getClickCount()==2 && D.inNippeEditor() && e.getSource()==EC) {
			D.edit(e.getSource());
		} else if (e.getClickCount()==2 && !BasicStr.osIsWindows() && e.getSource()==EC) {
			D.edit(e.getSource());
		} else if (e.getClickCount()==1 && (BasicStr.CtrlOn&&!BasicStr.ShiftOn) && D.canEdit()) {
			D.edit(e.getSource());
		}
		BasicStr.CtrlOn=false;
		BasicStr.ShiftOn=false;
	}
	public void mouseEntered(MouseEvent e) {
	}
	public void mouseExited(MouseEvent e) {
	}

	/* -------------- Key Listener ------------- */
	
	public void keyPressed(KeyEvent e) {
		BasicStr.CtrlOn=e.isControlDown();
		BasicStr.ShiftOn=e.isShiftDown();
	}
	public void keyTyped(KeyEvent e) {
	}
	public void keyReleased(KeyEvent e) {
		BasicStr.CtrlOn=false;
		BasicStr.ShiftOn=false;
	}

	/* -------------- Adjustment Listener ------------- */

	public void adjustmentValueChanged(AdjustmentEvent e) {
		EC.containerSize=jp.getSize().height;
		jsb.setMaximum(Math.max(EC.getCanvasHeight()-EC.containerSize+250,0));
		jsb.setUnitIncrement(20);
		if (!e.getValueIsAdjusting() || EC.getCanvasHeight()<5000 ) {
			jsb.setVisibleAmount(250);
			EC.setLocation(0,-jsb.getValue());
			EC.paint(EC.getGraphics());
			paintSceneControls();
			paintSceneSpaces();
			paintControls();
		}
	}
}