/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.descartes;

import java.awt.*;
import java.net.*;
import java.awt.event.*;
import java.awt.image.*;
import com.mja.math.*;
import com.mja.util.*;
import com.mja.gui.*;
import com.mja.descgui.*;
import com.mja.parser.*;
import com.mja.textedit.*;
import com.mja.lang.*;
import com.mja.file.*;
import com.mja.cmp.*;

import javax.swing.*;

public abstract class Space extends Panel implements KeyListener, MouseListener, MouseMotionListener, LiveOrDie {

	public boolean TextBound=false;
	protected Descartes D;
	public spaceConfig spcfg;
	TextAreaControl[] TAC;
	MediaControl[] MC;
	mjaColor backColor;
	Image main;
	protected Image back;
	private Image bgimage;
	protected Graphics2D gmain;
	Graphics2D gback,G;
	protected boolean pleaseStop;
	protected Thread thr;
	static final int bck=0, net=1, net10=2, axe=3, tex=4;
	int w=-1,h=-1;
	int Ox,Oy;
	private int Oxa,Oya;
	private double esca;
	Node NOx,NOy,Nesc;

	private int xa, ya;
	private controlConfig aC;
	private int acix;
	private controlConfig[] grctrl;
	private controlConfig[] txtctrl;
	private controlConfig[] mediactrl;
	private String SpaceID="";
	protected Node visibilityN;

	abstract protected void stopMoving();
	abstract protected void flushImages();
	abstract protected boolean isMoving();
	abstract protected void doCalc();
	abstract protected boolean onDrag(boolean right,int dx,int dy);
	abstract protected void writeCoordinates(Graphics g,int x,int y);
	abstract protected R2 injection(double x,double y);
	abstract protected R2 projection(R2 p);
	abstract public String toTeX(boolean grayscale);
	
	private boolean isPainted=false;
	private Node pos_x,pos_y;
	Scene scene;

	public Space(Descartes D,int index) {
		this(D,D.cfg.sc[index]);
	}
	
	public Space(Descartes D,spaceConfig sc) {
	/* check size=0 */
		setLayout(null);
		this.D=D;
		spcfg=sc;
		visibilityN=D.p.Analyse(spcfg.s_visicond,"1");
		setVisible(visibilityN.Evaluate(1)>0);
		backColor=spcfg.DC[0];
		int nCG=0;
		for (int i=0;i<D.cfg.cc.length;i++) {
			if (D.cfg.cc[i].type==data.c_graphic && D.cfg.cc[i].s_value[controlConfig.ixspace].equals(spcfg.getName())) {
				nCG++;
			}
		}
		grctrl=new controlConfig[nCG];
		int k=0;
		for (int i=0;i<D.cfg.cc.length;i++) {
			if (D.cfg.cc[i].type==data.c_graphic && D.cfg.cc[i].s_value[controlConfig.ixspace].equals(spcfg.getName())) {
				grctrl[k]=D.cfg.cc[i];
				grctrl[k].setSpace(this);
				k++;
			}
		}
		int nCT=0;
		for (int i=0;i<D.cfg.cc.length;i++) {
			if (D.cfg.cc[i].type==data.c_string && D.cfg.cc[i].s_value[controlConfig.ixspace].equals(spcfg.getName())) {
				nCT++;
			}
		}
		txtctrl=new controlConfig[nCT];
		k=0;
		for (int i=0;i<D.cfg.cc.length;i++) {
			if (D.cfg.cc[i].type==data.c_string && D.cfg.cc[i].s_value[controlConfig.ixspace].equals(spcfg.getName())) {
				txtctrl[k]=D.cfg.cc[i];
				txtctrl[k].setSpace(this);
				k++;
			}
		}
		TAC=new TextAreaControl[nCT];
		for (int i=0;i<nCT;i++) {
			TAC[i]=new TextAreaControl(D,txtctrl[i]);
			add(TAC[i]);
		}
		int nCM=0;
		for (int i=0; i<D.cfg.cc.length; i++) {
			if ( ((D.cfg.cc[i].type==data.c_audio)||(D.cfg.cc[i].type==data.c_video))
				&& D.cfg.cc[i].s_value[controlConfig.ixspace].equals(spcfg.getName())) {
				nCM++;
			}
		}
		mediactrl=new controlConfig[nCM];
		k=0;
		for (int i=0;i<D.cfg.cc.length;i++) {
			if ( ((D.cfg.cc[i].type==data.c_audio)||(D.cfg.cc[i].type==data.c_video))
				&& D.cfg.cc[i].s_value[controlConfig.ixspace].equals(spcfg.getName())) {
				mediactrl[k]=D.cfg.cc[i];
				mediactrl[k].setSpace(this);
				k++;
			}
		}
		MC=new MediaControl[nCM];
		for (int i=0;i<nCM;i++) {
			MC[i]=new MediaControl(D,mediactrl[i]);
			add(MC[i]);
		}		
		setSpaceID(spcfg.getSpaceID());
	}
	
//	public Image getImage() {
//		return main;
//	}
	
	public Image getBackImage() {
		return back;
	}

	public void registerTextsInParser() {
		for (int i=0;i<TAC.length;i++) {
			TAC[i].registerTextInParser();
		}		
	}
	
	public void updateTextsFromNodes() {
		for (int i=0;i<TAC.length;i++) {
			TAC[i].updateVisualComponent();
		}		
	}
	
	public void showSolutions() {
		for (int i=0;i<TAC.length;i++) {
			TAC[i].showSolution();
		}		
	}
	
	public void showUserTexts() {
		for (int i=0;i<TAC.length;i++) {
			TAC[i].showUserText();
		}		
	}
	
	public String getSpaceID() {
		return SpaceID;
	}
	
	public void setSpaceID(String sid) {
		SpaceID=sid;
	}
	
/* --------- LiveOrDie --------- */

	private boolean isUsed=false;
	
	public void setIsUsed(boolean b) {
		isUsed=b;
		if (isUsed) {
			D.cfg.addSpace(spcfg);
		} else {
			if (getParent() !=null) {
				getParent().remove(this);
			}
			D.cfg.removeSpace(spcfg);
		}
	}
	
	public boolean isUsed() {
		return isUsed;
	}

	public Component cloneComponent(String cid) {
		spaceConfig nspcfg=spcfg.cloneSpaceConfig();
		Space spc=nspcfg.makeSpace(D);
		spc.preSize();
		D.scene.addEditObjectToText(nspcfg,cid);
		return spc;
	}
	
	public String getSpaceName() {
		return spcfg.getName();
	}

/* -------- END LiveOrDie ------- */


	private Dimension d;

	public void preSize() {
		d=D.getSceneSize();
		d.height=d.height-D.scene.getTitleHeight();
		if (d.width<=0 || d.height<=0) {
			System.out.println("premature sizing");
//			d.width=200;
//			d.height=50;
			(new Exception("premature preSize 2")).printStackTrace();
		}		
		if (!BasicStr.hasContent(spcfg.left)) {
			spcfg.left="0";
		}
		if (!BasicStr.hasContent(spcfg.top)) {
			spcfg.top="0";
		}
		try {
			pos_x=D.p.Analyse("Space "+getName()+".x",spcfg.left,true,false);
		} catch (Exception e) {
			int Lx=(int)Math.round(BasicStr.parsePercentageOf(spcfg.left,d.width,0,0));
			pos_x=D.p.Analyse(Integer.toString(Lx),Lx);
		}
		try {
			pos_y=D.p.Analyse("Space "+getName()+".y",spcfg.top,true,false);
		} catch (Exception e) {
			int Ly=(int)Math.round(BasicStr.parsePercentageOf(spcfg.top,d.height,0,0));
			pos_y=D.p.Analyse(Integer.toString(Ly),Ly);
		}
		w=(int)Math.round(BasicStr.parsePercentageOf(spcfg.width,d.width,d.width,d.width));
		h=(int)Math.round(BasicStr.parsePercentageOf(spcfg.height,d.height,d.height,d.height));
		if (w<=0 || h<=0) {
			System.out.println("spcfg.width="+spcfg.width+" d.width="+d.width);
			System.out.println("spcfg.height="+spcfg.height+" d.height="+d.height);
			System.out.println("Space.presize w="+w);
			System.out.println("Space.presize h="+h);
			(new Exception("espacio sin dimensiones")).printStackTrace();
		}
		int Lx=(int)Math.round(pos_x.Evaluate(0));
		int Ly=(int)Math.round(pos_y.Evaluate(0));
		if (!(D instanceof Arquimedes)) {
			// clip space to scene
			if (Lx<0) { w=w+Lx; Lx=0;}
			if (Ly<0) { h=h+Ly; Ly=0;}
			int right=Math.min(Lx+w,d.width);
			int bottom=Math.min(Ly+h,d.height);
			w=right-Lx;
			h=bottom-Ly;
		}
	
		getSpaceParameters(new Dimension(w,h));
		locate();
		if (BasicStr.hasContent(spcfg.bgimage)&&w>0&&h>0) {
			bgimage=mjaFile.loadImage(D.imagesHT,D,spcfg.bgimage);
		}
		D.p.Var(spcfg.prefix()+"_w").setDouble(w);
		D.p.Var(spcfg.prefix()+"_h").setDouble(h);
	}
	
	void getSpaceParameters(Dimension d) {
		String vname=spcfg.prefix()+"escala";
		double esc;
		if (D.p.isVar(vname) || D.p.isAux(vname)) {
			esc=D.p.rvalue(vname);
		} else {
			esc=BasicStr.parsePercentageOf(spcfg.getScale(),d.width,D.cfg.escala0,D.cfg.escala0);
		}
		Nesc=D.p.Var(vname);
		Nesc.setDouble(esc);
		vname=spcfg.prefix()+"Ox";
		if (D.p.isVar(vname) || D.p.isAux(vname)) {
			Ox=(int) Math.round(D.p.rvalue(vname)+(double)w/2.0);
		} else {
			Ox=(int) Math.round(BasicStr.parsePercentageOf(spcfg.s_Ox,d.width,0,0)+(double)w/2.0);
			D.p.Var(vname).setDouble(Ox-(double)w/2.0);
		}
		NOx=D.p.Var(vname);
		vname=spcfg.prefix()+"Oy";
		if (D.p.isVar(vname) || D.p.isAux(vname)) {
			Oy=(int) Math.round(D.p.rvalue(vname)+(double)h/2.0);
		} else {
			Oy=(int) Math.round(BasicStr.parsePercentageOf(spcfg.s_Oy,d.height,0,0)+(double)h/2.0);
			D.p.Var(vname).setDouble(Oy-(double)h/2.0);
		}
		NOy=D.p.Var(vname);
		String[] xa={"x"};
		String expr="(x-"+spcfg.prefix()+"Ox-"+Integer.toString(w/2)+")/"+spcfg.prefix()+"escala";
		D.p.defineFunc(spcfg.prefix()+"_x",xa,expr,false);
		String[] ya={"y"};
		expr="("+Integer.toString(h/2)+"-y+"+spcfg.prefix()+"Oy)/"+spcfg.prefix()+"escala";
		D.p.defineFunc(spcfg.prefix()+"_y",ya,expr,false);
	}

	void locate() {
		if (TextBound) {
			setSize(w,h);
		} else {
			double dLx=pos_x.Evaluate(0);
			double dLy=pos_y.Evaluate(0);
			int Lx=(int)Math.round(dLx);
			int Ly=(int)Math.round(dLy);
			int ww=w;
			int hh=h;
			setBounds(Lx,Ly,ww,hh);
		}
		for (int i=0;i<TAC.length;i++) {
			TAC[i].locate();
		}
		for (int i=0;i<MC.length;i++) {
			MC[i].locate();
		}
	}

	protected boolean started=false;
	
	private Image TransformGrayToTransparency(BufferedImage image) {
	    ImageFilter filter = new RGBImageFilter() {
      		public final int filterRGB(int x, int y, int rgb) {
        		return (rgb << 8) & 0xFF000000;
      		}
    	};
    	ImageProducer ip = new FilteredImageSource(image.getSource(), filter);
    	return Toolkit.getDefaultToolkit().createImage(ip);
  	}

	Image createTranspImage(int w,int h) {
		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice gs = ge.getDefaultScreenDevice();
		GraphicsConfiguration gc = gs.getDefaultConfiguration();
		Image ima=gc.createCompatibleImage(w, h, Transparency.TRANSLUCENT);
		try {
			Thread.sleep(10);
		} catch (InterruptedException e) {
		}
		Graphics2D g2=(Graphics2D) ima.getGraphics();
		g2.setFont(mjaFont.AWTMono);
		return ima;
	}

	void prepareMainImage() {
		main=createTranspImage(w,h);
		gmain=(Graphics2D) main.getGraphics();
		gmain.setFont(mjaFont.AWTMono);
		buffer=createImage(w,h);
		gb=(Graphics2D) buffer.getGraphics();
		gb.setFont(mjaFont.AWTMono);
	}

	public void start() {
		setBackground(backColor.getAdaptedColor());
		if (isVisible()) {
			for (int i=0;i<spcfg.DC.length;i++) {
				spcfg.DC[i].analyse(D.p);
			}
			G=(Graphics2D) getGraphics();
			int n=0;
//			if ((this instanceof SpaceTA)) {
//				main=null;
//				gmain=(Graphics2D) getGraphics();
//			} else {
				while (n<10) { // Esto se hace porque en ocasiones el pointer a la imagen queda nulo
					try {
						prepareMainImage();
						break;
					} catch (Exception e) {
						n++;
						System.out.println("Space.start waiting "+n);
					}
				}
				if (gmain==null) {
					System.out.println("Graphics not initialized in "+this);
				}
//			}
			n=0;
//			if ((this instanceof SpaceTA)) {
//				back=null;
//				gback=(Graphics2D) getGraphics();
//			} else {
				while (n<10) { // Esto se hace porque en ocasiones el pointer a la imagen queda nulo
					try {
//						back=createImage(Math.max(w,1),Math.max(h,1));
						back=createImage(w,h);
						gback=(Graphics2D) back.getGraphics();
						break;
					} catch (Exception e) {
						n++;
						System.out.println("Space.start waiting "+n);
					}
				}
//			}
			pleaseStop=false;
			started = true;

			addMouseListener(this);
			addMouseMotionListener(this);
			addKeyListener(this);
		}
		for (int i=0;i<TAC.length;i++) {
			TAC[i].start();
		}
	}

	private boolean variable_back_drawn=false;
	private boolean space_variable=false;
	
	protected void updateForTransparency() {
		int xi=-getLocation().x;
		int yi=-getLocation().y;
		if (D.scene.EC!=null) {
			gback.drawImage(D.scene.EC.getBufferedImage(),xi,yi,this);
		}
		for (int i=0;i<D.scene.spa.length;i++) {
			if (D.scene.spa[i]==this) {
				break;
			}
			if (D.scene.spa[i].isVisible()) {
				if (buffer!=null) {
					xi=D.scene.spa[i].getLocation().x-getLocation().x;
					yi=D.scene.spa[i].getLocation().y-getLocation().y;
					gback.drawImage(D.scene.spa[i].buffer,xi,yi,this);
				}
			}
		}
	}

	void doBack(boolean clear) {
		setVisible(visibilityN.Evaluate(1)>0);
		if (isVisible()) {
			if (!started) {
				start();
			}
			Stroke bs=new BasicStroke(1,BasicStroke.CAP_BUTT,BasicStroke.JOIN_ROUND);
		    gback.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
			gback.setStroke(bs);
			if ((backColor.variable()&&!variable_back_drawn) || (backColor.getAdaptedColor().getAlpha()<0xff)) {
				gback.setColor(Color.lightGray);
				gback.fillRect(0,0,w,h);
				updateForTransparency();
				clear=true;
			} 
			if (clear) {
				if (backColor.variable()) {
					if (!variable_back_drawn) {  // modified 2006-12-13
						Color bc=backColor.getAdaptedColor();
						gback.setColor(bc);
						Node X=D.p.Var("x");
						Node Y=D.p.Var("y");
						synchronized (X) {
							synchronized (Y) {
								double savex=X.getDouble(),savey=Y.getDouble();
								for (int j=0;j<h;j++) {
									Y.setDouble(y(j));
									for (int i=0;i<w;i++) {
										X.setDouble(x(i));
										Color bc1=backColor.getAdaptedColor();
										if (!bc.equals(bc1)) {
											gback.setColor(bc1);
											space_variable=true;
										}
										gback.fillRect(i,j,1,1);
									}
								}
								X.setDouble(savex); Y.setDouble(savey);
							}
						}
						variable_back_drawn=true;
					} else if (!space_variable) {
						gback.setColor(backColor.getAdaptedColor());
						gback.fillRect(0,0,w,h);
					}
				} else {
					gback.setColor(backColor.getAdaptedColor());
					gback.fillRect(0,0,w,h);
				}
				if (bgimage!=null) {
					int bgw=bgimage.getWidth(this);
					int bgh=bgimage.getHeight(this);
					switch (spcfg.bgdisplay) {
					case spaceConfig.stretch:
						gback.drawImage(bgimage,0,0,w,h,this);
						break;
					case spaceConfig.center:
						gback.drawImage(bgimage,(w-bgw)/2,(h-bgh)/2,this);
						break;
					case spaceConfig.patch:
						for (int j=0;j<h;j+=bgh) {
							for (int i=0;i<w;i+=bgw) {
								gback.drawImage(bgimage,i,j,this);
							}
						}
						break;
					default :
						gback.drawImage(bgimage,0,0,this);
						break;
					}
				}
			}
		}
	}
	
	protected void doFront(boolean calc) {
		if (isVisible()) {
			Component[] ca=getComponents();
			if (buffer!=null) {
				for (int i=0;i<ca.length;i++) {
					if (ca[i] instanceof ButtonControl) {
						((mjaComponent) ((ButtonControl) ca[i]).getMjaButton()).setBackImage(buffer);
					}
				}
			}
			if (D instanceof Arquimedes && !backColor.isEqualTo(Scene.sceneBackColor) ) {
				int gr=4;
				for (int i=0;i<gr;i++) {
					gmain.setColor(new Color(0,0,0,(3*(gr-i)*(gr-i))/2));
					gmain.drawRect(i,i,getSize().width-1-2*i,getSize().height-1-2*i);
				}
			}

		}

	}
	
/* ------------------------- update and paint ------------------------ */

	protected Image buffer;
	private Graphics2D gb;

	protected void drawBackControls() {
		Graphics2D [] bga={(Graphics2D) back.getGraphics()};
		for (int i=0;i<grctrl.length;i++) {
			if (grctrl[i].hasTrace) {
				grctrl[i].drawControl(true,bga);
			}
		}
	}

	private void drawFrontControls() {
		Graphics2D [] fga={(Graphics2D) buffer.getGraphics()};
		for (int i=0;i<grctrl.length;i++) {
			grctrl[i].drawControl(false,fga);
		}
	}

	public synchronized void paint(Graphics g) {
		if (g!=null && back!=null && main!=null && buffer!=null) {
//			drawBackControls();
			gb.drawImage(back,0,0,this);
			gb.drawImage(main,0,0,this);
			drawFrontControls();
			g.drawImage(buffer,0,0,this);
		}
	}
	
	public Image getBuffer() {
		return buffer;
	}

	private boolean updating;
	
	public synchronized void updateSpace(boolean clear,boolean calc) {
		if (!updating) {
			updating=true;
			doBack(clear);
			doFront(calc);
			updating=false;
		}
	}
	
	synchronized void paintSpace() {
		paint(G);
	}

/* --------------------------- Runnable ------------------------------ */

	private boolean dragging;
	
	boolean isDragging() {
		return dragging;
	}
	
	private volatile boolean stopping=false;
	protected boolean isStopping() {
		return stopping;
	}
	
	public void stop() {
		for (int i=0;i<TAC.length;i++) {
			TAC[i].stop();
		}
		stopping=true;
		pleaseStop=true;
		try {
			while (isRunning()) {
				Thread.sleep(10);
			}
		} catch (InterruptedException e) {
		}
		flushImages();
		if (main!=null) {
			main.flush();
			main=null;
		}
		if (back!=null) {
			back.flush();
			back=null;
		}
		if (bgimage!=null) {
			bgimage.flush();
			bgimage=null;
		}
		if (buffer!=null) {
			buffer.flush();
			buffer=null;
		}
		stopping=false;
	}
	
	protected boolean pleaseStop() {
		return pleaseStop;
	}
	
	private boolean isRunning() {
		return thr!=null;
	}

	double x(int m) {
		return (Nesc!=null)?(double)(m-Ox)/Nesc.getDouble():0;
	}
	double y(int n) {
		return (Nesc!=null)?(double)(Oy-n)/Nesc.getDouble():0;
	}

	boolean inside(R2 q) {
		return -w<=q.ix()&&q.ix()<=2*w&&-h<=q.iy()&&q.iy()<=2*h ;
	}

	boolean closeEnough(R2 t,R2 cp,int w,int h,boolean useRect) {
		R2 r=projection(cp);
		if (useRect) {
			return Math.abs(r.x-t.x)<w && Math.abs(r.y-t.y)<h;
		} else {
			return (r.x-t.x)*(r.x-t.x)/(w*w)+(r.y-t.y)*(r.y-t.y)/(h*h)<1;
		}
//		return (.distance(t)<d);
	}

	void adjustControls() {
		for (int i=0;i<grctrl.length;i++) {
			grctrl[i].adjust();
		}
	}

	void setActiveControl(int ix) {
		if (ix<0) { ix=grctrl.length-1; }
		if (ix>=grctrl.length) { ix=0; }
		for (int i=0;i<grctrl.length;i++) {
			grctrl[i].isSelected=false;
		}
		grctrl[ix].isSelected=true;
		aC=grctrl[ix];
		aC.isPressed=true;
		acix=ix;
	}

	boolean spaceChanged() {
		Ox=(int)Math.round(D.p.rvalue(spcfg.prefix()+"Ox")+(double)w/2.0);
		Oy=(int)Math.round(D.p.rvalue(spcfg.prefix()+"Oy")+(double)h/2.0);
		correctScale();
		boolean changed=Nesc.getDouble()!=esca||Ox!=Oxa||Oy!=Oya;
		esca=Nesc.getDouble(); Oxa=Ox; Oya=Oy;
		return changed;
	}

	protected void correctScale() {
		// la escala est limitada al interval [E-6,1E6]
		if (Nesc.getDouble()<0.000001) {
			Nesc.setDouble(0.000001);
		}
		if (Nesc.getDouble()>1000000) {
			Nesc.setDouble(1000000);
		}
	}

/* ----------------- MouseListener ----------------- */

	private boolean dragged;
	private R2 ds=new R2(0,0);
	private long mouse_pressed_at,mouse_released_at;

	public void mousePressed(MouseEvent ev) {
		dragging=true;
		xa=ev.getX();
		ya=ev.getY();
		if (ev.getModifiers()==ev.BUTTON1_MASK) {
			D.updateNodesFromControls();	// agregado el 29 de mayo de 2009   Va a haber refresh seguro.
			stopMoving();	// agregado el 11 de julio de 2007
			mouse_pressed_at=System.currentTimeMillis();
			updateMouseVariables(ev);
			dragged=false;
			if (aC!=null) {
				aC.isSelected=false;
			}
			aC=null;
			R2 t=new R2(xa,ya);
			boolean controlTouched=false;
			for (int i=grctrl.length-1;i>=0;i--) {	// modificado el 18 de Junio de 2007
				D.p.Var(grctrl[i].label+".active").setDouble(0);
				D.p.Var(grctrl[i].label+".activo").setDouble(0);
			}
			for (int i=grctrl.length-1;i>=0;i--) {
				if (grctrl[i].isActive()) {
					if (closeEnough(t,grctrl[i].cpos,grctrl[i].size_x+1,grctrl[i].size_y+1,grctrl[i].hasImage())) {
						R2 r=projection(grctrl[i].cpos);
						ds.x=r.x-t.x;
						ds.y=r.y-t.y;
						setActiveControl(i);
						D.p.Var(grctrl[i].label+".active").setDouble(1);	// modificado el 18 de Junio de 2007
						D.p.Var(grctrl[i].label+".activo").setDouble(1);	// modificado el 18 de Junio de 2007
						controlTouched=true;
						D.refresh(false);
						return;
					}
				}
			}
			if (!controlTouched) {
				D.p.Var(spcfg.prefix()+"mouse_pressed").setDouble(1);
				D.p.Var(spcfg.prefix()+"ratn_apretado").setDouble(1);
			}
			D.refresh(false);
			if (!controlTouched) {
				writeCoordinates(G,ev.getX(),ev.getY());
			}
		}
	}

	long mouse_dragged_at;

	public void mouseReleased(MouseEvent ev) {
		if (ev.getModifiers()==ev.BUTTON1_MASK) {
			if (aC!=null) {
				aC.isPressed=false;
			}
			mouse_released_at=System.currentTimeMillis();
			if (mouse_dragged_at+250<mouse_released_at) {
				stopMoving();
			}
			updateMouseVariables(ev);
			boolean wasDragged=dragged;
			if (dragged) {
				dragged=false;
				D.setCursor(null);
				dragged=false;
			}
			if (aC!=null) {
				requestFocus();
			}
			if (D.p.Var(spcfg.prefix()+"mouse_pressed").getDouble()==1) {
				D.p.Var(spcfg.prefix()+"mouse_pressed").setDouble(0);
				D.p.Var(spcfg.prefix()+"ratn_apretado").setDouble(0);
				D.refresh(true);
			} else {
				D.refresh(false);
			}
		}
		dragging=false;
		requestFocus();
	}

	public void mouseClicked(MouseEvent ev) {
		if (
			((ev.getClickCount()==1)&&BasicStr.CtrlOn&&(!BasicStr.ShiftOn))
			) {
			D.edit(this);
		} else if (ev.getClickCount()==1) {
			// modificado el 04 de mayo de 2007 para detectar un clic derecho falso en un pzarrn electrnico.
			if (((ev.getModifiers()==ev.BUTTON3_MASK)|| (BasicStr.CtrlOn&&BasicStr.ShiftOn)) && (System.currentTimeMillis()-mouse_released_at>1000)) {
				R2 t=new R2(xa,ya);
				boolean controlTouched=false;
				for (int i=0;i<grctrl.length;i++) {
					if (grctrl[i].isActive()) {
						if (closeEnough(t,grctrl[i].cpos,grctrl[i].size_x+1,grctrl[i].size_y+1,grctrl[i].hasImage())) {
							if (BasicStr.hasContent(grctrl[i].explanation) ) {
								controlTouched=true;
								paint(G);
								mjaText.msg((Component) ev.getSource(),
											grctrl[i].label,
											grctrl[i].explanation,
											grctrl[i].expl_font);
							}
						}
					}
				}
				if (!controlTouched) {
					D.showExternalSpinners();
				}
			} else {
				stopMoving();
				D.p.Var(spcfg.prefix()+"mouse_clicked").setDouble(1);
				D.p.Var(spcfg.prefix()+"clic_izquierdo").setDouble(1);
//	 2014-11-17
//				D.refresh(false);
//				D.p.Var(spcfg.prefix()+"mouse_clicked").setDouble(0);
//				D.p.Var(spcfg.prefix()+"clic_izquierdo").setDouble(0);
			}
		}
	}
	public void mouseEntered(MouseEvent ev) { 
//		D.updateNodesFromControls();	// agregado el 27 de mayo de 2009
	}
	public void mouseExited(MouseEvent ev) {
	 }

/* ---------------- MouseMotionListener --------------- */

	public void mouseDragged(MouseEvent ev) {
		if (ev.getModifiers()==ev.BUTTON1_MASK) {
			updateMouseVariables(ev);
		}
		if (aC!=null) {
			if (ev.getModifiers()==ev.BUTTON1_MASK) {
				aC.setControlPos(ev.getX()+ds.x,ev.getY()+ds.y);
			}
		} else if (!spcfg.fixed) {
			mouse_dragged_at=System.currentTimeMillis();
			dragged|=onDrag(ev.getModifiers()==ev.BUTTON3_MASK,ev.getX()-xa,ev.getY()-ya);
			D.updateTextFieldsFromNodes();
			if (dragged) {
				D.setCursor(new Cursor(Cursor.MOVE_CURSOR));
			}
		}
		xa=ev.getX();
		ya=ev.getY();
		D.refresh(false);
		if (aC==null && ev.getModifiers()==ev.BUTTON1_MASK) {
			writeCoordinates(G,ev.getX(),ev.getY());
		}
	}

	private controlConfig ttc;

	public void mouseMoved(MouseEvent ev) {
		Mouse_X=ev.getX();
		Mouse_Y=ev.getY();
		if (spcfg.mouseMotionActive) {
			updateMouseVariables(ev);
		}
		if (ttc!=null) {
			ttc.isOver=false;
		}
		boolean touched=false;
		R2 t=new R2(ev.getX(),ev.getY());
		for (int i=0;i<grctrl.length;i++) {
			grctrl[i].isOver=false;
			if (grctrl[i].isActive()) {
				if (closeEnough(t,grctrl[i].cpos,grctrl[i].size_x,grctrl[i].size_y,grctrl[i].hasImage())) {
					touched=true;
					if (grctrl[i]!=ttc) {
						ttc=grctrl[i];
						if (BasicStr.hasContent(grctrl[i].tooltip) ) {
							Component C=(Component) ev.getSource();
							mjaGui.showTooltip(D,C,
							new Point(C.getLocationOnScreen().x+ev.getX(),C.getLocationOnScreen().y+ev.getY()+16),
							grctrl[i].tooltip,D.Tr.getTr(data.rightClick),
							grctrl[i].tooltip_font);
						}
					}
					break;
				}
			}
		}
		boolean refreshed=false;
		if (touched) {
			setCursor(new Cursor(Cursor.HAND_CURSOR));
			if (!ttc.isOver) {
				ttc.isOver=true;
//		Eliminado 2014-06-27
//				D.refresh(true);
//				refreshed=true;
			}
		} else {
			setCursor(null);
			mjaGui.hideTooltip();
			if (ttc!=null) {
				ttc.isOver=false;
//		Eliminado 2014-06-27
//				D.refresh(true);
//				refreshed=true;
			}
			ttc=null;
		}
		if (!refreshed && spcfg.mouseMotionActive) {
				D.refresh(true);
		}
	}
	
	private int Mouse_X,Mouse_Y;

	private void updateMouseVariables(MouseEvent ev) {
		D.p.Var(spcfg.prefix()+"mouse_x").setDouble(x(ev.getX()));
		D.p.Var(spcfg.prefix()+"mouse_y").setDouble(y(ev.getY()));
		D.p.Var(spcfg.prefix()+"ratn_x").setDouble(x(ev.getX()));
		D.p.Var(spcfg.prefix()+"ratn_y").setDouble(y(ev.getY()));
	}

/* -------------------- KeyListener ------------------- */

	public void keyPressed(KeyEvent e) {
		BasicStr.CtrlOn =e.isControlDown();
		BasicStr.ShiftOn =e.isShiftDown();
		if (e.getSource()==this && grctrl.length>0) {
			if (aC==null) {
				setActiveControl(0);
			}
			if (aC!=null && aC.isActive()) {
				controlConfig aC0=aC;
				R2 p=aC.getControlPos();
				switch (e.getKeyCode()) {
				case KeyEvent.VK_LEFT: p.x-=1; break;
				case KeyEvent.VK_RIGHT: p.x+=1; break;
				case KeyEvent.VK_UP: p.y-=1; break;
				case KeyEvent.VK_DOWN: p.y+=1; break;
				case KeyEvent.VK_TAB:
					if (e.getModifiers()==KeyEvent.SHIFT_MASK) {
						acix-=1;
						if (acix<0) {
							acix=grctrl.length-1;
						}
					} else {
						acix+=1;
						if (acix>=grctrl.length) {
							acix=0;
						}
					}
					setActiveControl(acix);
					e.consume();
					updateSpace(false,false);
					return;
				default: return;
				}
				aC0.setControlPos(p.x,p.y);
				D.scene.refreshScene(true);
			}
		}
	}

	public void keyReleased (KeyEvent e) {
		BasicStr.CtrlOn=false;
		BasicStr.ShiftOn=false;
	}
	public void keyTyped (KeyEvent e) {
	}

}