/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.descartes;

import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.util.*;
import com.mja.util.*;
import com.mja.math.*;
import com.mja.math3D.*;
import com.mja.gui.*;
import com.mja.descgui.*;
import com.mja.parser.*;
import com.mja.poly.*;
import com.mja.lang.*;
import com.mja.text.*;
import com.mja.descartes.*;

public class SpaceR3 extends Space implements Runnable {

	static final mjaColor dfc=new mjaColor(Color.white), dbc=new mjaColor(Color.pink);
	static final double oneDegree = Math.PI/180;

	private mjaSurface[][] pS;
	private Vector text,text_par,tx,ty,tfont,tcol,tcond;
	private Surface Poly;
	private viewR3 view;
	private R3 rot, vr=new R3(0,0,0);
	private Graphics[] bga=new Graphics[0],
					   fga=new Graphics[1];
	static final int sortAndPaint=0, painterAlgorithm=1, rayTracing=2;
	private int renderMethod;
	private Node NRotx,NRoty,NRotz, Ndepth, NUIx, NUIy, NUIz, NUIdim,NUIInt;
	private parametrization[] par;

	public SpaceR3(Descartes D,spaceConfig sc) {
		super(D,sc);
		initUI();
	}

	public SpaceR3 (Descartes D,int index) {
		super(D,index);
		initUI();
		renderMethod=sortAndPaint;
	}
	
	public String toTeX(boolean grayscale) {
		double SCL=Nesc.Evaluate(48);
		double DX=(getWidth()/2+NOx.Evaluate(0))/SCL;
		double DY=(getHeight()/2-NOy.Evaluate(0))/SCL;		
		return Poly.toTeX(this,grayscale,SCL,getWidth()/2,getHeight()/2,DX,DY,textsToTeX(grayscale));
	}

	private String textsToTeX(boolean grayscale) {
		String str="";
		if (text!=null) {
			for (int i=0;i<text.size();i++) {
				parametrization param=(parametrization)text_par.elementAt(i);
				double saveP=param.Np.getDouble();
				double s1=param.Ni.Evaluate(0);
				double s2=param.Nf.Evaluate(1);
				int N=(int) Math.round(param.NN.Evaluate(0));
				for (int j=0;j<=N;j++) {
					param.Np.setDouble(s1+((double) j * (s2-s1))/(double) N);
					Node cond=(Node) tcond.elementAt(i);
					if (cond.Evaluate(1)>0) {
						MathText mthtxt=(MathText) text.elementAt(i);
						Node nx=(Node) tx.elementAt(i);
						Node ny=(Node) ty.elementAt(i);
						int x0=(int) Math.round(nx.Evaluate(20))-20;
						int y0=(int) Math.round(ny.Evaluate(20));
						mjaColor c=(mjaColor) tcol.elementAt(i);
						str=str+BasicStr.DefineColor(grayscale,"fc",c.getAdaptedColor());
						str=str+mthtxt.getTeXColors(grayscale);
						str=str+"\\uput{1pt}[30]("+BasicStr.DblStr(x0)+","+BasicStr.DblStr(getHeight()-mthtxt.getTextY(this))+"){\\shortstack[l]{"+mthtxt.toTeX()+"}}\r\n";
					}
				}
				param.Np.setDouble(saveP);
			}
		}
		return str;
	}


	void initUI() {
		String nuidim=spcfg.prefix()+"userIlum.dim";
		if (!D.p.isVar(nuidim)) {
			D.p.Var(nuidim).setDouble(1.0);
		}
		NUIdim=D.p.Var(nuidim);
		NUIx=D.p.Var(spcfg.prefix()+"userIlum.x");
		NUIy=D.p.Var(spcfg.prefix()+"userIlum.y");
		NUIz=D.p.Var(spcfg.prefix()+"userIlum.z");
		NUIInt=D.p.Var(spcfg.prefix()+"userIlum.I");
	}

/* -----------------------   Overwrites  -------------------------*/

	void getSpaceParameters(Dimension d) {
		super.getSpaceParameters(d);
		String dname=spcfg.prefix()+"observador";
		double depth;
		if (D.p.isVar(dname) || D.p.isAux(dname)) {
			depth=D.p.rvalue(dname);
			depth=(depth<0.25*(w+h))?2.5*(w+h):depth; 
		} else {
			depth=2.5*(w+h);
		}
		Ndepth=D.p.Var(dname);
		Ndepth.setDouble(correctedDepth(depth));
		view=new viewR3(Ox,Oy,Ndepth.getDouble(),Nesc.getDouble());
		NRotx=D.p.Var(spcfg.prefix()+"rot.x"); //NRotx.r=0;
		NRoty=D.p.Var(spcfg.prefix()+"rot.y"); //NRoty.r=0;
		NRotz=D.p.Var(spcfg.prefix()+"rot.z"); //NRotz.r=0;
		rot=new R3(NRotx.getDouble(),NRoty.getDouble(),NRotz.getDouble());
	}

	public void start() {
		if (isVisible()) {
			super.start();
			pS=new mjaSurface[D.cfg.allgR3c.length][1];
			tx=new Vector();
			ty=new Vector();
			text=new Vector();
			text_par=new Vector();
			tfont=new Vector();
			tcol=new Vector();
			tcond=new Vector();
			par=new parametrization[pS.length];
			for (int i=0;i<pS.length;i++) {
				graphR3Config gc=D.cfg.allgR3c[i];
				if (gc!=null && gc.s_value!=null && gc.s_value[graphR3Config.space]!=null 
					&& gc.s_value[graphR3Config.space].equals(spcfg.getName())) {
					par[i]=new parametrization(D.p,"s","[0,0]","0");
					double saveP=par[i].Np.getDouble();
					if (gc.isFamily) {
						par[i]=new parametrization(D.p,gc.s_value[gc.ixf_parameter],gc.s_value[gc.ixf_interval],gc.s_value[gc.ixf_steps]);
						par[i].Np.setDouble(par[i].Ni.Evaluate(0));
					}
					double s1=par[i].Ni.Evaluate(0);
					double s2=par[i].Nf.Evaluate(1);
					int N=(int) Math.round(par[i].NN.Evaluate(0));
					if (D.Tr.equals(gc.s_value[graphR3Config.type],data.t_text)) {
						// parse Text configuration.
						for (int j=0;j<=N;j++) {
							par[i].Np.setDouble(s1+((double) j * (s2-s1))/(double) N);
							try {
								String[] st=BasicStr.getTokens(gc.s_value[graphR3Config.expr]);
								tx.addElement(D.p.Analyse(st[0]));
								ty.addElement(D.p.Analyse(st[1]));
								mjaColor c=D.Tr.parseColor(gc.s_value[graphR3Config.color],new mjaColor(Color.gray),null);
								MathText mthtxt=new MathText(D.p,D.Tr,gc.s_value[graphR3Config.text],D.p.Analyse(gc.s_value[graphR3Config.dec],2),D.Tr.isNotFalse(gc.s_value[graphR3Config.fixed]),gc.font,c.getAdaptedColor());
								text.addElement(mthtxt);
								text_par.addElement(par[i]);
								tfont.addElement(gc.font);
								c.analyse(D.p);
								tcol.addElement(c);
								tcond.addElement(D.p.Analyse(gc.getCondition(),1));
							} catch (Exception exc) {
								System.out.println("bad expression "+gc.s_value[graphR3Config.expr]);
							}
						}
					} else {
						pS[i]=new mjaSurface[N+1];
						for (int j=0;j<=N;j++) {
							par[i].Np.setDouble(s1+((double) j * (s2-s1))/(double) N);
							pS[i][j]=SurfaceFactory.getMjaSurface(D.p,
								canonicType(D.Tr,gc.s_value[graphR3Config.type]),
								gc.s_value[graphR3Config.expr],
								gc.s_value[gc.Nu],
								gc.s_value[gc.Nv],
								gc.s_value[gc.grueso],
								gc.s_value[gc.ancho],
								gc.s_value[gc.largo],
								gc.s_value[gc.alto],
								D.Tr.parseColor(gc.s_value[gc.color],dfc,null),
								D.Tr.parseColor(gc.s_value[gc.backColor],dbc,null),
								D.Tr.isNotFalse(gc.s_value[gc.edges]),
								gc.s_value[gc.inirot],
								gc.s_value[gc.inipos],
								gc.s_value[gc.finrot],
								gc.s_value[gc.finpos],
								gc.s_value[gc.rotation],
								gc.s_value[gc.location],
								getModel(D.Tr,gc.s_value[gc.model]),
								gc.s_value[gc.ixcond],
								D.Tr.isTrue(gc.s_value[gc.ixbackground])
								);
							pS[i][j].split=D.Tr.isTrue(gc.s_value[graphR3Config.split]);
							if (D.Tr.equals(gc.s_value[graphR3Config.type],data.t_point)) {
								// parse Text configuration.
								try {
									mjaColor c=D.Tr.parseColor(gc.s_value[graphR3Config.color],new mjaColor(Color.gray),null);
									MathText mthtxt=new MathText(D.p,D.Tr,gc.s_value[graphR3Config.text],D.p.Analyse(gc.s_value[graphR3Config.dec],2),D.Tr.isNotFalse(gc.s_value[graphR3Config.fixed]),gc.font,c.getAdaptedColor());
									pS[i][j].setMathText(mthtxt);
								} catch (Exception exc) {
									System.out.println("bad expression "+gc.s_value[graphR3Config.expr]);
								}
							}
						}
						par[i].Np.setDouble(saveP);
					}
				}
			}
			if (D.Tr.equals(spcfg.s_value[spaceConfig.render],data.painter)) {
				renderMethod=painterAlgorithm;
			} else if (D.Tr.equals(spcfg.s_value[spaceConfig.render],data.raytrace)) {
				renderMethod=rayTracing;
			} else {
				renderMethod=sortAndPaint;
			}
			fga[0]=gmain;
			thr= new Thread(this);
			thr.setPriority(Thread.MIN_PRIORITY);
			thr.start();

		}
	}

	public void run() {
		updateSpace(true,false);
		if (main!=null) {
			paintSpace(); 
		}
		pleaseStop=false;
		try {
			while (!pleaseStop()) {
				thr.sleep(50);
				if (isMoving()&&!isDragging()) {
					moveOneStep();
					D.updateTextFieldsFromNodes(false); // 1o de Mayo de 2009, se pas a false, porque no deben actualizarse un campo de texto si tienen el foco.
					if (!isStopping()) {
						updateSpace(false,false);
						if (main!=null) {
							paintSpace(); 
						}
					}
				}
			}
		} catch(InterruptedException e) {
			e.printStackTrace();
		}
		thr=null;
	}

/* ------------------ Implementation of Space --------------------*/

	protected R2 injection(double x,double y) {
		view.setView(Ox,Oy,view.getDepth());
		R3 vp=view.getView();
		return new R2((double)(x-vp.x)/view.getScale(),(double)(vp.y-y)/view.getScale());
	}

	protected R2 projection(R2 p) {
		view.setView(Ox,Oy,view.getDepth());
		R3 vp=view.getView();
		return new R2(vp.x+view.getScale()*p.x,vp.y-view.getScale()*p.y);
	}

	protected void flushImages() {
	}
	protected void stopMoving() {
		vr.x=0; vr.y=0; vr.z=0;
	}
	protected boolean isMoving() {
		return ((vr.x!=0 || vr.y!=0 || vr.z!=0));
	}

	private void drawTexts() {
		if (text!=null) {
			for (int i=0;i<text.size();i++) {
				parametrization param=(parametrization)text_par.elementAt(i);
				double saveP=param.Np.getDouble();
				double s1=param.Ni.Evaluate(0);
				double s2=param.Nf.Evaluate(1);
				int N=(int) Math.round(param.NN.Evaluate(0));
				for (int j=0;j<=N;j++) {
					param.Np.setDouble(s1+((double) j * (s2-s1))/(double) N);
					Node cond=(Node) tcond.elementAt(i);
					if (cond.Evaluate(1)>0) {
						mjaColor c=(mjaColor) tcol.elementAt(i);
						gmain.setColor(c.getAdaptedColor());
						MathText mthtxt=(MathText) text.elementAt(i);
						Node nx=(Node) tx.elementAt(i);
						Node ny=(Node) ty.elementAt(i);
						int x0=(int) Math.round(nx.Evaluate(20));
						int y0=(int) Math.round(ny.Evaluate(20));
						Font f=(Font) tfont.elementAt(i);
						gmain.setFont(f);
						TFont tf=new TFont(f);
						Graphics[] g=new Graphics[1];
						g[0]=gmain;
						R2 q=new R2(x0,y0-tf.getAscent());
						mthtxt.draw(this,q.ix(),q.iy(),g);
					}
				}
				param.Np.setDouble(saveP);
			}
		}
	}

	private boolean allFixed;

	synchronized protected void doCalc() {	// synchronized para que no interfieran el run y el mouseDragged
		if (!allFixed ) {
			Poly=null;
			Surface IS=null;
			allFixed=true;
			for (int i=0;i<pS.length;i++) {
				if (par[i]!=null && par[i].Np!=null && D.cfg!=null && D.cfg.allgR3c[i]!=null) {
					double saveP=par[i].Np.getDouble();
					double s1=par[i].Ni.Evaluate(0);
					double s2=par[i].Nf.Evaluate(1);
					graphR3Config gc=D.cfg.allgR3c[i];
					if (!D.Tr.equals(gc.s_value[graphR3Config.type],data.t_text)) {
						int N=(int) Math.round(par[i].NN.Evaluate(0));
						if ( N+1 != pS[i].length) {
							if (gc.s_value[graphR3Config.space].equals(spcfg.getName())) {
								pS[i]=new mjaSurface[N+1];
								for (int j=0;j<=N;j++) {
									par[i].Np.setDouble(s1+((double) j * (s2-s1))/(double) N);
									pS[i][j]=SurfaceFactory.getMjaSurface(D.p,
										canonicType(D.Tr,gc.s_value[graphR3Config.type]),
										gc.s_value[graphR3Config.expr],
										gc.s_value[gc.Nu],
										gc.s_value[gc.Nv],
										gc.s_value[gc.grueso],
										gc.s_value[gc.ancho],
										gc.s_value[gc.largo],
										gc.s_value[gc.alto],
										D.Tr.parseColor(gc.s_value[gc.color],dfc,null),
										D.Tr.parseColor(gc.s_value[gc.backColor],dbc,null),
										D.Tr.isNotFalse(gc.s_value[gc.edges]),
										gc.s_value[gc.inirot],
										gc.s_value[gc.inipos],
										gc.s_value[gc.finrot],
										gc.s_value[gc.finpos],
										gc.s_value[gc.rotation],
										gc.s_value[gc.location],
										getModel(D.Tr,gc.s_value[gc.model]),
										gc.s_value[gc.ixcond],
										D.Tr.isTrue(gc.s_value[gc.ixbackground])
										);
									pS[i][j].split=D.Tr.isTrue(gc.s_value[graphR3Config.split]);
									if (D.Tr.equals(gc.s_value[graphR3Config.type],data.t_point)) {
										// parse Text configuration.
										try {
											mjaColor c=D.Tr.parseColor(gc.s_value[graphR3Config.color],new mjaColor(Color.gray),null);
											MathText mthtxt=new MathText(D.p,D.Tr,gc.s_value[graphR3Config.text],D.p.Analyse(gc.s_value[graphR3Config.dec],2),D.Tr.isNotFalse(gc.s_value[graphR3Config.fixed]),gc.font,c.getAdaptedColor());
											pS[i][j].setMathText(mthtxt);
										} catch (Exception exc) {
											System.out.println("bad expression "+gc.s_value[graphR3Config.expr]);
										}
									}
								}
							}
						}
					}
					for (int j=0;j<pS[i].length;j++) {
						if (pS[i][j]!=null) {
							par[i].Np.setDouble(s1+((double) j * (s2-s1))/(double) (pS[i].length-1));
							allFixed&=pS[i][j].isFixed();
							if (pS[i][j].cond.Evaluate(1)>0) {
								Surface newS=pS[i][j].create();
								newS.setSNode(par[i].Np);
								newS.setSValue(par[i].Np.getDouble());
								if (newS!=null) {
									if (spcfg.split||pS[i][j].split) {
										if (IS!=null) {
											newS=IS.split(newS);
										}
										IS=Surface.add(IS,newS);
									} else {
										Poly=Surface.add(Poly,newS);
									}
								}
							}
						}
					}
					par[i].Np.setDouble(saveP);
				}
			}
			if (Poly!=null) {
				Poly=Surface.add(Poly,IS);
			} else {
				Poly=IS;
			}
		}
	}
	
	private LinearTransformation SR=R3SpaceRotation(new R3(0,0,0));
	
	boolean dragEnhanced=false;

	synchronized protected void doFront(boolean calc) {
		if (isVisible()) {
			correctScale();
			view.setView(Ox,Oy,view.getDepth());
			view.setScale(Nesc.getDouble());
			Ndepth.setDouble(correctedDepth(Ndepth.getDouble()));
			view.setDepth(Ndepth.getDouble());
			prepareMainImage();
//			gmain.drawImage(back,0,0,null);
			drawTexts();
			adjustControls();
			if (calc) {
				doCalc();
			}
			if (Poly!=null) {
				Poly.setUserLightPos((int) NUIx.getDouble(),(int) NUIy.getDouble(),(int) NUIz.getDouble());
				Poly.setUserIllumination(NUIdim.getDouble(),NUIInt.getDouble());
				Poly.resetInitialPosition();
				try {
					if (dragEnhanced) {
						SR=LinearTransformation.Multiply(R3SpaceRotation(vr),SR);
						Poly.apply(SR);
					} else {
						Poly.resetInitialPosition();
						rot.x=NRotx.getDouble()*Math.PI/180;
						rot.y=NRoty.getDouble()*Math.PI/180;
						rot.z=NRotz.getDouble()*Math.PI/180;
						Poly.apply(R3SpaceRotation(rot));
					}
				} catch (Exception e) {
				}

				Poly.updateView(view);
				
				Graphics2D g2=(Graphics2D)gmain;
				if (BasicStr.antialias) {
				    g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
				    g2.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
				    g2.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
				    g2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE);
			    	g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
					g2.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
				    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
				}


				if (renderMethod==sortAndPaint) {
					Poly.sortAndPaint(D,g2);
				} else if (renderMethod==painterAlgorithm) {
					Poly.painterAlgorithm(D,g2);
				} else if (renderMethod==rayTracing) {
					Poly.rayTrace(D,g2,w,h);
				}
	
				if (BasicStr.antialias) {
				    g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);
				    g2.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_DEFAULT);
				    g2.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_DEFAULT);
				    g2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_DEFAULT);
			    	g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
					g2.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_DEFAULT);
				    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
				}
			}
			drawBackControls();
//			drawControls(bga,fga);
			super.doFront(calc);
		}
	}

	private synchronized void moveOneStep() {
		if (!dragEnhanced) {
			NRotx.setDouble(NRotx.getDouble()+vr.x*180/Math.PI);
			NRoty.setDouble(NRoty.getDouble()+vr.y*180/Math.PI);
			NRotz.setDouble(NRotz.getDouble()+vr.z*180/Math.PI);
		}
	}

	double correctedDepth(double depth) {
		return Math.max(depth,0.25*(double)(w+h)); // cambiada de 0.5 a 0.1 el 12 de julio de 2011 
	}

	protected boolean onDrag(boolean right,int dx,int dy) {
		if (right) {
			Nesc.setDouble(Nesc.getDouble()*((double)h-(double)dy)/(double)h);
			correctScale();
			view.setScale(Nesc.getDouble());
			Ndepth.setDouble(correctedDepth(Ndepth.getDouble()*((double)w+(double)dx)/(double)w));
			view.setView(Ox,Oy,Ndepth.getDouble());
			D.updateTextFieldsFromNodes();
			return backColor.variable();
		} else if (dx*dx+dy*dy>0) {
			vr.x=0;
			vr.y=((double)dy*oneDegree)/2.0;	
			vr.z=((double)dx*oneDegree)/2.0;
			moveOneStep();
			return true;	// cambiado a true en lugar de false el 12 de julio de 2007
		} else {
			return false;
		}
	}

	protected void writeCoordinates(Graphics g,int x,int y) {}

/* ----------------------- Private Methods ------------------------ */

	private static LinearTransformation R3SpaceRotation(R3 r) {
//		LinearTransformation Rx=LinearTransformation.Rotation(0,r.x);
		LinearTransformation Ry=LinearTransformation.Rotation(1,r.y);
		LinearTransformation Rz=LinearTransformation.Rotation(2,r.z);
		return LinearTransformation.Multiply(Ry,Rz);
//		return LinearTransformation.Multiply(Rx,LinearTransformation.Multiply(Ry,Rz));
	}

	public static String canonicType(translator Tr,String s) {
		for (int i=0;i<data.gtR3ix.length;i++) {
			if (Tr.equals(s,data.gtR3ix[i])) {
				return Tr.getTr(data.gtR3ix[i],1);
			}
		}
		return s;
	}

	static int getModel(translator Tr,String model) {
		if (Tr.equals(model,data.light)) {
			return viewModel.LIGHT;
		} else if (Tr.equals(model,data.metal)) {
			return viewModel.METAL;
		} else if (Tr.equals(model,data.wire)) {
			return viewModel.WIRE;
		} else { 		// or if (Tr.equals(model,data.plain))
			return viewModel.COLOR;
		}
	}

}