/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.descedit;

import java.awt.*;
import java.awt.event.*;
import com.mja.gui.*;
import com.mja.file.mjaFont;

public class editTextDialog extends editDialog implements ItemListener {

	private static final String[] fNames={"SansSerif","Serif","Monospaced"};
	private CheckboxGroup cbgName;
	private Checkbox[] cbName;
	private Checkbox cbItalic, cbBold;
	private Choice chSize;

	private Font font;
	private TextArea TA;

	private Panel NP;

	public editTextDialog(Frame parent,String title,boolean modal,String okLabel,String cancelLabel) {
		super(parent,title,modal,okLabel,cancelLabel);
		add("Center",TA=new TextArea(8,48));
		font=mjaFont.AWTMono;
		TA.setFont(font);
		NP=new Panel();
		NP.setLayout(new GridLayout(1,3));
		add("North",NP);
		Panel nameP=new Panel();
		nameP.setLayout(new GridLayout(fNames.length,1));
		cbgName = new CheckboxGroup();
		cbName=new Checkbox[fNames.length];
		for (int i=0;i<cbName.length;i++) {
			nameP.add(cbName[i]=new Checkbox(fNames[i], cbgName, false));
		}
		NP.add(nameP);
		Panel styleP=new Panel();
		styleP.setLayout(new GridLayout(2,1));
		styleP.add(cbItalic=new Checkbox("K"));
		styleP.add(cbBold=new Checkbox("N"));
		NP.add(styleP);
		NP.add(chSize=new Choice());
		makeChSz();
		for (int i=0;i<cbName.length;i++) {
			cbName[i].addItemListener(this);
			cbName[i].setFont(mjaFont.AWTMono);
		}
		cbItalic.addItemListener(this);
		cbItalic.setFont(mjaFont.AWTMono.deriveFont(Font.ITALIC,24));
		cbBold.addItemListener(this);
		cbBold.setFont(mjaFont.AWTMono.deriveFont(Font.BOLD,24));
		updateAll();
	}
	
	private void makeChSz() {
		NP.remove(chSize);
		NP.add(chSize=new Choice());
		chSize.addItemListener(this);
		chSize.setFont(mjaFont.AWTMono);
		for (int i=6;i<24;i++) {
			chSize.addItem(String.valueOf(i));
		}
		for (int i=24;i<48;i+=2) {
			chSize.addItem(String.valueOf(i));
		}
		for (int i=48;i<=80;i+=4) {
			chSize.addItem(String.valueOf(i));
		}
		NP.invalidate();
		NP.repaint();
	}

	public Font getSelectedFont() {
		return font;
	}

	public void setSelectedFont(Font font) {
		this.font=font;
		updateAll();
	}

	public void setText(String s) {
		TA.setText(s);
	}

	public String getText() {
		return TA.getText();
	}

	private void updateAll() {
		cbgName.setSelectedCheckbox(cbName[0]);
		for (int i=0;i<cbName.length;i++) {
			if (cbName[i].getLabel().equals(font.getName())) {
				cbgName.setSelectedCheckbox(cbName[i]);
				break;
			}
		}
		cbItalic.setState(font.isItalic());
		cbBold.setState(font.isBold());
		chSize.select(String.valueOf(font.getSize()));
		TA.setFont(font);
	}

	private String selectedName() {
		Checkbox cb=cbgName.getSelectedCheckbox();
		for (int i=0;i<cbName.length;i++) {
			if (cbName[i]==cb) {
				return cbName[i].getLabel();
			}
		}
		return cbName[0].getLabel();
	}

	private int selectedStyle() {
		int style=Font.PLAIN;
		if (cbItalic.getState()) {
			style+=Font.ITALIC;
		}
		if (cbBold.getState()) {
			style+=Font.BOLD;
		}
		return style;
	}
		
	private int selectedSize() {
		return Integer.parseInt(chSize.getSelectedItem());
	}

	public void itemStateChanged(ItemEvent e) {
		int ssz=selectedSize();
		font=new Font(selectedName(),selectedStyle(),ssz);
		TA.setFont(font);
	}

    public void windowActivated(WindowEvent e){
		setSelectedFont(getSelectedFont());
		super.windowActivated(e);
	}

}