/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.math;
/*------------------------------------------------------------*/
/*                            R2                              */
/*                         (c) MJA                            */
/*                     Septiembre de 1998                     */
/*------------------------------------------------------------*/
import com.mja.parser.*;

public class R2 {

	public double x,y;

	public R2() {
		x=0;y=0;
	}
	
	public R2(double x,double y) { 
		this.x=x; this.y=y; 
	}

	public R2 copy() { 
		return new R2(x,y); 
	}

	public int ix() { 
		return (int)Math.round(Math.max(Math.min(x,32000),-32000));
	}

	public int iy() { 
		return (int)Math.round(Math.max(Math.min(y,32000),-32000));
	}

	public boolean equals(R2 p) { 
		return ((x==p.x)&&(y==p.y)); 
	}

	public double norm2() { 
		return x*x+y*y; 
	}

	public double norm() { 
		return Math.sqrt(norm2()); 
	}

	public double distance(R2 p) { 
		R2 q=copy(); 
		q.sub(p); 
		return q.norm(); 
	}

	public double dot(R2 p) { 
		return x*p.x+y*p.y; 
	}

	public double det(R2 p) { 
		return x*p.y-y*p.x; 
	}

	public void mul(double m) { 
		x*=m; y*=m; 
	}

	public void div(double d) { 
		x/=d; y/=d; 
	}

	public void add(R2 p) { 
		x+=p.x; y+=p.y; 
	}

	public void sub(R2 p) { 
		x-=p.x; y-=p.y; 
	}

	public void normalize() {
		double abs=norm(); 
		if (abs!=0.0) { 
			div(abs); 
		} 
	}

	public void rotR90() { 
		double aux=x; x=y; y=-aux; 
	}

	public void rotL90() { 
		double aux=x; x=-y; y=aux; 
	}

	public void rot(double t) { 
		R2 p=copy();
		double cost=Math.cos(t);
		double sint=Math.sin(t);
		x = p.x*cost - p.y*sint;
		y = p.x*sint + p.y*cost;
	}

	public void rot(int g) { 
		rot((double)g*Math.PI/180.0); 
	}

	
	public static R2 intersection(R2 p1,R2 p2,R2 q1,R2 q2) {
/*
		double dpx=p2.x-p1.x;
		double dpy=p2.y-p1.y;
		double dqx=q2.x-q1.x;
		double dqy=q2.y-q1.y;
		double delta=dpx*dqy-dpy*dqx;
		if (Math.abs(delta)>0.000001){
			double t=(dqx*(q1.y-p1.y)+dqy*(p1.x-q1.x))/delta;
			if ((0<=t)&&(t<=1)) {
				return new R2(p1.x+t*dpx,p1.y+t*dpy);
			}
		} else {
			return null;
		}
		return null;
*/		
		double A11=(p2.x-p1.x), A12=(q1.x-q2.x), B1=q1.x-p1.x;
		double A21=(p2.y-p1.y), A22=(q1.y-q2.y), B2=q1.y-p1.y;
		double Det=A11*A22-A12*A21;
		if (Math.abs(Det)>0.000001) {
			double s=( B1*A22-B2*A12)/Det;
			double t=(-B1*A21+B2*A11)/Det;
			if (0<=s && s<=1 && 0<=t && t<=1) {
				return new R2(p1.x+A11*s,p1.y+A21*s);
			} else {
				return null;
			}
		} else if ( (p2.x-q1.x)*B2 != (p2.y-q1.y)*B1 ) {
			return null; // no estn alineados
		} else { // estn alineados
			if (p1.x!=p2.x) {
				double mp=Math.min(p1.x,p2.x);
				double Mp=Math.max(p1.x,p2.x);
				if (mp<=q1.x && q1.x<=Mp) {
					return q1;
				} else if (mp<=q2.x && q2.x<=Mp) {
					return q2;
				}
				return null;
			} else if (q1.x!=q2.x) {
				double mq=Math.min(q1.x,q2.x);
				double Mq=Math.max(q1.x,q2.x);
				if (mq<=p1.x && p1.x<=Mq) {
					return p1;
				} else if (mq<=p2.x && p2.x<=Mq) {
					return p2;
				}
				return null;
			} else if (p1.y!=p2.y) {
				double mp=Math.min(p1.y,p2.y);
				double Mp=Math.max(p1.y,p2.y);
				if (mp<=q1.y && q1.y<=Mp) {
					return q1;
				} else if (mp<=q2.y && q2.y<=Mp) {
					return q2;
				}
				return null;
			} else if (q1.y!=q2.y) {
				double mq=Math.min(q1.y,q2.y);
				double Mq=Math.max(q1.y,q2.y);
				if (mq<=p1.y && p1.y<=Mq) {
					return p1;
				} else if (mq<=p2.y && p2.y<=Mq) {
					return p2;
				}
				return null;
			} else if (p1.x==q1.x && p1.y==q1.y) {
				return p1;
			} else {
				return null;
			}
		}
	}


}