/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.math3D;

public class R3 { 

	public static final double epsilon=0.00000001;

	public double x,y,z;
	
	public String toString() {
		return "("+x+","+y+","+z+")";
	}

	public R3(double x,double y, double z) {
		this.x=x; this.y=y; this.z=z; 
	}
	
	public R3 cloneR3() {
		return new R3(x,y,z);
	}

	/* ---- operaciones internas ---- */

	public void add(R3 q) { 
		x+=q.x; y+=q.y; z+=q.z;
	}

	public void subtract(R3 q) { 
		x-=q.x; y-=q.y; z-=q.z;
	}

	public void rotX(double alfa) { 
		double c=Math.cos(alfa); 
		double s=Math.sin(alfa);
		double aux=y; 
		y = y*c-z*s; z = aux*s+z*c;
	}

	public void rotY(double alfa) { 
		double c=Math.cos(alfa);
		double s=Math.sin(alfa);
		double aux=z; 
		z = z*c-x*s; x = aux*s+x*c;
	}

	public void rotZ(double alfa) { 
		double c=Math.cos(alfa);
		double s=Math.sin(alfa);
		double aux=x; 
		x = x*c-y*s; y = aux*s+y*c;
	}

	/* ----------- funciones ------------- */

	public boolean equals(R3 p) { 
		return (x==p.x && y==p.y && z==p.z);
	}

	public boolean almostEquals(R3 p) { 
		return equals(p,epsilon);
	}

	public boolean equals(R3 p,double epsilon) { 
		return (Math.abs(x-p.x)<=epsilon && Math.abs(y-p.y)<=epsilon && Math.abs(z-p.z)<=epsilon);
	}

	public double Norm() { 
		return Math.sqrt(x*x+y*y+z*z);
	}

	public R3 Negative() { 
		return new R3(-x,-y,-z);
	}

	public R3 Add(R3 q) { 
		return new R3(x+q.x,y+q.y,z+q.z);
	}

	public R3 Subtract(R3 q) { 
		return new R3(x-q.x,y-q.y,z-q.z);
	}

	public R3 ScalarProduct(double sc) { 
		return new R3(sc*x,sc*y,sc*z);
	} 

	public R3 CrossProduct(R3 q) { 
		return new R3(y*q.z-z*q.y,z*q.x-x*q.z,x*q.y-y*q.x);
	}

	public double DotProduct(R3 q) { 
		return x*q.x+y*q.y+z*q.z;
	}

	public static double DotProduct(R3 p,R3 q) { 
		return p.x*q.x+p.y*q.y+p.z*q.z;
	}

	public static double Determinant(R3 U,R3 V,R3 W) {
		return U.DotProduct(V.CrossProduct(W));
	}

	public R3 RotX(double alfa) { 
		R3 r=new R3(x,y,z);
		double c=Math.cos(alfa);
		double s=Math.sin(alfa);
		r.y= y*c-z*s;
		r.z= y*s+z*c;
		return r;
	}

	public R3 RotY(double alfa) { 
		R3 r=new R3(x,y,z);
		double c=Math.cos(alfa);
		double s=Math.sin(alfa);
		r.z= z*c-x*s;
		r.x= z*s+x*c;
		return r;
	}

	public R3 RotZ(double alfa) { 
		R3 r=new R3(x,y,z);
		double c=Math.cos(alfa);
		double s=Math.sin(alfa);
		r.x= x*c-y*s;
		r.y= x*s+y*c;
		return r;
	}

	public static R3 reflectedRay(R3 l,R3 uN) { 
		R3 t=l.Subtract(uN.ScalarProduct(l.DotProduct(uN))); 
		R3 r=l.Add(t.ScalarProduct(-2.0));
		double N=r.Norm();
		if (N!=0.0) { 
			return r.ScalarProduct(1.0/N);
		} else {
			return l.Negative();
		}
	} 

}