/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.poly;

import java.util.*;
import java.awt.*;
import com.mja.math3D.*;
import com.mja.text.*;
import com.mja.gui.*;
import com.mja.parser.Node;
import com.mja.util.BasicStr;

public class Surface { 

	private Face[] face, ordface, drawface;
	private R3[] vertex, vertex0;
	
	public Surface(Face[] f) {
		Vector V=new Vector();
		for (int i=0;i<f.length;i++) {
			for (int j=0;j<f[i].P.length;j++) {
				if (!V.contains(f[i].P[j])) {
					boolean add=true;
					Enumeration e=V.elements();
					while (e.hasMoreElements()) {
						R3 test=(R3) e.nextElement();
						if (test.Subtract(f[i].P[j]).Norm()<=Face.epsilon) {
							add=false;
							f[i].P[j]=test;
							break;
						}
					}
					if (add) {
						V.addElement(f[i].P[j]);
					}
				}
			}
		}
		vertex=new R3[V.size()];
		vertex0=new R3[vertex.length];
		Enumeration e=V.elements();
		for (int i=0;i<vertex.length;i++) {
			vertex[i]=(R3) e.nextElement();
			vertex0[i]=vertex[i].cloneR3(); 
		}
		face=f;
		ordface=new Face[face.length];
		for (int i=0;i<face.length;i++) { 
			ordface[i]=face[i]; 
		}
	}
	
	private Surface() {
	}
	
	public Surface cloneSurface() {
		Face[] f=new Face[face.length];
		for (int i=0;i<f.length;i++) {
			f[i]=face[i].cloneFace();
		}
		Surface S=new Surface(f);
		return S;
	}
	
	public void setSValue(double sv) {
		for (int i=0;i<face.length;i++) {
			face[i].setSValue(sv);
		}
	}
	public void setSNode(Node sN) {
		for (int i=0;i<face.length;i++) {
			face[i].setSNode(sN);
		}
	}

	void setMathText(MathText mt) {
		if (mt!=null) {
			for (int i=0;i<face.length;i++) {
				face[i].setMathText(mt);
			}
		}
	}
	
	public int countFaces() {
		return face.length;
	}

	public int countVertices() {
		return vertex.length;
	}

	public void fixAsInitialPosition() {
		for (int i=0;i<vertex.length;i++) { 
			vertex0[i].x=vertex[i].x;
			vertex0[i].y=vertex[i].y;
			vertex0[i].z=vertex[i].z; 
		}
	}
	public void resetInitialPosition() {
		for (int i=0;i<vertex0.length;i++) { 
			vertex[i].x=vertex0[i].x;
			vertex[i].y=vertex0[i].y;
			vertex[i].z=vertex0[i].z; 
		}
	}

/* ---------------------- Face by Face Operations -------------------- */	
	
	public void apply(LinearTransformation LT) { 
		for (int i=0; i<vertex.length; i++) { 
			LT.Apply(vertex[i]); 
		}
		for (int i=0;i<face.length;i++) {
			face[i].calculateUnitNormal();
		}
	}
	
	private viewR3 view;
	
	public void updateView(viewR3 view) { 
		this.view=view;
		for (int i=0;i<face.length;i++) {
			face[i].updateView(view);
		}
	}
	
	public void setViewModel(int model) {
		viewModel vM=new viewModel();
		vM.setModel(model);
		for (int i=0;i<face.length;i++) {
			face[i].setViewModel(vM);
		}
	}

	public void setUserLightPos(int x,int y,int z) {
		for (int i=0;i<face.length;i++) {
			face[i].getViewModel().setUserLightPos(x,y,z);
		}
	}
	
	public void setUserIllumination(double dim,double uI) {
		for (int i=0;i<face.length;i++) {
			face[i].getViewModel().setUserIllumination(dim,uI);
		}
	}

	void setEdgeColor(Color c) {
		for (int i=0;i<face.length;i++) {
			face[i].setEdgeColor(c);
		}
	}
	
	void setFrontColor(Color c) {
		for (int i=0;i<face.length;i++) {
			face[i].setFrontColor(c);
		}
	}

	void setBackColor(Color c) {
		for (int i=0;i<face.length;i++) {
			face[i].setBackColor(c);
		}
	}

	public void setDrawEdges(boolean oe) {
		for (int i=0;i<face.length;i++) {
			face[i].setDrawEdges(oe);
		}
	}
	
/* -------------------------- Surface Operations -------------------- */	

	public static Surface add(Surface s1,Surface s2) {
		if (s1==null) {
			return s2;
		} else if (s2==null) {
			return s1;
		} else {
			return add(s1,s2.face);
		}
	}
	
	public static Surface add(Surface S,Face[] fa) {
		if (S==null) {
			return new Surface(fa);
		} else if (fa==null) {
			return S;
		} else {
			Face[] Fa=new Face[S.face.length+fa.length];
			int i=0;
			while(i<S.face.length) { 
				Fa[i]=S.face[i++]; 
			}
			int j=0;
			while(j<fa.length) { 
				Fa[i++]=fa[j++]; 
			}
			return new Surface(Fa);
		}
	}

	public Surface split(Surface S) {
		for (int i=0;i<face.length;i++) {
			Vector VF=new Vector();
			for (int j=0;j<S.face.length;j++) {
				Face[] fa=face[i].split(S.face[j]);
				for (int k=0;k<fa.length;k++) {
					VF.addElement(fa[k]);
				}
			}
			Face[] FA=new Face[VF.size()];
			Enumeration e=VF.elements();
			for (int n=0;n<FA.length;n++) {
				FA[n]=(Face) e.nextElement();
			}
			S=new Surface(FA);
		}
		return S;
	}

/* ----------------------- Simple Sort Algorithm ---------------------- */


/*
	void sort() {
		int klo=0, khi=0, k=0, j=1;
		while (true) { 
			if (j>=face.length) break;
			if (!LTHOE(j-1,j)) {  
				if (!LTHOE(0,j)) { 
					INSERT(j,0); 
				} else { 
					klo=0; khi=j-1;
					while (true) {
						if (klo>khi) {
							k=klo; 
							break; 
						}
						k=(klo+khi)/2;
						if (LTHOE(k,j)) {
							klo=k+1;
						} else {
							khi=k-1;
						}
					}
					INSERT(j,k); 
				}
			}
			j++;
		}
//		System.out.println("BI");
//		for (int i=0;i<ordface.length;i++) {
//			System.out.println(ordface[i]);
//		}

	}

	void INSERT(int i,int j) { 
		Face aux=ordface[i];
		System.arraycopy(ordface,j,ordface,j+1,i-j);
		ordface[j]=aux; 
	}

	boolean LTHOE(int i,int k) { 
		if (k>=face.length) { 
			return true; 
		} else {
			return ordface[k].isCloserThan(ordface[i]);
		}
	} 
	
*/

	public String toTeX(Component C,boolean grayscale,double SCL, int W, int H, double DX, double DY,String textsInTeX) {
		String str=BasicStr.getTeX_Header(W,H);
		str=str+BasicStr.getTeX_BeginDoc(SCL,W,H,DX,DY);
		for (int n=0;n<drawface.length;n++) {
			str=str+drawface[n].toTeX(C,n,2*H-10,grayscale);
		}
		str=str+textsInTeX;
		str=str+BasicStr.TeX_Tail;
		return str;
	}
	
	private void QuickSort() {
		QS(0,face.length-1);
//		System.out.println("QS");
//		for (int i=0;i<ordface.length;i++) {
//			System.out.println(ordface[i]);
//		}
	}

	private void QS(int i1, int i2) {
		int j1=i1,j2=i2;
		double pivot=ordface[i1].aveDistanceToEye+(ordface[i2].aveDistanceToEye-ordface[i1].aveDistanceToEye)/2;
		boolean fin=false;
		while (!fin) {
			boolean LR=(pivot>ordface[j1].aveDistanceToEye);
			boolean RR=(ordface[j2].aveDistanceToEye>pivot);
			boolean AR=LR&RR;
			if (AR) {
				Face temp=ordface[j1];
				ordface[j1]=ordface[j2];
				ordface[j2]=temp;
			}
			if (AR||!LR) j1++;
			if (AR||!RR) j2--;
			fin=(j2<=j1);
			if (fin&&(i1<j2)&&(j2!=i2)) QS(i1,j2);
			if (fin&&(j1<i2)&&(j1!=i1)) QS(j1,i2);
		}	
	}

	
/* ----------------------- Order Algorithm ---------------------- */

	public void sortAndPaint(Container C,Graphics2D g) { 
		QuickSort();
//		drawface=ordface;
		for (int i=0;i<face.length;i++) {
			ordface[i].draw(C,g);
		}
	}

/* ----------------------- Painter's Algorithm ---------------------- */

	public void painterAlgorithm(Container C,Graphics2D g) { 
		setDrawingOrder();
		for (int i=0;i<face.length;i++) {
			drawface[i].draw(C,g);
		}
	}

	private void setDrawingOrder() {
		QuickSort();
		Vector V=new Vector();
		for (int i=0;i<ordface.length;i++) {
			ordface[i].drawn=false;
		}
		drawface=new Face[ordface.length];
		int drawix=0;
		int NC=ordface.length;
		double epsilon0=0.001;
		double epsilon=epsilon0;
		while (true) {
			int NCa=NC;
			boolean oneDrawn=false;
			for (int i=0;i<ordface.length;i++) {
				if (!ordface[i].drawn) {
					boolean canDraw=true;
					for (int j=0;j<ordface.length;j++) {
						if ((j!=i) 
							&& (!ordface[j].drawn) 
							&& (ordface[i].inFrontOf(V,ordface[j],epsilon))
							) {
							canDraw=false;
							break;
						}
					}
					if (canDraw) {
						NC--;
						drawface[drawix++]=ordface[i];
						ordface[i].drawn=true;
						oneDrawn=true;
					}
				}
			}
			if (NC==0) { // PA ended CORRECTLY"
				break;
			} else if (NC==NCa) { // Can't continue;
				epsilon=epsilon*10;
				if (epsilon>0.1) {
					System.out.println("Error in Painter Algorithm");				
					for (int i=0;i<ordface.length;i++) {
						if (!ordface[i].drawn) {
							drawface[drawix++]=ordface[i];
							ordface[i].drawn=true;
						}
					}
					break;
				}
			} else {
				epsilon=epsilon0;
			}
		}
	}

/* -------------------------- Ray Tracing ----------------------- */	

	public void rayTrace(Container C,Graphics2D g,int w, int h) { 
		QuickSort();
		refine(g,0,0,w,h,64);
		for (int i=0;i<ordface.length;i++) {
			ordface[i].drawText(C,g);
		}
	}
	
	private void refine(Graphics g,int x0, int y0, int w, int h, int sz) {
		if (sz<4) { 
			sz=1; 
		}
		for (int n=y0;n<y0+h;n+=sz) {
			for (int m=x0;m<x0+w;m+=sz) {
				if ( touches(m,n,sz) ) {
					if (!atomPainted(g,m,n,sz) ) {
						refine(g,m,n,sz,sz,sz/2);
					}
				}
			}
		}
	}

	private boolean touches(int x0, int y0, int sz) {
		for (int i=ordface.length-1;i>=0;i--) {
			if (ordface[i].touches(x0,y0,sz)) {
				return true;
			}
		}
		return false;
	}
	
	private boolean atomPainted(Graphics g,int x0, int y0, int sz) {
		double d=Double.MAX_VALUE;
		int ix=-1;
		if (sz==1) {
			for (int i=ordface.length-1;i>=0;i--) {
				if (ordface[i].contains(x0,y0,sz)) {
					double t=ordface[i].intersectionToEye(x0,y0);
					if (t<d) {
						d=t;
						ix=i;
					}
				}
			}
		} else {
			double[] min=new double[ordface.length];
			double[] max=new double[ordface.length];
			boolean[] contained=new boolean[ordface.length];
			for (int i=ordface.length-1;i>=0;i--) {
				contained[i]=ordface[i].contains(x0,y0,sz);
				if (contained[i]) {
					if ( ordface[i].getDrawEdges() && 
						(ordface[i].isOnEdge(x0,y0)||ordface[i].isOnEdge(x0+sz-1,y0)||
						 ordface[i].isOnEdge(x0,y0+sz-1)||ordface[i].isOnEdge(x0+sz-1,y0+sz-1)) ) {
						return false;
					}
					double t00=ordface[i].intersectionToEye(x0     ,y0   );
					double t10=ordface[i].intersectionToEye(x0+sz-1,y0   );
					double t01=ordface[i].intersectionToEye(x0     ,y0+sz-1);
					double t11=ordface[i].intersectionToEye(x0+sz-1,y0+sz-1);
					min[i]=Math.min(t00,Math.min(t10,Math.min(t01,t11)));
					max[i]=Math.max(t00,Math.max(t10,Math.max(t01,t11)));
					for (int j=i+1;j<ordface.length;j++) {
						if (contained[j]) {
							if ( (min[i]<=min[j]&&min[j]<=max[i])
							  || (min[i]<=max[j]&&max[j]<=max[i])
							  || (min[j]<=min[i]&&min[i]<=max[j])
							  || (min[j]<=max[i]&&max[i]<=max[j]) ) {
								return false;
							}
						}
					}
					if (ordface[i].contains(x0,y0,sz)) {
						if (t00<d) {
							d=t00;
							ix=i;
						}
					}
				} else if (ordface[i].touches(x0,y0,sz) ) {
					return false;
				} 
			}
		}
		if (ix>=0) {
			g.setColor(ordface[ix].getColorAt(x0,y0,sz==1));
			g.fillRect(x0,y0,sz,sz);
		}
		return true;
	}
	
}