/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.text;

import java.awt.*;
import java.util.*;
import java.awt.datatransfer.*;

import com.mja.parser.*;
import com.mja.util.*;
import com.mja.file.mjaFont;

public class BasicText implements ClipboardOwner {

	protected Vector info;
	private TFont font;
	private TColors colors=new TColors(Color.white,Color.black);
	public Parser parser;
	private int initialTNodes,finalTNodes,segBegin,segEnd;
	private boolean segMarked=false;
	public static Vector textBuffer;
	private int cursorPos;

	public BasicText(Parser p) {
		this(p,null,null,null);
	}

	public BasicText(Parser p,TFont f) {
		this(p,f,null,null);
	}

	public BasicText(Parser p,TFont f,Color bc,Color tc) {
		this.parser=p;
		colors=new TColors(bc,tc);
		if (f!=null) {
			font=f;
		}  else {
			font=new TFont(mjaFont.makeFont("Serif",Font.PLAIN,24));
		}
		info=new Vector();
		initialTNodes=0;
		cursorPos=initialTNodes;
		info.addElement(new TNode(this,"\u001A"));
		finalTNodes=1;
//		insert(new TNode(this,color),cursorPos);
//		insert(new TNode(this,font),cursorPos);
	}

	public void println() {
		for (int j=0;j<info.size();j++) {
			System.out.print(j+" "+TNode(j).toExpresion(true)+";");
		}
		System.out.println("");
	}

	public void setParser(Parser parser) {
		this.parser=parser;
		for (int i=0;i<info.size();i++) {
			TNode tn=(TNode) info.elementAt(i);
			tn.setParser(parser);
		}
	}

	Parser getParser() {
		return parser;
	}

	public String toString() {
		String s=super.toString()+"\n";
		for (int i=0;i<info.size();i++) {
			TNode tn=(TNode) info.elementAt(i);
			s+=tn.toString();
		}
		return s;
	}

	void setInfo(Vector v) {
		info=v;
	}

	Vector getInfo() {
		return info;
	}


    Font getFont() {
    	return font.getMjaFont().getAwtFont();
    }

	public TFont getTFont() {
		return font;
	}

	public void setTFont(TFont tf) {
		font=tf;
	}

	public TColors getColors() {
		return colors;
	}

	public void setColors(Color bc,Color fc) {
		colors=new TColors(bc,fc);
	}

	public TNode TNode(int i) {
		synchronized (info) {
			if (i<info.size()) {
				return (TNode) info.elementAt(i);
			} else {
				new Exception("info.size()="+info.size()+" <= "+i).printStackTrace();
				return null;
			}
		}
	}

	public String getTeXColors(boolean grayscale) {
		String str="";
		int ncolor=0;
		for (int i=0;i<end();i++) {
			TNode tn=TNode(i);
			if (tn.isColors()) {
				TColors tc=tn.getColors();
				Color textcolor=tc.getTextColor();
				String colorname="tcolor"+BasicStr.justify(""+ncolor,-3,'0');
				ncolor++;
				str=str+BasicStr.DefineColor(grayscale,colorname,textcolor);
//				"\\definecolor{"+colorname+"}{rgb}{"+DblStr(textcolor.getRed()/256.0)+","+DblStr(textcolor.getGreen()/256.0)+","+DblStr(textcolor.getBlue()/256.0)+"}\r\n";
			}
		}
		return str;
	}

	public int getCursorPos() {
		return cursorPos;
	}

	public void setCursorPos(int pos) {
		cursorPos=Math.min(pos,info.size()-1);;
	}

	public int indexOf(TNode tn) {
		for (int i=begin();i<end();i++) {
			if (tn==(TNode) info.elementAt(i)) {
				return i;
			}
		}
		return begin();
	}

	public void setCursorPos(TNode tn) {
		setCursorPos(indexOf(tn));
	}

	/* -------------------- marked segment ------------------ */

	public int begin() {
		return initialTNodes;
	}

	public int end() {
		synchronized (info) {
			return info.size()-finalTNodes;
		}
	}

	public int infoSize() {
		return info.size();
	}

	public int segBegin() {
		return segBegin;
	}

	public int segEnd() {
		return segEnd;
	}

	private void setSegment(int n1,int n2) {
		int min=Math.min(n1,n2);
		int max=Math.max(n1,n2);
		while (max-1>=begin() && !TNode(max-1).isChar() ) {
			max--;
		}
		if (n1==min) {
			segBegin=n1;
			segEnd=max;
		} else {
			segBegin=max;
			segEnd=n2;
		}
	}

	public boolean segMarked() {
		return segMarked;
	}

	public void markSegment(int n1,int n2) {
		setSegment(n1,n2);
		segMarked=(n1!=n2);
	}

	public void updateMarkedSegment() {
		setSegment(segBegin,cursorPos);
	}

	public void unmarkSegment() {
   		segMarked=false;
	}

    public int segMin() {
    	return Math.min(segBegin,segEnd);
    }

    public int segMax() {
    	return Math.max(segBegin,segEnd);
    }

    /* ---------------- Char and Line structure --------------- */

	/**
	 * Returns the first TNode of type CHAR in the first line.
	 */
    public int firstLine() {
    	return firstChar(initialTNodes);
    }

	/**
	 * Return the first TNode of type CHAR starting at N.
	 */
    public int firstChar(int N) {
//    	return N;
    	int n=N;
		while (n<end() && !TNode(n).isChar()) {
			n++;
		}
		if (n==end()) {
			return N;
		}
		return n;
    }

	/**
	 * Return the first TNode of type FORMULA starting at N.
	 */
    public int firstFormula(int N) {
    	int n=N;
		while (n<end() && !TNode(n).isFormula()) {
			n++;
		}
		if (n==end()) {
			return -1;
		}
		return n;
    }

	/**
	 * Returns true if N is the last TNode of type CHAR
	 */
    public boolean afterLastChar(int N) {
    	if (TNode(N).isChar()) {
    		return N==end();
    	}
    	int n=N+1;
		while (n<end()) {
			if (TNode(n).isChar()) {
				return false;
			}
			n++;
		}
		return true;
    }
	/**
	 * Returns the first TNode of type CHAR in the same line of TNode N.
	 */
    public int startOfLine(int N) {
    	int n=N;
    	while (n>initialTNodes &&(TNode(n-1)!=null)&& (!TNode(n-1).isEndOfLine()) ) {
    		n--;
    	}
    	return firstChar(n);
    }

    public int endOfLine(int n) {
    	int i=n;
    	while (i<end() && (!TNode(i).isEndOfLine()) ) {
    		i++;
    	}
    	return i;
    }

    public int nextChar(int N) {
    	int n=firstChar(N);
    	if (n<end()) {
    		n++;
    	}
    	return firstChar(n);
    }

    public int prevChar(int N) {
    	int n=N;
    	if (n>initialTNodes) {
    		n--;
    	}
		while (n>initialTNodes && !TNode(n).isChar()) {
			n--;
		}
		return firstChar(n);
    }

    public int nextLine(int n) {
    	return firstChar(startOfLine(endOfLine(Math.min(end(),endOfLine(n)+1))));
    }

    public int previousLine(int n) {
    	return firstChar(startOfLine(Math.max(begin(),prevChar(startOfLine(n)))));
    }

	/* -------------------- editing functions ------------------- */

    public void insertString(String s) {
    	String str=new String(s);
    	if (BasicStr.classExists("com.mja.parser.Node") && Node.inAlgebraTutor) {
	    	str=BasicStr.replace(str,"eqn","#");
	    	str=BasicStr.replace(str,"eqa","@");
	    	str=BasicStr.replace(str,"eqx","\u007e");
    	}
		for (int i=0;i<str.length();i++) {
			char ch=str.charAt(i);
			if (ch!='\r') {
				insertChar(ch);
			}
		}
    }

    void insertChar(char ch) {
    	insertText((new Character(ch)).toString());
    }

    void insertText(String s) {
		insert(new TNode(this,s),cursorPos);
	}

    boolean delete() {
		int N=firstChar(cursorPos);
		if (N<end()) {
			return removeNode(N);
		}
    	return false;
    }

    boolean backSpace() {
		int N=prevChar(firstChar(cursorPos));
		if (N!=cursorPos && N<end()) {
			return removeNode(N);
		}
    	return false;
    }

	void removeMarkedSegment() {
		if (segMarked) {
			remove(segMin(),segMax());
	    	segMarked=false;
		}
	}

    /* ---------------- editing methods ------------- */

    public void clean () {
    	TFont f=getTFont();
    	TColors tc=getColors();
		int i=begin();
		int ia=i;
		while ( i<end()) {
			TNode ti=TNode(i);
			if (ti.isColors()) {
				TColors tc1=ti.getColors();
				if (tc1.equals(tc)) {
					removeNode(i);
				} else {
					tc=tc1;
				}
			} else if (ti.isTFont()) {
				TFont f1=ti.getTFont();
				if (f1.equals(f)) {
					removeNode(i);
				} else {
					f=f1;
				}
			} else {
				ti.clean();
			}
			if (TNode(i)==ti) {
				i++;
			}
		}
		boolean validFont=true;
		boolean validColor=true;
		for (i=end();i>=begin();i--) {
			TNode tn=TNode(i);
			if (tn.isColors()) {
				if (!validColor) {
					removeNode(i);
				}
				validColor=false;
			} else if (tn.isTFont()) {
				if (!validFont) {
					removeNode(i);
				}
				validFont=false;
			} else if (tn.isChar()) {
				validFont=true;
				validColor=true;
			}
		}
    }

	private static boolean localcopy=false;
	
	public void lostOwnership(Clipboard clipboard, Transferable contents) {
		localcopy=false;
	}
 
	public void copy(Component C) {
		textBuffer=new Vector();
		int n1=segMin();
		int n2=segMax();
		textBuffer.addElement(new TNode(this,getColorsAtTNode(n1)));
		textBuffer.addElement(new TNode(this,getTFontAtTNode(n1)));
		localcopy=false;
		String str="";
		for (int n=n1;n<n2;n++) {
			localcopy=true;
			textBuffer.addElement(TNode(n).cloneTNode(this));
			if (TNode(n).isPureChar()) {
				str+=TNode(n).getChar();
			} else if (TNode(n).isFormula()) {
				str+=TNode(n).toExpresion(true);
			}
			try {
				Clipboard c=C.getToolkit().getSystemClipboard();
				Transferable t=new StringSelection(str);
				c.setContents(t,this);
			} catch (Exception ex) {
			}
		}
	}

	public void paste(Component C) {
		if (localcopy) {
			if (textBuffer!=null) {
				int N=getCursorPos();
				TColors tc=getColorsAtTNode(getCursorPos());
				TFont f=getTFontAtTNode(getCursorPos());
				for (int n=0;n<textBuffer.size();n++) {
					insert(((TNode) textBuffer.elementAt(n)).cloneTNode(this),getCursorPos());
				}
				insert(new TNode(this,tc),getCursorPos());
				insert(new TNode(this,f),getCursorPos());
//				clean();
			}
		} else {
			String str="";
			try {
				Clipboard c=C.getToolkit().getSystemClipboard();
				Transferable t=c.getContents(this);
				str=(String) t.getTransferData(DataFlavor.stringFlavor);
//				System.out.println(str);
				if (BasicStr.hasContent(str)) {
					textBuffer=new Vector();
					insertString(str);
				}
			} catch (Exception ex) {
			}
		}
	}

	public BasicText getTextBuffer() {
		BasicText bt=new BasicText(parser,font,colors.getBackColor(),colors.getTextColor());
		for (int n=0;n<textBuffer.size();n++) {
			bt.insert(((TNode) textBuffer.elementAt(n)).cloneTNode(this),bt.cursorPos);
		}
		return bt;
	}

    /* ---------------- private methods ------------- */

    private boolean charInSeg(int n) {
    	return segMarked && segMin()<=n && n<segMax();
    }

    protected Vector cloneInfo() {
    	Vector v=new Vector();
    	for (int i=0;i<info.size();i++) {
    		v.addElement(((TNode)info.elementAt(i)).cloneTNode(this));
    	}
    	return v;
    }

    /* ---------------- editing methods ------------- */

	protected void insert(TNode tn,int n) {
		if (tn!=null) {
			info.insertElementAt(tn,n);
			if (tn.isComponent()) {
				if (((LiveOrDie) tn.getComponent()).isUsed()) {
					String cid=Long.toString(System.currentTimeMillis(),16);
					tn.setNewComponent(((LiveOrDie) tn.getComponent()).cloneComponent(cid),cid);
				} else {
					((LiveOrDie) tn.getComponent()).setIsUsed(true);
				}
			}
			tn.setParser(parser);
			if (n<=cursorPos) { cursorPos++; }
			if (n<=segBegin) { segBegin++; }
			if (n<=segEnd) { segEnd++; }
		} else {
			new Exception("null TNode").printStackTrace();
		}
	}

	protected boolean removeNode(int n) {
		if (n>=begin() && n<=end()) {	// 7 de abril d 2009: se incluyeron los extremos porque en las frmulas el principio es necesario para camiar color
			TNode tn=(TNode) info.elementAt(n);
			tn.removeFI();
			if (tn.isComponent()) {
				((LiveOrDie) tn.getComponent()).setIsUsed(false);
			}
			info.removeElementAt(n);
			if (n<cursorPos) { cursorPos--; }
			if (n<segBegin) { segBegin--; }
			if (n<segEnd) { segEnd--; }
			return true;
		}
		return false;
	}

	public void remove(int n1,int n2) {
		boolean somethingRemoved=false;
		int min=Math.min(n1,n2);
		int max=Math.max(n1,n2);
//		min=Math.max(min,begin()+1);
//		min=Math.max(min,begin());
//		max=Math.min(max,end());
		if (min<max) {
			for (int i=max-1;i>=min;i--) {
				somethingRemoved|=removeNode(i);
			}
		}
//		return somethingRemoved;
	}

	/* -------------------- Color & Font ------------------- */

    public TColors getColorsAtTNode(int n) {
    	return getColorsAtTNode(n,true);
    }
    public TColors getColorsAtTNode(int n,boolean initColor) {
		for (int i=n-1;i>=0;i--) {
			if (TNode(i).isColors()) {
				return TNode(i).getColors();
			}
		}
		if (initColor) {
			return getColors();
		} else {
			return null;
		}
    }

    public TFont getTFontAtTNode(int n) {
    	if (this instanceof Formula) {
    		return font;
    	}
		for (int i=n-1;i>=0;i--) {
			if (TNode(i).isTFont()) {
				return TNode(i).getTFont();
			}
		}
		return getTFont();
    }

    public void calcValues(int dec,boolean fixed,int aL,boolean editing,boolean learnerMode) {
		for (int i=begin();i<end();i++) {
			if (TNode(i).isFormula()) {
				TNode(i).getFI().getFormula(0).calcValues(dec,fixed,aL,editing,learnerMode);
			}
		}
    }

	/* -------------------- Font Metrics ------------------- */

//    public FontMetrics getFontMetricsAtTNode(Component C,int n) {
//    	return getFontMetrics(C,getTFontAtTNode(n));
//    }

    public int getLineAscentAtTNode(Component C,int n) {
    	int i=startOfLine(n);
		TFont tf=getTFontAtTNode(i);
		int a=tf.getAscent();
    	while (TNode(i)!=null && !TNode(i).isEndOfLine()) {
			TNode tn=TNode(i);
	    	if (tn.isChar()) {
	    		int a0=a;
				a=Math.max(a,tn.getAscent(tf));
	    	} else if (tn.isTFont()) {
				tf=tn.getTFont();
			}
    		i++;
    	}
    	return a;
    }

    public int getLineDescentAtTNode(Component C,int n) {
    	int i=startOfLine(n);
		TFont tf=getTFontAtTNode(i);
		int d=tf.getDescent();
    	while (!TNode(i).isEndOfLine()) {
			TNode tn=TNode(i);
	    	if (tn.isChar()) {
				d=Math.max(d,tn.getDescent(tf));
	    	} else if (tn.isTFont()) {
				tf=tn.getTFont();
			}
    		i++;
    	}
    	return d;
    }

    public int getLineHeightAtTNode(Component C,int n) {
    	int A=getLineAscentAtTNode(C,n);
    	int D=getLineDescentAtTNode(C,n);
    	return A+D+1;
    }

    public int getCharWidthAtTNode(Component C,int n,boolean useExtra) {
    	TNode tn=TNode(n);
    	if (tn.isChar()) {
//	   		FontMetrics fm=getFontMetricsAtTNode(C,n);
    		return tn.getCharWidth(getTFontAtTNode(n),C,useExtra);
    	}
    	return 0;
    }

    public int getCharVisualLimitAtTNode(Component C,int n,boolean useExtra) {
    	return getCharVisualLimitAtTNode(C,n,useExtra,false);
    }
    
    public int getCharVisualLimitAtTNode(Component C,int n,boolean useExtra,boolean strict) {
    	TNode tn=TNode(n);
    	if (tn.isChar()) {
	   		TFont tf=getTFontAtTNode(n);
	   		int w=tn.getCharWidth(tf,C,useExtra);
	   		if (tn.isFormula()) {
	   			return w-(int)Math.round((1.0*tf.getMjaFont().stringWidth(C,"a"))/3.0);
	   		} else {
	   			if (strict) {
	    			return (int) Math.round(w);
	   			} else {
	    			return (int) Math.round((2.0*w)/3.0);
	   			}
    		}
    	}
    	return 0;
    }

    public static Color bestBG(Color bg) {
    	if (bg.getRed()+bg.getGreen()+bg.getBlue()<382) {
    		return Color.white;
    	} else {
    		return Color.black;
    	}
    }

    public static Color bestFG(Color bg) {
    	if (bg.getRed()+bg.getGreen()+bg.getBlue()<382) {
    		return Color.black;
    	} else {
    		return Color.white;
    	}
    }

}