/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.text;

import java.awt.*;
import java.util.*;

import com.mja.parser.*;
import com.mja.util.TFont;
import com.mja.util.LiveOrDie;

public class Formula extends Text {

	public static boolean drawBox=true;

	TNode parent;
	private boolean active,assisted;
	private Node Fnode;

	public Formula(Parser p) {
		super(p);
	}

	public Formula(Parser p,TFont f) {
		super(p,f);
	}

	public Formula(Parser p,TFont f,Color bc,Color tc) {
		super(p,f,bc,tc);
	}

	public void setAssisted(boolean b) {
		assisted=b;
		int n1=begin();
    	int n2=end();
    	for (int i=n1;i<n2;i++) {
    		TNode tn=TNode(i);
    		if (tn.isChar()&&tn.getFI()!=null) {
   				FormulaItem FI=tn.getFI();
   				FI.setAssisted(assisted);
   			}
   		}
   	}

	public boolean isAssisted() {
		return assisted;
	}

	public TNode getParent() {
		return parent;
	}

	public void setNode(Node n) {
		Fnode=n;
	}

	public Node getNode() {
		return Fnode;
	}

	public Formula subFormulaWithNode(Node N1) {
		if (getNode()!=null) {
			if (getNode()==N1) {
				return this;
			} else {
				if (getNode().getLeft()!=null&&getNode().getLeft().F!=null) {
					Formula auxF=getNode().getLeft().F.subFormulaWithNode(N1);
					if (auxF!=null) {
						return auxF;
					}
				}
				if (getNode().getRight()!=null&& getNode().getRight()[0]!=null && getNode().getRight()[0].F!=null) {
					Formula auxF=getNode().getRight()[0].F.subFormulaWithNode(N1);
					if (auxF!=null) {
						return auxF;
					}
				}
			}
		}
		return null;
	}

	public static Formula makeFormulaFromBuffer(Formula FF,Vector buffer) {
		Formula f=new Formula(FF.parser,FF.getTFont(),FF.getColors().getBackColor(),FF.getColors().getTextColor());
		if (buffer!=null) {
			for (int n=0;n<buffer.size();n++) {
				f.insert(((TNode) buffer.elementAt(n)).cloneTNode(FF),f.getCursorPos());
			}
		}
		return f;
	}

	/* -------------------- editing functions ------------------- */

	public void insertSuperIndex(String s) {
		insert(new TNode(this,s,TNode.SUPIX),getCursorPos());
	}
	public void insertSuperIndex(Formula f) {
		insert(new TNode(this,f,TNode.SUPIX),getCursorPos());
	}
	public void insertSubIndex(String s) {
		insert(new TNode(this,s,TNode.SUBIX),getCursorPos());
	}
	void insertSubIndex(Formula f) {
		insert(new TNode(this,f,TNode.SUBIX),getCursorPos());
	}
	public void insertSum(String from,String to,String what) {
		insert(new TNode(this,new FormulaItem(getParser(),from,to,what),TNode.SUM),getCursorPos());
	}
	public void insertSum(Formula from,Formula to,Formula what) {
		insert(new TNode(this,new FormulaItem(from,to,what),TNode.SUM),getCursorPos());
	}

	public void insertIntegral(String from,String to,String what) {
		insert(new TNode(this,new FormulaItem(getParser(),from,to,what),TNode.INTEGRAL),getCursorPos());
	}
	public void insertIntegral(Formula from,Formula to,Formula what) {
		insert(new TNode(this,new FormulaItem(from,to,what),TNode.INTEGRAL),getCursorPos());
	}

	public void insertLimit(String from,String to,String what) {
		insert(new TNode(this,new FormulaItem(getParser(),from,to,what),TNode.LIMIT),getCursorPos());
	}
	public void insertLimit(Formula from,Formula to,Formula what) {
		insert(new TNode(this,new FormulaItem(from,to,what),TNode.LIMIT),getCursorPos());
	}
    public void insertMatrix(String[][] sa) {
		insert(new TNode(this,sa.length,sa[0].length,new FormulaItem(getParser(),sa)),getCursorPos());
    }
    public void insertMatrix(Formula[][] fa) {
		insert(new TNode(this,fa.length,fa[0].length,new FormulaItem(fa)),getCursorPos());
    }
    public void insertDefParts(String[] sa) {
		insert(new TNode(this,sa.length,new FormulaItem(getParser(),sa)),getCursorPos());
    }
    public void insertDefParts(Formula[] fa) {
		insert(new TNode(this,fa.length,new FormulaItem(fa)),getCursorPos());
    }


//	void insertExpr(String expr) {
//		insert(new TNode(this,parser,expr),getCursorPos());
//	}

	public void insertExpr(String expr,String dec_expr,boolean fixed) {
		TNode tn=new TNode(this,parser,expr);
		if (dec_expr!=null&&dec_expr.length()>0) {
			tn.setDecExpr(dec_expr);
			tn.setFixed(fixed);
		}
		insert(tn,getCursorPos());
	}

	public void insertRadical(String index,String radicand) {
		insert(new TNode(this,new FormulaItem(getParser(),index,radicand),TNode.RADICAL),getCursorPos());
	}

	public void insertRadical(Formula index,Formula radicand) {
		insert(new TNode(this,new FormulaItem(index,radicand),TNode.RADICAL),getCursorPos());
	}

	public void insertSqrt(Formula radicand) {
		insert(new TNode(this,new FormulaItem(new Formula(getParser()),radicand),TNode.RADICAL),getCursorPos());
	}

	public void insertFraction(String num,String den) {
		insert(new TNode(this,new FormulaItem(parser,num,den),TNode.FRACT),getCursorPos());
	}

	public void insertFraction(Formula num,Formula den) {
		insert(new TNode(this,new FormulaItem(num,den),TNode.FRACT),getCursorPos());
	}

	public void insertProduct(Formula m,Formula n) {
		insert(new TNode(this,new FormulaItem(m,n),TNode.MULT),getCursorPos());
	}

	public void insertFormula(Formula f,boolean separate) {
		if (separate) {
			insert(new TNode(this,f,TNode.FORMULA),getCursorPos());
		} else {
			for (int i=f.begin();i<f.end();i++) {
				insert(f.TNode(i),getCursorPos());
			}
		}
	}

//	public void insertFormula(Formula f) {
//		insert(new TNode(this,f,TNode.FORMULA),getCursorPos());
//		for (int i=f.begin();i<f.end();i++) {
//			insert(f.TNode(i),getCursorPos());
//		}
//	}

	public void removeMarkedSegment() {
		removeSegment(segMin(),segMax());
	}

	public void removeSegment(int n1,int n2) {
		if (segMarked()) {
			TColors tc=getColorsAtTNode(n2);
			TFont f=getTFontAtTNode(segMax());
//			for (int i=n1;i<=n2;i++) {
//				TNode tn=TNode(i);
//				tn.removeFI();
//			}
			remove(n1,n2);
			insertColors(tc,false);
			insertTFont(f,false);
		}
		unmarkSegment();
	}

//	public void removeContents() {
////		for (int i=0;i<info.size();i++) {
////			TNode tn=TNode(i);
////			tn.removeFI();
////			
////		}
//		remove(0,info.size());
//	}

    /* ------------------ Graphic Text Metrics ------------------ */

	int getAscent() {
		int a=getTFont().getAscent();
    	int i=begin();
    	while (i<end()) {
			TNode tn=TNode(i);
	    	if (tn.isChar()) {
	    		a=Math.max(a,tn.getAscent(getTFont()));
	    	}
    		i++;
    	}
    	return a;
    }

    int getDescent() {
		int d=getTFont().getDescent();
    	int i=begin();
    	while (i<end()) {
			TNode tn=TNode(i);
	    	if (tn.isChar()) {
	    		d=Math.max(d,tn.getDescent(getTFont()));
	    	}
    		i++;
    	}
    	return d;
    }

    public Formula getFormulaAt(int x,int y) {
    	if (contains(x,y)) {
    		Formula form=null;
	    	int i=begin();
	    	while (form==null && i<end()) {
				TNode tn=TNode(i);
				if (tn.getFI()!=null) {
    				for (int j=0;j<tn.getFI().size();j++) {
    					form=tn.getFI().getFormula(j).getFormulaAt(x,y);
    					if (form!=null) {
    						break;
    					}
    				}
    			}
	    		i++;
	    	}
	    	if (form!=null) {
	    		return form;
	    	} else {
	    		return this;
	    	}
    	} else {
    		return null;
    	}
    }

    int getHeight() {
    	return getAscent()+getDescent();
    }

    public int getIndexOfTNodeAtPixel(Component C,int x,int y) {
    	int o=getOffsetOfTNodeAtPixel(C,0,x);
		return firstChar(o);
    }

    public int getOffsetOfTNodeAtPixel(Component C,int n,int x) {
    	int i0=begin();
    	int i=i0;
    	int xx=x0;
    	while (i<end() && !TNode(i).isEndOfLine() && xx+((2.0*getCharWidthAtTNode(C,i,true))/3.0)<x ) {
   			xx+=getCharWidthAtTNode(C,i++,true);
    	}
    	return i-i0;
    }


    public int getFormulaWidth(Component C) {
		TFont tf=getTFont();
		int x=0;
		for (int i=begin();i<end();i++) {
			TNode tn=TNode(i);
			if (tn.isChar()) {
				x+=tn.getCharWidth(tf,C,true);
			}
		}
		return Math.max(x,tf.getMjaFont().stringWidth(C," "));
    }
    
    public int getStringWidth(Component C,String s) {
		TNode tn=new TNode(this,s);
		return tn.getCharWidth(getTFont(),C,true);
    }

    public int getX(Component C,int n) {
		int x=0;
		for (int i=0;i<n;i++) {
			TNode tn=TNode(i);
			if (tn.isChar()) {
				x+=tn.getCharWidth(getTFont(),C,true);
			}
		}
		return x0+x;
    }

    public int getY() {
    	return y0;
    }

    boolean contains(int x,int y) {
		if (parent.isFraction()) {
			int wnum=parent.getFI().getFormula(FormulaItem.NUM).W;
			int wden=parent.getFI().getFormula(FormulaItem.DEN).W;
			int w=Math.max(wnum,wden);
			int dw=(w-W)/2;
    		return posX-dw<=x && x<posX-dw+w && posY-A<=y && y<posY+D;
//		} else if (parent.isProduct()) {
//			int wnum=parent.getFI().getFormula(FormulaItem.NUM).W;
//			int wden=parent.getFI().getFormula(FormulaItem.DEN).W;
//			int w=wnum+12+wden;
//    		return x0-dw<=x && x<x0-dw+w && y0-A<=y && y<y0+D;
		} else {
    		return posX<=x && x<posX+W && posY-A<=y && y<posY+D;
		}
    }

    public void calcValues(int dec,boolean fixed,int aL,boolean editing,boolean learnerMode) {
		for (int i=begin();i<end();i++) {
			TNode tn=TNode(i);
			tn.calcValues(dec,fixed,aL,editing,learnerMode);
		}
    }

    public void setActive(boolean a) {
    	active=a;
    }

    /* ----------------- drawing methods ---------------- */

    void calcAndSetFonts(TFont font) {
   		setFonts(font);
   		if (useMinFont) {
	   		double factor=1.0;
	   		int n=1;
	   		while ((getMinFontSize())<min_font_size) {
				TFont newtf=new TFont(font.getName(),font.getStyle(),font.getSize()+n++);
		   		setFonts(newtf);
	   		}
   		}
    }
    
    void draw(Graphics2D g,Container C,TFont font,int x,int y,boolean editing,boolean marked,boolean main,boolean showholes) {
		locateToFindOrDraw(C,x,y,true);
		g.setColor(getColors().getTextColor());
    	doDraw(g,C,editing,marked,main,showholes);
    }

    private static final int min_font_size=11;

	int getMinFontSize() {
		int minsz=getTFont().getSize();
    	for (int i=begin();i<end();i++) {
    		TNode tn=TNode(i);
    		if (tn.isChar()&&tn.getFI()!=null) {
				FormulaItem FI=tn.getFI();
				for (int j=0;j<FI.size();j++) {
					minsz=Math.min(minsz,FI.getFormula(j).getMinFontSize());
				}
    		}
    	}
    	return minsz;
    }

    public boolean updateNodes(Component Cmp,String cID) {
    	for (int i=begin();i<end();i++) {
    		TNode tn=TNode(i);
    		if (tn.isChar()&&tn.getFI()!=null) {
    			if (tn.getFI().updateNodes(Cmp,cID)) {
    				return true;
    			};
    		} else if (tn.updateComponent(Cmp,cID)) {
    			return true;
    		}
    	}
    	return false;
    }

    private int x0,x1,x2,x3,xe,y0,A,D,W;

    public static double sixp=0.66;
    public static double ixp=0.80;
    public static double frp=0.90;
    public static boolean useMinFont=true;

    public void setFonts(TFont font) {
    	setTFont(font);
		int n1=begin();
    	int n2=end();
    	for (int i=n1;i<n2;i++) {
    		TNode tn=TNode(i);
    		if (tn.isChar()&&tn.getFI()!=null) {
   				FormulaItem FI=tn.getFI();
    			if (tn.isSubIndex()) {
    				FI.getFormula(0).setFonts(font.scale(sixp));
    			} else if (tn.isSuperIndex()) {
    				FI.getFormula(0).setFonts(font.scale(sixp));
    			} else if (tn.isFormula()) {
    				FI.getFormula(0).setFonts(font);
    			} else if (tn.isFraction()) {
   					FI.getFormula(FormulaItem.NUM).setFonts(font.scale(frp));
   					FI.getFormula(FormulaItem.DEN).setFonts(font.scale(frp));
    			} else if (tn.isProduct()) {
   					FI.getFormula(FormulaItem.NUM).setFonts(font.scale(frp));
   					FI.getFormula(FormulaItem.DEN).setFonts(font.scale(frp));
    			} else if (tn.isRadical()) {
   					FI.getFormula(FormulaItem.INDEX).setFonts(font.scale(ixp));
   					FI.getFormula(FormulaItem.RADICAND).setFonts(font);
    			} else if (tn.isSum()||tn.isIntegral()||tn.isLimit()) {
   					FI.getFormula(FormulaItem.FROM).setFonts(font.scale(ixp));
   					FI.getFormula(FormulaItem.TO).setFonts(font.scale(ixp));
   					FI.getFormula(FormulaItem.WHAT).setFonts(font);
    			} else {
    				for (int k=0;k<FI.size();k++) {
    					FI.getFormula(k).setFonts(font);
    				}
				}
    		}
    	}
    }

    public void setColors(TColors c) {
    	super.setColors(c.getBackColor(),c.getTextColor());
    	for (int i=begin();i<end();i++) {
    		TNode tn=TNode(i);
    		tn.setColors(getColors());
    	}
    }
    
    private int posX,posY;
    
    public void locate(Component C,int x,int y) {
    	locateToFindOrDraw(C,x,y,false);
    }
    
    public Point getLocation() {
    	return new Point(posX,posY);
    }

    private FontMetrics fm;

    private void locateToFindOrDraw(Component C,int xx,int yy,boolean toDraw) {
		if (toDraw) {
	    	x0=xx;
    		y0=yy;
		} else {
			posX=xx;
			posY=yy;
		}
    	A=getAscent();
    	D=getDescent();
		W=getFormulaWidth(C);
		TFont tf=getTFont();
		int a=getTFont().getAscent();
		int d=getTFont().getDescent();
		int h=a+d;
		int he=(a-d)/2;
		int n1=begin();
    	int n2=end();
    	int x=xx,y=yy;
    	for (int i=n1;i<n2;i++) {
    		TNode tn=TNode(i);
    		if (tn.isChar()) {
	    		tn.setLocation(x,y);
	    		int dx=tn.getCharWidth(tf,C,true);
   				Formula form=null;
   				if (tn.getFI()!=null) {
   					form=tn.getFI().getFormula(0);
   				}
    			if (tn.isSubIndex()) {
    				form.locateToFindOrDraw(C,x+tn.getdx2(),y-he+form.getAscent(),toDraw);
    			} else if (tn.isSuperIndex()) {
    				form.locateToFindOrDraw(C,x+tn.getdx2(),y-he-form.getDescent(),toDraw);
    			} else if (tn.isFormula()) {
    				form.locateToFindOrDraw(C,x+tn.getdx2(),y,toDraw);
    			} else if (tn.isFraction()) {
    				Formula num=tn.getFI().getFormula(FormulaItem.NUM);
    				Formula den=tn.getFI().getFormula(FormulaItem.DEN);
    				int dxnum=num.getFormulaWidth(C);
    				int dxden=den.getFormulaWidth(C);
    				int numd=num.getDescent();
    				int numh=num.getHeight();
    				int dena=den.getAscent();
    				int denh=den.getHeight();
    				num.locateToFindOrDraw(C,x+(dx-dxnum)/2,y-he-numd,toDraw);
    				den.locateToFindOrDraw(C,x+(dx-dxden)/2,y-he+dena,toDraw);
    			} else if (tn.isProduct()) {
    				Formula num=tn.getFI().getFormula(FormulaItem.NUM);
    				Formula den=tn.getFI().getFormula(FormulaItem.DEN);
//    				int dxnum=num.getFormulaWidth(C);
    				num.locateToFindOrDraw(C,x,y,toDraw);
    				den.locateToFindOrDraw(C,x+tn.getdX1()+tn.getdX2(),y,toDraw);
    			} else if (tn.isRadical()) {
    				Formula index=tn.getFI().getFormula(FormulaItem.INDEX);
    				Formula radicand=tn.getFI().getFormula(FormulaItem.RADICAND);
    				index.locateToFindOrDraw(C,x,y-he-index.getDescent(),toDraw);
    				radicand.locateToFindOrDraw(C,x+tn.getX1()+tn.getdx2(),y,toDraw);
    			} else if (tn.isSum()) {
    				Formula from=tn.getFI().getFormula(FormulaItem.FROM);
    				Formula to=tn.getFI().getFormula(FormulaItem.TO);
    				Formula what=tn.getFI().getFormula(FormulaItem.WHAT);
    				to.locateToFindOrDraw(C,x+(tn.getX1()-to.getFormulaWidth(C))/2,y-a-to.getDescent()+1,toDraw);
    				from.locateToFindOrDraw(C,x+(tn.getX1()-from.getFormulaWidth(C))/2,y+d+from.getAscent()-2,toDraw);
    				what.locateToFindOrDraw(C,x+tn.getX1(),y,toDraw);
    			} else if (tn.isIntegral()) {
    				Formula from=tn.getFI().getFormula(FormulaItem.FROM);
    				Formula to=tn.getFI().getFormula(FormulaItem.TO);
    				Formula what=tn.getFI().getFormula(FormulaItem.WHAT);
    				int extrato=(tf.isItalic())?h/2:h/6;
    				int extrafrom=(tf.isItalic())?h/6:h/12;
    				to.locateToFindOrDraw(C,x+tn.getdX2()+2+extrato,y-a-to.getDescent()+1,toDraw);
    				from.locateToFindOrDraw(C,x+tn.getdX2()+3-extrafrom,y+d+from.getAscent()-2,toDraw);
    				what.locateToFindOrDraw(C,x+tn.getX1()+2,y,toDraw);
    			} else if (tn.isLimit()) {
    				Formula from=tn.getFI().getFormula(FormulaItem.FROM);
    				Formula to=tn.getFI().getFormula(FormulaItem.TO);
    				Formula what=tn.getFI().getFormula(FormulaItem.WHAT);
    				int dy=Math.max(from.getAscent(),to.getAscent());
    				from.locateToFindOrDraw(C,x,y+d+dy-2,toDraw);
    				to.locateToFindOrDraw(C,x+tn.getdX2(),y+d+dy-2,toDraw);
    				what.locateToFindOrDraw(C,x+tn.getX1()+3,y,toDraw);
    			} else if (tn.isMatrix()) {
    				for (int k=0;k<tn.getFI().size();k++) {
    					tn.getFI().getFormula(k).locateToFindOrDraw(C,
    							x+4*tn.getdx2()+(k%tn.getColumns())*((tn.getW()-2*tn.getdx2())/tn.getColumns()),
    							y-he+tn.getMaxA()-tn.getMatrixHeight()/2+(k/tn.getColumns())*(tn.getMatrixHeight()/tn.getRows()),
    							toDraw);
    				}
    			} else if (tn.isDefParts()) {
    				for (int k=0;k<tn.getFI().size();k++) {
    					int xL=x+4*tn.getdx2();
    					int dy=(tn.getDefPartsHeight())/(tn.getParts());
    					int yL=y-he+tn.getMaxA()-tn.getDefPartsHeight()/2+5+k*dy;
    					tn.getFI().getFormula(k).locateToFindOrDraw(C,xL,yL,toDraw);
    				}
				}
				x+=dx;
    		}
    	}
		if (segMarked()) {
			x1=getX(C,Math.max(n1,segMin()));
			x2=getX(C,Math.min(n2,segMax()));
			if (x1==x0) {
				x1-=2;
			}
			x3=getX(C,n2);
		} else {
			x1=getX(C,n1);
			x2=getX(C,n2);
			x3=getX(C,n2);
		}
    }
    
//    public String toString() {
//    	String s=super.toString()+"\n";
//    	for (int i=begin();i<end();i++) {
//    		TNode tn=TNode(i);
//    		s+=" "+tn.toString();
//    	}
//    	return s;
//    }

    boolean hasChar() {
    	for (int i=begin();i<end();i++) {
    		TNode tn=TNode(i);
    		if (tn.isChar()) {
    			return true;
    		}
    	}
    	return false;
    }

    public boolean hasHoles(boolean check) {
    	if (check && !hasChar() ) {
    		return true;
    	}
    	for (int i=begin();i<end();i++) {
    		TNode tn=TNode(i);
			FormulaItem FI=tn.getFI();
			if (FI!=null) {
				if (FI.hasHoles(tn.getType()==tn.RADICAL)) {
					return true;
				}
			}
    	}
    	return false;
    }
    
	protected void updateComponentLocation(Container C,int n) {
		TNode tn=TNode(n);
		if (tn.isComponent()) {
			TFont tf=getTFont();
			int x=tn.getX()+tn.getdx2();
//			int y=tn.getY();
			int y=posY;
			int a=tf.getAscent();
			int d=tf.getDescent();
			int H=a+d-2;
			Component Cmp =tn.getComponent();
			int W=(Cmp.getSize().width==Cmp.getSize().height)?H:Cmp.getSize().width;
	    	if (x!=Cmp.getLocation().x||(y-a)!=Cmp.getLocation().y||Cmp.getSize().width!=W||Cmp.getSize().height!=H) {
    			Cmp.setBounds(x,y-a,W,H);
	    	}
	    	if (Cmp.getFont()!=null) {
	    		if ( !Cmp.getFont().getName().equals(tf.getName())
	    			||Cmp.getFont().getStyle()!=(tf.getStyle()%4)
	    			||Cmp.getFont().getSize()!=tf.getSize()
	    		) {
	    			Cmp.setFont(new Font(tf.getName(),tf.getStyle()%4,tf.getSize()));
	    		}
	    	}
			if (!(Cmp.getParent()==C)) {
    			C.add(Cmp);
    			Cmp.setVisible(true);
    			if (Cmp instanceof LiveOrDie) {
    				((LiveOrDie) Cmp).start();
    			}
    		}
		}
	}

    void updateComponentLocations(Container C) {
    	for (int i=0;i<info.size();i++) {
    		TNode tn=TNode(i);
    		if (tn.isComponent()) {
	   			updateComponentLocation(C,i);
    		} else if (tn.getFI()!=null) {
    			FormulaItem FI=tn.getFI();
    			for (int j=0;j<FI.size();j++) {
    				Formula F=FI.getFormula(j);
    				F.updateComponentLocations(C);
    			}
    		}
    	}
    }

    public void doDraw(Graphics2D g,Container C) {
    	doDraw(g,C,false,false,false,false);
    }
    
    private static final Color exprBC=new Color(0xffffaa);

    void doDraw(Graphics2D g,Container C,boolean editing,boolean marked,boolean main,boolean showholes) {
 		Color gc=g.getColor();
		TColors currentColors=getColors();
		gc=currentColors.getTextColor();
		if (segMarked()) {
			g.setColor(bestBG(Color.white));
			g.fillRect(x1,y0-A,x2-x1,A+D);
		}
		g.setColor(gc);
    	for (int i=begin();i<end();i++) {
    		TNode tn=TNode(i);
    		if (tn.isColors()) {
    			currentColors=tn.getColors();
				if (currentColors!=null) {
					if (currentColors.getTextColor()!=null) {
						gc=currentColors.getTextColor();
		    			g.setColor(gc);
		    		}
		    	}
    		} else if (tn.isTFont()) {
    		} else if (tn.isChar()) {
				boolean isMarked=marked || ( segMarked() && (segMin()<=i) && (i<segMax()) );
				fillBack(isMarked,g,currentColors,tn,A,D);
		    	if (editing && tn.isExpr() /* && drawBox*/ ) {
		    		g.setColor(exprBC);
					g.fillRect(tn.getX(),tn.getY()-tn.getA(),tn.getW(),tn.getA()+tn.getD());
					g.setColor(gc);
		    	}
    			if (isMarked) {
    				g.setColor(bestFG(Color.white));
//g.setColor(Color.red);
    			} else {
    				g.setColor(gc);
//g.setColor(Color.blue);
    			}
				FormulaItem FI=tn.getFI();
    			if (tn.isFormula()||tn.isSubIndex()||tn.isSuperIndex()) {
//   					fillBack(isMarked,g,currentColors,tn);
   					FI.doDraw(g,C,editing,isMarked,showholes,false);
    			} else if (tn.isFraction()) {
//   					fillBack(isMarked,g,currentColors,tn);
   					Color sc=g.getColor();
   					FI.doDraw(g,C,editing,isMarked,showholes,false);
   					g.setColor(sc);
//g.setColor(Color.green);
   					int xx1=tn.getX()+tn.getdx2(), yy=tn.getY()-tn.gethe(),xx2=tn.getX()+tn.getW()-tn.getdx2();
			    	TFont.setAntialias(g,false);
    				g.drawLine(xx1,yy,xx2,yy);
    				if (fm!=null) {
    					int fmh=fm.getHeight();
    					if (fmh>32) {
	    					g.drawLine(xx1,yy+1,xx2,yy+1);
	    					if (fmh>45) {
		    					g.drawLine(xx1,yy-1,xx2,yy-1);
		    					if (fmh>60) {
			    					g.drawLine(xx1,yy+2,xx2,yy+2);
			    					if (fmh>75) {
				    					g.drawLine(xx1,yy-2,xx2,yy-2);
				    					if (fmh>90) {
					    					g.drawLine(xx1,yy+3,xx2,yy+3);
					    				}
					    			}
				    			}
			    			}
		    			}
    				}
			    	TFont.setAntialias(g,getTFont());
    			} else if (tn.isRadical()) {
    				int x=tn.getX(),y=tn.getY();
    				int x0=x;
    				int x3=x+tn.getX1();
    				int x1=x3-tn.getdX2();
    				int x2=(x1+x3)/2;
    				int x4=x+tn.getW();
    				int y0=y-tn.gethe()+1;
    				int y1=y0;
    				int y2=y+tn.getD()-2;
    				int y3=y-FI.getFormula(FormulaItem.RADICAND).getAscent()-1;
    				int y4=y3;
//   					fillBack(isMarked,g,currentColors,tn);
   					Color sc=g.getColor();
   					FI.doDraw(g,C,editing,isMarked,showholes,true);
			    	TFont.setAntialias(g,true);
   					g.setColor(sc);
//g.setColor(Color.green);
    				g.drawLine(x0+2,y0,x1+1,y1);
    				g.drawLine(x1,y1,x2,y2);
    				g.drawLine(x2,y2,x3,y3);
					g.drawLine(x3,y3,x4,y4);
    			} else if (tn.isSum()) {
//   					fillBack(isMarked,g,currentColors,tn);
			    	TFont.setAntialias(g,true);
					getTFont().getMjaFont().drawString(C,g,"\u2211",tn.getX()+(tn.getX1()-tn.getdX2())/2,tn.getY());
			    	TFont.setAntialias(g,getTFont());
   					FI.doDraw(g,C,editing,isMarked,showholes,false);
    			} else if (tn.isIntegral()) {
//   					fillBack(isMarked,g,currentColors,tn);
   					TFont tf=getTFont();
			    	g.setFont(tf.getMjaFont().getAwtFont());
   					FI.doDraw(g,C,editing,isMarked,showholes,false);
			    	TFont.setAntialias(g,true);
					com.mja.file.mjaFont mjaf=tf.scale(1.2).getMjaFont();
					mjaf.drawString(C,g,"\u222b",tn.getX(),tn.getY()+(int) Math.round(0.07*g.getFont().getSize()));
			    	TFont.setAntialias(g,tf);
    			} else if (tn.isLimit()) {
//   					fillBack(isMarked,g,currentColors,tn);
    				Formula from=tn.getFI().getFormula(FormulaItem.FROM);
    				Formula to=tn.getFI().getFormula(FormulaItem.TO);
    				Formula what=tn.getFI().getFormula(FormulaItem.WHAT);
    				int xf=from.x0+from.getFormulaWidth(C);
    				int yf=from.y0/*+Math.max(from.A,to.A)*/;
					getTFont().getMjaFont().drawString(C,g,"lm",xf+2,tn.getY());
			    	TFont.setAntialias(g,true);
					getTFont().getMjaFont().drawString(C,g," \u2192",xf,yf);
			    	TFont.setAntialias(g,getTFont());
			    	g.setFont(getTFont().getMjaFont().getAwtFont());
   					FI.doDraw(g,C,editing,isMarked,showholes,false);
    			} else if (tn.isMatrix()) {
//   					fillBack(isMarked,g,currentColors,tn,A,D);
    				int x=tn.getX()+tn.getdx2();
    				int y=tn.getY();
    				int dx=tn.getW()-2*tn.getdx2();
    				int x1=x,x2=x+dx,y1=y-tn.getA()+4,y2=y+tn.getD()-4;
			    	TFont.setAntialias(g,false);
	   				g.drawLine(x1,y1,x1,y2);
	   				g.drawLine(x1,y1,x1+2,y1-2);
	   				g.drawLine(x1,y2,x1+2,y2+2);
	   				g.drawLine(x2,y1,x2,y2);
	   				g.drawLine(x2,y1,x2-2,y1-2);
	   				g.drawLine(x2,y2,x2-2,y2+2);
			    	TFont.setAntialias(g,getTFont());
   					FI.doDraw(g,C,editing,isMarked,showholes,false);
    			} else if (tn.isDefParts()) {
//   					fillBack(isMarked,g,currentColors,tn,A,D);
    				int x=tn.getX();
    				int y=tn.getY();
    				int h=4;
    				int x1=x,y1=y-tn.getA()+4,y2=y+tn.getD()-4;
    				int y0=(y1+y2)/2;
			    	TFont.setAntialias(g,true);
	   				g.drawLine(x1+2*h,y1,x1+h,y1+h);
	   				g.drawLine(x1+h,y1+h,x1+h,y0-h);
	   				g.drawLine(x1+h,y0-h,x1,y0);
	   				g.drawLine(x1,y0,x1+h,y0+h);
	   				g.drawLine(x1+h,y0+h,x1+h,y2-h);
	   				g.drawLine(x1+h,y2-h,x1+2*h,y2);
			    	TFont.setAntialias(g,getTFont());
   					FI.doDraw(g,C,editing,isMarked,showholes,false);
    			} else if (tn.isProduct()) {
//   					fillBack(isMarked,g,currentColors,tn);
   					Color sc=g.getColor();
   					FI.doDraw(g,C,editing,isMarked,showholes,false);
   					g.setColor(sc);
//g.setColor(Color.green);
					getTFont().getMjaFont().drawString(C,g,"  ",tn.getX()+tn.getdX1(),tn.getY());
    			} else {
			    	TFont.setAntialias(g,getTFont());
    				int x=tn.getX()+tn.getdx2();
    				int y=tn.getY();
    				int dx=tn.getW()-2*tn.getdx2();
					int uy=2;
    				int A=tn.getA();
    				int D=tn.getD();
    				if (tn.isComponent()) {
    					Component Cmp=tn.getComponent();
			    		if (isMarked) {
				    		Cmp.setForeground(bestFG(Color.white));
				    		Cmp.setBackground(Color.black);
				    	} else {
			    			Cmp.setForeground(currentColors.getTextColor());
				    		Cmp.setBackground(null);
				    	}
    					updateComponentLocation(C,i);
    				} else if (tn.isChar() && (tn.getChar().equals("(")||tn.getChar().equals(")"))) {
	    				if (tn.getChar().equals("(")) { // buscar el que realmente cierre
	    					boolean closureFound=false;
	    					int countP=1;
		    				for (int j=i+1;j<=end();j++) {
		    					TNode ttnn=TNode(j);
		    					if (")".equals(ttnn.getChar())) {
		    						countP--;
		    						if (countP==0) {
			    						closureFound=true;
			    						break;
			    					}
		    					} else if ("(".equals(ttnn.getChar())) {
		    						countP++;
		    					}
		    					A=Math.max(A,ttnn.getA());
		    					D=Math.max(D,ttnn.getD());
		    				}
	    				} else { // buscar el que realmente abre
	    					boolean openingFound=false;
	    					int countP=-1;
		    				for (int j=i-1;j>=begin();j--) {
		    					TNode ttnn=TNode(j);
		    					if ("(".equals(ttnn.getChar())) {
		    						countP++;
		    						if (countP==0) {
		    							openingFound=true;
		    							break;
		    						}
		    					} else if (")".equals(ttnn.getChar())) {
		    						countP--;
		    					}
		    					A=Math.max(A,ttnn.getA());
		    					D=Math.max(D,ttnn.getD());
		    				}
	    				}
						int ya=y-(A-D)/2;
						int R=(int) Math.round((double)(dx*dx+(A+D)*(A+D))/(4.0*(double)dx));
						int alfa=(int) Math.round(0.75*180.0*Math.asin((double)(A+D)/(2.0*(double)R))/Math.PI);
//    					fillBack(isMarked,g,currentColors,tn,A,D);
	    				if (tn.getChar().equals("(")) {
		    				g.drawArc(x+dx/3,ya-R,2*R,2*R,180-alfa,2*alfa);
    					} else if (tn.isChar()&&tn.getChar().equals(")")) {
	    					g.drawArc(x+2*dx/3-2*R,ya-R,2*R,2*R,-alfa,2*alfa);
	    				}
    				} else {
//    					fillBack(isMarked,g,currentColors,tn);
    					getTFont().getMjaFont().drawString(C,g,tn.getChar(),x,y);
					}
					if (getTFont().isUnderline()) {
						g.drawLine(x,y+uy,x+dx-1,y+uy);
						if (getTFont().isBold()) {
							g.drawLine(x,y+uy+1,x+dx-1,y+uy+1);
						}
					}
					if (getTFont().isOverline()) {
						g.drawLine(x,y-A,x+dx-1,y-A);
						if (getTFont().isBold()) {
							g.drawLine(x,y-A+1,x+dx-1,y-A+1);
						}
					}
				}
    		}
    	}
    	TFont.setAntialias(g,false);
    	if (editing&& drawBox &&(active||(getParent()!=null &&getParent().isFormula()))) {
    		int dx1=1;
    		if (getParent().isFormula()) {
    			dx1=2;
    		}
    		if (getParent()!=null) {
    			dx1+=getParent().getdX1();
    		}
    		g.setColor(lightLineColor);
 	    	g.drawRect(x0-dx1+1,y0-A+1,W+2*dx1-2,A+D-2);
    	}
		g.setColor(gc);
    }
    
    public static final Color formulaCursorColor=new Color(0xd0dd1830,true);
    private static final Color lightLineColor=new Color(0x70dd1830,true);

	private static void fillBack(boolean isMarked,Graphics g,TColors cc,TNode tn) {
		fillBack(isMarked,g,cc,tn,tn.getA(),tn.getD());
	}
	private static void fillBack(boolean isMarked,Graphics g,TColors cc,TNode tn,int A,int D) {
		if (!isMarked&&cc!=null&&cc.getBackColor()!=null) {
			g.setColor(cc.getBackColor());
			g.fillRect(tn.getX(),tn.getY()-A,tn.getW(),A+D);
			g.setColor(cc.getTextColor());
		}
	}

/* ----- DUDA : Se usa en alguna parte?
	public static Formula redoFormula(Formula F,Parser p,boolean separate,String por_str,boolean msa,boolean par_all) {
		Node N=Node.formulaToNode(F,p);
		return N.toFormula(p,separate,por_str,msa,par_all);
	}
------------------------------------------- */

    public Formula cloneFormula() {
    	RTF rtf=new RTF(this);
    	String s=rtf.toRTF(false);
    	Formula F=new Formula(getParser(),getTFont().cloneTFont(),getColors().getBackColor(),getColors().getTextColor());
    	F.parser=parser;
    	RTF.parseRTF(s,F);
		return F;
    }
    
    public String toExpresion(boolean content) {
    	return toExpresion(begin(),end(),content);
    }

    public String toExpresion(int p1,int p2,boolean content) {
		String s="";
		boolean prevIsAT=false;
		boolean prevIsNum=false;
		String prevChar="";
		int n=0;
		TNode ta=null;
		for (int i=p1;i<p2;i++) {
			n++;
			TNode tn=(TNode) info.elementAt(i);
			if (tn.isChar()&&!tn.isWhiteSpace()) {
				String ch=tn.getChar();
				if (tn.isAlgebraicToken()) {
					if ((prevIsAT || (ta!=null&&ta.isSuperIndex()))
						&& !(
						  (prevIsNum&&tn.isNumSymbol()) // para juntar varios dgitos y el punto decimal en un nmero
						||("(".equals(prevChar))
						||(")".equals(ch))
						||("[".equals(prevChar))
						||("]".equals(ch))
						||(ch!=null&&ch.startsWith("_")) // para los subndices
						))
					{
						s+="*";
					}
					prevIsNum=tn.isNumSymbol();
					prevIsAT=true;
					prevChar=ch;
					if (prevChar==null) {
						prevChar="";
					}
				} else {
					prevIsAT=false;
					prevIsNum=false;
					prevChar="";
				}
				s+=tn.toExpresion(content);
			}
			ta=tn;
		}
		return s;
    }

	private void cleanSigns() {
    	TFont f=getTFont();
    	TColors tc=getColors();
		int i=begin();
		TNode ta=null;
		while ( i<end()) {
			TNode tn=TNode(i);
			if (ta!=null&&ta.isPureChar()&&tn.isPureChar()) {
				String s=tn.getChar();
				if (s.startsWith("-")&&s.length()>1) {
					if (ta.getChar().equals("+")) {
						ta.setChar("-");
						tn.setChar(s.substring(1));
					} else if (ta.getChar().equals("-")) {
						TNode taa=null;
						if (i>=begin()+2) {
							taa=TNode(i-2);
						}
						if (taa!=null&&taa.isPureChar()&&(taa.getChar().equals("=")||taa.getChar().equals("("))) {
							if (removeNode(--i)) {
								
							} else {
								System.out.println("Formula.couldNotRemoveNode "+TNode(i).toString());
							}
						} else {
							ta.setChar("+");
						}
						tn.setChar(s.substring(1));
					} else { // in this non-empty case don't change the sign!
					}
				}
			}
			ta=tn;
			i++;
		}
	}

}