/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.text;

import java.awt.Graphics2D;
import java.awt.Component;
import java.awt.Container;
import com.mja.parser.*;

public class FormulaItem {

	public static final int NUM=0, DEN=1;
	public static final int INDEX=0, RADICAND=1;
	public static final int FROM=0, TO=1, WHAT=2;
	private int rows=1, columns=1, parts=2;

	private TNode parent;
	private Formula[] F;

	public FormulaItem(Parser p,String[] sa) {
		if (sa!=null) {
			parts=sa.length;
			F=new Formula[sa.length];
			for (int i=0;i<F.length;i++) {
				F[i]=new Formula(p);
				if (sa[i]!=null) {
					F[i].insertString(sa[i]);
//				} else {
//					F[i].insertString("");
				}
			}
		} else {
			F=new Formula[0];
		}
	}

	public FormulaItem(Parser p,String[][] sa) {
		columns=sa.length;
		rows=sa[0].length;
		F=new Formula[rows*columns];
		for (int j=0;j<columns;j++) {
			for (int i=0;i<rows;i++) {
				F[i+j*rows]=new Formula(p);
				if (sa[j][i]!=null) {
					F[i+j*rows].insertString(sa[j][i]);
				}
			}
		}
	}

	public FormulaItem(Parser p,String s0,String s1,String s2) {
		F=new Formula[3];
		F[0]=new Formula(p);
		F[0].insertString(s0);
		F[1]=new Formula(p);
		F[1].insertString(s1);
		F[2]=new Formula(p);
		F[2].insertString(s2);
	}

	public FormulaItem(Parser p,String s0,String s1) {
		F=new Formula[2];
		F[0]=new Formula(p);
		F[0].insertString(s0);
		F[1]=new Formula(p);
		F[1].insertString(s1);
	}

	public FormulaItem(Parser p,String s0) {
		F=new Formula[1];
		F[0]=new Formula(p);
		F[0].insertString(s0);
	}

	public FormulaItem(Formula[] f) {
		parts=f.length;
		F=new Formula[parts];
		for (int j=0;j<parts;j++) {
			F[j]=f[j];
		}
	}

	public FormulaItem(Formula f0,Formula f1,Formula f2) {
		F=new Formula[3];
		F[0]=f0;
		F[1]=f1;
		F[2]=f2;
	}

	public FormulaItem(Formula[][] fa) {
		columns=fa.length;
		rows=fa[0].length;
		F=new Formula[rows*columns];
		for (int j=0;j<columns;j++) {
			for (int i=0;i<rows;i++) {
				F[i+j*rows]=fa[j][i];
			}
		}
	}

	public FormulaItem(Formula f0,Formula f1) {
		F=new Formula[2];
		F[0]=f0;
		F[1]=f1;
	}

	public FormulaItem(Formula f0) {
		F=new Formula[1];
		F[0]=f0;
	}

	public TNode getParent() {
		return parent;
	}

	public int size() {
		return F.length;
	}

	void setParent(TNode p) {
		parent=p;
		for (int i=0;i<F.length;i++) {
			F[i].parent=parent;
		}
	}

	void setColors(TColors tc) {
		for (int i=0;i<F.length;i++) {
			F[i].setColors(tc);
		}
	}

	boolean updateNodes(Component C,String cID) {
		for (int i=0;i<F.length;i++) {
			if (F[i].updateNodes(C,cID)) {
				return true;
			}
		}
		return false;
	}

	public void setAssisted(boolean b) {
		for (int i=0;i<F.length;i++) {
			F[i].setAssisted(b);
		}
	}

	public void setFormula(int i,Formula f) {
		if (i<F.length) {
			F[i]=f;
			F[i].parent=parent;
		}
	}

	public Formula getFormula(int i) {
		if (i<F.length) {
			return F[i];
		} else {
			return null;
		}
	}

	public FormulaItem cloneItem() {
		Formula[] f=new Formula[F.length];
		for (int i=0;i<F.length;i++) {
			f[i]=F[i].cloneFormula();
		}
		return new FormulaItem(f);
	}
	
//	void clearColors() {
//		for (int i=0;i<F.length;i++) {
//			F[i].clearColors();
//		}
//	}
//
//	void clearFonts() {
//		for (int i=0;i<F.length;i++) {
//			F[i].clearFonts();
//		}
//	}

    void doDraw(Graphics2D g,Container C,boolean ed,boolean marked,boolean showholes,boolean isRadical) {
		for (int i=0;i<F.length;i++) {
			F[i].doDraw(g,C,ed,marked,false,showholes&&(!isRadical||i!=0));
		}
    }

    boolean hasHoles(boolean isRadical) {
		for (int i=0;i<F.length;i++) {
			if (F[i].hasHoles(true)) {
				if ((isRadical)&&i==0) {
				} else {
					return true;
				}
			}
		}
		return false;
	}

    public String toString() {
    	String s="";
		for (int i=0;i<F.length;i++) {
			s+=" "+F[i].toString();
		}
		return s;
    }

}