/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.text;

import java.awt.*;
import java.util.*;

import com.mja.parser.*;
import com.mja.lang.*;
import com.mja.gui.mjaText;
import com.mja.util.BasicStr;
import com.mja.util.TFont;

public class MathText {

	private Text T;
	private translator Tr;
	private Node decimals;
	private boolean fixed;

	public MathText(Parser p,translator Tr,String str,Node decimals,boolean fixed,Font font,Color color) {
		this.Tr=Tr;
		T=new Text(p,new TFont(font.getName(),font.getStyle(),font.getSize()),null,color);
		RTF.parseRTF(DescartesToRTF(font,str),T);
		T.setLeftMargin(3);
		T.setTopAndBottomMargin(0);
		this.decimals=decimals;
		this.fixed=fixed;
	}
	
	public void setColors(Color bc,Color fc) {
		T.setColors(bc,fc);
	}
	
	public int ancho(Component C) {
		return T.getWidth(C);
	}

	public int alto(Component C) {
		return T.getHeight(C);
	}

	public int getTextY(Component C) {
		return T.getTextY(C);
	}

	public int getAscent(Component C) {
		return T.getY(C,0);
	}

    public String toRTF(boolean plain) {
    	RTF rtf=new RTF(T);
    	String s=rtf.toRTF(plain);
    	return s;
    }

    public String toTeX() {
    	RTF rtf=new RTF(T);
    	return rtf.toTeX(Color.black,true);
	}
    public String toTeX(Color defaultColor,boolean useOwnColors) {
    	RTF rtf=new RTF(T);
    	return rtf.toTeX(defaultColor,useOwnColors);
    }

	public String getTeXColors(boolean grayscale) {
		return T.getTeXColors(grayscale);
	}

	public static String DescartesToRTF(Font f,String txt) {
		if (txt.startsWith(RTF.RTF_header)) {
			return txt;
		} else {
			String s=RTF.RTF_header+"\\uc0 \n";
			s+="{\\fonttbl\\f0\\fcharset0 "+RTF.WfontName(f.getName())+";}\n";
			s+="\\f0\\fs"+(f.getSize()*2);
			if (f.isBold()) {
				s+="\\b";
			}
			if (f.isItalic()) {
				s+="\\i";
			}
			s+=" "+parseSquareBrackets(BasicStr.replace(txt,"\\n","\n"));
			s+="\n}";
			return s;
		}
	}

	public static String parseSquareBrackets(String txt) {
		StringBuffer sb=new StringBuffer();
		int state=0;
		int i=0;
		while (i<txt.length()) {
			char ch=txt.charAt(i++);
			if (ch=='[') {
				if (state==0) {
					sb.append(RTF.RTF_mjaformula+RTF.RTF_expr+" ");
				} else {
					sb.append('[');
				}
				state++;
			} else if (ch==']') {
				if (state==1) {
					sb.append("}}");
				} else {
					sb.append(']');
				}
				state--;
				if (state<0) {
					state=0;
//					(new Exception(txt.substring(0,i+1)+"\ntoo many ] ")).printStackTrace();
					(new Exception(txt+"\ntoo many ] ")).printStackTrace();
				}
			} else if (ch=='\n') {
				sb.append("\\par ");
			} else {
				if (ch=='\\'&&i<txt.length()) {
					ch=txt.charAt(i++);
				}
				RTF.appendAsChar(sb,ch);
			}
		}
		return new String(sb);
	}

	public void draw(Component cs,int ix,int iy,Graphics[] g) {
		if (cs instanceof Container) {
			draw((Container) cs,ix,iy,g);
		}
	}
	
	private void draw(Container cs,int ix,int iy,Graphics[] g) {
		int dec=2;
		if (BasicStr.classExists("com.mja.parser.Node") && decimals!=null) {
			dec=(int)Math.round(decimals.Evaluate(2));
		}
		T.calcValues(dec,fixed,Tr.getActiveLanguage(),false,false);
		T.setLeftMargin(ix);
		T.setTopAndBottomMargin(iy);
		for (int i=0;i<g.length;i++) {
			T.drawSegment(g[i],cs,false,T.firstLine(),T.end(),false);
		}
	}

}