/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.text;

import java.awt.*;
import java.util.*;

//import com.mja.parser.*;
import com.mja.file.mjaHtml;
import com.mja.util.BasicStr;
import com.mja.util.TFont;
import org.unam.matem.Attribute;
import com.mja.util.LivingCanvas;

public class RTF {

	public static final String RTF_header="{\\rtf1", RTF_mjaformula="{\\*\\mjaformula",
		RTF_component="{\\*\\component",
		RTF_hyperlink="{\\*\\hyperlink",
		RTF_expr="{\\expr", RTF_decimals="\\decimals", RTF_fixed="\\fixed",
		RTF_fraction="{\\fraction",RTF_product="{\\product",RTF_num="{\\num", RTF_den="{\\den",
		RTF_radical="{\\radical", RTF_index="{\\index", RTF_radicand="{\\radicand",
		RTF_sum="{\\sum", RTF_integral="{\\integral", RTF_limit="{\\limit", RTF_from="{\\from", RTF_to="{\\to", RTF_what="{\\what",
		RTF_matrix="{\\matrix",RTF_rows="\\rows", RTF_columns="\\columns",RTF_element="{\\element",
		RTF_defparts="{\\defparts",RTF_parts="\\parts",
		RTF_subix="{\\subix",
		RTF_supix="{\\supix";
	public static final int Monospaced=0, Serif=1, SansSerif=2;
	public static final String[] Java2FontNames={"Monospaced","Serif","SansSerif"};
	public static final String[] Java1FontNames={"Courier","TimesRoman","Helvetica"};
	public static final String[] WindowsFontNames={"Courier New","Times New Roman","Arial"};
	public static float defaultFontSize = 12f, fontScale=2f;
	private static final char[] hexdigits = { '0', '1', '2', '3', '4', '5', '6', '7',
				      '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };
	private	StringBuffer RTFsb;
	private boolean RTFafterKeyword;
    private Dictionary RTFcolorTable, RTFfontTable, RTFstyleTable;
    private int RTFcolorCount, RTFfontCount, RTFstyleCount, RTFusedFontCount;
    /* the value of the last \\ucN keyword emitted */
    private static int RTFunicodeCount;

    private int[] outputConversion;

    private BasicText T;

 	public RTF (BasicText T) {
 		this.T=T;
 		if (T instanceof Text) {
    		((Text)this.T).clean();
    	}
 		textKeywords=new CharacterKeywordPair[0];
 	}

    /* An array of character-keyword pairs. This could be done
       as a dictionary (and lookup would be quicker), but that
       would require allocating an object for every character
       written (slow!). */
    static class CharacterKeywordPair
      { public char character; public String keyword; };
    static protected CharacterKeywordPair[] textKeywords;

    static String WfontName(String JfontName) {
    	for (int i=0;i<Java1FontNames.length;i++) {
    		if (WindowsFontNames[i].equals(JfontName)) {
    			return WindowsFontNames[i];
    		}
    		if (Java1FontNames[i].equals(JfontName)) {
    			return WindowsFontNames[i];
    		}
    		if (Java2FontNames[i].equals(JfontName)) {
    			return WindowsFontNames[i];
    		}
    	}
    	return WindowsFontNames[0];
    }

    static String JfontName(String WfontName) {
    	for (int i=0;i<WindowsFontNames.length;i++) {
    		if (WindowsFontNames[i].equals(WfontName)) {
    			return Java2FontNames[i];
    		}
    	}
    	return Java2FontNames[0];
    }

	private void writeRTFBegingroup() {
	    RTFsb.append('{');
	    RTFafterKeyword = false;
	}

	private void writeRTFEndgroup() {
	    RTFsb.append('}');
	    RTFafterKeyword = false;
	}

	private void writeRTFRawString(String str) {
		int strlen = str.length();
		for (int offset = 0; offset < strlen; offset ++) {
			RTFsb.append(str.charAt(offset));
		}
	}

	private void writeRTFControlWord(String keyword) {
		RTFsb.append('\\');
		writeRTFRawString(keyword);
		RTFafterKeyword = true;
	}

	private void writeRTFControlWord(String keyword, int arg) {
		RTFsb.append('\\');
		writeRTFRawString(keyword);
		writeRTFRawString(String.valueOf(arg)); /* TODO: correct in all cases? */
		RTFafterKeyword = true;
	}

	private void writeRTFLineBreak() {
//		writeRTFRawString("\n");
		RTFafterKeyword = false;
	}

	private void writeRTFText(String s) {
		int pos, end;

		pos = 0;
		end = s.length();
		for( ; pos < end; pos ++)
		writeRTFCharacter(s.charAt(pos));
	}

	private void writeRTFCharacter(char ch) {
		/* Nonbreaking space is in most RTF encodings, but the keyword is
		preferable; same goes for tabs */
		if (ch == 0xA0) { /* nonbreaking space */
			RTFsb.append((char)0x5C);  /* backslash */
			RTFsb.append((char)0x7E);  /* tilde */
			RTFafterKeyword = false; /* non-alpha keywords are self-terminating */
			return;
		}

		if (ch == 0x09) { /* horizontal tab */
			writeRTFControlWord("tab");
			return;
		}

		if (ch == 10 || ch == 13) { /* newline / paragraph */
		/* ignore CRs, we'll write a paragraph element soon enough */
			return;
		}

		int b = convertCharacter(outputConversion, ch);
		if (b == 0||b>127) {
			/* Unicode characters which have corresponding RTF keywords */
//			int i;
//			for(i = 0; i < textKeywords.length; i++) {
//				if (textKeywords[i].character == ch) {
//					writeRTFControlWord(textKeywords[i].keyword);
//					return;
//				}
//			}
			/* In some cases it would be reasonable to check to see if the
			glyph being written out is in the Symbol encoding, and if so,
			to switch to the Symbol font for this character. TODO. */
			/* Currently all unrepresentable characters are written as
			Unicode escapes. */
			String approximation = approximationForUnicode(ch);
			if (approximation.length() != RTFunicodeCount) {
				RTFunicodeCount = approximation.length();
				writeRTFControlWord("uc", RTFunicodeCount);
			}
			writeRTFControlWord("u", (int)ch);
			writeRTFRawString(" ");
			writeRTFRawString(approximation);
			RTFafterKeyword = false;
			return;
		}

//		if (b > 127) {
//			int nybble;
//			RTFsb.append('\\');
//			RTFsb.append('\'');
//			nybble = ( b & 0xF0 ) >>> 4;
//			RTFsb.append(hexdigits[nybble]);
//			nybble = ( b & 0x0F );
//			RTFsb.append(hexdigits[nybble]);
//			RTFafterKeyword = false;
//			return;
//		}

		switch (b) {
		case '}':
		case '{':
		case '\\':
			RTFsb.append((char)0x5C);  /* backslash */
			RTFafterKeyword = false;  /* in a keyword, actually ... */
			/* fall through */
		default:
			if (RTFafterKeyword) {
				RTFsb.append((char)0x20);  /* space */
				RTFafterKeyword = false;
			}
			RTFsb.append((char)b);
			break;
		}
	}

	private	String approximationForUnicode(char ch)	{
		/* TODO: Find reasonable approximations for all Unicode characters
		in all RTF code pages... heh, heh... */
//		return "?";
		return "";
	}

	/** Takes a translation table (a 256-element array of characters)
	* and creates an output conversion table for use by
	* convertCharacter(). */
	/* Not very efficient at all. Could be changed to sort the table
	for binary search. TODO. (Even though this is inefficient however,
	writing RTF is still much faster than reading it.) */
	private static int[] outputConversionFromTranslationTable(char[] table) {
		int[] conversion = new int[2 * table.length];

		int index;

		for(index = 0; index < table.length; index ++) {
			conversion[index * 2] = table[index];
			conversion[(index * 2) + 1] = index;
		}

		return conversion;
	}

	private	static int[] outputConversionForName(String name) {
//		char[] table = (char[])RTFReader.getCharacterSet(name);
		char[] table =new char[256];
		for (int i=0;i<table.length;i++) {
			table[i]=(char)i;
		}
		return outputConversionFromTranslationTable(table);
	}

	/** Takes a char and a conversion table (an int[] in the current
	* implementation, but conversion tables should be treated as an opaque
	* type) and returns the
	* corresponding byte value (as an int, since bytes are signed).
	*/
	/* Not very efficient. TODO. */
	private	static int convertCharacter(int[] conversion, char ch) {
		int index;

		for(index = 0; index < conversion.length; index += 2) {
			if(conversion[index] == ch) {
				return conversion[index + 1];
			}
		}

		return 0;  /* 0 indicates an unrepresentable character */
	}

	private void putColorInTable(Color c) {
		if (c!=null&&RTFcolorTable.get(c) == null) {
			RTFcolorTable.put(c, new Integer(RTFcolorCount));
			RTFcolorCount ++;
		}
	}

	private boolean endFont;

    private void prepareRTFHeader() {
	    RTFcolorTable = new Hashtable();
	    RTFcolorCount = 0;
//	    RTFcolorTable.put(T.getColor(), new Integer(0));
//	    RTFcolorCount++;

	    RTFfontTable = new Hashtable();
	    RTFfontCount = 0;
//	    RTFfontTable.put(T.getTFont().getName(), new Integer(0));
	    RTFfontTable.put(T.getTFont().toString(), new Integer(0));
	    RTFfontCount++;
//	    RTFusedFontCount=0;
	    RTFusedFontCount++;

	    RTFstyleTable = new Hashtable();
	    /* TODO: put default style in style table */
	    RTFstyleCount = 0;

	    RTFunicodeCount = 0;

//    	TFont f=T.getTFont();
//    	Color c=T.getColor();
    	endFont=false;
    	prepareColors(T);
		if (endFont) {
			RTFusedFontCount--;
		}
    	RTFsb=new StringBuffer();
    	TEXsb=new StringBuffer();
    }

    private void prepareColors(BasicText T) {
		for (int i=0;i<T.end();i++) {
			endFont=false;
			TNode tn=T.TNode(i);
			if (tn.isColors()) {
				putColorInTable(tn.getColors().getTextColor());
				putColorInTable(tn.getColors().getBackColor());
			} else if (tn.isTFont()) {
				TFont f1=tn.getTFont();
//				if (RTFfontTable.get(f1.getName()) == null) {
//					RTFfontTable.put(f1.getName(), new Integer(RTFfontCount));
				if (RTFfontTable.get(f1.toString()) == null) {
					RTFfontTable.put(f1.toString(), new Integer(RTFfontCount));
					RTFfontCount ++;
					RTFusedFontCount++;
					endFont=true;
				} else {
					endFont=false;
				}
			} else if (tn.isMath()) {
				if (tn.isExpr()) {
				} else if (tn.getFI()!=null) {
					for (int j=0;j<tn.getFI().size();j++) {
					    prepareColors(tn.getFI().getFormula(j));
					}
				}
			}
		}
    }

    private void writeRTFHeader() {
		writeRTFBegingroup();
		writeRTFControlWord("rtf", 1);
//		writeRTFControlWord("ansi");
		writeRTFControlWord("uc", 0);
	    outputConversion = outputConversionForName("ansi");
		writeRTFLineBreak();

		/* write font table */
		String[] sortedFontTable = new String[RTFfontCount];
		Enumeration fonts = RTFfontTable.keys();
		String font;
		while(fonts.hasMoreElements()) {
			font = (String)fonts.nextElement();
			Integer num = (Integer)(RTFfontTable.get(font));
			sortedFontTable[num.intValue()] = TFont.getNameFromString(font);
		}
		writeRTFBegingroup();
		writeRTFControlWord("fonttbl");
		for(int index = 0; index < RTFfontCount; index ++) {
			writeRTFControlWord("f", index);
//			writeRTFControlWord("fnil");  /* TODO: supply correct font style */
//			writeRTFControlWord("fprq", 1);
			writeRTFControlWord("fcharset", 0);
			writeRTFText(WfontName(sortedFontTable[index]));
			writeRTFText(";");
		}
		writeRTFEndgroup();
		writeRTFLineBreak();

		/* write color table */
		if (RTFcolorCount > 0) {
			Color[] sortedColorTable = new Color[RTFcolorCount];
			Enumeration colors = RTFcolorTable.keys();
			Color color;
			while(colors.hasMoreElements()) {
				color = (Color)colors.nextElement();
				Integer num = (Integer)(RTFcolorTable.get(color));
				sortedColorTable[num.intValue()] = color;
			}
			writeRTFBegingroup();
			writeRTFControlWord("colortbl");
			for(int index = 0; index < RTFcolorCount; index ++) {
				color = sortedColorTable[index];
				if (color != null) {
					writeRTFControlWord("red", color.getRed());
					writeRTFControlWord("green", color.getGreen());
					writeRTFControlWord("blue", color.getBlue());
				}
				writeRTFRawString(";");
			}
			writeRTFEndgroup();
			writeRTFLineBreak();
		}

//		/* write the style sheet */
//		if (RTFstyleCount > 1) {
//			writeRTFBegingroup();
//			writeRTFControlWord("stylesheet");
//			Enumeration styles = RTFstyleTable.keys();
//			while(styles.hasMoreElements()) {
//				Style style = (Style)styles.nextElement();
//				int styleNumber = ((Integer)styleTable.get(style)).intValue();
//				writeRTFBegingroup();
//				String styleType = (String)style.getAttribute(Constants.StyleType);
//				if (styleType == null) {
//					styleType = Constants.STParagraph;
//				}
//				if (styleType.equals(Constants.STCharacter)) {
//					writeRTFControlWord("*");
//					writeRTFControlWord("cs", styleNumber);
//				} else if(styleType.equals(Constants.STSection)) {
//					writeRTFControlWord("*");
//					writeRTFControlWord("ds", styleNumber);
//				} else {
//					writeRTFControlWord("s", styleNumber);
//				}
//
//				AttributeSet basis = style.getResolveParent();
//				MutableAttributeSet goat;
//				if (basis == null) {
//					goat = new SimpleAttributeSet();
//				} else {
//					goat = new SimpleAttributeSet(basis);
//				}
//
//				updateSectionAttributes(goat, style, false);
//				updateParagraphAttributes(goat, style, false);
//				updateCharacterAttributes(goat, style, false);
//
//				basis = style.getResolveParent();
//				if (basis != null && basis instanceof Style) {
//					Integer basedOn = (Integer)styleTable.get(basis);
//					if (basedOn != null) {
//						writeRTFControlWord("sbasedon", basedOn.intValue());
//					}
//				}
//
//				Style nextStyle = (Style)style.getAttribute(Constants.StyleNext);
//				if (nextStyle != null) {
//					Integer nextNum = (Integer)styleTable.get(nextStyle);
//					if (nextNum != null) {
//						writeRTFControlWord("snext", nextNum.intValue());
//					}
//				}
//
//				Boolean hidden = (Boolean)style.getAttribute(Constants.StyleHidden);
//				if (hidden != null && hidden.booleanValue()) {
//					writeRTFControlWord("shidden");
//				}
//
//				Boolean additive = (Boolean)style.getAttribute(Constants.StyleAdditive);
//				if (additive != null && additive.booleanValue()) {
//					writeRTFControlWord("additive");
//				}
//
//
//				writeRTFText(style.getName());
//				writeRTFText(";");
//				writeRTFEndgroup();
//			}
//			writeRTFEndgroup();
//			writeRTFLineBreak();
//		}
    }

    private void writeRTFColor(String control,Color c) {
    	if (c!=null) {
			Object o=RTFcolorTable.get(c);
			if (o!=null) { // por si hay un color copiado a una frmula
				writeRTFControlWord(control,((Number)o).intValue());
			}
		}
    }

    private void writeRTFDocument(BasicText bt) {
    	TFont f=null;
    	int fontNum=0;
    	float fontSize=0;
    	boolean bold=false,italic=false,underline=false,overline=false;
//	    writeRTFControlWord("pard");
		for (int i=0;i<bt.end();i++) {
			TNode tn=bt.TNode(i);
			if (tn.isColors()) {
				writeRTFColor("cf",tn.getColors().getTextColor());
				writeRTFColor("cb",tn.getColors().getBackColor());
			} else if (tn.isTFont()) {
				TFont f1=tn.getTFont();
//				Object o = RTFfontTable.get(f1.getName());
				Object o = RTFfontTable.get(f1.toString());
				if (o!=null) { // por si hay un cambio de fuente copiado a una frmula
					fontNum = ((Number)o).intValue();
					if (f==null) {
						fontSize=f1.getSize();
						bold=f1.isBold();
						italic=f1.isItalic();
						underline=f1.isUnderline();
						overline=f1.isOverline();
						writeRTFControlWord("f", fontNum);
						writeRTFControlWord("fs",Math.round(fontSize * fontScale));
						if (bold) {
							writeRTFControlWord("b");
						}
						if (italic) {
							writeRTFControlWord("i");
						}
						if (underline) {
							writeRTFControlWord("ul");
						}
						if (overline) {
							writeRTFControlWord("ol");
						}
					} else {
						if (!f1.getName().equals(f.getName())) {
							writeRTFControlWord("f", fontNum);
							fontSize=-1;
						}
						if (f1.getSize()!=fontSize) {
							fontSize=f1.getSize();
							writeRTFControlWord("fs",Math.round(fontSize * fontScale));
						}
						if (f1.isBold()!=bold) {
							if (bold) {
								writeRTFControlWord("b0");
							} else {
								writeRTFControlWord("b");
							}
							bold=f1.isBold();
						}
						if (f1.isItalic()!=italic) {
							if (italic) {
								writeRTFControlWord("i0");
							} else {
								writeRTFControlWord("i");
							}
							italic=f1.isItalic();
						}
						if (f1.isUnderline()!=underline) {
							if (underline) {
								writeRTFControlWord("ulnone");
							} else {
								writeRTFControlWord("ul");
							}
							underline=f1.isUnderline();
						}
						if (f1.isOverline()!=overline) {
							if (overline) {
								writeRTFControlWord("olnone");
							} else {
								writeRTFControlWord("ol");
							}
							overline=f1.isOverline();
						}
					}
					f=f1;
				}
			} else if (tn.isComponent()) {
				writeRTFBegingroup();
			    writeRTFControlWord(RTF_component.substring(2));
			    writeControl(tn);
				writeRTFEndgroup();
			} else if (tn.isHyperlink()) {
				writeRTFBegingroup();
			    writeRTFControlWord(RTF_hyperlink.substring(2));
			    RTFafterKeyword=false;
	    		writeRTFRawString(" "+tn.getChar()+"|"+tn.getUri());
				writeRTFEndgroup();
			} else if (tn.isMath()) {
				if (tn.isFormula()) {
					writeRTFBegingroup();
				    writeRTFControlWord(RTF_mjaformula.substring(2));
				}
				writeMath(tn);
				if (tn.isFormula()) {
					writeRTFEndgroup();
				}
			} else if (tn.isChar()) {
				String s=tn.getChar();
				if (s.equals("\n")) {
				    writeRTFControlWord("par ");
    				writeRTFLineBreak();
    			} else {
		    		writeRTFText(s);
		    	}
			}
		}
    }
    

    private void writeControl(TNode tn) {
    	if (tn.isComponent()) {
    		writeRTFControlWord(tn.getComponentType()+" "+tn.getComponentId());
    	}
    }
    
    private void writeMath(TNode tn) {
		if (tn.isExpr()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_expr.substring(2));
		    writeRTFText(tn.getExpr());
		    if (tn.getDecExpr()!=null) {
			    writeRTFControlWord(RTF_decimals.substring(1));
			    writeRTFText(tn.getDecExpr());
			    if (tn.getFixed()) {
				    writeRTFControlWord(RTF_fixed.substring(1)+"1");
				} else {
				    writeRTFControlWord(RTF_fixed.substring(1)+"0");
				}
		    }
			writeRTFEndgroup();
		} else if (tn.isFraction()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_fraction.substring(2));
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_num.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.NUM));
			writeRTFEndgroup();
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_den.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.DEN));
			writeRTFEndgroup();
			writeRTFEndgroup();
		} else if (tn.isProduct()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_product.substring(2));
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_num.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.NUM));
			writeRTFEndgroup();
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_den.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.DEN));
			writeRTFEndgroup();
			writeRTFEndgroup();
		} else if (tn.isRadical()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_radical.substring(2));
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_index.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.INDEX));
			writeRTFEndgroup();
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_radicand.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.RADICAND));
			writeRTFEndgroup();
			writeRTFEndgroup();
		} else if (tn.isSum()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_sum.substring(2));
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_from.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.FROM));
			writeRTFEndgroup();
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_to.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.TO));
			writeRTFEndgroup();
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_what.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.WHAT));
			writeRTFEndgroup();
			writeRTFEndgroup();
		} else if (tn.isIntegral()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_integral.substring(2));
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_from.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.FROM));
			writeRTFEndgroup();
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_to.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.TO));
			writeRTFEndgroup();
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_what.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.WHAT));
			writeRTFEndgroup();
			writeRTFEndgroup();
		} else if (tn.isLimit()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_limit.substring(2));
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_from.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.FROM));
			writeRTFEndgroup();
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_to.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.TO));
			writeRTFEndgroup();
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_what.substring(2));
		    writeRTFDocument(tn.getFI().getFormula(FormulaItem.WHAT));
			writeRTFEndgroup();
			writeRTFEndgroup();
		} else if (tn.isSubIndex()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_subix.substring(2));
	    	writeRTFDocument(tn.getFI().getFormula(0));
			writeRTFEndgroup();
		} else if (tn.isSuperIndex()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_supix.substring(2));
	    	writeRTFDocument(tn.getFI().getFormula(0));
			writeRTFEndgroup();
		} else if (tn.isMatrix()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_matrix.substring(2));
		    writeRTFControlWord(RTF_rows.substring(1));
		    writeRTFText(String.valueOf(tn.getRows()));
		    writeRTFControlWord(RTF_columns.substring(1));
		    writeRTFText(String.valueOf(tn.getColumns()));
			for (int i=0;i<tn.getFI().size();i++) {
				writeRTFBegingroup();
			    writeRTFControlWord(RTF_element.substring(2));
			    writeRTFDocument(tn.getFI().getFormula(i));
				writeRTFEndgroup();
			}
			writeRTFEndgroup();
		} else if (tn.isDefParts()) {
			writeRTFBegingroup();
		    writeRTFControlWord(RTF_defparts.substring(2));
		    writeRTFControlWord(RTF_parts.substring(1));
		    writeRTFText(String.valueOf(tn.getParts()));
			for (int i=0;i<tn.getFI().size();i++) {
				writeRTFBegingroup();
			    writeRTFControlWord(RTF_element.substring(2));
			    writeRTFDocument(tn.getFI().getFormula(i));
				writeRTFEndgroup();
			}
			writeRTFEndgroup();
		} else {
	    	writeRTFDocument(tn.getFI().getFormula(0));
		}
    }

	private void writeRTFTrailer() {
	    writeRTFEndgroup();
	    writeRTFLineBreak();
	}
	
    public String toRTF (boolean plain) {
 		if (T instanceof Text) {
    		((Text)T).clean();
    	}
    	prepareRTFHeader();
    	boolean isPlain=false;
    	if (plain) {
//System.out.println("RTFcolorCount="+RTFcolorCount);
//System.out.println("RTFusedFontCount="+RTFusedFontCount);
//    		isPlain=(RTFcolorCount==0&&RTFusedFontCount<2);
    		isPlain=(RTFcolorCount==0&&RTFusedFontCount==0);
    		if (isPlain) {
  System.out.println("isPlain");
				for (int i=0;i<T.end();i++) {
					TNode tn=T.TNode(i);
					if (tn.isColors()) {
					} else if (tn.isTFont()) {
					} else if (tn.isFormula()) {
						Formula F=tn.getFI().getFormula(0);
						if (F.end()==F.begin()+1) {
							tn=F.TNode(F.begin());
							if (tn.isExpr()) {
							    if (tn.getDecExpr()==null) {
							    	RTFsb.append("[");
							    	RTFsb.append(tn.getExpr());
							    	RTFsb.append("]");
								} else {
//System.out.println("DecExpr="+tn.getDecExpr());
									isPlain=false;
									break;
							    }
							} else {
//System.out.println("Math="+tn.toString());
								isPlain=false;
								break;
							}
						} else {
//System.out.println("long Formula="+F.begin()+","+F.end());
							isPlain=false;
							break;
						}
					} else if (tn.isChar()) {
						String s=tn.getChar();
						if (s.equals("\n")) {
					    	RTFsb.append("\\n");
		    			} else {
					    	RTFsb.append(s);
					    }
					} else {
//System.out.println("ERROR="+tn.toString());
						isPlain=false;
						break;
					}
				}
			}
    	}
    	if (isPlain) {
    	} else {
    		RTFsb=new StringBuffer();
	    	writeRTFHeader();
	    	writeRTFDocument(T);
	    	writeRTFTrailer();
	    }
	    String str=RTFsb.toString();
		str=BasicStr.replace(str,"\"","&quote;");
		str=BasicStr.replace(str,"<","&lt;");
		str=BasicStr.replace(str,">","&gt;");
    	return str;
    }

    public static void parseRTF(String rtfstr,BasicText T) {
		String str=new String(rtfstr);
		str=BasicStr.replace(str,"&gt;",">");
		str=BasicStr.replace(str,"&lt;","<");
		str=BasicStr.replace(str,"&quote;","\"");
    	RTFunicodeCount=1;
    	Hashtable fonts=new Hashtable();
    	Hashtable colors=new Hashtable();
    	if (BasicStr.hasContent(str)) {
    		if (str.startsWith(RTF_header)) {
    			addText(str.substring("{\\rtf1".length()),T,fonts,colors);
    			int pos="{\\rtf1".length();
    		} else {
    			T.insertString(BasicStr.replace(str,"\\n","\n"));
    		}
    	}
    	if (T instanceof Text) {
    		((Text)T).clean();
    	}
   		T.setCursorPos(T.firstChar(T.begin()));
    }

    private static void addText(String str,BasicText T,Hashtable fonts,Hashtable colors) {
    	TFont f=T.getTFont();
    	Color fgcolor=T.getColors().getTextColor();
    	Color bgcolor=T.getColors().getBackColor();
		int pos=0;
//		boolean par=false;
		while (pos<str.length()) {
			Attribute token=getNextToken(str,pos);
			if (token.name.length()==0) {
				break;
			} else if (token.name.equals("}")) {
				break;
			} else if (token.name.equals("\\")) {
//					if (par) {
//						T.insertChar('\n');
//						par=false;
//					}
					T.insertString(token.value);
			} else if (token.name.startsWith("{")) {
//					if (par) {
//						T.insertChar('\n');
//	    				par=false;
//					}
	    			if (token.name.startsWith("{\\fonttbl")) {
	    				boolean f0in=fonts.get("\\f0")!=null;
	    				addFonts(fonts,token.name.substring("{\\fonttbl".length()).trim());
	    				if (!f0in&&fonts.get("\\f0")!=null) {
	    					f.setName(JfontName((String) fonts.get("\\f0")));
	    					if ((T instanceof Text)&&!(T instanceof Formula)) {
	    						((Text)T).insertTFont(f.cloneTFont(),false);
	    					}
	    				}
	    			} else if (token.name.startsWith("{\\colortbl")) {
	    				addColors(colors,token.name.substring("{\\colortbl".length()).trim());
	    			} else if (token.name.startsWith(RTF_component)) {
	    				if (T instanceof Text) {
	    					String info=token.name.substring(RTF_component.length()+1).trim();
	    					StringTokenizer st=new StringTokenizer(info," }");
	    					String ctype=st.nextToken();
	    					String cid=st.nextToken();
	    					LivingCanvas Can=new LivingCanvas();
	    					((Text)T).insertComponent(Can,ctype,cid);
	    				}
	    			} else if (token.name.startsWith(RTF_hyperlink)) {
	    				if (T instanceof Text) {
	    					String info=token.name.substring(RTF_hyperlink.length()+1).trim();
	    					StringTokenizer st=new StringTokenizer(info,"|}");
	    					String label=st.nextToken();
	    					String url=st.nextToken();
	    					((Text)T).insertHyperlink(label,url);
	    				}
	    			} else if (token.name.startsWith(RTF_mjaformula)) {
	    				Formula F=null;
	    				if (T instanceof Formula) {
	    					F=(Formula)T;
	    				} else if (T instanceof Text) {
							F=new Formula(T.getParser(),T.getTFont(),bgcolor,fgcolor);
	    					((Text)T).insertFormula(F);
	    				}
						addText(token.name.substring(RTF_mjaformula.length()).trim(),F,fonts,colors);
						if (T instanceof Text) {
							F.setFonts(f.cloneTFont());
						}
					} else if (token.name.startsWith(RTF_expr)) {
						addMath((Formula)T,token.name.substring(RTF_expr.length()).trim(),TNode.EXPR,colors);
					} else if (token.name.startsWith(RTF_fraction)) {
						addMath((Formula)T,token.name.substring(RTF_fraction.length()).trim(),TNode.FRACT,colors);
					} else if (token.name.startsWith(RTF_product)) {
						addMath((Formula)T,token.name.substring(RTF_product.length()).trim(),TNode.MULT,colors);
					} else if (token.name.startsWith(RTF_radical)) {
						addMath((Formula)T,token.name.substring(RTF_radical.length()).trim(),TNode.RADICAL,colors);
					} else if (token.name.startsWith(RTF_sum)) {
						addMath((Formula)T,token.name.substring(RTF_sum.length()).trim(),TNode.SUM,colors);
					} else if (token.name.startsWith(RTF_integral)) {
						addMath((Formula)T,token.name.substring(RTF_integral.length()).trim(),TNode.INTEGRAL,colors);
					} else if (token.name.startsWith(RTF_limit)) {
						addMath((Formula)T,token.name.substring(RTF_limit.length()).trim(),TNode.LIMIT,colors);
					} else if (token.name.startsWith(RTF_matrix)) {
						addMath((Formula)T,token.name.substring(RTF_matrix.length()).trim(),TNode.MATRIX,colors);
					} else if (token.name.startsWith(RTF_defparts)) {
						addMath((Formula)T,token.name.substring(RTF_parts.length()).trim(),TNode.DEFPARTS,colors);
					} else if (token.name.startsWith(RTF_subix)) {
						addMath((Formula)T,token.name.substring(RTF_subix.length()).trim(),TNode.SUBIX,colors);
					} else if (token.name.startsWith(RTF_supix)) {
						addMath((Formula)T,token.name.substring(RTF_supix.length()).trim(),TNode.SUPIX,colors);
					}
			} else if (token.name.startsWith("\\")) {
				if (token.name.equals("\\f")) {
					Object o=fonts.get(token.name+token.value.trim());
					if (o!=null) {
						f.setName(JfontName((String) o));
	    				if (T instanceof Text) {
		    				((Text)T).insertTFont(f.cloneTFont(),false);
		    			}
					}
				} else if (token.name.equals("\\fs")) {
					try {
						f.setSize(Integer.parseInt(token.value.trim())/2);
	    				if (T instanceof Text) {
		    				((Text)T).insertTFont(f.cloneTFont(),false);
		    			}
					} catch (Exception e) {
					}
				} else if (token.name.equals("\\cf")||token.name.equals("\\cb")) {
					Object o=colors.get("\\cf"+token.value.trim());
					if (o!=null) {
						if (token.name.equals("\\cf")) {
							fgcolor=(Color)o;
							bgcolor=null;
						} else {
							bgcolor=(Color)o;
						}
	    				if (T instanceof Text) {
	    					TColors tc=new TColors(bgcolor,fgcolor);
	    					((Text)T).insertColors(tc,false);
	    				}
					}
				} else if (token.name.equals("\\u")) {
					try {
						char ch=(char)Integer.parseInt(token.value.trim());
//	    				if (par) {
//	    					T.insertChar('\n');
//		    				par=false;
//	    				}
	    				T.insertChar(ch);
					} catch (Exception e) {
					}
				} else if (token.name.startsWith("\\b")) {
					f.setBold(token.value.trim().length()==0);
					if (T instanceof Text) {
	    				((Text)T).insertTFont(f.cloneTFont(),false);
	    			}
				} else if (token.name.startsWith("\\i")) {
					f.setItalic(token.value.trim().length()==0);
					if (T instanceof Text) {
	    				((Text)T).insertTFont(f.cloneTFont(),false);
	    			}
				} else if (token.name.startsWith("\\ulnone")) {
					f.setUnderline(false);
					if (T instanceof Text) {
	    				((Text)T).insertTFont(f.cloneTFont(),false);
	    			}
				} else if (token.name.startsWith("\\ul")) {
					f.setUnderline(true);
					if (T instanceof Text) {
	    				((Text)T).insertTFont(f.cloneTFont(),false);
	    			}
				} else if (token.name.startsWith("\\olnone")) {
					f.setOverline(false);
					if (T instanceof Text) {
	    				((Text)T).insertTFont(f.cloneTFont(),false);
	    			}
				} else if (token.name.startsWith("\\ol")) {
					f.setOverline(true);
					if (T instanceof Text) {
	    				((Text)T).insertTFont(f.cloneTFont(),false);
	    			}
				} else if (token.name.startsWith("\\uc")) {
					try {
						RTFunicodeCount=Integer.parseInt(token.value.trim());
					} catch (Exception e) {
					}
				} else if (token.name.startsWith("\\par")) {
//					if (par) {
						T.insertChar('\n');
//					}
//					par=true;
				}
			} else if (isWhiteSpace(token.name.charAt(0))) {
			} else {
//				if (par) {
//					T.insertChar('\n');
//					par=false;
//				}
				T.insertString(token.name);
			}
			pos+=token.name.length()+token.value.length();
		}
		T.setCursorPos(T.firstChar(T.begin()));
//System.out.println("FINAL NODES "+T.TNode(T.end()-1).toString()+","+T.TNode(T.end()).toString());
    }

    private static void addMath(Formula T,String str,int type,Hashtable colors) {
    	Color fgcolor=T.getColors().getTextColor();
    	Color bgcolor=T.getColors().getBackColor();
    	Formula auxT=null,numT=null,denT=null,indexT=null,radicandT=null,fromT=null,toT=null,whatT=null;
    	Formula[][] matrixT=null;
    	Formula[] defpartsT=null;
    	int nrows=-1,ncolumns=-1,irow=-1,icolumn=-1;
    	int nparts=-1,ipart=-1;
		auxT=new Formula(T.getParser(),T.getTFont(),T.getColors().getBackColor(),T.getColors().getTextColor());
		if (type==TNode.FRACT) {
			numT=new Formula(T.getParser(),T.getTFont(),T.getColors().getBackColor(),T.getColors().getTextColor());
			denT=new Formula(T.getParser(),T.getTFont(),T.getColors().getBackColor(),T.getColors().getTextColor());
		} else if (type==TNode.RADICAL) {
			indexT=new Formula(T.getParser(),T.getTFont(),T.getColors().getBackColor(),T.getColors().getTextColor());
			radicandT=new Formula(T.getParser(),T.getTFont(),T.getColors().getBackColor(),T.getColors().getTextColor());
		} else if (type==TNode.SUM||type==TNode.INTEGRAL||type==TNode.LIMIT) {
			fromT=new Formula(T.getParser(),T.getTFont(),T.getColors().getBackColor(),T.getColors().getTextColor());
			toT=new Formula(T.getParser(),T.getTFont(),T.getColors().getBackColor(),T.getColors().getTextColor());
			whatT=new Formula(T.getParser(),T.getTFont(),T.getColors().getBackColor(),T.getColors().getTextColor());
//		} else if (type==TNode.COMPONENT) {
//System.out.println("type==COMPONENT");
//			auxT=T;
		} else if (type==-1) {
			auxT=T;
		}
		/* for type=TNode.EXPR */
    	String expr="";
    	String decimals=null,rows=null,columns=null,parts=null;
    	/* ------------------- */
    	boolean fixed=false;
    	int pos=0;
		while (pos<str.length()) {
			Attribute token=getNextToken(str,pos);
//			if (type==TNode.RADICAL&&pos==0&&!token.name.startsWith("{")) {
			if (type==TNode.RADICAL&&pos==0&&!token.name.startsWith(RTF_index)) {
				auxT=radicandT;
			}
			if (token.name.length()==0) {
				break;
			} else if (token.name.equals("}")) {
				break;
			} else if (token.name.startsWith("{")) {
				if (token.name.startsWith(RTF_component)) {
					String info=token.name.substring(RTF_component.length()+1).trim();
					StringTokenizer st=new StringTokenizer(info," }");
					String ctype=st.nextToken();
					String cid=st.nextToken();
					LivingCanvas Can=new LivingCanvas();
					T.insertComponent(Can,ctype,cid);
				} else if (token.name.startsWith(RTF_expr)) {
					addMath(auxT,token.name.substring(RTF_expr.length()).trim(),TNode.EXPR,colors);
				} else if (token.name.startsWith(RTF_fraction)) {
					addMath(auxT,token.name.substring(RTF_fraction.length()).trim(),TNode.FRACT,colors);
				} else if (token.name.startsWith(RTF_product)) {
					addMath(auxT,token.name.substring(RTF_product.length()).trim(),TNode.MULT,colors);
				} else if (token.name.startsWith(RTF_radical)) {
					addMath(auxT,token.name.substring(RTF_radical.length()).trim(),TNode.RADICAL,colors);
				} else if (token.name.startsWith(RTF_sum)) {
					addMath(auxT,token.name.substring(RTF_sum.length()).trim(),TNode.SUM,colors);
				} else if (token.name.startsWith(RTF_integral)) {
					addMath(auxT,token.name.substring(RTF_integral.length()).trim(),TNode.INTEGRAL,colors);
				} else if (token.name.startsWith(RTF_limit)) {
					addMath(auxT,token.name.substring(RTF_limit.length()).trim(),TNode.LIMIT,colors);
				} else if (token.name.startsWith(RTF_matrix)) {
					addMath(auxT,token.name.substring(RTF_matrix.length()).trim(),TNode.MATRIX,colors);
				} else if (token.name.startsWith(RTF_defparts)) {
					addMath(auxT,token.name.substring(RTF_defparts.length()).trim(),TNode.DEFPARTS,colors);
				} else if (token.name.startsWith(RTF_subix)) {
					addMath(auxT,token.name.substring(RTF_subix.length()).trim(),TNode.SUBIX,colors);
				} else if (token.name.startsWith(RTF_supix)) {
					addMath(auxT,token.name.substring(RTF_supix.length()).trim(),TNode.SUPIX,colors);
				} else if (type==TNode.FRACT) {
					if (token.name.startsWith(RTF_num)) {
						addMath(numT,token.name.substring(RTF_num.length()).trim(),-1,colors);
					} else if (token.name.startsWith(RTF_den)) {
						addMath(denT,token.name.substring(RTF_den.length()).trim(),-1,colors);
					}
				} else if (type==TNode.RADICAL) {
					if (token.name.startsWith(RTF_index)) {
						addMath(indexT,token.name.substring(RTF_index.length()).trim(),-1,colors);
					} else if (token.name.startsWith(RTF_radicand)) {
						addMath(radicandT,token.name.substring(RTF_radicand.length()).trim(),-1,colors);
					}
				} else if (type==TNode.SUM||type==TNode.INTEGRAL||type==TNode.LIMIT) {
					if (token.name.startsWith(RTF_from)) {
						addMath(fromT,token.name.substring(RTF_from.length()).trim(),-1,colors);
					} else if (token.name.startsWith(RTF_to)) {
						addMath(toT,token.name.substring(RTF_to.length()).trim(),-1,colors);
					} else if (token.name.startsWith(RTF_what)) {
						addMath(whatT,token.name.substring(RTF_what.length()).trim(),-1,colors);
					}
				} else if (type==TNode.MATRIX) {
					if (token.name.startsWith(RTF_element)) {
						if (matrixT==null) {
							nrows=Integer.parseInt(rows);
							ncolumns=Integer.parseInt(columns);
							matrixT=new Formula[ncolumns][nrows];
							irow=0;icolumn=0;
						}
						matrixT[icolumn][irow]=new Formula(T.getParser(),T.getTFont(),T.getColors().getBackColor(),T.getColors().getTextColor());
						addMath(matrixT[icolumn][irow],
							token.name.substring(RTF_element.length()).trim(),-1,colors);
						if (irow+1<nrows) {
							irow++;
						} else if (icolumn+1<ncolumns) {
							irow=0;
							icolumn++;
						}
					}
				} else if (type==TNode.DEFPARTS) {
					if (token.name.startsWith(RTF_element)) {
						if (defpartsT==null) {
							nparts=Integer.parseInt(parts);
							defpartsT=new Formula[nparts];
							ipart=0;
						}
						defpartsT[ipart]=new Formula(T.getParser(),T.getTFont(),T.getColors().getBackColor(),T.getColors().getTextColor());
						addMath(defpartsT[ipart],
							token.name.substring(RTF_element.length()).trim(),-1,colors);
						if (ipart+1<nparts) {
							ipart++;
						}
					}
				}
			} else if (token.name.startsWith("\\")) {
				if (token.name.equals("\\cf")||token.name.equals("\\cb")) {
					Object o=colors.get("\\cf"+token.value.trim());
					if (o!=null) {
						if (token.name.equals("\\cf")) {
							fgcolor=(Color)o;
							bgcolor=null;
						} else {
							bgcolor=(Color)o;
						}
    					TColors tc=new TColors(bgcolor,fgcolor);
    					auxT.insertColors(tc,false);
					}
	    		} else if (token.name.equals("\\u")) {
					try {
						String s=token.value;
						if (s.endsWith("}")) {
							s=s.substring(0,s.length()-1);
						}
						char ch=(char)Integer.parseInt(s.trim());
						if (type==TNode.EXPR) { // agregado el 13 de marzo de 2008, ver 3.919-3.920
	    					expr+=ch;
						} else {
	    					auxT.insertChar(ch);
						}
					} catch (Exception e) {
					}
	    		} else if (type==TNode.EXPR) {
	    			if (token.name.startsWith(RTF_decimals)) {
	    				decimals="";
	    			} else if (token.name.startsWith(RTF_fixed)) {
	    				fixed=("1".equals(token.value.trim()));
	    			}
	    		} else if (type==TNode.MATRIX) {
	    			if (token.name.startsWith(RTF_rows)) {
	    				rows="";
	    			} else if (token.name.startsWith(RTF_columns)) {
	    				columns="";
	    			}
	    		} else if (type==TNode.DEFPARTS) {
	    			if (token.name.startsWith(RTF_parts)) {
	    				parts="";
	    			}
				}
			} else if (isWhiteSpace(token.name.charAt(0))) {
			} else {
				switch (type) {
				case -1:
	   					auxT.insertString(token.name);
	   					break;
				case TNode.EXPR:
						if (decimals==null) {
	   						expr+=token.name.trim();
						} else {
	   						decimals+=token.name.trim();
						}
						break;
				case TNode.MATRIX:
						if (columns!=null) {
	   						columns+=token.name.trim();
						} else if (rows!=null) {
	   						rows+=token.name.trim();
						} else {
	   						expr+=token.name.trim();
	   					}
						break;
				case TNode.DEFPARTS:
						if (parts!=null) {
	   						parts+=token.name.trim();
						} else {
	   						expr+=token.name.trim();
	   					}
						break;
				default:
	   					auxT.insertString(token.name);
				}
			}
			pos+=token.name.length()+token.value.length();
		}
		switch (type) {
		case -1:
				break;
		case TNode.EXPR:
			T.insertExpr(expr,decimals,fixed);
			break;
		case TNode.FRACT:
			T.insertFraction(numT,denT);
			break;
		case TNode.MULT:
			T.insertProduct(numT,denT);
			break;
		case TNode.RADICAL:
			T.insertRadical(indexT,radicandT);
			break;
		case TNode.SUM:
			T.insertSum(fromT,toT,whatT);
			break;
		case TNode.INTEGRAL:
			T.insertIntegral(fromT,toT,whatT);
			break;
		case TNode.LIMIT:
			T.insertLimit(fromT,toT,whatT);
			break;
		case TNode.MATRIX:
			T.insertMatrix(matrixT);
			break;
		case TNode.DEFPARTS:
			T.insertDefParts(defpartsT);
			break;
		case TNode.SUBIX:
			T.insertSubIndex(auxT);
			break;
		case TNode.SUPIX:
			T.insertSuperIndex(auxT);
			break;
		}
    }

    private static void addFonts(Hashtable ht,String str) {
    	int pos=0;
    	String fontKey="";
		while (pos<str.length()) {
			Attribute token=getNextToken(str,pos);
			if (token.name.length()==0) {
				break;
			} else if (token.name.equals("}")) {
				break;
			} else if (token.name.equals("\\")) {
			} else if (token.name.startsWith("{")) {
			} else if (token.name.startsWith("\\")) {
				if (token.name.equals("\\f")) {
					fontKey=token.name+token.value.trim();
				}
			} else if (isWhiteSpace(token.name.charAt(0))) {
			} else {
				String fontName=token.name;
				if (fontName.endsWith(";")) {
					fontName=fontName.substring(0,fontName.length()-1);
				}
				if (fontKey.length()>0) {
					ht.put(fontKey,fontName);
				}
				fontKey="";
			}
			pos+=token.name.length()+token.value.length();
		}
    }

    private static void addColors(Hashtable ht,String str) {
    	int pos=0;
    	String red="0",green="0",blue="0";
		while (pos<str.length()) {
			Attribute token=getNextToken(str,pos);
			if (token.name.length()==0) {
				break;
			} else if (token.name.equals("}")) {
				break;
			} else if (token.name.equals("\\")) {
			} else if (token.name.startsWith("{")) {
			} else if (token.name.startsWith("\\")) {
				if (token.name.startsWith("\\red")) {
					red=token.value.trim();
				} else if (token.name.startsWith("\\green")) {
					green=token.value.trim();
				} else if (token.name.startsWith("\\blue")) {
					blue=token.value.trim();
				}
			} else if (isWhiteSpace(token.name.charAt(0))) {
			} else {
				if (token.name.startsWith(";")) {
					try {
						int r=Integer.parseInt(red);
						int g=Integer.parseInt(green);
						int b=Integer.parseInt(blue);
						Color c=new Color(r,g,b);
						String key="\\cf"+String.valueOf(ht.size());
						ht.put(key,c);
					} catch (Exception e) {
					}
					red="0"; green="0"; blue="0";
				}
			}
			pos+=token.name.length()+token.value.length();
		}
    }

    private static Attribute getNextToken(String str,int pos) {
    	if (str.charAt(pos)=='}') {
    		return new Attribute("}","");
    	} else if (str.charAt(pos)=='{') {
    		int i=pos+1;
    		int keylevel=1;
    		while (i<str.length()&&keylevel>0) {
    			if (str.charAt(i)=='}') {
    				keylevel--;
    			} else if (str.charAt(i)=='{') {
    				keylevel++;
    			}
    			i++;
    		}
    		return new Attribute(str.substring(pos,i),"");
    	} else if (isWhiteSpace(str.charAt(pos))) {
    		return new Attribute((new Character(str.charAt(pos))).toString(),"");
    	} else if (str.charAt(pos)=='\\') {
	    	if (pos+1<str.length()&&isCtrlChar(str.charAt(pos+1))) {
	    		return new Attribute("\\",new Character(str.charAt(pos+1)).toString());
	    	} else {
		    	int i=pos+1;
	    		while (i<str.length()&&!isCtrlSep(str.charAt(i))) {
	    			i++;
	    		}
	    		if (i<str.length()&&str.charAt(i)==' ') {
	    			i++;
	    		}
	    		String ctrl=str.substring(pos,i);
	    		int j=1;
	    		while (j<ctrl.length()&&!isNumChar(ctrl.charAt(j))) {
	    			j++;
	    		}
	    		String name=ctrl.substring(0,j);
	    		String value=ctrl.substring(j);
    			if (name.equals("\\u")) {
    				value+=str.substring(i,i+RTFunicodeCount);
    			}
	    		return new Attribute(name,value);
	    	}
    	} else {
	    	int i=pos+1;
    		while (i<str.length()&&!isCtrlChar(str.charAt(i))&&!isWhiteSpace(str.charAt(i))) {
    			i++;
    		}
    		return new Attribute(str.substring(pos,i),"");
    	}
    }

    private static boolean isCtrlSep(char ch) {
    	return (ch==' '||ch==';'||isWhiteSpace(ch)||isCtrlChar(ch));
    }

    private static boolean isWhiteSpace(char ch) {
    	return (ch=='\r'||ch=='\n');
    }

    private static boolean isCtrlChar(char ch) {
    	return (ch=='\\'||ch=='{'||ch=='}');
    }

    private static boolean isNumChar(char ch) {
    	return (ch=='-'||(ch>='0'&&ch<='9'));
    }

	static void appendAsChar(StringBuffer sb,char ch) {
		if (ch=='{' || ch=='}'|| ch=='\\') {
			sb.append('\\');
		}
		sb.append(ch);
	}


/* --------------------------TeX-------------------------*/

	private	StringBuffer TEXsb;
	public static final int tiny=0, scriptsize=1,footnotesize=2,small=3,normalsize=4,large=5,Large=6,LARGE=7,huge=8,Huge=9;
	private boolean sizeDefined=false;
    private static int mathModeIx=0;
    private static final String 
    	ffamily_rm="\\rmfamily ", ffamily_tt="\\ttfamily ", ffamily_ss="\\sffamily ",
    	fsize_normal="\\normalsize ",
    	fseries_md="\\mdseries ", fseries_bf="\\bfseries ",
    	fshape_normal="\\normalfont ", fshape_sl="\\slshape ", fshape_sc="\\scshape " ;
	private static String fsize=fsize_normal,ffamily=ffamily_rm,fseries=fseries_md,fshape=fshape_normal;

	private Color defaultColor;
	private boolean useOwnColors;

	public String toTeX (Color defaultColor,boolean useOwnColors) {
		this.defaultColor=defaultColor;
		this.useOwnColors=useOwnColors;
   		TEXsb=new StringBuffer();
    	mathModeIx=0;
    	ffamily=ffamily_rm;
    	fsize=fsize_normal;
    	fseries=fseries_md;
    	fshape=fshape_normal;
   		writeTEXText(ffamily);
   		writeTEXText(fsize);
   		writeTEXText(fseries);
   		writeTEXText(fshape);
    	writeTEXDocument(T);
	    return TEXsb.toString();
	}
	
	private void writeTEXText(String s) {
		for(int pos=0; pos < s.length(); pos ++) {
			TEXsb.append(s.charAt(pos));
		}
	}

    private void writeTEXMath(TNode tn) {
		if (tn.isExpr()) {
			String str=tn.getEval();
			for (int i=0;i<str.length();i++) {
				char ch=str.charAt(i);
				writeTEXText(BasicStr.TeXchar(ch,mathModeIx>0));
			}
		} else if (tn.isFraction()) {
		    writeTEXText("\\frac{");
		    writeTEXDocument(tn.getFI().getFormula(FormulaItem.NUM));
			writeTEXText("}{");
		    writeTEXDocument(tn.getFI().getFormula(FormulaItem.DEN));
			writeTEXText("}");
		} else if (tn.isProduct()) {
		} else if (tn.isRadical()) {
		    if (BasicStr.hasContent(tn.getFI().getFormula(FormulaItem.INDEX).toExpresion(false))) {
			    writeTEXText("\\root {");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.INDEX));
				writeTEXText("} \\of {");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.RADICAND));
				writeTEXText("}");
		    } else {
			    writeTEXText("\\sqrt{");
	    		writeTEXDocument(tn.getFI().getFormula(FormulaItem.RADICAND));
				writeTEXText("}");
	    	}
		} else if (tn.isSum()) {
			    writeTEXText("\\sum\\sb {");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.FROM));
				writeTEXText("} \\sp {");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.TO));
				writeTEXText("}");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.WHAT));
		} else if (tn.isIntegral()) {
			    writeTEXText("\\int\\sb {");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.FROM));
				writeTEXText("} \\sp {");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.TO));
				writeTEXText("}");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.WHAT));
		} else if (tn.isLimit()) {
			    writeTEXText("\\lim\\sb {");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.FROM));
				writeTEXText(" \\rightarrow ");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.TO));
				writeTEXText("}");
			    writeTEXDocument(tn.getFI().getFormula(FormulaItem.WHAT));
		} else if (tn.isSubIndex()) {
			writeTEXText("\\sb{");
	    	writeTEXDocument(tn.getFI().getFormula(0));
			writeTEXText("}");
		} else if (tn.isSuperIndex()) {
			writeTEXText("\\sp{");
	    	writeTEXDocument(tn.getFI().getFormula(0));
			writeTEXText("}");
		} else if (tn.isMatrix()) {
		} else if (tn.isDefParts()) {
		} else {
	    	writeTEXDocument(tn.getFI().getFormula(0));
		}
    }
    
    private int ncolor=0;
    private void writeTEXDocument(BasicText bt) {
		String colorname="fc";
		for (int i=0;i<bt.end();i++) {
			TNode tn=bt.TNode(i);
			if (tn.isColors()) {
				TColors tc=tn.getColors();
				Color textcolor=defaultColor;
				if (useOwnColors) {
					textcolor=tc.getTextColor();
				} 
//				Color backcolor=tc.getBackColor();
				colorname="tcolor"+BasicStr.justify(""+ncolor,-3,'0');
				writeTEXText("\\color{"+colorname+"} ");
				ncolor++;
			} else if (tn.isTFont()) {
				String ffamily_a=ffamily;
				String fsize_a=fsize;
				String fseries_a=fseries;
				String fshape_a=fshape;
				TFont tfont=tn.getTFont();
				if (tfont.isSerif()) {
					ffamily=ffamily_rm;
				} else if (tfont.isMonospaced()) {
					ffamily=ffamily_tt;
				} else {
					ffamily=ffamily_ss;
				}
				int size=Math.min(Math.max(0,(int) Math.floor(((double) tfont.getSize()-6.0)/3.5)),9);
		   		switch (size) {
		   			case tiny:fsize="\\tiny "; break;
		   			case scriptsize: fsize="\\scriptsize "; break;
		   			case footnotesize: fsize="\\footnotesize "; break;
		   			case small: fsize="\\small "; break;
		   			case normalsize: fsize="\\normalsize "; break;
		   			case large: fsize="\\large "; break;
		   			case Large: fsize="\\Large "; break;
		   			case LARGE: fsize="\\LARGE "; break;
		   			case huge: fsize="\\huge "; break;
		   			case Huge: fsize="\\Huge "; break;
		   			default: fsize=fsize_normal; break;
		   		}
		   		int style=tfont.getStyle();
		   		if (style==0) {
		   			fshape=fshape_normal;
		   			fseries=fseries_md;
		   		} else if (style==1) {
		   			fshape=fshape_normal;
		   			fseries=fseries_bf;
		   		} else if (style==2) {
		   			fshape=fshape_sl;
		   			fseries=fseries_md;
		   		} else if (style==3) {
		   			fshape=fshape_sl;
		   			fseries=fseries_bf;
		   		} else {
		   			fshape=fshape_sc;
		   			fseries=fseries_md;
		   		}
		   		if (!fsize.equals(ffamily_a)) { writeTEXText(ffamily); }
		   		if (!fsize.equals(fsize_a)) { writeTEXText(fsize); }
		   		if (!fseries.equals(fseries_a)) { writeTEXText(fseries); }
		   		if (!fshape.equals(fshape_a)) { writeTEXText(fshape); }
			} else if (tn.isComponent()) {
			} else if (tn.isHyperlink()) {
			} else if (tn.isMath()) {
		    	if (mathModeIx==0) {writeTEXText("$");}
		    	mathModeIx++;
				writeTEXMath(tn);
				mathModeIx--;
		    	if (mathModeIx==0) {writeTEXText("$");}
			} else if (tn.isChar()) {
				String s=tn.getChar();
				if (s.equals("\n")) {
					writeTEXText("\\\\ ");
					writeTEXText("\\color{"+colorname+"} "); 
		   			writeTEXText(fshape); 
		   			writeTEXText(fseries);
		   			writeTEXText(fsize);
				   	writeTEXText(ffamily);
    			} else if ((s.charAt(0)>=0x0391)&&(s.charAt(0)<=0x03C9)) {
    				if (mathModeIx==0) { writeTEXText("$"); }
    				writeTEXText(BasicStr.TeXgreek(s.charAt(0)));
    				if (mathModeIx==0) { writeTEXText("$"); }
	    		} else {
					for (int j=0;j<s.length();j++) {
						char ch=s.charAt(j);
						writeTEXText(BasicStr.TeXchar(ch,mathModeIx>0));
					}
			   	}
			}
		}
    }
    
/* ----------------------- end TeX ------------------------- */


}