/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.text;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.net.URL;
import java.awt.image.BufferedImage;

import com.mja.parser.*;
import com.mja.lang.*;
import com.mja.img.*;

import com.mja.util.TFont;
import com.mja.text.BasicText;
import com.mja.text.Text;
import com.mja.text.Formula;
import com.mja.text.TColors;
import com.mja.text.TNode;
import com.mja.text.RTF;
import com.mja.util.BasicStr;
import com.mja.descartes.Descartes;
import com.mja.descartes.Space;
import com.mja.descartes.spaceConfig;

//import java.awt.datatransfer.*;

public class TextCanvas extends Panel implements MouseListener, MouseMotionListener {

	public static final String UpdateSelectors="UpdateSelectors",
								ActivateFormulaButtons="ActivateFormulaButtons",
								ActivateTextButtons="ActivateTextButtons";

    protected boolean editing, learnerMode, inVerticalMotion, onlyFormula, hasFocus, fixed=false;
    protected volatile boolean updating;
    protected volatile Formula mainF;
    protected volatile Formula F;
	protected Text T;
    protected int dec=2, aL=0;
    protected volatile int mainFIndex;
	protected Applet A;
    protected Frame parent;
    protected translator Tr;
    protected Parser parser;
	protected volatile BufferedImage buffer;

	private Graphics bG;
    private int minH;
    private boolean variableSize;

	public TextCanvas(Applet A,Frame parent,translator Tr,Parser p,TFont f,Color c,boolean variableSize,boolean onlyFormula) {
		this.A=A;
		this.parent=parent;
		this.Tr=Tr;
		this.parser=p;
		this.variableSize=variableSize;
		this.onlyFormula=onlyFormula;
		setLayout(null);
		setText(new Text(parser,f,Color.white,c));
		setBackground(Color.white);
		addMouseListener(this);
		addMouseMotionListener(this);
	}
	
	public void setText(String s,Font f,Color bc,Color c) {
		TFont tf=new TFont(f);
		minH=(tf.getAscent()+tf.getDescent());
		setText(new Text(parser,tf,bc,c));
	}

    public boolean updateTextComponent(Component Cmp,String cID) {
    	return T.updateNodes(Cmp,cID);
    }
    
    private boolean started=false;

	public boolean start() {
		if (!started) {
			removeMouseListener(this);
			removeMouseMotionListener(this);
			setCursorPos(0);
			if (updateSize(true)) {
				updateScreen();	
			}
			addMouseListener(this);
			addMouseMotionListener(this);
			started=(buffer!=null);
		}
		return started;
	}


	public void setFont(Font f) {
		super.setFont(f);
		T.setTFont(new TFont(f));
	}

	public void setForeground(Color c) {
		T.setColors(T.getColors().getBackColor(),c);
		super.setForeground(c);
	}

	public void setBackground(Color c) {
		T.setColors(c,T.getColors().getTextColor());
		super.setBackground(c);
	}

	public void setCursorPos(int pos) {
		if (F!=null) {
			F.setCursorPos(Math.min(Math.max(pos,F.begin()),F.end()));
		} else {
			T.setCursorPos(Math.min(Math.max(pos,T.begin()),T.end()));
		}
	}

	public void addActionListener(ActionListener al) {
		// necesaria para que EditorCanvas pueda sobreescribirla
	}

    public boolean onlyFormula() {
    	return onlyFormula;
    }
    
    public BufferedImage getBufferedImage() {
    	return buffer;
    }

    /* --------------------- General Public methods ----------------- */

	public boolean  isEditing() {
		return editing;
	}

	public void setExprDefaults(int dec,boolean fixed,int aL) {
		this.dec=dec;
		this.fixed=fixed;
		this.aL=aL;
	}

    public String toRTF(boolean plain) {
    	RTF rtf=new RTF(T);
    	String s=rtf.toRTF(plain);
    	return s;
    }

	public Formula getMainFormula() {
		return mainF;
	}

	public Formula getFormula() {
		return F;
	}

	public void setMainFormula(Formula F) {
		T.TNode(mainFIndex).getFI().setFormula(0,F);
		mainF=F;
		this.F=F;
	}

	public String getCharAtCursor() {
		TNode tn=null;
		if (F!=null) {
			tn=F.TNode(F.firstChar(F.getCursorPos()));
		} else {
			tn=T.TNode(T.firstChar(T.getCursorPos()));
		}
		if (tn.isPureChar()) {
			String s=tn.getChar();
			if (s!=null&&s.length()>=1&&!(s.charAt(0)=='\n')&&!(s.charAt(0)=='\u001A')) {
				char ch=s.charAt(0);
				return "0x"+Integer.toHexString((int)ch);
			}
		}
		return "";
	}

	/* ----------------- dimensions ---------------------- */

	public Dimension getPreferredSize() {
		return getMinimumSize();
	}

	public Dimension getMinimumSize() {
		return getSize();
	}

	public void setMinimumHeight(int h) {
		minH=h;
	}

	public int getMinimumHeight() {
		return minH;
	}

	public int getCanvasHeight() {
		return  2+Math.max(getTextHeight(),getMinimumHeight());
	}

    public int getTextHeight() {
    	return T.getHeight(this);
    }

    public int getTextWidth() {
    	return T.getWidth(this);
    }

	/*---------------- Text ----------------*/

	public void setText(Text T) {
		this.T=T;
		if (onlyFormula) {
			int N=T.firstFormula(T.begin());
			if (N<0) {
				T.setCursorPos(T.firstChar(T.begin()));
				T.insertFormula(new Formula(parser,T.getTFont()));
				N=T.firstFormula(T.begin());
			}
			mainFIndex=N;
			T.setCursorPos(mainFIndex);
			mainF=T.TNode(mainFIndex).getFI().getFormula(0);
			F=mainF;
		} else {
			mainFIndex=-1;
			mainF=null;
			F=null;
		}
		if (getFont()!=null) {
			T.setTFont(new TFont(getFont()));
		}
		T.setParser(T.parser);
	}

	public Text getText() {
		return T;
	}

	public TFont getFirstFont() {
		return T.getTFontAtTNode(T.firstChar(T.begin()));
	}

    /* ---------------------- Text and Formulas ------------------------- */

    public TColors getCursorColors() {
    	if (F!=null) {
	    	if (F.segMarked()) {
				return F.getColorsAtTNode(F.segMin());
	    	} else {
				return F.getColorsAtTNode(F.getCursorPos());
			}
		} else {
	    	if (T.segMarked()) {
				return T.getColorsAtTNode(T.segMin());
	    	} else {
				return T.getColorsAtTNode(T.getCursorPos());
			}
		}
    }

    public TFont getCursorTFont() {
    	if (F!=null) {
    		return T.getTFontAtTNode(mainFIndex);
		} else {
	    	if (T.segMarked()) {
				return T.getTFontAtTNode(T.segMin());
	    	} else {
	    		TFont f=T.getTFontAtTNode(T.getCursorPos());
				return f;
			}
		}
    }

	/* -------------------- editing functions ------------------- */

    public void insertColors(TColors tc) {
    	if (F!=null) {
	   		F.insertColors(tc,true);
			updateScreenFromTNode(mainFIndex);
    	} else {
	   		T.insertColors(tc,true);
			if (T.segMarked()) {
				T.clean();
				updateScreenSegment(T.segMin(),T.segMax());
			} else {
	    		updateScreenLineAtTNode(T.getCursorPos());
		    }
		  }
		if (updateSize()) {
			updateScreen();
		}
    }

    public void insertTFont(TFont f) {
    	if (F!=null) {
			T.markSegment(T.getCursorPos(),T.getCursorPos()+1);
			T.insertTFont(f,true);
			T.unmarkSegment();
			T.clean();
			mainFIndex=T.getCursorPos();
			updateScreenFromTNode(mainFIndex);
    	} else {
	    	T.insertTFont(f,true);
			if (T.segMarked()) {
				updateScreenFromTNode(T.segMin());
			} else {
				updateScreenFromTNode(T.getCursorPos());
			}
		}
		if (updateSize()) {
			updateScreen();
		}
    }

	public void insertFormula(String s) {
		TFont tf0=getCursorTFont();
		int fsz=tf0.getSize();
		TFont tf=new TFont(new Font("Serif",Font.ITALIC,fsz+2));
		T.insertTFont(tf,false);
   		Formula form=new Formula(parser,tf,getCursorColors().getBackColor(),getCursorColors().getTextColor());
   		form.insertString(s);
		insertFormula(form);
		T.insertTFont(tf0,false);
		if (updateSize()) {
			updateScreen();
		} else {
			updateScreenFromTNode(mainFIndex);
		}
	}

    private void insertFormula(Formula f) {
		if (T.segMarked()) {
			removeMarkedSegment();
		}
		mainFIndex=T.getCursorPos();
    	T.insertFormula(f);
    	f.setActive(true);
    	setMainFormula(f);
    }

    public void insertSpace(int type) {
    	Descartes D=(Descartes) A;
    	String newname=D.scene.getNewSpaceName();
    	String cID=Long.toString(System.currentTimeMillis(),16);
    	String spcfgstr="id='"+newname+"' cID='"+cID+"'";
    	Space sp=null;
    	switch (type) {
    		case spaceConfig.dosD : spcfgstr+="tipo='R2' x='0' y='0' ancho='360' alto='240' despl_imagen='arr-izq' fondo='f0f8fa' red='b8c4c8' red10='889498' ejes='405860' texto='no'";
	    							break;
    		case spaceConfig.tresD: spcfgstr+="tipo='R3' x='0' y='0' ancho='360' alto='240' escala='64' despl_imagen='arr-izq' fondo='e0ecef' R3='si'";
	    							break;
    		case spaceConfig.alg:   spcfgstr+="tipo='TA' x='0' y='0' ancho='360' alto='240' fuente puntos='17' smbolo de multiplicacin='punto' ecuacin='4x-8=2x+6'";
	    							break;
    		case spaceConfig.appl:  spcfgstr+="tipo='AP' x='0' y='0' ancho='360' alto='240' fuente puntos='17' despl_imagen='arr-izq' fondo='f0f8fa' red='b8c4c8' red10='889498' ejes='405860' ";
	    							break;
    		case spaceConfig.htmliframe:  spcfgstr+="tipo='HTMLIFrame' ancho='480' alto='360' archivo='http://arquimedes.matem.unam.mx'";
	    							break;
    				
    	}
    	D.cfg.addSpace(spcfgstr);
    	sp=D.cfg.sc[D.cfg.sc.length-1].makeSpace(D);
    	if (sp!=null) {
//    		sp.setVisible(false);
	    	sp.TextBound=true;
	    	sp.preSize();
	    	int h0=T.getLineHeightAtTNode(this,getCursorPos());
	    	T.insertComponent(sp,"Space",cID);
	    	if (D.getEditor()!=null) {
	    		D.getEditor().setInfo();
	    	}
		    if (updateSize(false)) {
		    	updateScreen();
		    } else {
		    	updateScreenLineAtTNode(getCursorPos());
		    	updateScreenLineAtTNode(getCursorPos());  // se repite para que se acomode bien
		    	int h1=T.getLineHeightAtTNode(this,getCursorPos());
		    	if (h1>h0) {
			    	scrollFromTNode(getCursorPos(),h0-h1);
			    	int nl=T.nextLine(getCursorPos());
	    			updateScreenLineAtTNode(nl);
		    	}
		    }
		    sp.updateSpace(true,true);
    	}
    	requestFocus();
    }

    public void insertChar(char ch) {
   		insertText((new Character(ch)).toString());
    }

    public void insertText(String s) {
    	if (F!=null) {
	    	if (editing) {
	   			if (F.segMarked()) {
	   				removeMarkedSegment();
	   			}
		    	int lh=T.getLineHeightAtTNode(this,mainFIndex);
		    	F.insertText(s);
		    	if (lh!=T.getLineHeightAtTNode(this,mainFIndex)) {
					updateScreenFromTNode(mainFIndex);
		    	} else {
					updateScreenLineAtTNode(mainFIndex);
		    	}
			}
    	} else {
			if (T.segMarked()) {
				removeMarkedSegment();
			}
	    	TNode tn=T.TNode(T.getCursorPos());
	    	int h0=T.getLineHeightAtTNode(this,T.getCursorPos());
			if (s.length()==1 && s.charAt(0)>255) {
				TFont tf0=getCursorTFont();
				int fsz=tf0.getSize();
				TFont tf=new TFont(new Font("Serif",Font.ITALIC,fsz+2));
				T.insertTFont(tf,false);
		   		T.insertText(s);
				T.insertTFont(tf0,true);
			} else {
	    		T.insertText(s);
	    	}
	    	if (updateSize()) {
	    		updateScreen();
	    	} else {
	    		int h1=T.getLineHeightAtTNode(this,T.getCursorPos()-1);
	    		int h2=T.getLineHeightAtTNode(this,T.getCursorPos());
		    	if (s.equals("\n")) {
					updateScreenLineAtTNode(T.getCursorPos()-1);
					scrollFromTNode(T.getCursorPos(),-Math.min(h1,h2));
				} else if (h0!=h1) {
					scrollFromTNode(T.getCursorPos(),h0-h2);
				}
				updateScreenLineAtTNode(T.getCursorPos());
			}
		}
	    if (updateSize(false)) {
	    	updateScreen();
	    }
	}

/* --------------------------- end insertions --------------------------- */

    protected void delete() {
    	if (F!=null) {
	    	if (editing) {
				int h1=T.getLineHeightAtTNode(this,mainFIndex);
				F.delete();
				updateScreenLineAtTNode(mainFIndex);
				int h2=T.getLineHeightAtTNode(this,mainFIndex);
				if (h1>h2) {
					scrollFromTNode(mainFIndex,h1-h2);
				}
			}
    	} else {
	    	TNode tn=T.TNode(T.getCursorPos());
	    	int h0=T.getLineHeightAtTNode(this,T.getCursorPos());
	    	int h1=T.getLineHeightAtTNode(this,T.getCursorPos()+1);
	    	boolean lineJump="\n".equals(tn.getChar());
	    	if (T.delete()) {
		    	updateScreenLineAtTNode(T.getCursorPos());
		    	int h2=T.getLineHeightAtTNode(this,T.getCursorPos());
		    	if (lineJump) {
			    	scrollFromTNode(T.getCursorPos(),h0+h1-h2);
		    	} else if (h0>h2) {
			    	scrollFromTNode(T.getCursorPos(),h0-h2);
			    }
			}
		}
    }

    protected void backSpace() {
    	if (F!=null) {
	    	if (editing) {
				int h1=T.getLineHeightAtTNode(this,mainFIndex);
				if (F.backSpace()) {
					updateScreenLineAtTNode(mainFIndex);
					int h2=T.getLineHeightAtTNode(this,mainFIndex);
					if (h1>h2) {
						scrollFromTNode(mainFIndex,h1-h2);
					}
				}
			}
    	} else {
	    	TNode tn=T.TNode(T.prevChar(T.getCursorPos()));
			int h1=T.getLineHeightAtTNode(this,T.getCursorPos());
			int h0=(T.getCursorPos()>0)?T.getLineHeightAtTNode(this,T.getCursorPos()-1):0;
	    	if (T.backSpace()) {
		    	if ("\n".equals(tn.getChar())) {
					updateScreenLineAtTNode(T.getCursorPos());
			    	int h2=T.getLineHeightAtTNode(this,T.getCursorPos());
			    	scrollFromTNode(T.getCursorPos(),h0+h1-h2);
			    } else {
					updateScreenLineAtTNode(T.getCursorPos());
					int h2=T.getLineHeightAtTNode(this,T.getCursorPos());
					if (h1>h2) {
						scrollFromTNode(T.getCursorPos(),h1-h2);
					}
			    }
			}
		}
    }

	protected void removeMarkedSegment() {
		if (F!=null) {
			if (editing && F.segMarked()) {
				int h1=T.getLineHeightAtTNode(this,mainFIndex);
				F.removeMarkedSegment();
				updateScreenLineAtTNode(mainFIndex);
				int h2=T.getLineHeightAtTNode(this,mainFIndex);
				if (h1>h2) {
					scrollFromTNode(mainFIndex,h1-h2);
				}
			}
		} else {
			if (T.segMarked()) {
				int h0=T.getLineHeightAtTNode(this,T.segMin());
				int y0=T.getY(this,T.segMin())-T.getLineAscentAtTNode(this,T.segMin())+h0;
				int h1=T.getLineHeightAtTNode(this,T.segMax());
				int y1=T.getY(this,T.segMax())-T.getLineAscentAtTNode(this,T.segMax())+h1;
				T.removeMarkedSegment();
				updateScreenLineAtTNode(T.getCursorPos());
				int h2=T.getLineHeightAtTNode(this,T.getCursorPos());
				if (h0+y1-y0-h2>0) {
					scrollFromTNode(T.getCursorPos(),h0+y1-y0-h2);
				}
			}
		}
	}

	protected void scrollFromTNode(int ix,int dh) {
		int h2=T.getLineHeightAtTNode(this,ix);
		int y=T.getY(this,ix)-T.getLineAscentAtTNode(this,ix)+h2;
		bG.clipRect(0,y,getSize().width,getSize().height);
		bG.drawImage(buffer,0,-dh,this);
		bG.setClip(null);
//		bG.setColor(Color.white);
		bG.setColor(getBackground());
		bG.fillRect(0,getTextHeight()-T.getBottomMargin(),getSize().width,getSize().height);
    	T.updateComponentLocations(this,bG);
		Graphics g=getGraphics();
		g.clipRect(0,y,getSize().width,getSize().height);
		g.drawImage(buffer,0,0,this);
//		paint(g);
    }


    /* ------------------ Graphic Navigation ------------------ */

    protected int getCursorPos() {
    	if (F!=null) {
    		return F.getCursorPos();
    	} else {
    		return T.getCursorPos();
    	}
    }

    /* ---------------- update methods -------------- */

    protected void updateScreenSegment(int ini,int fin) {
   		if (bG!=null) {
    		T.drawSegment(bG,this,editing,ini,fin,showholes);
    	}
    }

    protected void updateScreenLineAtTNode(int n) {
    	updateScreenSegment(n,n);
    }

    protected void updateScreenFromTNode(int N) {
    	if (N<=T.end()) {
	    	updateScreenSegment(N,T.end());
	    }
	    if (bG!=null) {
 	 	   int ymax=T.getY(this,T.end())+T.getLineDescentAtTNode(this,T.end());
// 	    	bG.setColor(Color.white);
			bG.setColor(getBackground());
			bG.fillRect(0,ymax,getSize().width,getSize().height-ymax);
			Graphics g=getGraphics();
			if (g!=null) {
				g.clipRect(0,ymax,getSize().width,getSize().height-ymax);
				g.drawImage(buffer,0,0,this);
			}
	    }
    }
 
	private boolean showholes;
	
	public void updateScreen() {
		updateScreen(showholes);
	}
	
	private volatile boolean updatingScreen=false;
	private int n=1;

	public void updateScreen(boolean showholes) {
//System.out.println("TextCanvas.updateScreen IN");
//long T0=System.currentTimeMillis();
//new Exception().printStackTrace();
		if (!updatingScreen) {
			updatingScreen=true;
			this.showholes=showholes;
			if (T!=null) {
				T.calcValues(dec,fixed,aL,isEditing(),learnerMode);
//				updateGraphics();
				updateSize(false);
			    if (bG!=null) {
			    	bG=buffer.getGraphics();
					bG.setColor(Color.white);
					bG.setColor(getBackground());
					bG.fillRect(0,0,getSize().width,getSize().height);
					updateScreenFromTNode(T.begin());
				}
			}
			updatingScreen=false;
		}
//long dT=System.currentTimeMillis()-T0;
//System.out.println("TextCanvas.updateScreen OUT time = "+dT);
	}

    /* ---------------- public paint methods -------------- */

    public void update(Graphics g) {
    	paint(g);
    }
    
    private boolean painting=false;
    private boolean isPainting() {
    	return painting;
    }

    public void paint(Graphics g) {
    	if (isPainting()) return;
		painting=true;
    	super.paint(g);
    	if (buffer!=null) {
			if (getSize().height<getCanvasHeight()) {
//				System.out.println("Changing TextCanvas height in paint. "+getSize().height+" --> "+getCanvasHeight());
				setSize(getSize().width,buffer.getHeight()); // Necesario para que pinte completo en el navegador
			}
			g.setClip(0,-getLocation().y,getSize().width,2000);
		   	g.drawImage(buffer,0,0,this);
		}
	   	painting=false;
   }

	public void stop() {
		flush();
	}

	protected void flush() {
		if (buffer!=null) {
			buffer.flush();
			buffer=null;
			System.gc();
		}
	}

	public Dimension bufferSize() {
		if (buffer!=null) {
			return new Dimension(buffer.getWidth(this),buffer.getHeight(this));
		}
		return null;
	}

   	public void compactCanvas(boolean doGraphics) {
   		setSize(getSize().width,getCanvasHeight());
   		if (doGraphics) {
//   			updateGraphics();
			updateSize(false);
			updateScreen();
   		}
   	}
   	
	private boolean updateGraphics(int w,int h) {
		if (w>0 & h>0) {
			if ((buffer==null || (buffer.getWidth(this)!=w) || (buffer.getHeight(this)<h) ) ) {
				flush();
				buffer=new BufferedImage(w,h,BufferedImage.TYPE_4BYTE_ABGR);
				bG=buffer.getGraphics();
				return true;
			}
		}
		return false;
	}
	
	public boolean updateSize() {
		return updateSize(false);
   	}
   	
	public boolean updateSize(boolean force) {
//System.out.println("updateSize to "+getCanvasHeight());
		if ( force || (buffer==null)
			|| (getSize().height<getCanvasHeight()) 
			|| (buffer.getHeight(this)<getCanvasHeight()) 
			) {
			int CH=20;
			Container C=getParent();
			if (C!=null && C.getSize()!=null) {
				CH=Math.max(CH,C.getSize().height);
			}
			int newHeight=Math.max(CH,getCanvasHeight()+((editing)?200:0) );
    		setSize(getSize().width,newHeight);
			boolean gu=updateGraphics(getSize().width,newHeight);
			return gu;
		}
   		return false;
   	}
   	
	public int getCursorY() {
		return T.getY(this,getCursorPos());
	}

    /* ---------------------- Mouse Motion Listener -------------------- */

    public void mouseMoved(MouseEvent e) {
    	TNode tn=T.getTNodeAtPixel(this,e.getX(),e.getY());
    	if (tn!=null && tn.isHyperlink()) {
			if (A!=null) {
				A.getAppletContext().showStatus(tn.getUri());
			}
	    	if (getCursor()!=Cursor.getPredefinedCursor(Cursor.HAND_CURSOR)) {
	    		setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
	    	}
    	} else {
			if (A!=null) {
				A.getAppletContext().showStatus("");
			}
    		if (isEditing()) {
		    	if (getCursor()!=Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR)) {
		    		setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
	    		}
	    	} else {
	    		setCursor(null);
	    	}
	    }
    }
    
    private int mouse_x,mouse_y;
    
    public Scrollbar jsb;
    public int containerSize=480;
    private volatile boolean dragging=false;

    public void mouseDragged(MouseEvent e) {
    	if (!dragging) {
    		dragging=true;
	    	int dy=e.getY()-mouse_y;
	   		if (jsb!=null && dy!=0) {
		    	int minY=-Math.max(getCanvasHeight()-containerSize+000,0);
	    		int newY=Math.max(Math.min(getLocation().y+dy,0),minY);
	    		if (getLocation().y!=newY) {
	    			dy=newY-getLocation().y;
			    	mouse_y=e.getY()-dy;
	 	    		setLocation(0,newY);
	    			invalidate();
	    			repaint();
					jsb.setMaximum(-minY);
					jsb.setVisibleAmount(250);
					jsb.setValue(-newY);
	    		}
		    } 
	    	dragging=false;
    	}
    	mouse_x=e.getX();
    }

/* ----------------- Mouse Listener ---------------- */

    public void mousePressed(MouseEvent e) {
    	mouse_x=e.getX();
    	mouse_y=e.getY();
    	setCursor(new Cursor(Cursor.MOVE_CURSOR));
    }
    public void mouseReleased(MouseEvent e) {
    	setCursor(null);
    }
    public void mouseClicked(MouseEvent e) {
		if (A!=null) {
			TNode tn=T.getTNodeAtPixel(this,e.getX(),e.getY());
			if (tn!=null && tn.isHyperlink() && tn.getUri()!=null)  {
				try {
					if (tn.getUri().toLowerCase().startsWith("http://")) {
						A.getAppletContext().showDocument(new URL(tn.getUri()),"_blank");
					} else if (tn.getUri().toLowerCase().startsWith("/resources")) {
						A.getAppletContext().showDocument(A.getClass().getResource(tn.getUri()),"_blank");
					} else {
						A.getAppletContext().showDocument(new URL(A.getDocumentBase(),tn.getUri()),"_self");
					}
				} catch (Exception ex) {
					System.out.println(ex.getMessage());
					System.out.println("in EditorCanvas.mouseClicked");
				}
			}
		}
    }
    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

}