/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package com.mja.textedit;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

import com.mja.parser.*;
import com.mja.lang.*;
import com.mja.img.*;
import com.mja.gui.*;
import com.mja.util.TFont;
import com.mja.util.BasicStr;
import com.mja.file.mjaFile;
import com.mja.file.mjaFont;
import com.mja.text.BasicText;
import com.mja.text.Text;
import com.mja.text.TColors;
import com.mja.text.Formula;
import com.mja.text.FormulaItem;
import com.mja.text.TNode;
import com.mja.text.TextCanvas;
import com.mja.text.RTF;
import com.mja.descartes.Descartes;
import com.mja.descartes.controlConfig;
import com.mja.descartes.Space;
import com.mja.descartes.SpaceR2;
import com.mja.descartes.SpaceR3;
import com.mja.descartes.spaceConfig;
import com.mja.descgui.*;


public class EditorCanvas extends TextCanvas implements Runnable, KeyListener, FocusListener, ComponentListener {

    private int cursorX;
    private MouseCursorHandler mch;
    private Vector AL=new Vector();
    private String por_str=null;
    private Vector localBuffer;
	private Thread thr;
    private boolean pleaseStop;
    
	public EditorCanvas(Applet A,Frame parent,translator Tr,MouseCursorHandler mch,
		Parser p,TFont f,Color c,
		boolean moveimg,
		boolean variableSize,boolean onlyFormula) {
		this(A,parent,Tr,mch,p,f,c,moveimg,false,variableSize,onlyFormula);
	}

	public EditorCanvas(Applet A,Frame parent,translator Tr,MouseCursorHandler mch,
		Parser p,TFont f,Color c,
		boolean moveimg,boolean moveimg2,
		boolean variableSize,boolean onlyFormula) {
		super(A,parent,Tr,p,f,c,variableSize,onlyFormula);
		this.mch=mch;
		addFocusListener(this);
//		addComponentListener(this);
	}
	
	protected void updateSelectors() {
		callActionListeners(UpdateSelectors,ActionEvent.ACTION_PERFORMED);
	}

	public boolean start() {
		removeComponentListener(this);
		boolean started=super.start();
		if (started) {
			updateCursorThread();
		}
		addComponentListener(this);
		return started;
	}
	
	public void setForeground(Color c) {
		super.setForeground(c);
		setFormulaColors(T.getColors());
	}

	public void setBackground(Color c) {
		super.setBackground(c);
		setFormulaColors(T.getColors());
	}

	public void setText(Text T) {
		if (isEditing()) {
			removeListeners();
		}
		super.setText(T);
		if (isEditing()) {
			addListeners();
		}
	}
	
	public void setMultStr(String por_str) {
		this.por_str=por_str;
	}

	public void addActionListener(ActionListener al) {
		if (al!=null) {
			AL.addElement(al);
		}
	}

	protected void callActionListeners(String ac) {
		callActionListeners(ac,0);
	}
	void callActionListeners(String ac,int action) {
		for (int i=0;i<AL.size();i++) {
			ActionListener al=(ActionListener) AL.elementAt(i);
		  	al.actionPerformed(new ActionEvent(this,action,ac));
		}
	}

	public void setLearnerMode(boolean b) {
		learnerMode=b;
		updateSize();
		updateScreen();
		updateCursorThread();
	}

    public boolean segIsMarked() {
    	if (F!=null) {
    		return F.segMarked();
    	} else {
    		return T.segMarked();
    	}
    }

	public void unmarkSegment() {
		if (F!=null && F.segMarked()) {
			F.unmarkSegment();
			updateScreenSegment(mainFIndex,mainFIndex);
		}
		if (T.segMarked()) {
			int n1=T.segMin(), n2=T.segMax();
			T.unmarkSegment();
			updateScreenSegment(n1,n2);
		}
	}
	
	public void updateComponents() {
		if (A instanceof com.mja.descartes.Descartes) {
			com.mja.descartes.Descartes D=(com.mja.descartes.Descartes) A;
			com.mja.descgui.NumericControl[] nc=D.getNumCtrls();
			for (int i=0;i<nc.length;i++) {
				if (nc[i]!=null) {
					updateTextComponent(nc[i],nc[i].getControlID());
				}
			}
		}
	}

	public void setEditing(boolean e) {
		if (editing) {
			unmarkSegment();
			removeListeners();
		}
		editing=e;
		if (editing) {
			addListeners();
		}
		updateSize();
		updateCursorThread();
		if (!editing) {
			setCursor(null);
		}
	}
	
	public void setActiveFormulaAtNode(Node N) {
		mainF.unmarkSegment();
		F=mainF.subFormulaWithNode(N);
		F.markSegment(F.begin(),F.end());
		updateScreen();
	}

    public void insertHyperlink() {
    	String label="HV";
    	String url="http://recursostic.educacion.es/descartes/web/";
    	if (T.segMarked()) {
    		label="";
    		int n=T.segBegin();
    		while (n<T.segEnd()) {
    			label+=T.TNode(n++).getChar();
    		}
    		T.removeMarkedSegment();
    	}
    	int n=T.getCursorPos();
    	T.insertHyperlink(label,url);
    	TNode tn=T.TNode(n);
    	tn.editHyperlink();
    	updateScreen();
	    requestFocus();
    }

    static final int pulsador=0,texto_corto=1,texto_medio=2,texto_largo=3, menu=4, barra=5, boton=6;
    
    public void insertControl(int interfaz,boolean inFormula) {
    	Descartes D=(Descartes) A;
		TFont tf=T.getTFontAtTNode(T.getCursorPos());
		int h=tf.getAscent()+tf.getDescent()+6;
    	String nom=D.scene.getNewControlName();
    	String cID=Long.toString(System.currentTimeMillis(),16);
    	controlConfig cc;
    	int H=h;
    	if (inFormula) {
    		H=h-2;
    	}
    	switch (interfaz) {
    		case pulsador : cc=D.cfg.addNumCtrlCfg("id='"+nom+"' nombre='"+nom+"' tipo='numrico' regin='escenario' expresin='(0,0,"+120+","+H+")' valor='0' decimales='2' visible='si' dibujar-si='' cID='"+cID+"' ");
    						break;
    		case texto_corto : if (inFormula) {
    							cc=D.cfg.addNumCtrlCfg("id='"+nom+"' nombre='' tipo='numrico' interfaz='campo de texto' solo_texto='s' regin='escenario' expresin='(0,0,"+H+","+H+")' valor='||' decimales='0' fijo='no' visible='si' dibujar-si='' cID='"+cID+"' ");
    						} else {
					    		cc=D.cfg.addNumCtrlCfg("id='"+nom+"' nombre='' tipo='numrico' interfaz='campo de texto' solo_texto='s' regin='escenario' expresin='(0,0,"+40+","+H+")' valor='||' decimales='0' fijo='no' visible='s' dibujar-si='' cID='"+cID+"' ");
    						}
    						break;
    		case texto_medio : cc=D.cfg.addNumCtrlCfg("id='"+nom+"' nombre='' tipo='numrico' interfaz='campo de texto' solo_texto='s' regin='escenario' expresin='(0,0,"+100+","+H+")' valor='||' decimales='0' fijo='no' visible='si' dibujar-si='' cID='"+cID+"' ");
    						break;
    		case texto_largo : cc=D.cfg.addNumCtrlCfg("id='"+nom+"' nombre='' tipo='numrico' interfaz='campo de texto' solo_texto='s' regin='escenario' expresin='(0,0,"+240+","+H+")' valor='||' decimales='0' fijo='no' visible='si' dibujar-si='' cID='"+cID+"' ");
    						break;
    		case menu : cc=D.cfg.addNumCtrlCfg("id='"+nom+"' nombre='' tipo='numrico' interfaz='men'  regin='escenario' expresin='(0,0,"+120+","+H+")' valor='0' decimales='0' opciones=' ,opcin1,,opcin2' visible='no' dibujar-si='' cID='"+cID+"' ");
    						break;
    		case barra : cc=D.cfg.addNumCtrlCfg("id='"+nom+"' nombre='' tipo='numrico'  interfaz='barra' regin='escenario' expresin='(0,0,"+120+","+H+")' valor='0' decimales='2' fijo='no' min='0' max='100' incr='1' visible='no' dibujar-si='' cID='"+cID+"' ");
    						break;
    		case boton : cc=D.cfg.addNumCtrlCfg("id='"+nom+"' nombre='"+nom+"' tipo='numrico' interfaz='botn'  regin='escenario' expresin='(0,0,"+120+","+H+")' valor='0' decimales='0' visible='no' dibujar-si='' cID='"+cID+"' ");
    						break;
    		default : cc=D.cfg.addNumCtrlCfg("id='"+nom+"' nombre='"+nom+"' tipo='numrico' regin='escenario' expresin='(0,0,"+120+","+h+")' valor='0' decimales='2' visible='si' dibujar-si='' cID='"+cID+"' ");
    						
    	}
    	if (D.getEditor()!=null) {
    		D.getEditor().setInfo();
    	}
    	NumericControl nc=cc.makeNumericControl(D);
    	nc.setVisible(false);
    	int h0=T.getLineHeightAtTNode(this,getCursorPos());
    	if (F!=null) {
	    	F.insertComponent(nc,"NumCtrl",cID);
    	} else {
	    	T.insertComponent(nc,"NumCtrl",cID);
    	}
    	nc.addMouseListener(D);
    	updateScreenLineAtTNode(getCursorPos());
    	updateScreenLineAtTNode(getCursorPos()); // se repite para que se vea bien.
    	int h1=T.getLineHeightAtTNode(this,getCursorPos());
    	if (h1>h0) {
	    	scrollFromTNode(getCursorPos(),h0-h1);
	    	int nl=T.nextLine(getCursorPos());
	    	updateScreenLineAtTNode(nl);
    	}
	    if (updateSize(false)||F!=null) {
	    	updateScreen();
	    }
	    nc.paintAll(nc.getGraphics());
		paintCursor();
	    requestFocus();
	    
    }

	/*---------------- Formulas ----------------*/

    public void insertExpr(String expr) {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				removeMarkedSegment();
			}
	   		F.insertExpr(expr,Integer.toString(dec),fixed); // 04/12/08 se agregaron los dos ltimos parmetros
			mainF.calcValues(dec,fixed,aL,editing,learnerMode);
			updateScreen();
			paintCursor();
		    updateSize();
		}
    }

    public void insertFunction(String funcname) {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				removeMarkedSegment();
			}
			int n=F.getCursorPos();
	   		F.insertText(funcname);
	   		F.setCursorPos(n);
	   		smallMoveRight();
			updateScreenFromTNode(mainFIndex);
			paintCursor();
		    updateSize();
		}
    }

    public void insertRadical(String index,String radicand) {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				removeMarkedSegment();
			}
			int n=F.getCursorPos();
	   		F.insertRadical(index,radicand);
	   		F.setCursorPos(n);
	   		smallMoveRight();
			updateScreenFromTNode(mainFIndex);
			paintCursor();
		    updateSize();
		}
    }

    public void insertSum(String from,String to,String what) {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				removeMarkedSegment();
			}
			int n=F.getCursorPos();
	   		F.insertSum(from,to,what);
	   		F.setCursorPos(n);
	   		smallMoveRight();
			updateScreenFromTNode(mainFIndex);
			paintCursor();
		    updateSize();
		}
    }

    public void insertIntegral(String from,String to,String what) {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				removeMarkedSegment();
			}
			int n=F.getCursorPos();
	   		F.insertIntegral(from,to,what);
	   		F.setCursorPos(n);
	   		smallMoveRight();
			updateScreenFromTNode(mainFIndex);
			paintCursor();
		    updateSize();
		}
    }

    public void insertLimit(String from,String to,String what) {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				removeMarkedSegment();
			}
			int n=F.getCursorPos();
	   		F.insertLimit(from,to,what);
	   		F.setCursorPos(n);
	   		smallMoveRight();
			updateScreenFromTNode(mainFIndex);
			paintCursor();
		    updateSize();
		}
    }

    public void insertMatrix(String[][] sa) {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				removeMarkedSegment();
			}
			int n=F.getCursorPos();
	   		F.insertMatrix(sa);
	   		F.setCursorPos(n);
	   		smallMoveRight();
			updateScreenFromTNode(mainFIndex);
			paintCursor();
		    updateSize();
		}
    }

    public void insertDefParts(String[] sa) {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				removeMarkedSegment();
			}
			int n=F.getCursorPos();
	   		F.insertDefParts(sa);
	   		F.setCursorPos(n);
	   		smallMoveRight();
			updateScreenFromTNode(mainFIndex);
			paintCursor();
		    updateSize();
		}
    }

    public void insertFraction(String num,String den) {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				removeMarkedSegment();
			}
			int n=F.getCursorPos();
	   		F.insertFraction(num,den);
	   		if (num==null||num.length()==0) {
		   		F.setCursorPos(n);
				n=F.getCursorPos();
		   		smallMoveRight();
	   		} else if (den==null||den.length()==0) {
		   		smallMoveLeft();
	   		}
			updateScreenFromTNode(mainFIndex);
			paintCursor();
		    updateSize();
		}
    }
    
    public void insertNewFraction() {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				String nstr=F.toExpresion(F.segBegin(),F.segEnd(),true);
				try {
					Node N=parser.Analyse("EditorCanvas",nstr,true,false);
					Formula nf=N.toFormula();
					Formula df=new Formula(parser);
					F.removeMarkedSegment();
					try {
						if (localBuffer!=null) {
							Formula auxf=Formula.makeFormulaFromBuffer(F,localBuffer);
							localBuffer=null;
							String dstr=auxf.toExpresion(true);
							Node D=parser.Analyse("EditorCanvas",dstr,true,false);
							df=D.toFormula();
							F.insertFraction(nf,df);
						} else {
							F.insertFraction(nf,df);
							smallMoveLeft();
						}
					} catch (Exception ex2) {
						F.insertFraction(nf,df);
						smallMoveLeft();
					}
					updateScreenFromTNode(mainFIndex);
					paintCursor();
				    updateSize();
				} catch (Exception ex) {
					insertFraction(nstr,"");
				}
			} else {
				insertFraction("","");
			}
		}
    }

    public void insertSubIndex(String s) {
    	if (editing && F!=null) {
			if (F.segMarked()) {
				removeMarkedSegment();
			}
			int n=F.getCursorPos();
	   		F.insertSubIndex(s);
	   		F.setCursorPos(n);
	   		smallMoveRight();
			updateScreenFromTNode(mainFIndex);
			paintCursor();
		    updateSize();
		}
    }

    public void insertSuperIndex(String s) {
    	if (editing && F!=null) {
   			if (F.segMarked()) {
   				removeMarkedSegment();
   			}
    		int n=F.getCursorPos();
	   		F.insertSuperIndex(s);
	   		F.setCursorPos(n);
	   		smallMoveRight();
			updateScreenFromTNode(mainFIndex);
			paintCursor();
		    updateSize();
		}
    }
    
    /* ----------------------- formula buffer -------------------- */

	void copyFormula() {
		if (F!=null && F.segMarked()) {
			localBuffer=new Vector();
			int n2=F.segMax();
			for (int n=F.segMin();n<n2;n++) {
				TNode tn=F.TNode(n).cloneTNode(F);
				localBuffer.addElement(tn);
			}
		}
	}    
    

    /* -------------------- Cursor Movements ------------------ */

    protected void setCursorPos(int x,int y) {
    	int n=T.getIndexOfLineAtPixel(this,y);
    	int o=T.getOffsetOfTNodeAtPixel(this,n,x,false);
		int index=T.firstChar(n+o);
		if (!onlyFormula) {
    		T.setCursorPos(index);
    	}
    	inVerticalMotion=false;
    	boolean inFormula=false;
    	if (T.TNode(index).isFormula()) {
    		mainF=T.TNode(index).getFI().getFormula(0);
    		Formula f=mainF.getFormulaAt(x,y);
    		if (f!=null) {
				inFormula=true;
				setActiveFormulaAt(index,f,x,y);
			}
    	}
		if (!onlyFormula&&!inFormula) {
    		mainF=null;
    		setEditingFormula(null);
    		mainFIndex=-1;
    	}
    }

    private void setCursorPosInText(int x,int y) {
    	int n=T.getIndexOfLineAtPixel(this,y);
    	int o=T.getOffsetOfTNodeAtPixel(this,n,x,false);
    	T.setCursorPos(T.firstChar(n+o));
    	inVerticalMotion=false;
    }

    private void setCursorPosInFormula(int x,int y) {
    	int o=F.getOffsetOfTNodeAtPixel(this,0,x);
    	F.setCursorPos(F.firstChar(o));
    	inVerticalMotion=false;
    }

	void smallMoveRight() {
    	int N=F.firstChar(F.getCursorPos());
    	TNode tn=F.TNode(N);
		if (tn.isSubIndex()||tn.isSuperIndex()||tn.isFraction()||tn.isProduct()||tn.isRadical() || tn.isSum() || tn.isIntegral() || tn.isLimit() || tn.isMatrix() || tn.isDefParts() ) {
	    	setEditingFormula(tn.getFI().getFormula(0));
		} else {
			BasicText p=tn.getParent();
			boolean done=false;
			if (F.afterLastChar(N) && p instanceof Formula) {
				TNode tnp=((Formula)p).getParent();
				if (tnp!=null) {
					if (tnp.isFraction()||tn.isProduct()||tnp.isRadical() || tnp.isSum()||tnp.isIntegral()||tnp.isLimit() || tnp.isMatrix() || tnp.isDefParts() ) {
						for (int j=0;j<tnp.getFI().size()-1;j++) {
							if (tnp.getFI().getFormula(j)==F) {
						    	setEditingFormula(tnp.getFI().getFormula(j+1));
						    	done=true;
						    	break;
						    }
						}
					}
				}
			}
			if (!done) {
				if (F.afterLastChar(N)) {
					doEscape();
				}
				if (F!=null) {
					moveRight();
				} else {
					moveCursor(false,true,right);
				}
			}
		}
	}

	void smallMoveLeft() {
    	int N=F.getCursorPos();
    	TNode tn=F.TNode(N);
	    if (N>F.firstChar(F.begin())) {
			moveLeft();
	    	N=F.getCursorPos();
			tn=F.TNode(N);
			if (tn.isSubIndex()||tn.isSuperIndex()||tn.isFraction()||tn.isProduct()||tn.isRadical() || tn.isSum() || tn.isIntegral() || tn.isLimit() || tn.isMatrix() || tn.isDefParts() ) {
		    	setEditingFormula(tn.getFI().getFormula(tn.getFI().size()-1));
		    	F.setCursorPos(F.end());
			}
		} else {
			BasicText p=tn.getParent();
			boolean done=false;
			if (p instanceof Formula) {
				TNode tnp=((Formula)p).getParent();
				if (tnp!=null) {
					if (tnp.isFraction()||tn.isProduct()||tnp.isRadical()||tnp.isSum()||tnp.isIntegral()||tnp.isLimit() || tnp.isMatrix() || tnp.isDefParts() ) {
						for (int j=tnp.getFI().size()-1;j>0;j--) {
							if (tnp.getFI().getFormula(j)==F) {
								setEditingFormula(tnp.getFI().getFormula(j-1));
						    	F.setCursorPos(F.end());
								done=true;
								break;
							}
						}
					}
				}
			}
			if (!done) {
				doEscape();
			}
		}
	}

    boolean doEscape() {
		TNode tn=F.getParent();
		if (tn!=null) {
			BasicText p=tn.getParent();
			if (p instanceof Formula) {
		    	setEditingFormula((Formula)p);
		    	F.setCursorPos(tn);
		    	return true;
		    }
		}
		if (!onlyFormula) {
			mainF=null;
			setEditingFormula(null);
			mainFIndex=-1;
		}
		return false;
	}

	boolean moveRight() {
		int N=F.getCursorPos();
		F.setCursorPos(F.firstChar(N));
		if (F.getCursorPos()<F.end()) {
			F.setCursorPos(F.getCursorPos()+1);
		}
		if (F.getCursorPos()!=N) {
			return true;
		}
		return false;
	}

	boolean moveLeft() {
		if (F!=null) {
			int N=F.getCursorPos();
			F.setCursorPos(F.prevChar(N));
			if (F.getCursorPos()!=N) {
				return true;
			}
		}
		return false;
	}

	boolean goToUpperFormula() {
    	int N=F.getCursorPos();
    	TNode tn=F.TNode(N);
		BasicText pu=tn.getParent();
		if (pu instanceof Formula) {
			TNode tnp=((Formula)pu).getParent();
			if (tnp!=null) {
				if (tnp.isFraction()||tnp.isProduct()) {
					if (tnp.getFI().getFormula(FormulaItem.DEN)==F) {
				    	setEditingFormula(tnp.getFI().getFormula(FormulaItem.NUM));
				    	return true;
					}
				} else if (tnp.isRadical()) {
					if (tnp.getFI().getFormula(FormulaItem.RADICAND)==F) {
				    	setEditingFormula(tnp.getFI().getFormula(FormulaItem.INDEX));
				    	return true;
					}
				} else if (tnp.isSum()||tnp.isIntegral()||tnp.isLimit()) {
					if (tnp.getFI().getFormula(FormulaItem.WHAT)==F||tnp.getFI().getFormula(FormulaItem.FROM)==F) {
				    	setEditingFormula(tnp.getFI().getFormula(FormulaItem.TO));
				    	return true;
					}
				}
			}
		}
		return false;
	}

	boolean goToLowerFormula() {
    	int N=F.getCursorPos();
    	TNode tn=F.TNode(N);
		BasicText pd=tn.getParent();
		if (pd instanceof Formula) {
			TNode tnp=((Formula)pd).getParent();
			if (tnp!=null) {
				if (tnp.isFraction()||tnp.isProduct()) {
					if (tnp.getFI().getFormula(FormulaItem.NUM)==F) {
				    	setEditingFormula(tnp.getFI().getFormula(FormulaItem.DEN));
				    	return true;
					}
				} else if (tnp.isRadical()) {
					if (tnp.getFI().getFormula(FormulaItem.INDEX)==F) {
				    	setEditingFormula(tnp.getFI().getFormula(FormulaItem.RADICAND));
				    	return true;
					}
				} else if (tnp.isSum()||tnp.isIntegral() ||tnp.isLimit() ) {
					if (tnp.getFI().getFormula(FormulaItem.WHAT)==F||tnp.getFI().getFormula(FormulaItem.TO)==F) {
				    	setEditingFormula(tnp.getFI().getFormula(FormulaItem.FROM));
				    	return true;
					}
				}
			}
		}
		return false;
	}

	public void setFormulaColors(TColors tc) {
		if (mainF!=null) {
			mainF.clearColors();
			mainF.setColors(tc);
		}
	}

    private static final int left=1,right=2,up=3,down=4,home=5,end=6;

	private void moveCursor(boolean ShiftOn,boolean CtrlOn,int m) {
   		eraseCursor();
   		if (F!=null) {
   			if ((m==down||m==up)&&CtrlOn) {
		    	switch (m) {
					case down : insertSubIndex("");
								break;
					case up   : insertSuperIndex("");
								break;
		    	}
		    	if (F.segMarked()) {
		    		F.unmarkSegment();
					updateScreenFromTNode(mainFIndex);
		    	}
   			} else if (ShiftOn) {
		    	int N=F.getCursorPos();
		    	switch (m) {
		    		case left : moveLeft();
		    					break;
		    		case right: moveRight();
		    					break;
		    	}
				if (N!=F.getCursorPos()) {
					if (!F.segMarked()) {
						F.markSegment(N,F.getCursorPos());
					} else {
						F.updateMarkedSegment();
					}
					updateScreenFromTNode(mainFIndex);
			   	}
	    	} else if (CtrlOn) {
		    	if (F.segMarked()) {
		    		F.unmarkSegment();
					updateScreenFromTNode(mainFIndex);
		    	}
		    	switch (m) {
		    		case left : if (!doEscape()) {
		    						moveLeft();
		    					}
		    					break;
		    		case right: doEscape();
								if (F!=null) {
									moveRight();
								} else {
									moveCursor(false,false,right);
								}
		    					break;
		    		case up   : goToUpperFormula();
		    					break;
		    		case down : goToLowerFormula();
		    					break;
		    	}
	    	} else {
		    	if (F.segMarked()) {
		    		F.unmarkSegment();
					updateScreenFromTNode(mainFIndex);
		    	}
		    	switch (m) {
		    		case left : smallMoveLeft();
		    					break;
		    		case right: smallMoveRight();
		    					break;
		    		case up   : goToUpperFormula();
		    					break;
		    		case down : goToLowerFormula();
		    					break;
		    		case home : F.setCursorPos(F.startOfLine(F.getCursorPos()));
		    					break;
		    		case end  : F.setCursorPos(F.endOfLine(F.getCursorPos()));
		    					break;
		    	}
		    }
   		} else {
	    	int N=T.getCursorPos();
	    	int ln;
	    	int CY=getCursorY();
	    	switch (m) {
	    		case left : 
							if (N>T.firstChar(T.begin())) {
	    						T.setCursorPos(T.prevChar(T.getCursorPos()));
						    	if (!CtrlOn&&T.TNode(T.getCursorPos()).isFormula()) {
						    		int index=T.getCursorPos();
						    		mainF=T.TNode(index).getFI().getFormula(0);
						    		setActiveFormulaAt(index,mainF,0,0);
									moveCursor(false,false,end);
						    	} else {
	    						}
	    					}
	    					break;
	    		case right: T.setCursorPos(T.firstChar(T.getCursorPos()));
	    					if (T.getCursorPos()<T.end()) {
						    	if (!CtrlOn&&T.TNode(T.getCursorPos()).isFormula()) {
						    		int index=T.getCursorPos();
						    		mainF=T.TNode(index).getFI().getFormula(0);
						    		setActiveFormulaAt(index,mainF,0,0);
						    	} else {
		    						T.setCursorPos(T.getCursorPos()+1);
		    					}
		    				}
	    					break;
	    		case up   : if (!inVerticalMotion) {
	    						cursorX=T.getX(this,T.getCursorPos());
	    						inVerticalMotion=true;
	    					}
							ln=T.previousLine(T.getCursorPos());
							T.setCursorPos(ln+T.getOffsetOfTNodeAtPixel(this,ln,cursorX,false));
	    					break;
	    		case down : if (!inVerticalMotion) {
					    		cursorX=T.getX(this,T.getCursorPos());
					    		inVerticalMotion=true;
					    	}
							ln=T.nextLine(T.getCursorPos());
							T.setCursorPos(ln+T.getOffsetOfTNodeAtPixel(this,ln,cursorX,false));
	    					break;
	    		case home : T.setCursorPos(T.startOfLine(T.getCursorPos()));
	    					break;
	    		case end  : T.setCursorPos(T.endOfLine(T.getCursorPos()));
	    					break;
	    	}
			if (N!=T.getCursorPos()) {
			   	if (ShiftOn) {
					if (!T.segMarked()) {
						T.markSegment(N,T.getCursorPos());
					} else {
						T.updateMarkedSegment();
					}
					updateScreenSegment(Math.min(N,T.segMin()),Math.max(N,T.segMax()));
			    } else {
			    	if (T.segMarked()) {
						int n1=T.segMin(), n2=T.segMax();
			    		T.unmarkSegment();
				   		updateScreenSegment(n1,n2);
			    	}
			    }
	    	}
		    if (mainF==null) {
		    	if (CY<getCursorY()) {
		    		callActionListeners("moved_down");
		    	} else if (CY>getCursorY()) {
		    		callActionListeners("moved_up");
		    	}
		    }
	    }
		paintCursor();
	}

	private boolean editNode(int x,int y) {
		if (F!=null) {
			int N=F.getCursorPos();
			int k=N;
			if (!F.TNode(k).isExpr() && k>F.begin()) {
				k--;
			}
			TNode tn=F.TNode(k);
			if (tn.isExpr()) {
				MathEditor me=new MathEditor(parent,Tr,tn.getExpr(),tn.getDecExpr(),tn.getFixed());
				me.display(true,true);
				if (me.ok) {
					tn.setExpr(me.getExpr());
					tn.setDecExpr(me.getDecExpr());
					tn.setFixed(me.getFixed());
					mainF.calcValues(dec,fixed,aL,editing,learnerMode);
					updateScreen();
				}
				me.dispose();
				return true;
			}
		} else {
			TNode tn=T.getTNodeAtPixel(this,x,y);
			if (tn!=null && tn.isHyperlink()) {
				tn.editHyperlink();
				updateScreen();
				return true;
			}
		}
		return false;
    }

	/* ---------------- add/remove listeners ---------------- */

	private void removeListeners() {
		removeKeyListener(this);
	}

	private void addListeners() {
		addKeyListener(this);
	}

    /* ----------------------- Key Listener ----------------------- */

    private char UC_base=0x0000;

    public void setUnicodeBase(int base) {
    	UC_base=(char)(base-0x0061);;
    }

    public void keyTyped(KeyEvent e) {
    	if (!isEnabled()) return;
    	if (mch!=null && mch.isHand()) {
    		return;
    	}
    	updating=true;
		if ((e.getKeyChar()>=(char) 0x20) &&
			!((e.getKeyChar()=='/') && (F!=null)) &&
			!((e.getKeyChar()=='_') && (F!=null)) &&
			!((e.getKeyChar()=='^') && (F!=null))) {
			if ( (e.getKeyChar() >= (char) 0x41 && e.getKeyChar()<(char)(0x41+26)) ||
				(e.getKeyChar() >= (char) 0x61 && e.getKeyChar()<(char)(0x61+26)) ) {
		    	saveText();
				insertChar((char)(e.getKeyChar()+UC_base));
			} else {
				if (e.getKeyChar()!=KeyEvent.VK_DELETE) { // para que no dibuje un cuadro al pulsar "del".
					char ch=(char)(e.getKeyChar());
					if (ch=='*' && por_str!=null) {
						ch=por_str.charAt(0);
					}
			    	saveText();
					insertChar(ch);
				}
			}
	    	if (F!=null) {
	   			if (F.getCursorPos()==F.end() && e.getKeyChar()=='(') {
			    	saveText();
	   				insertChar(')');
	   				moveLeft();
	   			}
	   		}
	   	}
    	updating=false;
    }

    /* ------------------- Low Level Keyboard Events -------------- */
    int TBSZ=25;
    String [] Sa=new String[TBSZ+1];
    int[] cposa=new int[TBSZ+1];
    
    public void saveText() {
    	for (int n=TBSZ-1;n>=0;n--) {
    		Sa[n+1]=Sa[n];
    		cposa[n+1]=cposa[n];
    	}
    	cposa[0]=T.getCursorPos();
    	RTF rtf=new RTF(T);
    	Sa[0]=rtf.toRTF(false);
    }
    
    public void popText() {
		if (Sa[0]!=null) {
			T=new Text(parser);
			RTF.parseRTF(Sa[0],T);
			T.setCursorPos(cposa[0]);
			for (int n=0;n<TBSZ-1;n++) {
				Sa[n]=Sa[n+1];
				cposa[n]=cposa[n+1];
			}
			Sa[TBSZ]=null;
		}
    }
    
    public void keyPressed(KeyEvent e) {
    	if (!isEnabled()) return;
    	if (mch!=null && mch.isHand()) {
    		return;
    	}
    	updating=true;
		if (!(e.getKeyCode()==KeyEvent.VK_UP||e.getKeyCode()==KeyEvent.VK_DOWN)) {
			inVerticalMotion=false;
		}
		BasicStr.ShiftOn=e.isShiftDown();
		BasicStr.CtrlOn =e.isControlDown();
		boolean AltOn  =e.isAltDown();
//		boolean AltGrOn =...
		boolean inserted=false;
		switch (e.getKeyChar()) {
			case '/'					:
			case '_'					: 
//											if (F==null) {	
//												insertFormula("");
//											}
											if (F!=null) {
										    	saveText();
												inserted=true;
												insertNewFraction();
											}
											break;
			case '^'					: 	
//											if (F==null) {	
//												insertFormula("");
//											}
											if (F!=null) {
												inserted=true;
										    	saveText();
												insertSuperIndex("");
											}
											break;
			default                         :
				switch (e.getKeyCode()) {
				case KeyEvent.VK_W				: if (BasicStr.CtrlOn&&F!=null) {System.out.println(F.toString());}; break;
				case KeyEvent.VK_Q				: if (BasicStr.CtrlOn) {
											    	saveText();
													if (F==null) {	
														insertFormula("");
													}
													inserted=true;
													insertNewFraction();
												  }
												  break;

				case KeyEvent.VK_F				: if (BasicStr.CtrlOn) {
											    	saveText();
													if (F==null) {	
														insertFormula("");
													} else {
														inserted=true;
														insertNewFraction();
													}
												  }
													break;
				case KeyEvent.VK_E				: if (BasicStr.CtrlOn) { // expression
											    	saveText();
													if (F==null) {	
														insertFormula("");
													}
													inserted=true;
													insertExpr("1");
												  }
												  break;
				case KeyEvent.VK_S				: if (BasicStr.CtrlOn) { // Sumatoria
											    	saveText();
													if (F==null) {	
														insertFormula("");
													}
													inserted=true;
													insertSum("","","");
												  };
												  break;
				case KeyEvent.VK_R				: if (BasicStr.CtrlOn) { // radical
											    	saveText();
													if (F==null) {	
														insertFormula("");
													}
													inserted=true;
													insertRadical("","");
												  }
												  break;
				case KeyEvent.VK_I				: if (BasicStr.CtrlOn) { // integral
											    	saveText();
													if (F==null) {	
														insertFormula("");
													}
													inserted=true;
									 			  	insertIntegral("","","");
												  }
												  break;
				case KeyEvent.VK_L				: if (BasicStr.CtrlOn) { // limit
											    	saveText();
													if (F==null) {	
														insertFormula("");
													}
									 			  	inserted=true;
									 			  	insertLimit("","","");
									 			  }
									 			  break;
//				case KeyEvent.VK_M				: if (BasicStr.CtrlOn) { // matrix
//													if (F==null) {	
//														insertFormula("");
//													}
//													insertMatrix();
//									 			  }
//									 			  break;
				case KeyEvent.VK_P				: if (BasicStr.CtrlOn&&F!=null) { }
									 			  break;
				case KeyEvent.VK_ESCAPE			: if (F!=null) {
													eraseCursor();
												  	doEscape();
												  	paintCursor();
												  }
												  break;
				case KeyEvent.VK_ENTER			: if (F==null) {
													insertText("\n");
											    	saveText();
												  	callActionListeners("moved_down");
												  } else {
												  	callActionListeners("Verify");
												  }
												  break;
				case KeyEvent.VK_LEFT 			: moveCursor(BasicStr.ShiftOn,BasicStr.CtrlOn,left); break;
				case KeyEvent.VK_RIGHT 			: moveCursor(BasicStr.ShiftOn,BasicStr.CtrlOn,right); break;
				case KeyEvent.VK_DOWN 			: if (BasicStr.CtrlOn) {
											    	saveText();
													if (F==null) {
														insertFormula("");
													}
													inserted=true;
													insertSubIndex("");
												  } else {
													moveCursor(BasicStr.ShiftOn,BasicStr.CtrlOn,down); 
												  }
												  break;
				case KeyEvent.VK_UP 			: if (BasicStr.CtrlOn) {
											    	saveText();
													if (F==null) {
														insertFormula("");
													}
													inserted=true;
													insertSuperIndex("");
												  } else {
													moveCursor(BasicStr.ShiftOn,BasicStr.CtrlOn,up); 
												  }
												  break;
				case KeyEvent.VK_END 			: moveCursor(BasicStr.ShiftOn,BasicStr.CtrlOn,end); break;
				case KeyEvent.VK_HOME 			: moveCursor(BasicStr.ShiftOn,BasicStr.CtrlOn,home); break;
				case KeyEvent.VK_DELETE			: 
											    	saveText();
												  if ((F!=null&&F.segMarked())||(F==null&&T.segMarked())) {
													removeMarkedSegment();
												  } else {
													delete();
												  }
												  	callActionListeners("moved");
													break;
				case KeyEvent.VK_BACK_SPACE		: 
											    	saveText();
													if ((F!=null&&F.segMarked())||(F==null&&T.segMarked())) {
													removeMarkedSegment();
												  } else {
													backSpace();
												  }
												  	callActionListeners("moved_up");
													break;
				case KeyEvent.VK_C				: if (F!=null) {
											    	saveText();
													if (BasicStr.CtrlOn) {
														if (BasicStr.ShiftOn) {
															insertFraction("","");
													    	if (F.segMarked()) {
													    		inserted=true;
													    	}
														} else { //if (F.segMarked() ) {
															F.copy(this);
														}
													}
												  } else {
													if (BasicStr.CtrlOn /*&& T.segMarked()*/) {
														T.copy(this);
													}
												  }
												  	callActionListeners("moved");
													break;
				case KeyEvent.VK_X				: if (F!=null) {
											    	saveText();
													if (BasicStr.CtrlOn) {
														if (F.segMarked()) {
															copyFormula();
															F.copy(this);
															removeMarkedSegment();
															updateSize();
														}
													}
												} else {
													if (BasicStr.CtrlOn && T.segMarked()) {
												    	saveText();
														T.copy(this);
														removeMarkedSegment();
														updateSize();
													}
												}
											  	callActionListeners("moved");
												break;
				case KeyEvent.VK_Z				: if (BasicStr.CtrlOn) {
													popText();
													updateScreen();
												}
												break;
				case KeyEvent.VK_V				: if (BasicStr.CtrlOn) {
											    	saveText();
													if (F!=null) {
													    if (F.segMarked()) {
															removeMarkedSegment();
														}
														int N=F.getCursorPos();
														if (editing) {
															int h0=T.getLineHeightAtTNode(this,mainFIndex);
															F.paste(this);
															updateComponents();
															updateScreenLineAtTNode(mainFIndex);
															if (updateSize(false)) {
																updateScreen();
//																updateScreen();
															} else {
																updateScreenLineAtTNode(mainFIndex);
																int h1=T.getLineHeightAtTNode(this,mainFIndex);
																if (h1>h0) {
																	scrollFromTNode(mainFIndex,h0-h1);
																}
																updateScreenLineAtTNode(T.nextLine(mainFIndex));
															}
														}
													} else {
														if (T.segMarked()) {
															removeMarkedSegment();
														}
														int N=T.getCursorPos();
														T.paste(this);
														updateComponents();
														updateScreenFromTNode(N);
														if (updateSize()) {
															updateScreen();
														}
													}
												  	callActionListeners("moved");
												}
												break;
			}
		}
		if (inserted&&F.segMarked()) {
			F.unmarkSegment();
			updateScreenFromTNode(mainFIndex);
		}
    	updating=false;
    	updateSelectors();
    }

    public void keyReleased(KeyEvent e) {
    	BasicStr.CtrlOn=false;
    }

    /* ---------------------- Mouse Listener -------------------- */

	private int mouseX,mouseY,activeImage=-1,activeImage2=-1;
	private boolean mousePressedWhileEditing=false, dragged=false;
	private Point point;
    private boolean falseDrag;

    public void mousePressed(MouseEvent e) {
		requestFocus();
    	falseDrag=false;
		dragged=false;
    	if (!isEnabled()) return;
    	if (mch!=null && mch.isHand()) {
    		return;
    	}
    	if (F!=null) {
    		int N=F.getIndexOfTNodeAtPixel(this,e.getX(),e.getY());
	    	if (isEditing()&&F.segMarked()&&F.segMin()<=N && N<=F.segMax()) {
				mousePressedWhileEditing=true;
				point=e.getComponent().getLocationOnScreen();
				F.copy(this);
			} else {
	    		doMousePressed(e.getX(),e.getY());
	    	}
    	} else {
    		int N=T.getIndexOfTNodeAtPixel(this,e.getX(),e.getY());
	    	if (isEditing()&&T.segMarked()&&T.segMin()<=N && N<=T.segMax()) {
				mousePressedWhileEditing=true;
				point=e.getComponent().getLocationOnScreen();
			} else {
	    		doMousePressed(e.getX(),e.getY());
	    	}
	    }
    }

    public void doMousePressed(int ex,int ey) {
    	if (mch!=null && mch.isHand()) {
    		return;
    	}
    	if (!hasFocus) {
    		requestFocus();
    	}
    	if (isEditing()) {
			eraseCursor();
			unmarkSegment();
			setCursorPos(ex,ey);
			T.markSegment(T.getCursorPos(),T.getCursorPos());
	    	if (F!=null) {
				Formula tempF=F.getFormulaAt(ex,ey);
				if (tempF!=null) {
					setActiveFormulaAt(mainFIndex,tempF,ex,ey);
					tempF.markSegment(tempF.getCursorPos(),tempF.getCursorPos());
				}
			}
			paintCursor();
			updateSelectors();
			requestFocus();
		}
    }

    public void mouseReleased(MouseEvent e) {
    	falseDrag=true;
    	if (!isEnabled()) return;
    	if (mch!=null && mch.isHand()) {
    		return;
    	}
    	if (F!=null) {
			if (mousePressedWhileEditing) {
				mousePressedWhileEditing=false;
				if (dragged) {
					dragged=false;
					Point l=getLocationOnScreen();
					int ex=(point.x+e.getX())-l.x;
					int ey=(point.y+e.getY())-l.y;
					boolean in=(0<=ex&&ex<getSize().width&&0<=ey&&ey<getSize().height);
					if (in && F.segMarked()) {
						Formula F0=F;
						int N=F.segMin(),M=F.segMax();
						doMousePressed(ex,ey);
						if (F!=null) {
							Formula tempF=F.getFormulaAt(ex,ey);
							if (tempF!=null) {
								if (!e.isControlDown()) {
									F0.remove(N,M);
								}
								tempF.paste(this);
								updateComponents();
								updateSize();
								updateScreen();
	//							updateScreenLineAtTNode(getCursorPos());
								updateScreen();
							}
						}
					}
					unmarkSegment();
					point=null;
					return;
				} else {
					point=null;
					eraseCursor();
					unmarkSegment();
					setCursorPos(e.getX(),e.getY());
					if (F!=null) {
						F.markSegment(F.getCursorPos(),F.getCursorPos());
					}
					paintCursor();
					updateSelectors();
					requestFocus();
					return;
				}
			}
    	} else {
			if (mousePressedWhileEditing) {
				mousePressedWhileEditing=false;
				if (dragged) {
					dragged=false;
					Point l=getLocationOnScreen();
					int ex=(point.x+e.getX())-l.x;
					int ey=(point.y+e.getY())-l.y;
					boolean in=(0<=ex&&ex<getSize().width&&0<=ey&&ey<getSize().height);
					if (in) {
						T.copy(this);
						int N=T.segMin(),M=T.segMax();
						doMousePressed(ex,ey);
						if (!e.isControlDown()) {
							T.remove(N,M);
						}
						N=Math.min(N,T.getCursorPos());
						T.paste(this);
						updateScreenFromTNode(N);
						if (updateSize(false)) {
							updateScreen();
						}
					} else {
						unmarkSegment();
					}
					point=null;
					return;
				} else {
					point=null;
					eraseCursor();
					unmarkSegment();
					setCursorPos(e.getX(),e.getY());
					T.markSegment(T.getCursorPos(),T.getCursorPos());
					paintCursor();
					updateSelectors();
					requestFocus();
					return;
				}
			}
	    }
    }
    
	protected void activateButtons() {
		String command="";
		if (F!=null) {
			command=ActivateFormulaButtons;
		} else {
			command=ActivateTextButtons;
		}
		callActionListeners(command,ActionEvent.ACTION_PERFORMED);
    }

    protected void setActiveFormulaAt(int ix,Formula f,int x,int y) {
    	if (F!=null) {
	    	F.setActive(false);
	    }
    	F=f;
    	mainFIndex=ix;
    	if (F!=null) {
	    	F.setActive(true);
			setCursorPosInFormula(x,y);
	    }
	    activateButtons();
	 }

    protected void setEditingFormula(Formula f) {
    	setActiveFormulaAt(mainFIndex,f,0,0);
    }

	protected void selectWordAt(int n) {
		if (F!=null) {
			F.markSegment(F.begin(),F.end());
			updateScreenFromTNode(mainFIndex);
		} else {
			int N1=n,N2=n;
			if (!T.TNode(n).isWordSeparator()) {
				while (N1>T.begin()&&!T.TNode(N1-1).isWordSeparator()) {
					N1--;
				}
				while (N1<n && !T.TNode(N1).isChar()) {
					N1++;
				}
				while (N2<T.end()&&!T.TNode(N2+1).isWordSeparator()) {
					N2++;
				}
				while (N2>n && !T.TNode(N2).isChar()) {
					N2--;
				}
				T.markSegment(N1,N2+1);
				updateScreen();
				for (int i=N1;i<=N2;i++) {
					if (T.TNode(i).isFormula()) {
						Parser.analyseEquation(new Parser(1),T.TNode(i).getFI().getFormula(0).toExpresion(true));
					}
				}
			}
		}
	}

	protected void selectLineAt(int n) {
		if (F!=null) {
			mainF.markSegment(mainF.begin(),mainF.end());
			updateScreenFromTNode(mainFIndex);
		} else {
			int N1=n,N2=n;
			if (!T.TNode(n).isLineSeparator()) {
				while (N1>T.begin()&&!T.TNode(N1-1).isLineSeparator()) {
					N1--;
				}
				while (N2<T.end()&&!T.TNode(N2+1).isLineSeparator()) {
					N2++;
				}
				T.markSegment(N1,N2+1);
				updateScreen();
			}
		}
	}


    /* ---------------------- paint/erase Cursor ------------------- */
    
    private int cX, cY, cW, cH;

    protected void paintCursor() {
		try {
	    	if ((isEditing() && hasTheFocus() && !T.segMarked()) ) {
				Graphics2D G=(Graphics2D) getGraphics();
				TFont.setAntialias(G,true);
				if (G!=null) {
					int X,Y,y,aC,dC,hC,ymin,ymax;
	    			int wC=2;
					TFont f;
					if (F!=null) {
			    		f=F.getTFont();
//						G.setColor(F.getColorsAtTNode(F.getCursorPos()).getTextColor());
						G.setColor(Formula.formulaCursorColor);
			    		X=F.getX(this,F.getCursorPos());
			    		Y=F.getY(this,F.getCursorPos());
			    		Y=F.getLocation().y;
						aC=Math.min(f.getAscent(),F.getLineAscentAtTNode(this,F.getCursorPos()));
						dC=Math.min(f.getDescent(),F.getLineDescentAtTNode(this,F.getCursorPos()));
						y=Y;
					} else {
			    		f=T.getTFontAtTNode(T.getCursorPos());
						G.setColor(T.getColorsAtTNode(T.getCursorPos()).getTextColor());
			    		X=T.getX(this,T.getCursorPos());
			    		Y=T.getY(this,T.getCursorPos());
						y=T.getY(this,T.getCursorPos());
						aC=f.getAscent();
						dC=f.getDescent();
				    }
	    			if (f.isBold()) {
	    				wC=3;
    				}
				    X=X-1;
					hC=aC+dC;
					ymin=y-aC;
					ymax=y+dC;
		    		if (f.isItalic()) {
		    			int X1=(int) Math.round(X+((double)aC)/4.5);
		    			int X2=(int) Math.round(X-((double)dC)/4.5);
		    			int xmin=X2;
		    			int xmax=X1+wC;
//		    			if (f.isBold()) {
			    			Polygon P=new Polygon();
			    			P.addPoint(X1   , ymin);
			    			P.addPoint(X1+wC, ymin);
			    			P.addPoint(X2+wC, ymax);
			    			P.addPoint(X2   , ymax);
			    			G.fillPolygon(P);
//			    		} else {
//			    			G.drawLine(X1,ymin,X2,ymax);
//			    		}
			    		cX=xmin;
			    		cY=ymin;
			    		cW=1+xmax-xmin;
			    		cH=1+ymax-ymin;
		    		} else {
			    		cX=X;
			    		cY=Y-aC;
			    		cW=1+wC;
			    		cH=1+hC;
						G.fillRect(X,Y-aC,wC,hC);
					}
			    }
		    }
		} catch (Exception e) {
		}
	}
	
    protected void eraseCursor() {
		Graphics G=getGraphics();
    	if (G!=null && isEditing() /* && hasTheFocus() && !T.segMarked()*/ ) {
    		G.setClip(cX,cY,cW,cH);
			G.drawImage(buffer,0,0,this);
    		G.setClip(0,0,getSize().width,getSize().height);
		}
    }


    /* ----------------------- Runnable ----------------------- */

    public void run() {
    	pleaseStop=false;
    	try {
	    	while (!pleaseStop()) {
	    		thr.sleep(200);
	    		if (pleaseStop()) {
	    			break;
	    		}
	    		boolean isupdtng=isUpdating();
	   			if (!isupdtng) {
		    		paintCursor();
	    		}
	    		for (int i=0;i<10;i++) {
	    			thr.sleep(50);
		    		if (pleaseStop()) {
		    			break;
		    		}
		    	}
	    		if (pleaseStop()) {
	    			break;
	    		}
	    		if (!isupdtng) {
		    		eraseCursor();
		    	}
	    	}
	    } catch (InterruptedException e) {
	    	e.printStackTrace();
	    }
	    thr=null;
    }
    
	protected void updateCursorThread() {
		if (isEditing()) {
			if (thr==null) {
				thr=new Thread(this);
				thr.start();
			} else {
			}
		} else {
			stop();
		}
	}

    private boolean isUpdating() {
    	return updating;
    }

    private boolean pleaseStop() {
    	return pleaseStop;
    }
    
    private boolean isRunning() {
    	return thr!=null;
    }

	public void stop() {
		pleaseStop=true;
		super.stop();
		while (isRunning()) {
			try {
				Thread.sleep(10);
			} catch (InterruptedException e) {
			}
		}
	}

    /* ---------------------- Mouse Listener -------------------- */

    private boolean dblclkusd;
    public boolean doubleClickUsed() {
    	return dblclkusd;
    }
    private int dblclckpos=-1;
    private long dblclckmillis=-1;
    
    public void mouseClicked(MouseEvent e) {
    	if (!isEnabled()) return;
    	if (mch!=null && mch.isHand()) {
    		return;
    	}
		if (isEditing()) {
	    	if (e.getClickCount()==2||(e.getClickCount()==1&&BasicStr.CtrlOn)) {
				dblclckmillis=System.currentTimeMillis();
	    		dblclkusd=false;
   				if (editNode(e.getX(),e.getY())) {
   					dblclkusd=true;
   				} else {
   					TNode tn=T.getTNodeAtPixel(this,e.getX(),e.getY());
   					if (tn!=null) {
						dblclckpos=getCursorPos();
						selectWordAt(dblclckpos);
						dblclkusd=T.segMarked();
					}
   				}
   				BasicStr.CtrlOn=false;
		    } else if (e.getClickCount()==1) {
		    	if ((System.currentTimeMillis()<(dblclckmillis+500)) && dblclckpos==getCursorPos()) {
		    		selectLineAt(getCursorPos());
		    	}
		    	dblclckmillis=-1;
		    	dblclckpos=-1;
		    }
		} else if (F==null && A!=null) {
//			TNode tn=T.getTNodeAtPixel(this,e.getX(),e.getY());
//			if (tn!=null && tn.isHyperlink() && tn.getUri()!=null)  {
//				try {
//					if (tn.getUri().toLowerCase().startsWith("http://")) {
//						A.getAppletContext().showDocument(new URL(tn.getUri()),"_blank");
//					} else {
//						A.getAppletContext().showDocument(new URL(A.getDocumentBase(),tn.getUri()),"_self");
//					}
//				} catch (Exception ex) {
//					System.out.println(ex.getMessage());
//					System.out.println("in EditorCanvas.mouseClicked");
//				}
//			}
		} else {
			if (e.getClickCount()==1) {
				if (mainF!=null) {
					Formula tempF=mainF.getFormulaAt(e.getX(),e.getY());
					if (tempF!=null) {
						mainF.unmarkSegment();
						if (e.getModifiers()==MouseEvent.BUTTON3_MASK) {
							F=tempF;
							F.markSegment(F.begin(),F.end());
							updateScreen();
							callActionListeners("right_click");
						} else {
							Node N=tempF.getNode();
							if (N!=null && mainF.getNode()!=null 
								&& (N.isVar()||(N.isUnaryOp()&&N.getRight()[0].isVar()) ) ) {
								Node auxN=(Node) mainF.getNode().getParent(N);
								if (auxN!=null && auxN.F!=null) {
									if (auxN.isUnaryOp()&&N.isVar()
										&&(mainF.getNode().getParent(auxN)!=null)
										&&(mainF.getNode().getParent(auxN).F!=null)) {
										auxN=(Node) mainF.getNode().getParent(auxN);
									}
									tempF=auxN.F;
								}
							}
							F=tempF;
							if (F!=null) {
								callActionListeners("single_click");
								mainF.unmarkSegment();
							// para que se recupere F para el doble clic.
								tempF=mainF.getFormulaAt(e.getX(),e.getY()); 
								if (tempF!=null) {
									N=tempF.getNode();
									if (N!=null && mainF.getNode()!=null 
										&& (N.isVar()||(N.isUnaryOp()&&N.getRight()[0].isVar())) ) {
										Node auxN=(Node) mainF.getNode().getParent(N);
										if (auxN!=null && auxN.F!=null) {
											if (auxN.isUnaryOp()&&N.isVar()
												&&(mainF.getNode().getParent(auxN)!=null)
												&&(mainF.getNode().getParent(auxN).F!=null)) {
												auxN=(Node) mainF.getNode().getParent(auxN);
											}
											tempF=auxN.F;
										}
									}
									F=tempF;
								}
								F.markSegment(F.begin(),F.end());
								updateScreen();
							}
						}
					}
				}
			} else if ((e.getClickCount()==2)||(BasicStr.CtrlOn&&(e.getClickCount()==1))) {
				if (F!=null) {
					Node N=F.getNode();
					if (N!=null) {
						callActionListeners("double_click");
					}
				}
			}
		}
    }

    public void mouseEntered(MouseEvent e) {
    	if (!isEnabled()) return;
    	if (mch!=null && mch.isHand()) {
    		return;
    	}
    	if (isEditing()) {
	    	setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
    	} 
    }

    public void mouseExited(MouseEvent e) {
    	if (!isEnabled()) return;
    	if (mch!=null && mch.isHand()) {
    		return;
    	}
    	setCursor(null);
    }

    /* ---------------------- Mouse Motion Listener -------------------- */

//    public void mouseMoved(MouseEvent e) {
//    	if (!isEnabled()) return;
//    	if (mch!=null && mch.isHand()) {
//    		return;
//    	}
//    }

    public void mouseDragged(MouseEvent e) {
    	if (!isEnabled()) return;
		if (falseDrag) return;
    	if (mch!=null && mch.isHand()) {
    		return;
    	}
   		dragged=true;
    	if (mousePressedWhileEditing) {
    		return;
    	}
    	if (isEditing()) {
	    	if (F!=null) {
		    	setCursorPosInFormula(e.getX(),e.getY());
		    	if (F.segEnd()!=F.getCursorPos()) {
			    	F.markSegment(F.segBegin(),F.getCursorPos());
			    	updateScreenLineAtTNode(mainFIndex);
			    }
			} else {
				int segmax=Math.max(T.segBegin(),T.segEnd());
		    	setCursorPosInText(e.getX(),e.getY());
		    	if (T.segEnd()!=T.getCursorPos()) {
			    	T.markSegment(T.segBegin(),T.getCursorPos());
			    	updateScreenSegment(T.segMin(),Math.max(T.segMax(),segmax));
			    }
			}
		}
    }

    /* ---------------------- Focus Listener -------------------- */

    public void focusGained(FocusEvent e) {
    	if (isEnabled()) {
			callActionListeners("focus",ActionEvent.ACTION_PERFORMED);
//			if (al!=null) { //linux
//				al.actionPerformed(new ActionEvent(this,ActionEvent.ACTION_PERFORMED,"focus"));
//			}
			hasFocus=true;
	    	if (F!=null) {
	    		F.setActive(true);
				updateScreenFromTNode(mainFIndex);
	    	}
	    	updateCursorThread();
	    	paintCursor();
    	}
    }

    public void focusLost(FocusEvent e) {
    	if (isEnabled()) {
	    	eraseCursor();
	    	hasFocus=false;
	    	if (F!=null) {
	    		F.setActive(false);
				updateScreenFromTNode(mainFIndex);
	    	}
	    }
    }
    
    boolean hasTheFocus() {
    	return hasFocus;
    }

/* ---------------- Component Listener -------------- */

	public void componentResized(ComponentEvent e) {
		if (updateSize(false)) {
			updateScreen();
		}
	}
	public void componentMoved(ComponentEvent e) {
	}
	public void componentShown(ComponentEvent e) {
		if (updateSize(false)) {
			updateScreen();
		}
	}
	public void componentHidden(ComponentEvent e) {
	}

}

class MathEditor extends editDialog implements ActionListener, MouseListener, Explainer {

	private Label l_dec;
	private TextField tf,tf_dec;
	private Checkbox ch_fixed;
	private Button b_help;
	private translator Tr;
	private MouseCursorHandler mch;

	MathEditor(Frame parent,translator Tr,String expr,String dec_expr,boolean fixed) {
		super(parent,Tr.getTr(data.expr),true,Tr.getTr(data.accept),Tr.getTr(data.cancel));
		this.Tr=Tr;
		mch=new MouseCursorHandler(this,this);
		setFont(mjaFont.AWTSansSerif);
		add("Center",tf=new TextField(40));
		tf.setFont(mjaFont.makeFont("SansSerif",Font.PLAIN,17));
		tf.setText(expr);
		tf.addActionListener(this);
		tf.addMouseListener(this);
		Panel P=new Panel();
		add("North",P);
		P.add(l_dec=new Label(Tr.getTr(data.decimals)));
		l_dec.addMouseListener(this);
		P.add(tf_dec=new TextField(20));
		tf_dec.setFont(mjaFont.AWTMono);
		tf_dec.setText(dec_expr);
		tf_dec.addMouseListener(this);
		P.add(ch_fixed=new Checkbox(Tr.getTr(data.fixed)));
		ch_fixed.setState(fixed);
		ch_fixed.addMouseListener(this);
		P.add(b_help=new Button("?"));
		b_help.addActionListener(this);
		b_help.addMouseListener(this);
	}

	String getExpr() {
		return tf.getText();
	}

	String getDecExpr() {
		String aux=tf_dec.getText();
		if (aux!=null && aux.length()>0) {
			return aux;
		} else {
			return null;
		}
	}

	boolean getFixed() {
		return ch_fixed.getState();
	}

	public void actionPerformed(ActionEvent e) {
		if (!mch.isHand()) {
			if (e.getSource()==tf) {
				ok=true;
				setVisible(false);
			} else if (e.getSource()==b_help) {
				mch.setHand();
			}
		}
	}

	public String getExplLabel(MouseEvent e) {
		Component c=(Component) e.getSource();
		if (c instanceof Label) {
			Label l=(Label) c;
			return l.getText();
		} else if (c instanceof TextField) {
			TextField t=(TextField) c;
			if (t==tf_dec) {
				return l_dec.getText();
			} else if (t==tf) {
				return Tr.getTr(data.expr);
			}
		} else if (c instanceof Button) {
			Button b=(Button) c;
			return b.getLabel();
		} else if (c instanceof Checkbox) {
			Checkbox chb=(Checkbox) c;
			return chb.getLabel();
		}
		return "";
	}

	public String getExplanation(MouseEvent e) {
		String[] sa=null;
		Component c=(Component) e.getSource();
		if (c instanceof Label) {
			Label l=(Label) c;
			if (l==l_dec) {
				sa=Expl.EditExpression[0];
			}
		} else if (c instanceof TextField) {
			TextField t=(TextField) c;
			if (t==tf_dec) {
				sa=Expl.EditExpression[0];
			} else if (t==tf) {
				sa=Expl.EditExpression[3];
			}
		} else if (c instanceof Checkbox) {
			Checkbox chb=(Checkbox) c;
			if (chb==ch_fixed) {
				sa=Expl.EditExpression[1];
			}
		} else if (c instanceof Button) {
			Button b=(Button) c;
			if (b==b_help) {
				sa=Expl.EditExpression[2];
			}
		}
		return Expl.get(sa,Tr.getActiveLanguage());
	}

/* ---------------- Mouse Listener -------------- */

	public void mousePressed(MouseEvent e) {}
	public void mouseReleased(MouseEvent e) {
		/* this should be in Mouse released to avoid selections to be performed and
		   on the other hand to avoid sececting parts in text areas */
		if (e.getModifiers()==MouseEvent.BUTTON3_MASK || mch.isHand() ) {
			mjaText.msg((Component) e.getSource(),getExplLabel(e),getExplanation(e),mjaFont.AWTMono);
			if (mch.isHand()) {
				e.consume();
				mch.setNull();
			}
		}
	}
	public void mouseClicked(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}

/* ---------------- paint -------------- */

}