/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package common;

import java.applet.*;
import java.util.*;
import org.unam.matem.Attribute;
import org.unam.matem.Communicator;

public interface DescartesInterface {

	/* ----- public interface for use with JavaScript ----- */

	public abstract Applet getApplet() ;

	/**
	  * Displays the "about" or Copyright dialog box.
	  */

	public abstract void about();

	/**
	  * Cleans the screen form any traces left by graphs. <br>
	  * It doesn't change parameter or auxiliary values. <br>
	  * The function may be called while the animation is running. <br>
	  */

	public abstract void clear();

	/**
	  * Returns in the first line the number of evaluation controls as 'questions=N'
	  * then in the second line the number of correct answers as 'correct=C'
	  * in the third line, the number of wrong answers as 'wrong=W'
	  * and N lines with the names and user defined values of all evaluation controls 
	  * followed by |1 if correct, by |0 if not correct and by | if not answered.
	  */
	public String getEvaluation();
	
	/**
	  * Returns the names and present values of all variables registered in the parser
	  * including arrays and matrices
	  */
	public String dump();

	/**
	  * Recovers the values of all variables from a previous dump and clears the scene
	  */

	public void undump(String dump);

	/**
	  * Puts into de evaluation controls the correct solutions as defined by the author 
	  */

	public void showSolution();

	/**
	  * Puts into de evaluation controls the user answers
	  */

	public void showAnswer();

	/**
	  * Recovers the values of all variables from a previous dump and clears the scene
	  */

	public String Evaluate(String expression) throws Exception;

	/**
	  * Refresheds the scene. <br>
	  * This method redraws the scene without cleaning the trace left by moving graphs.
	  */

	public abstract void refresh();

	/**
	  * Recovers the original configuration, i.e. the one provided by the applet parameters.
	  */

	public abstract void original();

	/**
	  * Continues or pauses animation. <br>
	  * If the animation has not started starts it.<br>
	  * If the animation is running it puts it in "pause" mode.<br>
	  * If the animation is in "pause" mode then it continues running.<br>
	  */

	public abstract void toggleAnimation();

	/**
	  * Initializes the animation. <br>
	  * It works both on a paused or a running animation.<br>
	  */

	public abstract void initAnimation();

	/**
	 * Opens the configuration editor window.
	 */

	public abstract void edit();

	/* ----------------- Buttons ----------------- */

	/**
	  * Sets the configuration of the buttons.<br>
	  * @param s Configuration String for the buttons.<br>
	  * <br>
	  * A typical form is<br>
	  * <br>
      * s="crditos=si config=no inicio=si limpiar=si"
	  */
	public abstract void setButtons(String s);

	/* ----------------- Space ----------------- */

	/**
	  * Sets the configuration of the Descartes space.<br>
	  * If the scene has more than one space, all other spaces are deleted
	  * and the only one remaining is the one defined by this method call.
	  * Note: in order to modify one space of several, use <a href="Descartes.html#replaceSpace(java.lang.String, java.lang.String)">replaceSpace(String, String).</a>
	  * @param s Configuration String for the space.<br>
	  * <br>
	  * A typical form is<br>
	  * <br>
      * s="fondo=blanco red=no red10=gris ejes=gris texto=grisObscuro"
	  */
	public abstract void setSpace(String s);

	/* ----------------- Spaces ----------------- */

	/**
	  * Removes all spaces from the configuration.<br>
	  */
	public abstract void removeSpaces();

	/**
	  * Adds a new space object.<br>
	  * @param s Configuration String for the space object.<br>
	  * <br>
	  * Typical forms for s are:<br>
	  * <br>
	  * s="id='E00' R3='false' fondo='negro' red='grisObscuro' red10='gris' nmeros='false' despliegue_del_fondo='arriba-izquierda'" <br>
	  * s="id='E01' R3='true' x='59%' y='53%' ancho='32%' alto='44%' escala='24' O.y='20' fondo='turquesa' despliegue_del_fondo='arriba-izquierda' despliegue='pintor' cortar='true'" <br>
	  * s="id='E02' x='5%' y='5%' ancho='30%' alto='30%' escala='10%' fondo='blanco' red10='gris' ejes='gris'" <br>
	  * s="id='E03' R3='si' x='5%' y='53%' ancho='32%' alto='44%' escala='32' fondo='turquesa'" <br>
	  */
	public abstract void addSpace(String s);

	/**
	  * Replaces the num space object.<br>
	  * @param num String representing the number of the space object.<br>
	  * @param s Configuration String for the new space object.<br>
	  * <br>
	  * Typical forms for s are:<br>
	  * s="id='E00' R3='false' fondo='negro' red='grisObscuro' red10='gris' nmeros='false' despliegue_del_fondo='arriba-izquierda'" <br>
	  * s="id='E01' R3='true' x='59%' y='53%' ancho='32%' alto='44%' escala='24' O.y='20' fondo='turquesa' despliegue_del_fondo='arriba-izquierda' despliegue='pintor' cortar='true'" <br>
	  * s="id='E02' x='5%' y='5%' ancho='30%' alto='30%' escala='10%' fondo='blanco' red10='gris' ejes='gris'" <br>
	  * s="id='E03' R3='si' x='5%' y='53%' ancho='32%' alto='44%' escala='32' fondo='turquesa'" <br>
	  * <br>
	  */
	public abstract void replaceSpace(String num,String s);

	/**
	  * Returns the num-th space configuration String.<br>
	  * @param num String representing the number of the space object.<br>
	  */
	public abstract String getSpace(String num);

	/**
	  * Deletes the num-th space object.<br>
	  * @param num String representing the number of the space object to be deleted.<br>
	  */
	public abstract void deleteSpace(String num);

	/* ----------------- Controls ----------------- */

	/**
	  * Removes all numerical and graphics controls from the configuration.<br>
	  */
	public abstract void removeControls();

	/**
	  * Adds a new control object.<br>
	  * @param s Configuration String for the control object.<br>
	  * <br>
	  * Typical forms for s are:<br>
	  * <br>
      * s="id=p tipo=numrico valor='1.5' decimales='2' incr='0.1'"<br>
      * s="id=C tipo=grfico expresin='[1,0]' constriccin='x^2+y^2=1' text='([C.x],[C.y])'"<br>
	  */
	public abstract void addControl(String s);
	/**
	  * Replaces the num control object.<br>
	  * @param num String represinting the number of the control object.<br>
	  * @param s Configuration String for the new control object.<br>
	  * <br>
	  * Typical forms for s are:<br>
      * s="id=p tipo=numrico valor='1.5' decimales='2' incr='0.1'"<br>
      * s="id=C tipo=grfico expresin='[1,0]' constriccin='x^2+y^2=1' text='([C.x],[C.y])'"<br>
	  * <br>
	  */
	public abstract void replaceControl(String num,String s);

	/**
	  * Returns the num-th control configuration String.<br>
	  * @param num String represinting the number of the control object.<br>
	  */
	public abstract String getControl(String num);

	/**
	  * Deletes the num-th control object.<br>
	  * @param num String representing the number of the control object to be deleted.<br>
	  */
	public abstract void deleteControl(String num);

	/* ----------------- Auxiliars ----------------- */

	/**
	  * Removes all auxiliar objects from the configuration.<br>
	  */
	public abstract void removeAux();

	/**
	  * Adds a new auxiliar object.<br>
	  * @param s Configuration String for the auxiliar object.<br>
	  * <br>
	  * Typical forms for s are:<br>
	  * <br>
      * s="id=Ta expresin='0' constante=si evaluar=una-sola-vez"<br>
      * s="id=f(x) expresin='P*cos(x)'"<br>
	  */
	public abstract void addAux(String s);

	/**
	  * Replaces the num-th auxiliar object.<br>
	  * @param num String represinting the number of the auxiliar object.<br>
	  * @param s Configuration String for the new auxiliar object.<br>
	  * <br>
	  * Typical forms for s are:<br>
      * s="id=Ta expresin='0' constante=si evaluar=una-sola-vez"<br>
      * s="id=f(x) expresin='P*cos(x)'"<br>
	  * <br>
	  */
	public abstract void replaceAux(String num,String s);

	/**
	  * Returns the num-th auxiliar configuration String.<br>
	  * @param num String represinting the number of the auxiliar object.<br>
	  */
	public abstract String getAux(String num);

	/**
	  * Deletes the num-th auxiliar object.<br>
	  * @param num String representing the number of the auxiliar object to be deleted.<br>
	  */
	public abstract void deleteAux(String num);

	/* ----------------- Graphs ----------------- */

	/**
	  * Removes all graphic objects from the configuration.<br>
	  */
	public abstract void removeGraphs();

	/**
	  * Adds a new graphic object.<br>
	  * @param s Configuration String for the graphic object.<br>
	  * <br>
	  * Typical forms for s are:<br>
	  * <br>
      * s="tipo=ecuacin expresin='y=f(x)' rastro=turquesa"<br>
      * s="tipo=segmento expresin='(-3,-1)(4,3)' color=magenta"<br>
      * s="tipo=curva expresin='(R*t*cos(2*pi*t),R*t*sen(2*pi*t))' color=rojo visible=no parmetro=t t.intervalo=[0,1] t.pasos=32 ancho=3"<br>
      * s="tipo=polgono expresin='(0,0)(1,1)(2,-1)(0,0)' color=gris"<br>
	  */
	public abstract void addGraph(String s);

	/**
	  * Replaces the num-th graphic object.<br>
	  * @param num String representing the number of the graphic object.<br>
	  * @param s Configuration String for the new graphic object.<br>
	  * <br>
	  * Typical forms for s are:<br>
	  * <br>
      * s="tipo=ecuacin expresin='y=f(x)' rastro=turquesa"<br>
      * s="espacio='Principal' tipo=segmento expresin='(-3,-1)(4,3)' color=magenta"<br>
      * s="espacio='E01_2D' tipo=curva expresin='(R*t*cos(2*pi*t),R*t*sen(2*pi*t))' color=rojo visible=no parmetro=t t.intervalo=[0,1] t.pasos=32 ancho=3"<br>
      * s="espacio='E02' tipo=polgono expresin='(0,0)(1,1)(2,-1)(0,0)' color=gris"<br>
	  */
	public abstract void replaceGraph(String num,String s);
	/**
	  * Returns the num-th graphic configuration String.<br>
	  * @param num String representing the number of the graphic object.<br>
	  */
	public abstract String getGraph(String num);
	/**
	  * Deletes the num-th graphic object.<br>
	  * @param num String representing the number of the graphic object to be deleted.<br>
	  */
	public abstract void deleteGraph(String num);

	/* ----------------- Graphs 3D ----------------- */

	/**
	  * Removes all 3D graphic objects from the configuration.<br>
	  */
	public abstract void remove3DGraphs();

	/**
	  * Adds a new 3D graphic object.<br>
	  * @param s Configuration String for the graphic object.<br>
	  * <br>
	  * Typical forms for s are:<br>
	  * <br>
	  * s="espacio='E01' tipo='Octaedro' color='naranja' color_reverso='rosa' aristas='true' rotini='(180,0,30*(1+C.x))' posini='(0,0,1.5*C.y+1.5)' modelo='luz' cortar='false'" <br>
	  * s="espacio='E01' tipo='Superficie' expresin='x=3*u-1.5 y=3*v-1.5 z=p*(y^2-x^2)/1.5' Nu='Nu' Nv='Nu' color_reverso='rosa' aristas='false' rotini='(0,0,30)' modelo='metal' cortar='false'" <br>
	  * s="espacio='E03' tipo='Superficie' expresin='x=3*u-1.5 y=3*v-1.5 z=0' Nu='Nu' Nv='Nu' color_reverso='rosa' aristas='false' rotini='(0,0,30)' posini='(0,0,-1.5)' modelo='alambre' cortar='false'" <br>
	  * s="espacio='E03' tipo='Paraleleppedo' ancho='3' largo='3' alto='3' color_reverso='rosa' aristas='false' rotini='(0,0,30)' modelo='alambre' cortar='false'" <br>
	  * s="espacio='E03' tipo='Esfera' Nu='Nu' Nv='2*Nu' ancho='2.5' color='naranja' color_reverso='rosa' aristas='true' modelo='luz' cortar='false'" <br>
	  * s="espacio='E01' tipo='macro' expresin='ejes' color_reverso='rosa' aristas='false' rot='(0,0,30)' cortar='false' nombre='mis_ejes'" <br>
	  */
	public abstract void add3DGraph(String s);

	/**
	  * Replaces the num-th 3D graphic object.<br>
	  * @param num String representing the number of the graphic object.<br>
	  * @param s Configuration String for the new 3D graphic object.<br>
	  * <br>
	  * Typical forms for s are:<br>
	  * <br>
	  * s="espacio='E01' tipo='Octaedro' color='naranja' color_reverso='rosa' aristas='true' rotini='(180,0,30*(1+C.x))' posini='(0,0,1.5*C.y+1.5)' modelo='luz' cortar='false'" <br>
	  * s="espacio='E01' tipo='Superficie' expresin='x=3*u-1.5 y=3*v-1.5 z=p*(y^2-x^2)/1.5' Nu='Nu' Nv='Nu' color_reverso='rosa' aristas='false' rotini='(0,0,30)' modelo='metal' cortar='false'" <br>
	  * s="espacio='E03' tipo='Superficie' expresin='x=3*u-1.5 y=3*v-1.5 z=0' Nu='Nu' Nv='Nu' color_reverso='rosa' aristas='false' rotini='(0,0,30)' posini='(0,0,-1.5)' modelo='alambre' cortar='false'" <br>
	  * s="espacio='E03' tipo='Paraleleppedo' ancho='3' largo='3' alto='3' color_reverso='rosa' aristas='false' rotini='(0,0,30)' modelo='alambre' cortar='false'" <br>
	  * s="espacio='E03' tipo='Esfera' Nu='Nu' Nv='2*Nu' ancho='2.5' color='naranja' color_reverso='rosa' aristas='true' modelo='luz' cortar='false'" <br>
	  * s="espacio='E01' tipo='macro' expresin='ejes' color_reverso='rosa' aristas='false' rot='(0,0,30)' cortar='false' nombre='mis_ejes'" <br>
	  */
	public abstract void replace3DGraph(String num,String s);
	/**
	  * Returns the num-th 3D graphic configuration String.<br>
	  * @param num String representing the number of the 3D graphic object.<br>
	  */
	public abstract String get3DGraph(String num);
	/**
	  * Deletes the num-th 3D graphic object.<br>
	  * @param num String representing the number of the 3D graphic object to be deleted.<br>
	  */
	public abstract void delete3DGraph(String num);

	/* ----------------- Animation ----------------- */

	/**
	  * Removes the animation.<br>
	  */
	public abstract void removeAnimation();
	/**
	  * Returns the animation configuration String.<br>
	  */
	public abstract String getAnimation();

	/**
	  * Defines a new animation.<br>
	  * @param s Configuration String for the new animation.<br>
	  * <br>
	  * A typical form for s:<br>
	  * <br>
      * s="id=Animacin algoritmo=si inicio='Ta=0' hacer='Ta=Ta+0.05' mientras='Ta<2' auto=si repetir=si controles=no evaluar=siempre"<br>
	  */
	public abstract void setAnimation(String s);

	/* ----------------- Applet Code ----------------- */

	/**
	  * Returns the applet configuration i.e. the html code that defines the applet configuration.<br>
	  */
	public abstract String getCode();

	/**
	  * Returns the applet configuration i.e. the html code that defines the applet configuration using the standard HTML encoding of accents and "foreign" characters.<br>
	  */
	public abstract String getHTMLEncodedCode();

	public abstract void setDescartesWeb2_0(boolean b);

/* ---------------- Communicator Interface --------------- */
	public abstract Attribute[] getState();
	public abstract void changed(Communicator cc);
	public abstract void addCommunicator(Communicator comm);
	public abstract void sendChanges();
/* ------------------------------------------------------- */	


}