/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package edu.ite.items.beans;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.StringBufferInputStream;
import java.io.StringReader;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;



public class Metadato {
	private String label;
	private String descripcion;
	private String[] defaultVal;
	private String value;
	private Boolean requiered;
	
	
	/**
	 * @return the requiered
	 */
	public Boolean getRequiered() {
		return requiered;
	}
	/**
	 * @param requiered the requiered to set
	 */
	public void setRequiered(Boolean requiered) {
		this.requiered = requiered;
	}
	public Metadato(String label, String descrip, String[] defVals, Boolean required) {
		this.label = label;
		this.descripcion = descrip;
		this.defaultVal = defVals; 
		
		if(this.defaultVal == null)
			this.defaultVal = new String[0];
		
		this.requiered = required;
	}
	public String getLabel() {
		if(label == null)
			return "";
		return label;
	}
	public void setLabel(String label) {
		this.label = label;
	}
	public String getDescripcion() {
		return descripcion;
	}
	public void setDescripcion(String descripcion) {
		this.descripcion = descripcion;
	}
	public String[] getDefaultVal() {
		return defaultVal;
	}
	public void setDefaultVal(String[] defaultVal) {
		this.defaultVal = defaultVal;
	}
	

	public String toString(){
		return "{"+label+","+descripcion+","+defaultVal.length+","+requiered+"}";
	}
	
	
	
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	@Override
	public Object clone() throws CloneNotSupportedException {
		Metadato clon = new Metadato(this.label.toString(),this.descripcion.toString(),this.defaultVal.clone(),this.requiered);
		return clon;
	}
	/**
	 * @param value the value to set
	 */
	public void setValue(String value) {
		this.value = value;
	}
	/**
	 * @return the value
	 */
	public String getValue() {
		if(value==null)
			return "";
		return value;
	}
	
	/**
	 * 
	 * @param metadatos Una cadena en formato xml que contien la informacion de los metadatos.
	 * <pre>
	 * 	&lt;metadatos&gt;
	 * 		&lt;metadato label="" descripcion="" valoresDefault="" obligatorio="" &gt;
	 * 		... 
	 * 	&lt;/metadatos&gt;
	 * </pre>
	 * @return Regresa los metadatos obtenidos de la cadena xml
	 * @throws SAXException
	 * @throws IOException
	 */
	public static final Metadato[] getMetadatosFromXMLString(String metadatos)
			throws SAXException {
		DocumentBuilderFactory builderFactory = DocumentBuilderFactory
				.newInstance();
		DocumentBuilder builderXML;
		Document domDoc;
		try {
			if (metadatos == null || metadatos.length() <= 0)
				throw new ParserConfigurationException("The string is null");

			StringReader reader = new StringReader(metadatos);
			InputSource inputSource = new InputSource(reader);
			builderXML = builderFactory.newDocumentBuilder();
			
			domDoc = builderXML.parse(inputSource);
			NodeList lista = domDoc.getElementsByTagName("metadatos");
			if (lista.getLength() <= 0){
				throw new SAXException("No se encuentra el tag metadatos en:"+metadatos);
			}
			Element metadatosList = (Element) lista.item(0);
			return getMetadatosFromXMLNode(metadatosList);
		} catch (ParserConfigurationException e) {
			System.out.println("Error desconocio. No puede saltar esta excepcion aqui. La configuracion para el xmlbuilder es segura");
			e.printStackTrace();
		} catch (Exception e) {
			System.out.println("Error desconocio. No puede saltar una excepcion aqui. EL input stream esta en memoria. Metadatos:\n"+metadatos);
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * 
	 * @param nodeMetadatos En Element xml que contien la informacion de los metadatos.
	 * <pre>
	 * 	&lt;metadatos&gt;
	 * 		&lt;metadato label="" descripcion="" valoresDefault="" obligatorio="" &gt;
	 * 		... 
	 * 	&lt;/metadatos&gt;
	 * </pre>
	 * @return Lo objetos Metadatos creados apartir del nodo xml
	 * @throws SAXException
	 * @throws IOException
	 */
	public static final Metadato[] getMetadatosFromXMLNode(Element nodeMetadatos) throws SAXException {
		NodeList metadatosNodes = nodeMetadatos.getElementsByTagName("metadato");
		Metadato[] resMeta = new Metadato[metadatosNodes.getLength()];
		for (int i = 0; i < metadatosNodes.getLength(); i++) {
			Element elementAct = (Element) metadatosNodes.item(i);
			String label = elementAct.getAttribute("name");
			String descrip = elementAct.getAttribute("descripcion");
			String defaultVal = elementAct.getAttribute("valoresDefault");
			String value = elementAct.getAttribute("value");
			Boolean required = elementAct.getAttribute("obligatorio").equals(
					"1");
			String[] vals = defaultVal.split(",");

			for (int j = 0; j < vals.length; j++) {
				vals[j] = vals[j].trim();
			}
			resMeta[i] = new Metadato(label, descrip, vals, required);
			resMeta[i].setValue(value);
		}
		return resMeta;
	}
	

	
	public String printToXMLString(){
		String res = "<metadato name='"+getLabel()+"' value='"+getValue()+"' />";
		return res;
	} 
	
	/**
	 * Regresa la representacion en XML de la lista de metadatos contenidos en 
	 * el arreglo param
	 * @param param la lista de los metadatos que se quieren pasar a XML
	 * @return  Una cadena en formato XML que representa a los metadatos
	 */
	public static String getMetadatoXMLfromArray(Metadato[] param){
		String machote = "<metadato name='{@name@}' value='{@value@}' />";
		String res = "<metadatos>\n";
		for (int i = 0; i < param.length; i++) {
			String tmpRep = machote.replace("{@name@}", param[i].getLabel());
			tmpRep = tmpRep.replace("{@value@}", param[i].getValue());
			res += "\t"+tmpRep+"\n";
		}
		res+="</metadatos>";
		return res;
	}
}