/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package edu.mec.descartesserver;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.Proxy;
import java.net.URL;
import java.net.URLEncoder;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import com.mja.file.mjaFile;

import edu.mec.descartesserver.beans.EsceneBean;

public class DescartesServerConector {
	
	private URL host;
	private String user;
	private String password;
	
	public DescartesServerConector(String host,String usr,String password) throws MalformedURLException{
		String shost=(host.toLowerCase().startsWith("http://"))?host:"http://"+host;
		setHost(new URL(shost+"/descartesServer.php"));
		setUser(usr);
		setPassword(password);
	}
	
	public String uploadEscene(String proyecto, String esceneName,EsceneBean esceneBean){
		String result = "Sin operacion ? ";
		try{
		ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
		esceneBean.preparaZip(byteArrayOutputStream); 
		
		String lineEnd = "\r\n";
		String twoHyphens = "--";
		String boundary = "******************";

		URL hostService = getHost();
		Proxy tmpProxy = mjaFile.getProxyFor(hostService);
		HttpURLConnection conn = (HttpURLConnection) hostService
				.openConnection(tmpProxy);
		
		conn.setRequestMethod("POST");
		conn.setUseCaches(false);
		conn.setDoInput(true);
		conn.setDoOutput(true);
		conn.setRequestProperty("Content-Type", "multipart/form-data;boundary="
				+ boundary);

		DataOutputStream dos = new DataOutputStream(conn.getOutputStream());
		
		Hashtable<String, Object> dataHash = new Hashtable<String, Object>();
		dataHash.put("service",		URLEncoder.encode("upload",			"UTF-8"));
		dataHash.put("user",		URLEncoder.encode(getUser(),		"UTF-8"));
		dataHash.put("password",	URLEncoder.encode(getPassword(),	"UTF-8"));
		dataHash.put("project",		URLEncoder.encode(proyecto,			"UTF-8"));
		dataHash.put("file",		URLEncoder.encode(esceneName,		"UTF-8"));
		if(byteArrayOutputStream.size()>0)
			dataHash.put("contents", 	byteArrayOutputStream.toByteArray());
		
		
		Enumeration<String> keys = dataHash.keys();
		while(keys.hasMoreElements()){
			String key	= keys.nextElement();
			Object value = dataHash.get(key);
			boolean esFile = (value instanceof byte[]); 

			dos.writeBytes(twoHyphens + boundary + lineEnd);
			String disposition =
				"Content-Disposition: form-data; name=\"" + key	+ "\""; 
			if(esFile){
				String fileName = key+"_GEN_BY_ARQ_SRV";
				disposition+= "; filename=\""+fileName+".zip\"";
			} 
			dos.writeBytes(disposition+lineEnd);
			dos.writeBytes(lineEnd);
			
			if(esFile)
				dos.write((byte[])value);
			else
				dos.write(value.toString().getBytes("UTF-8"));
			dos.writeBytes(lineEnd);
		}
		dos.writeBytes(twoHyphens + boundary + twoHyphens + lineEnd);
		dos.flush();
		dos.close();

		InputStream ioStream = conn.getInputStream();
		StringBuffer buffer = new StringBuffer(); 
		int readed; 
		while ((readed = ioStream.read()) != -1) {
			buffer.append((char)readed);
		}
		ioStream.close();
		result = buffer.toString();
		conn.disconnect();
		}catch (Exception e){
			e.printStackTrace();
			result = e.getMessage();
		}
		result = result.trim();
		return result;
	}
	
	
	public EsceneBean downloadEscena(String user,String proyecto, String escenaName, File dirToSave) throws IOException{
			
		URL hostService = getHost();
		String rawData = "service=download";
		rawData += "&project="+URLEncoder.encode(proyecto, "UTF-8");
		rawData += "&file="+URLEncoder.encode(escenaName, "UTF-8");
		rawData += "&user="+URLEncoder.encode(user, "UTF-8");
		
		Proxy tmpProxy = mjaFile.getProxyFor(hostService);
		HttpURLConnection conn = (HttpURLConnection) hostService
				.openConnection(tmpProxy);
		conn.setRequestMethod("POST");
		conn.setUseCaches(false);
		conn.setDoInput(true);
		conn.setDoOutput(true);
		conn.setRequestProperty("Content-Type","application/x-www-form-urlencoded");
		OutputStreamWriter wr = new OutputStreamWriter(conn.getOutputStream());
		wr.write(rawData);
		wr.flush();
		
		InputStream ioStream = conn.getInputStream();
		String contentType = conn.getContentType();
		Vector<String> files = new Vector<String>();
		
		if ("application/octet-stream".equals(contentType)) {
			ZipInputStream zipInputStream = new ZipInputStream(ioStream);
			ZipEntry entry;
			int BUFFER = 2048;
			BufferedOutputStream dest = null;

			while ((entry = zipInputStream.getNextEntry()) != null) {
				String destPath = dirToSave.getAbsolutePath()
						+ File.separatorChar + entry.getName();

				File parentDir = new File(destPath);
				parentDir = parentDir.getParentFile();
				if (!parentDir.isDirectory()) {
					if(!parentDir.mkdirs()){
						System.out.println("Error al crear el directorio padre '"+parentDir.getAbsolutePath()+"'");
					} 

				}
				if (destPath.lastIndexOf("/") >= destPath.length() - 1) {
					File destFile = new File(destPath);
					if (!destFile.isDirectory()) {
						destFile = destFile.getAbsoluteFile();
						boolean creado = destFile.mkdirs();
						if(!creado){
							System.out.println("Error al crear el directorio '"+destPath+"'");
						}
					}
					continue;
				}
				
				if(entry.getName().indexOf("index.html")<0)
					files.add(entry.getName());
				 
				
				
				FileOutputStream fos = new FileOutputStream(destPath);
				dest = new BufferedOutputStream(fos, BUFFER);
				int count;
				byte data[] = new byte[BUFFER];
				while ((count = zipInputStream.read(data, 0, BUFFER)) != -1) {
					dest.write(data, 0, count);
				}
				dest.flush();
				dest.close();
			}
			conn.disconnect();
			String codePath = dirToSave.getAbsolutePath()
			+ File.separatorChar + "index.html";
			
			EsceneBean tmpEscena = new EsceneBean();
			tmpEscena.setLocalPath(dirToSave.getAbsolutePath());
			tmpEscena.setFiles(files.toArray(new String[files.size()]));
			tmpEscena.setName(escenaName);
			tmpEscena.setHtmlCode(mjaFile.readFile(codePath));
			return tmpEscena;
		} else {
			int character;
			StringBuffer bufferString = new StringBuffer();
			Reader reader = new InputStreamReader(ioStream);
			while ((character = reader.read()) != -1)
				bufferString.append((char) character);

			System.out.println("El servidor respondio :\n"
					+ bufferString.toString() + "\n\n");
			conn.disconnect();
		}
		return null;
	}

	
	
	
	
	/**
	 * @return the host
	 */
	public URL getHost() {
		return host;
	}

	/**
	 * @param host the host to set
	 */
	public void setHost(URL host) {
		this.host = host;
	}

	/**
	 * @return the user
	 */
	public String getUser() {
		return user;
	}

	/**
	 * @param user the user to set
	 */
	public void setUser(String user) {
		this.user = user;
	}

	/**
	 * @return the password
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * @param password the password to set
	 */
	public void setPassword(String password) {
		this.password = password;
	}
}