/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package org.cnice.lms.client.admin;

import java.awt.*;
import java.awt.event.*;
import org.cnice.lms.admin.*;
import org.cnice.lms.client.*;

public class AdminTool extends Panel implements ActionListener, ItemListener {
	
	private static int users=0,courses=1,tutors=2,learners=3,activities=4;
	private static final String[][] cardName={
		{"New/Search Users","New/Search Courses","New/Search Tutors for Courses","New/Search Learners for Courses","Search Activities"},
		{"Dar de alta/Buscar Usuarios","Dar de alta/Buscar Cursos","Dar de alta/Buscar Tutores de Cursos","Dar de alta/Buscar Alumnos de Cursos","Buscar Actividades"}
	};
	private static final String[][] shortCardName={
		{"Users","Courses","Tutors","Learners","Activities"},
		{"Usuarios","Cursos","Tutores","Alumnos","Actividades"}
	};


	private AdminClient client;
	private boolean connectionOK, loginOK, exceptionCheckingLogin;
	
	private UsersPanel usersP;
	private CoursesPanel coursesP;
	private TutorsPanel tutorsP;
	private LearnersPanel learnersP;
	private ActivitiesPanel activitiesP;

	private Panel CP;
	private CardLayout cL;
	private Panel[] card;
	private Checkbox[] chbx_p;

	public AdminTool(String db_host,String db_initial_port,String db_final_port) {
		setBackground(Color.lightGray);
		setLayout(new BorderLayout());
		setFont(util.SansSerif);

		client=new AdminClient(db_host,(new Integer(db_initial_port)).intValue(),(new Integer(db_final_port)).intValue());
		connect();
		if (connectionOK && loginOK && !exceptionCheckingLogin) {
			normalConstruct();
		} else {
			exceptionConstruct();
		}
	}
	
	private void connect() {
		if (client.connect()) {
			connectionOK=true;
			ClientDialog cd=new ClientDialog(client);
			loginOK=(cd.getUserLogin()!=null);
		} else {
			connectionOK=false;
			System.out.println("No server listening");
		}
	}
		
	private void normalConstruct() {
		Panel NP=new Panel();
		NP.setLayout(new GridLayout(2,1));
		Panel NP0=new Panel();
		NP0.setLayout(new FlowLayout());
		Label lb_user=new Label(client.getUserTypeName()+": "+client.getUserName());
		NP0.add(lb_user);
		lb_user.setAlignment(Label.CENTER);
		lb_user.setFont(util.Monospaced);
		lb_user.setBackground(AdminPanel.veryLightGray);
		lb_user.setForeground(Color.black);
		NP.add(NP0);
		Panel NP1=new Panel();
		NP1.setLayout(new GridLayout(1,cardName[0].length+1));
		NP1.add(new Label());
		CheckboxGroup chbxg_p=new CheckboxGroup();
		chbx_p=new Checkbox[cardName[0].length];
		for (int i=0;i<cardName[0].length;i++) {
			NP1.add(chbx_p[i]=new Checkbox(shortCardName[util.Lang][i],false,chbxg_p));
			chbx_p[i].addItemListener(this);
			chbx_p[i].setEnabled(false);
		}
		NP.add(NP1);
		add("North",NP);
		CP=new Panel();
		add("Center",CP);
		CP.setLayout(cL=new CardLayout());
		card=new Panel[cardName[0].length];
		for (int i=0;i<cardName[0].length;i++) {
			card[i]=new Panel();
			CP.add(card[i],shortCardName[util.Lang][i]);

			card[i].setLayout(new BorderLayout());
			Panel cardNP=new Panel();
			card[i].add("North",cardNP);
			cardNP.setLayout(new FlowLayout(FlowLayout.CENTER));
			Label cardLbl=new Label(cardName[util.Lang][i]);
			cardLbl.setFont(util.SansSerif16);
			cardLbl.setAlignment(Label.CENTER);
			cardNP.add(cardLbl);
			if (i==users) {
				card[i].add("Center",usersP=new UsersPanel(client,this));
				usersP.addActionListener(this);
			} else if (i==courses) {
				card[i].add("Center",coursesP=new CoursesPanel(client,this));
				coursesP.addActionListener(this);
			} else if (i==tutors) {
				card[i].add("Center",tutorsP=new TutorsPanel(client,this));
				tutorsP.addActionListener(this);
			} else if (i==learners) {
				card[i].add("Center",learnersP=new LearnersPanel(client,this));
				learnersP.addActionListener(this);
			} else if (i==activities) {
				card[i].add("Center",activitiesP=new ActivitiesPanel(client,this));
				activitiesP.addActionListener(this);
			}
		}
		Label lb;
		add("East",lb=new Label("  "));
		add("West",lb=new Label("  "));
		add("South",lb=new Label("  "));
		for (int i=0;i<cardName[0].length;i++) {
			chbx_p[i].setEnabled(false);
		}
		chbx_p[activities].setEnabled(true);
		if (client.getUserType().equals(AdminClient.userTypeNames[0][AdminClient.admin])) {
			for (int i=0;i<cardName[0].length;i++) {
				chbx_p[i].setEnabled(true);
			}
			chbx_p[users].setState(true);
			cL.show(CP,shortCardName[util.Lang][users]);
		} else if (client.getUserType().equals(AdminClient.userTypeNames[0][AdminClient.tutor])) {
			chbx_p[tutors].setEnabled(true);
			chbx_p[learners].setEnabled(true);
			chbx_p[activities].setState(true);
			cL.show(CP,shortCardName[util.Lang][activities]);
		} else if (client.getUserType().equals(AdminClient.userTypeNames[0][AdminClient.learner])) {
			chbx_p[learners].setEnabled(true);
			chbx_p[activities].setState(true);
			cL.show(CP,shortCardName[util.Lang][activities]);
		}
	}
	
	public void refreshCourses() {
		tutorsP.refreshCourses();	
		learnersP.refreshCourses();	
	}
	
	public void refreshUsers() {
		tutorsP.refreshUsers();	
		learnersP.refreshUsers();	
		activitiesP.refreshUsers();	
	}
	
	public void goToActivities(String userid,String courseid,String tutor_userid) {
		chbx_p[activities].setState(true);
		cL.show(CP,chbx_p[activities].getLabel());
		activitiesP.setFromLearners(userid,courseid,tutor_userid);	
	}
	
/* ------------------- ActionListener ---------------- */

	public void actionPerformed(ActionEvent e){
		if (e.getSource()==usersP) {
			if (AdminClient.userTypeNames[util.Lang][AdminClient.learner].equals(usersP.getUserType())) {
				learnersP.selectLearnerById(usersP.getSelectedId());
				activitiesP.selectLearnerById(usersP.getSelectedId());
			} else if (AdminClient.userTypeNames[util.Lang][AdminClient.tutor].equals(usersP.getUserType())) {
				tutorsP.selectTutorById(usersP.getSelectedId());
				learnersP.selectTutorById(usersP.getSelectedId());
				activitiesP.selectTutorById(usersP.getSelectedId());
			}
		} else if (e.getSource()==coursesP) {
			tutorsP.selectCourseById(coursesP.getSelectedId());
			learnersP.selectCourseById(coursesP.getSelectedId());
			activitiesP.selectCourseById(coursesP.getSelectedId());
		} else if (e.getSource()==tutorsP) {
			learnersP.selectCourseById(tutorsP.getSelectedCourseId());
			learnersP.selectTutorById(tutorsP.getSelectedTutorId());
			activitiesP.selectCourseById(tutorsP.getSelectedCourseId());
			activitiesP.selectTutorById(tutorsP.getSelectedTutorId());
		} else if (e.getSource()==learnersP) {
			activitiesP.selectLearnerById(learnersP.getSelectedLearnerId());
			activitiesP.selectCourseById(learnersP.getSelectedCourseId());
			activitiesP.selectTutorById(learnersP.getSelectedTutorId());
//			activitiesP.getValues(learnersP.getLastActivity());
		} else if (e.getSource()==activitiesP) {
		}
	}

	/* ------------------- ItemListener ---------------- */

	public void itemStateChanged(ItemEvent e) {
		if (e.getStateChange()==e.SELECTED) {
			for (int i=0;i<cardName[0].length;i++) {
				if (e.getSource()==chbx_p[i]) {
					cL.show(CP,chbx_p[i].getLabel());
					return;
				}
			}
		}
	}

	/* ------------------- Exception construction ----------------- */
	
	private static final String BadConnection=
	"\n El programa no puede conectarse a la base de datos del LMS.\n"+
	"\n Si se trata de un applet lo ms probable es que el servidor\n"+
	"\n de la base de datos no est funcionando.\n"+
	"\n Si el programa no es un applet tambin es posible que haya\n"+
	"\n un error en el URL del servidor LMS o los puertos asignados.\n"+
	"\n Verifique que los datos que aparecen en LMS_Admin.bat\n"+
	"\n son vigentes.\n"
	;
	private static final String BadLogin=
	"\n Este programa necesita una combinacin usuario-contrasea\n"+
	"\n registrada en la base de datos del LMS.\n"+
	"\n Como el usuario no proporcion datos correctos de\n"+
	"\n usuario-contrasea el programa no puede funcionar.\n"
	;
	private static final String ExceptionCheckingLogin=
	"\n Se produjo una excepcin al verificar la contrasea.\n"+
	"\n Probablemente se trata de un problema transitorio.\n"+
	"\n Sugerimos reintentar entrar al programa ms tarde.\n"
	;

	private void exceptionConstruct() {
		TextArea TA=new TextArea(25,80);
		TA.setEditable(false);
		add("Center",TA);
		if (!connectionOK) {
			TA.setBackground(new Color(0xaaffaa));
			TA.setText(BadConnection);
		} else if (!loginOK) {
			TA.setBackground(new Color(0xffaaaa));
			TA.setText(BadLogin);
		} else if (exceptionCheckingLogin) {
			TA.setBackground(new Color(0xaaaaff));
			TA.setText(ExceptionCheckingLogin);
		}
	}
	
}