/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package org.cnice.lms.client.learner;
/*
 * @(#)LearnerClient.java    2.0  3 Sep 2002
 *
 * This software is property of CNICE
 *
 */

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;

import org.cnice.lms.learner.*;
import org.cnice.lms.common.*;
import org.cnice.lms.client.*;


/**
 *
 * @version	2.0  3 Sep 2002
 * @author Jose-Luis Abreu
 */
public class LearnerClient extends Client {

	private static volatile LearnerClient cl;
	private static volatile ClientDialog cd;

	public static LearnerClient client() {
		return cl;
	}

	public static synchronized void initLMS(String host) {
//System.out.println("initLMS in ...");
		if (cl==null) {
			try {
				cl=new LearnerClient(host);
				cl.connect();
				if (cl.isConnected()) {
					if (cd==null) {
						cd=new ClientDialog(cl);
					}
					String login=cd.getUserLogin();
					if (login!=null && login.length()>0) {
						cl.setLog(System.out);
					} else {
						cl=null;
					}
				} else {
					System.out.println("LMS not connected");
					cl=null;
				}
			} catch (Exception e) {
				e.printStackTrace();
				cl=null;
			}
		}
//System.out.println("initLMS out");
	}

	private PrintStream logps;
	private StringBuffer loginfo;

	public LearnerClient(String host) {
		super(host);
	}

	public LearnerClient(String host,int port) {
		super(host,port);
	}

	public LearnerClient(String host,int initial_port,int final_port) {
		super(host,initial_port,final_port);
	}

	public void setUser(String user_login,String user_password) {
		super.setUser(user_login,user_password);
		if (user_login!=null&&user_login.length()>0) {
			loginfo = new StringBuffer();
		}
	}

	public void setLog(String fileName) {
		if (fileName!=null) {
			try {
				fileName=fileName.replace('/',File.separatorChar);
				String dname=fileName;
				int p=dname.lastIndexOf(File.separatorChar);
				if (p>=0) {
					dname=dname.substring(0,p+1);
				}
				File f=new File(dname);
				f.mkdirs();
				logps = new PrintStream(
					new BufferedOutputStream(
					new FileOutputStream(fileName,true)));
			} catch (Exception e) {
				System.out.println(e);
			}
		}
	}

	public void setLog(PrintStream ps) {
		logps=ps;
	}

	private void logprintln(String id,String s) {
		if (logps!=null) {
			logps.println(s);
		}
		if (loginfo!=null) {
			loginfo.append(s+"\n");
			if (isConnected()&&id!=null) {
				updateActivity(id,s+"\n");
			}
		}
	}

	public void ProgramFinish() {
		if (logps!=null) {
			try {
				logps.close();
			} catch (Exception e) {
				System.out.println(e);
			}
		}
	}

	public void LMSFinish(Activity a) {
		LMSAddAction(a,"Finish",SqlMetadata.getDateTimeString());
		String auxs="	</ACTIVITY>\n"+
			"</LMS>";
		logprintln (a.getID(),auxs);
		if (isConnected()&&a.getID()!=null) {
			closeActivity(a.getID());
		}
	}

	public Activity LMSInitialize(String LearningProgram,String course,String unit,String exercise,String location) {
		String date=SqlMetadata.getDateTimeString();
		String[] info=null;
		if (isConnected()) {
			info=insertActivity(LearningProgram,date,course,unit,exercise,location);
		}
		String auxs="<?xml version=\"1.0\"?>\n"+
			"<!DOCTYPE LMS SYSTEM \"LMS.dtd\">\n"+
			"<LMS>\n"+
			"	<ACTIVITY"+" LearningProgram=\""+LearningProgram+"\"\n"+
			"		exercise=\""+exercise+"\"\n"+
			"		unit=\""+unit+"\"\n"+
			"		course=\""+course+"\"\n"+
			"		location=\""+location+"\"\n"+
			"	>";
		logprintln(info[0],auxs);
		Activity a=new Activity(info[0],info[1]);
		LMSAddAction(a,"Start",date);
		return a;
	}

	public void LMSAddAction(Activity a,String name, String msg) {
		String time=String.valueOf((System.currentTimeMillis()-a.getInitialMillis())/1000);
		String auxs="		<EVENT time=\""+time+"\" name=\""+name+"\"";
		if (msg!=null&&msg.length()>0) {
			auxs+=">";
			auxs+="\n			"+msg;
			auxs+="\n		</EVENT>";
		} else {
			auxs+="/>";
		}
		logprintln (a.getID(),auxs);
	}

	private String[] insertActivity(String LearningProgram,String date,String course,String unit,String exercise,String location) {
		String[] info= null;
		try{
			String insertStr="(LearningProgram,exercise,location,unit,userid,courseid,date,state) "+
				"VALUES ("
				+"'"+LearningProgram+"'"+","
				+"'"+exercise+"'"+","
				+"'"+location+"'"+","
				+"'"+unit+"'"+","
				+"?,"
				+"?,"
				+"'"+date+"'"+","
				+"0"+")";
			info = executeInsertActivity(insertStr,course);
//			System.out.println("El id es: "+info[0]);
		} catch (Exception ex) {
			tracePrint(ex);
		}
		return info;
	}

	private void updateActivity(String id,String appendStr) {
		try{
		    executeUpdateActivity(id,appendStr);
		} catch (Exception ex) {
			tracePrint(ex);
		}
	}

	private void closeActivity(String id) {
		try{
		    executeCloseActivity(id);
		} catch (Exception ex) {
			tracePrint(ex);
		}
	}

    private String[] executeInsertActivity(String insertStr,String course) throws Exception {
		Message request = new InsertActivityRequest(user_login,user_password,
			insertStr,user_login,course);
		Message response = sendRequest(request);
		if(response==null) throw new Exception("Unreachable destination");

		if(response.getType()==Message.INSERTACTIVITY_RESPONSE){
		    return ((InsertActivityResponse) response).getInfo();
		}
		if(response.getType()==Message.EXCEPTION_RESPONSE){
			throw new Exception(((ExceptionResponse) response).getExceptionMessage());
		}
		tracePrint("Unexpected Response Type");
		return null;
    }

    private void executeUpdateActivity(String id,String appendStr) throws Exception {
		Message request = new UpdateActivityRequest(user_login,user_password,
			id,appendStr);
		Message response = sendRequest(request);
		if(response==null) throw new Exception("Unreachable destination");

		if(response.getType()==Message.UPDATEACTIVITY_RESPONSE){
		    return;
		}
		if(response.getType()==Message.EXCEPTION_RESPONSE){
			throw new Exception(((ExceptionResponse) response).getExceptionMessage());
		}
		tracePrint("Unexpected Response Type");
    }

    private void executeCloseActivity(String id) throws Exception {
		Message request = new CloseActivityRequest(user_login,user_password,
			id);
		Message response = sendRequest(request);
		if(response==null) throw new Exception("Unreachable destination");

		if(response.getType()==Message.CLOSEACTIVITY_RESPONSE){
		    return;
		}
		if(response.getType()==Message.EXCEPTION_RESPONSE){
			throw new Exception(((ExceptionResponse) response).getExceptionMessage());
		}
		tracePrint("Unexpected Response Type");
    }

}