/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package org.cnice.rad.common;

/*
 * @(#)SqlMetadata.java
 */

/**
 * @version	1.0  Ene 2007
 * @administrator: Jose-Luis Abreu
 */

import java.io.Serializable;
import java.sql.Types;
import java.util.*;

public class SqlMetadata implements Serializable {

	public static final int INTEGER=Types.INTEGER, 	// and TINYINT
		VARCHAR=Types.VARCHAR, 						// and CHAR
		DATE=Types.DATE, 							// out YYYY-MM-DD
		// in YY-MM-DD, YYYY-MM-DD, YYMMDD, YYYYMMDD-DHHMSS ignoring time portion
		TIME=Types.TIME, 							// out HH:MM:SS
		// in HH:MM:SS, HHMMSS, HHMM, HH
		TIMESTAMP=Types.TIMESTAMP, 			// NULL sets the current time
		// out depends on length YYYYMMDDHHMMSS, YYMMDDHHMMSS, YYYYMMDD, YYMMDD
		// DATETIME, in and out 'YYYY-MM-DD HH:MM:SS'
		LONGVARBINARY=Types.LONGVARBINARY
		;
	private int type; // INTEGER, VARCHAR, DATE, TIME, TIMESTAMP, LONGVARBINARYB
	private String name;

	public SqlMetadata(int type,String name) {
		this.type=type;
		this.name=name;
	}

	public int getType() {
		return type;
	}

	public String getName() {
		return name;
	}

	public static int getSqlMetadataType(int sqltype) throws Exception {
		if (sqltype==Types.INTEGER||sqltype==Types.TINYINT) {
			return INTEGER;
		} else if (sqltype==Types.VARCHAR||sqltype==Types.CHAR) {
			return VARCHAR;
		} else if (sqltype==Types.DATE) {
			return DATE;
		} else if (sqltype==Types.TIME) {
			return TIME;
		} else if (sqltype==Types.TIMESTAMP) {
			return TIMESTAMP;
		} else if (sqltype==Types.LONGVARBINARY) {
			return LONGVARBINARY;
		} else {
		    throw new Exception("Not supported sql type "+sqltype);
		}
	};

	public static String getDateTimeString() {
//	    Calendar calendar = new GregorianCalendar();
	    Calendar calendar = Calendar.getInstance();
//	    System.out.println("Revisando hora "+ calendar);
		String date=justify(calendar.get(Calendar.YEAR),-4,'0')+
			"-"+
			justify(calendar.get(Calendar.MONTH)+1,-2,'0')+
			"-"+
			justify(calendar.get(Calendar.DAY_OF_MONTH),-2,'0');
		date+=" "+
			justify(calendar.get(Calendar.HOUR_OF_DAY),-2,'0')+
			":"+
			justify(calendar.get(Calendar.MINUTE),-2,'0')+
			":"+
			justify(calendar.get(Calendar.SECOND),-2,'0');
		return date;
		
	}

	public static String justify(int val,int j,char ch) {
		return justify(String.valueOf(val),j,ch);
	}
	
	public static String justify(String s,int j,char ch) {
		StringBuffer sb=new StringBuffer(s);
		if (j>0) {
			while (sb.length()<j) {
				sb.append(ch);
			}
		} else {
			while (sb.length()<-j) {
				sb.insert(0,ch);
			}
		}
		String auxs=new String(sb);
		if (j>0) {
			if (auxs.length()>j) {
				auxs=auxs.substring(0,j);
			}
		} else {
			if (auxs.length()>-j) {
				auxs=auxs.substring(auxs.length()+j);
			}
		}
		return auxs;
	}

}