/*
_____________________________________________

Descartes. Interactive Mathematics for Teaching and Learning
This program is free software; you can redistribute it and/or 
modify it under the terms of the European Public License 
(EUPL) v 1.1 as published by the IDABC (see http://ec.europa.eu/idabc/). 

Developing institutions
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Software development: Jos Luis Abreu Len, Marta Oliver Serrat
Functional Design: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Copyright by Ministerio de Educacin de Espaa.

This programme is distributed in good faith, but WITHOUT ANY WARRANTY, not even
with the implicit warranties of USABILITY or FITNESS FOR A CONCRETE PURPOSE.
For more information consult the EUPL License European Public License. 

You should have received a copy of the EUPL European Public License along with this 
program. If not, see http://ec.europa.eu/idabc/servlets/Docbb6d.pdf?id=31979
_____________________________________________

Descartes. Matemticas interactivas para la enseanza y el aprendizaje
Este programa es software libre; usted tiene derecho a redistribuirlo y/o
a modificarlo bajo los trminos de la Licencia pblica europea 
(EUPL) v 1.1 publicados por la IDABC (ver  http://ec.europa.eu/idabc/).

Instituciones desarrolladoras:
- Instituto de Tecnologas Educativas del Ministerio de Educacin de Espaa
- Instituto de Matemticas de la Universidad Nacional Autnoma de Mxico

Desarrollo de software: Jos Luis Abreu Len, Marta Oliver Serrat
Diseo funcional: Jos Luis Abreu Len, Juan Madrigal Muga, Jos R. Galo Snchez

Derechos de explotacin propiedad del Ministerio de Educacin de Espaa.

Este programa se distribuye de buena fe, pero SIN NINGUNA GARANTA, incluso sin 
las presuntas garantas implcitas de USABILIDAD o ADECUACIN A PROPSITO CONCRETO. 
Para mas informacin consulte la Licencia EUPL European Public License.

Usted debe haber recibido una copia de la Licencia EUPLjunto con este 
programa, si por algn motivo no le es posible visualizarla, puede consultarla en
la siguiente URL: http://ec.europa.eu/idabc/servlets/Docb4f4.pdf?id=31980
_____________________________________________

*/
package org.cnice.rad.server;

/*
 * @(#)ServeOneRequest.java
 */

/**
 * @version	1.0  Ene 2007
 * @administrator: Jose-Luis Abreu
 */

import java.io.*;
import java.util.Vector;
import java.net.*;
import java.sql.*;
import org.cnice.rad.common.*;
import org.cnice.rad.admin.*;
import org.cnice.rad.learner.*;

class ServeOneRequest implements Runnable {

	DatabaseConnection dbc;

	private static final int admin=0,teacher=1,learner=2;
	private Server server;
    private Socket socket;
    private ObjectInputStream in;
    private ObjectOutputStream out;

    ServeOneRequest(Server server,DatabaseConnection dbc,Socket s) throws IOException {
    	this.server=server;
    	this.dbc=dbc;
		socket = s;
		out =
		    new ObjectOutputStream( socket.getOutputStream());
		in =
		    new ObjectInputStream( socket.getInputStream());
		// If any of the above calls throw an
		// exception, the caller is responsible for
		// closing the socket. Otherwise the thread
		// will close it.
    }

    public void run() {
    	Message msg=null, rmsg=null;
    	Vector queueMsg=null;;
		try {
			Object objMsg = in.readObject();
			
			if (objMsg instanceof Message){
				queueMsg = new Vector();
				queueMsg.add(objMsg);
			} else {
				queueMsg = (Vector) objMsg;
			}
			
			for (int i = 0;i <queueMsg.size();i++) {
				msg = (Message) queueMsg.get(i);
				System.out.print(msg.getMessageName()+"...");
				rmsg = processRequest(msg);
				System.out.print(" Respond "+rmsg.getMessageName());
				
				if (msg.getType() != Message.UPDATEACTIVITY_REQUEST) {
					out.writeObject(rmsg);
				}

				
				System.out.println(".");	
			}	
						
		} catch(ClassNotFoundException ee) {
		    ee.printStackTrace();
		} catch(OptionalDataException ode) {
		    ode.printStackTrace();
		} catch(EOFException eoe){
			eoe.printStackTrace();
		} catch(SocketException soe){
			soe.printStackTrace();
		} catch (IOException ioe){
		    ioe.printStackTrace();
		}finally {
		    try {
				socket.close();
		    } catch(IOException e) {
				System.err.println("Socket not closed");
		    }
		}
		if (rmsg!=null&&rmsg.getType()==Message.CLOSESERVER_RESPONSE) {
			server.shutdown();
		}
    }


    private int getUserType(String login,String password) {
    	int type=-1;
System.out.println("Connected: "+dbc.isConnected());
    	if (dbc.isConnected()) {
			try{
				String query="SELECT tipo FROM usuario WHERE codigo_a='"+login+"' AND contrasenya='"+password+"'";
			    Statement stmt = dbc.getConnection().createStatement();
			    ResultSet rs = stmt.executeQuery(query);
				if (rs.next()) {
					String stype=rs.getString("tipo");
System.out.println("stype="+stype);
					if (stype!=null) {
						if (stype.equals("Alumno")) {
							type=learner;
						} else if (stype.equals("Profesor")) {
							type=teacher;
						} else if (stype.equals("Administrador")) {
							type=admin;
						}
					} else {
						type=-1;
					}
				}
			    rs.close();
			    stmt.close();
			} catch(SQLException se) {
System.out.println(se);
				type=-1;
			}
    	}
    	return type;
    }

    private synchronized Message processRequest(Message msg) {
		Message rmsg=null;
		String emsg="";
		try{
			if (dbc.isConnected()) {
				if (msg.getType()==Message.HELLO_REQUEST){
					rmsg = new HelloResponse();
				} else {
		    		int usertype=getUserType(msg.login,msg.password);
		    		if (usertype>=0) {
						if (msg.getType()==Message.USERDATA_REQUEST){
						    rmsg = new UserDataResponse(msg.login,msg.password,getPublicUserData(msg.login,msg.password));
						} else if (msg.getType()==Message.QUERY_REQUEST){
							String query=((QueryRequest)msg).getQuery();
						    SqlResult values = executeQuery(query);
						    rmsg = new QueryResponse(msg.login,msg.password,values);
						} else if(msg.getType()==Message.UPDATE_REQUEST){
							if (usertype==admin) {
								String updateStr=((UpdateRequest)msg).getUpdateString();
			    				SqlValue[] values=((UpdateRequest)msg).getValues();
					    		int n=executeUpdate(updateStr,values);
					    		rmsg = new UpdateResponse(msg.login,msg.password,n);
							} else {
					    		emsg="User not allowed error.";
							}
						} else if (msg.getType()==Message.CLOSESERVER_REQUEST){
							if (usertype==admin) {
							    rmsg = new CloseServerResponse(msg.login,msg.password);
							} else {
					    		emsg="User not allowed error.";
							}
						} else if (msg.getType()==Message.CREATESESIONGRUPO_REQUEST){
					    	String inserStr=((CreateSesionGrupoRequest)msg).getInsertString();
					    	String grupo=((CreateSesionGrupoRequest)msg).getGrupo();
				    		String[] info = executeCreateSesionGrupo(msg.login,msg.password,inserStr,grupo);
							if (info==null || info[0].equals("")) {
					    		rmsg = new ExceptionResponse(msg.login,msg.password,
				    					"Error of the assignation of the identifier");
					    	} else {
				    			rmsg = new CreateSesionGrupoResponse(
				    				msg.login,msg.password,info);
				    		}
						} else if (msg.getType()==Message.CLOSESESIONGRUPO_REQUEST){
				    		executeCloseSesionGrupo(msg.login,msg.password,((CerrarSesionGrupoRequest)msg).getGrupo());
						    rmsg = new CerrarSesionGrupoResponse(msg.login,msg.password);
						} else if (msg.getType()==Message.INSERTACTIVITY_REQUEST){
							InsertActivityRequest request=(InsertActivityRequest)msg;
				    		String[] info = executeInsertActivity(msg.login,msg.password,
					    			request.getId_sesion_equipo(),
					    			request.getDate(),
					    			request.getCodebase(),
					    			request.getFilePath(),
					    			request.getName(),
					    			request.getAppletCode(),
					    			request.getInfoind(),
					    			request.getInfoest());
							if (info==null || info[0].equals("")) {
					    		rmsg = new ExceptionResponse(msg.login,msg.password,
				    					"Error of the assignation of the identifier");
					    	} else {
				    			rmsg = new InsertActivityResponse(msg.login,msg.password,info);
				    		}
						} else if(msg.getType()==Message.CLOSEACTIVITY_REQUEST){
							CloseActivityRequest request=(CloseActivityRequest)msg;
				    		executeCloseActivity(
				    			request.getID(),
				    			request.getDate(),
				    			request.getNota(),
				    			request.getAciertos(),
				    			request.getErrores());
				    		rmsg = new CloseActivityResponse(msg.login,msg.password);
						} else if (msg.getType()==Message.UPDATEACTIVITY_REQUEST){
							UpdateActivityRequest request=(UpdateActivityRequest)msg;
				    		executeUpdateActivity(
				    			request.getID(),request.getTipo(),request.getName(),request.getSeconds(),
				    			request.getCentiseconds(),request.getParametros());
				    		rmsg = new UpdateActivityResponse(msg.login,msg.password);
						} else if (msg.getType()==Message.CLOSESESIONEQUIPO_REQUEST){
							CloseSesionEquipoRequest request=(CloseSesionEquipoRequest)msg;
				    		executeCloseSesionEquipo(msg.login,msg.password,
				    			request.getID());
						    rmsg = new CloseSesionEquipoResponse(msg.login,msg.password);
						} else if (msg.getType()==Message.CREATESESIONEQUIPO_REQUEST){
					    	CreateSesionEquipoRequest request=(CreateSesionEquipoRequest)msg;
				    		String[] info = executeCreateSesionEquipo(
				    				msg.login,msg.password,
					    			request.getNom_equipo(),
					    			request.getID_grupo(),
					    			request.getID_alumno());
							if (info==null || info[0].equals("")) {
					    		rmsg = new ExceptionResponse(msg.login,msg.password,
				    					"Error of the assignation of the identifier");
					    	} else {
				    			rmsg = new CreateSesionEquipoResponse(
				    				msg.login,msg.password,info);
				    		}
						} else {
							emsg="Fatal error, Unknown Request type: " + msg.getType();
							System.err.println(emsg);
						}
					} else {
						emsg="Authentication error.";
					}
				}
			} else {
				emsg="Connection to database failed.";
			}
		} catch (Exception ex) {
		    rmsg = null;
		    ex.printStackTrace();
		    emsg=ex.getMessage();
		}
		if (rmsg==null) {
		    rmsg = new ExceptionResponse(msg.login,msg.password,emsg);
			System.out.println(emsg);
		}
		return rmsg;
    }

    private String getID(String table,String field,String value) {
    	String id=null;
		try{
			String query="SELECT id FROM "+table+" WHERE "+
				field+"='"+value+"'";
		    Statement stmt = dbc.getConnection().createStatement();
		    ResultSet rs = stmt.executeQuery(query);
			if (rs.next()) {
				id=rs.getString(1);
			}
		    rs.close();
		    stmt.close();
		} catch(SQLException se) {
			id=null;
		}
	    return id;
    }

/*
 *
id int(6) unsigned NOT NULL auto_increment, 
tipo varchar(15) NOT NULL DEFAULT 'Alumno', 
grupo int(6) unsigned NOT NULL DEFAULT '0', 
codigo_a varchar(10) NOT NULL DEFAULT '', 
contrasenya varchar(12) NOT NULL DEFAULT '', 
apellido_p varchar(30) NOT NULL DEFAULT '', 
apellido_m varchar(30) NOT NULL DEFAULT '', 
nombres varchar(40) NOT NULL DEFAULT '', 
alias varchar(20), 
correo_e varchar(40) DEFAULT '', 
*/
    private String[] getPublicUserData(String login,String password) {
    	String[] data=new String[0];
		try{
			String query="SELECT * FROM usuario WHERE codigo_a='"+login+"' AND contrasenya='"+password+"'";
		    Statement stmt = dbc.getConnection().createStatement();
		    ResultSet rs = stmt.executeQuery(query);
			if (rs.next()) {
				data=new String[10];
				data[0]=rs.getString("id");
				data[1]=rs.getString("tipo");
				data[2]=String.valueOf(rs.getInt("grupo"));
				data[3]=rs.getString("codigo_a");
				data[4]=rs.getString("contrasenya");
				data[5]=rs.getString("apellido_p");
				data[6]=rs.getString("apellido_m");
				data[7]=rs.getString("nombres");
				data[8]=rs.getString("alias");
				data[9]=rs.getString("correo_e");
			}
		    rs.close();
		    stmt.close();
		} catch(SQLException se) {
			se.printStackTrace();
System.out.println(" ERROR ");
			data=new String[0];
		}
    	return data;
    }

    private SqlResult executeQuery(String query) throws Exception {
    	SqlResult values=new SqlResult();
		try{
		    Statement stmt = dbc.getConnection().createStatement();
		    ResultSet rs = stmt.executeQuery(query);
			ResultSetMetaData rsmd = rs.getMetaData ();
			int numCols = rsmd.getColumnCount ();
			SqlMetadata[] metadata= new SqlMetadata[numCols];
			for (int i=1; i<=numCols; i++) {
				int type = SqlMetadata.getSqlMetadataType(rsmd.getColumnType(i));
				metadata[i-1]=new SqlMetadata(type,rsmd.getColumnName(i));
			}
			while (rs.next ()) {
				SqlValue[] value=new SqlValue[numCols];
				for (int i=1; i<=numCols; i++) {
					switch (metadata[i-1].getType()) {
					case SqlMetadata.INTEGER :
						value[i-1]=new SqlValue(metadata[i-1].getType(),
							rs.getInt(i));
							break;
					case SqlMetadata.VARCHAR :
						value[i-1]=new SqlValue(metadata[i-1].getType(),
							rs.getString(i));
							break;
					case SqlMetadata.DATE :
						value[i-1]=new SqlValue(metadata[i-1].getType(),
							rs.getString(i));
							break;
					case SqlMetadata.TIME :
						value[i-1]=new SqlValue(metadata[i-1].getType(),
							rs.getString(i));
							break;
					case SqlMetadata.TIMESTAMP :
						value[i-1]=new SqlValue(metadata[i-1].getType(),
							rs.getString(i));
							break;
					case SqlMetadata.LONGVARBINARY :
						value[i-1]=new SqlValue(metadata[i-1].getType(),
							rs.getBytes(i));
							break;
					}
				}
				values.addValue(new SqlValueArray(metadata,value));
			}
		    rs.close();
		    stmt.close();
		} catch(SQLException se) {
		    System.err.println(se);
		    throw new Exception(se.getMessage());
		}

		return values;
    }

    private int executeUpdate(String updateStr,SqlValue[] values) throws Exception {
    	int n=-1;
		try{
		    PreparedStatement pstmt = dbc.getConnection().prepareStatement(updateStr);
//System.out.println("Server executeUpdate:"+updateStr);
		    if (values!=null) {
		    	for (int i=0;i<values.length;i++) {
		    		switch (values[i].getType()) {
		    			case SqlMetadata.INTEGER:
		    					pstmt.setInt(i+1,values[i].getInt());
		    					break;
		    			case SqlMetadata.DATE:
		    					pstmt.setString(i+1,values[i].getString());
		    					break;
		    			case SqlMetadata.TIME:
		    					pstmt.setString(i+1,values[i].getString());
		    					break;
		    			case SqlMetadata.TIMESTAMP:
		    					pstmt.setString(i+1,values[i].getString());
		    					break;
		    			case SqlMetadata.VARCHAR:
		    					pstmt.setString(i+1,values[i].getString());
		    					break;
		    			case SqlMetadata.LONGVARBINARY:
		    					pstmt.setBytes(i+1,values[i].getBytes());
//System.out.println(i+" "+new String(values[i].getBytes()));
		    					break;
		    		}
		    	}
		    }
		    n=pstmt.executeUpdate();
		    pstmt.close();
		    return n;
		} catch (SQLException se) {
		    se.printStackTrace();
		    System.err.println(se);
		    throw new Exception(se.getMessage());
		}
    }

    private String[] executeCreateSesionGrupo(String login,String password,
    	String insertStr,String grupo) throws Exception {
    	String grupoid=getID("grupo","nombre",grupo);
		String[] info=null;
		try{
			String s="INSERT INTO sesion_grupo "+insertStr;
		    PreparedStatement pstmt = dbc.getConnection().prepareStatement(s);
    		pstmt.setString(1,grupoid);
		    pstmt.executeUpdate();
		    pstmt.close();
	
		    Statement stmt = dbc.getConnection().createStatement();
		    ResultSet rs = stmt.executeQuery("SELECT LAST_INSERT_ID()");
		    if(!rs.next()){
				throw new Exception("Unable to retrieve identifier");
		    }
		    String identifier=rs.getString(1);
		    rs.close();
		    stmt.close();
			info=new String[1];
			info[0]=identifier;
			System.out.println("Register "+identifier+" inserted");
		} catch (SQLException se) {
		    se.printStackTrace();
		    System.err.println(se);
		    throw new Exception(se.getMessage());
		}
		return info;
    }

    private String[] executeCreateSesionEquipo(String login,String password,
    	String nom_equipo,String id_grupo,String[] id_alumno) throws Exception {
    		
//System.out.println("creando sesin de equipo");
		String[] info=null;
		try{

		    Statement stmte = dbc.getConnection().createStatement();
		    ResultSet rse = stmte.executeQuery("SELECT ID FROM equipo WHERE NOMBRE='"+nom_equipo+"'");
		    if (!rse.next()) {
				String insert="INSERT INTO equipo "
						+"(nombre) "
						+"VALUES ("
						+"'"+nom_equipo+"')";
			    PreparedStatement pstmt = dbc.getConnection().prepareStatement(insert);
			    pstmt.executeUpdate();
			    pstmt.close();
			    stmte = dbc.getConnection().createStatement();
			    rse = stmte.executeQuery("SELECT ID FROM equipo WHERE NOMBRE='"+nom_equipo+"'");
			    rse.next();
		    }
			String id_equipo = rse.getString(1);
//System.out.println("id_equipo="+id_equipo);
			String date=SqlMetadata.getDateTimeString();
			String insert="INSERT INTO sesion_equipo "
					+"(s_grupo,equipo,inicio,estado) "
					+"VALUES ("
					+"'"+id_grupo+"',"
					+"'"+id_equipo+"',"
					+"'"+date+"',"
					+"0)";
		    PreparedStatement pstmt = dbc.getConnection().prepareStatement(insert);
		    pstmt.executeUpdate();
		    pstmt.close();
	
		    Statement stmt = dbc.getConnection().createStatement();
		    ResultSet rs = stmt.executeQuery("SELECT LAST_INSERT_ID()");
		    if(!rs.next()){
				throw new Exception("Unable to retrieve identifier");
		    }
		    String identifier=rs.getString(1);
		    rs.close();
		    stmt.close();
		    int sz=1;
		    if (id_alumno!=null) {
		    	sz+=id_alumno.length;
		 	}
			info=new String[sz];
			info[0]=identifier;
			System.out.println("Register sesion_equipo "+identifier+" inserted");
		    if (id_alumno!=null) {
			    for (int i=0;i<id_alumno.length;i++) {
					insert="INSERT INTO sesion_alumno "
							+"(s_equipo,alumno) "
							+"VALUES ("
							+"'"+identifier+"',"
							+"'"+id_alumno[i]+"')";
				    pstmt = dbc.getConnection().prepareStatement(insert);
				    pstmt.executeUpdate();
				    pstmt.close();
			
				    stmt = dbc.getConnection().createStatement();
				    rs = stmt.executeQuery("SELECT LAST_INSERT_ID()");
				    if(!rs.next()){
						throw new Exception("Unable to retrieve identifier");
				    }
				    info[i+1]=rs.getString(1);
				    rs.close();
				    stmt.close();
					System.out.println("Register sesion_alumno "+info[i+1]+" inserted");
			    }
			}
		} catch (SQLException se) {
		    se.printStackTrace();
		    System.err.println(se);
		    throw new Exception(se.getMessage());
		}
		return info;
    }

    private void executeCloseSesionGrupo(String login,String password,String grupo) throws Exception {
		try{
		// cerrar sesiones de equipo
    		String s="UPDATE sesion_equipo SET estado=1 "+"WHERE estado=0";
    		if (grupo!=null && grupo.length()>0) {
    			 s+=" AND s_grupo IN (SELECT id FROM sesion_grupo WHERE grupo IN (SELECT id FROM grupo WHERE nombre ='"+grupo+"'))";
    		}
		    PreparedStatement pstmt = dbc.getConnection().prepareStatement(s);
		    pstmt.executeUpdate();
		    pstmt.close();	
		// cerrar sesin de grupo
			String date=SqlMetadata.getDateTimeString();
    		s="UPDATE sesion_grupo SET estado=1, "+"fin='"+date+"' "+"WHERE estado=0";
    		if (grupo!=null && grupo.length()>0) {
    			 s+=" AND grupo IN (SELECT id FROM grupo WHERE nombre ='"+grupo+"')";
    		}
		    pstmt = dbc.getConnection().prepareStatement(s);
		    pstmt.executeUpdate();
		    pstmt.close();	
		} catch (SQLException se) {
		    se.printStackTrace();
		    System.err.println(se);
		    throw new Exception(se.getMessage());
		}
    }

    private String[] executeInsertActivity(String login,String password,
    	String id_sesion_equipo,String date,
    	String codebase,String filepath,String name,String code,String infoind,String infoest) throws Exception {
    	String userid=getID("usuario","codigo_a",login);
		String[] info=null;
    	if (userid!=null) {
			try{
			    PreparedStatement pstmt = null;
			    String insert=null;
				String query="SELECT * FROM escena "+
					"WHERE codebase='"+codebase+"' "+
					"AND archivo='"+filepath+"' "+
					"AND nombre='"+name+"'";
			    Statement stmt = dbc.getConnection().createStatement();
			    ResultSet rs = stmt.executeQuery(query);
			    String id_escena=null;
				while (rs.next()) {
				    String s_applet=new String(rs.getBytes("applet"));
				    if (s_applet.equals(code)) {
				    	id_escena=rs.getString("id");
				    	break;
				    }
				}
			    rs.close();
			    stmt.close();
			    if (id_escena==null) {
					insert="INSERT INTO escena "
							  	+"(codebase,archivo,nombre,applet,infoind,infoest) "
								+"VALUES ("
								+"'"+codebase+"',"
								+"'"+filepath+"',"
								+"'"+name+"',"
								+"?,?,?)";
								
				    pstmt = dbc.getConnection().prepareStatement(insert);
		    		pstmt.setBytes(1,code.getBytes());
		    		pstmt.setBytes(2,infoind.getBytes());
		    		pstmt.setBytes(3,infoest.getBytes());
				    pstmt.executeUpdate();
				    pstmt.close();
				    stmt = dbc.getConnection().createStatement();
				    rs = stmt.executeQuery("SELECT LAST_INSERT_ID()");
				    if(!rs.next()){
						throw new Exception("Unable to retrieve identifier");
				    }
				    id_escena=rs.getString(1);
				    rs.close();
				    stmt.close();
			    }
				insert="INSERT INTO actividad "
						  	+"(s_equipo,escena,inicio) "
							+"VALUES ("
							+"'"+id_sesion_equipo+"',"
							+"'"+id_escena+"',"
							+"'"+date+"')";
			    pstmt = dbc.getConnection().prepareStatement(insert);
//	    		pstmt.setString(1,id_escena);
			    pstmt.executeUpdate();
			    pstmt.close();
		
			    stmt = dbc.getConnection().createStatement();
			    rs = stmt.executeQuery("SELECT LAST_INSERT_ID()");
			    if(!rs.next()){
					throw new Exception("Unable to retrieve identifier");
			    }
			    String identifier=rs.getString(1);
			    rs.close();
			    stmt.close();
				info=new String[1];
				info[0]=identifier;
				System.out.println("Register "+identifier+" inserted");
			} catch (SQLException se) {
			    se.printStackTrace();
			    System.err.println(se);
			    throw new Exception(se.getMessage());
			}
		} else {
			String s="";
			if (userid==null) {
				s+=" useird is null ";
			}
		    System.err.println(s);
		    throw new Exception(s);
		}
		return info;
    }

    private void executeCloseActivity(
    	String id,String date,
    	int nota,int aciertos,int errores) throws Exception {
		try{
			String query="SELECT * FROM actividad WHERE id='"+id+"'";
		    Statement stmt = dbc.getConnection().createStatement();
		    ResultSet rs = stmt.executeQuery(query);
		    if(!rs.next()){
				throw new Exception("Unable to retrieve object");
		    }
//		    if (rs.getInt("state")>0) {
//			    rs.close();
//			    stmt.close();
//		    	throw new Exception("Error with state field");
//		    }
		    rs.close();
		    stmt.close();


    		String s="UPDATE actividad SET "+
    			"fin='"+date+"',"+
    			"nota=?,aciertos=?,errores=? "+
    			"WHERE id='"+id+"'";
		    PreparedStatement pstmt = dbc.getConnection().prepareStatement(s);
		    pstmt.setInt(1,nota);
		    pstmt.setInt(2,aciertos);
		    pstmt.setInt(3,errores);
		    pstmt.executeUpdate();
		    pstmt.close();	
		} catch (SQLException se) {
		    se.printStackTrace();
		    System.err.println(se);
		    throw new Exception(se.getMessage());
		}
    }

    private void executeUpdateActivity(
    	String id,String tipo,String name,String seconds,String centiseconds,String[][] parametros) throws Exception {
		try{
			String insert="INSERT INTO accion "
					  	+"(actividad,segundos,centesimas,tipo,nombre) "
						+"VALUES ("
						+"'"+id+"',"
						+"'"+seconds+"',"
						+"'"+centiseconds+"',"
						+"'"+tipo+"',"
						+"'"+name+"')";
//System.out.println(insert);
		    PreparedStatement pstmt = dbc.getConnection().prepareStatement(insert);
		    pstmt.executeUpdate();
		    pstmt.close();
	
		    Statement stmt = dbc.getConnection().createStatement();
		    ResultSet rs = stmt.executeQuery("SELECT LAST_INSERT_ID()");
		    if(!rs.next()){
				throw new Exception("Unable to retrieve identifier");
		    }
		    String id_accion=rs.getString(1);
		    rs.close();
		    stmt.close();
		    if (parametros!=null) {
				for (int i=0;i<parametros.length;i++) {
					if ((parametros[i]!=null) && (parametros[i].length==2)) {
						insert="INSERT INTO parametro "
								  	+"(accion,identificador,valor) "
									+"VALUES ("
									+"'"+id_accion+"',"
									+"'"+parametros[i][0]+"',"
									+"'"+parametros[i][1]+"')";
					    pstmt = dbc.getConnection().prepareStatement(insert);
					    pstmt.executeUpdate();
					    pstmt.close();
					}
				}
		  	}
		} catch (SQLException se) {
		    se.printStackTrace();
		    System.err.println(se);
		    throw new Exception(se.getMessage());
		}
    }

    private void executeCloseSesionEquipo(String login,String password,
    	String id_sesion_equipo) throws Exception {
		try{
    		String s="UPDATE sesion_equipo SET estado=1 "+
    		"WHERE id='"+id_sesion_equipo+"'";
		    PreparedStatement pstmt = dbc.getConnection().prepareStatement(s);
		    pstmt.executeUpdate();
		    pstmt.close();	
		} catch (SQLException se) {
		    se.printStackTrace();
		    System.err.println(se);
		    throw new Exception(se.getMessage());
		}
    }
/*
 * sesion_equipo y las sesion_alumno
 CREATE TABLE sesion_equipo ( 
id int(10) unsigned NOT NULL auto_increment, 
s_grupo int(8) unsigned NOT NULL DEFAULT '0', 
equipo int(8) unsigned NOT NULL DEFAULT '0', 
inicio datetime NOT NULL DEFAULT '2007-01-01', 
estado tinyint(2) NOT NULL DEFAULT '0', 
PRIMARY KEY (id), 
foreign key (s_grupo) references sesion_grupo(id), 
foreign key (equipo) references equipo(id) 
) type = InnoDB;

CREATE TABLE sesion_alumno ( 
id int(10) unsigned NOT NULL auto_increment, 
s_equipo int(10) unsigned NOT NULL DEFAULT '0', 
alumno int(6) unsigned NOT NULL DEFAULT '0', 
PRIMARY KEY (id), 
foreign key (s_equipo) references sesion_equipo(id), 
foreign key (alumno) references usuario(id) 
) type = InnoDB;
*/
}